from __future__ import annotations

import itertools
import math
from typing import TYPE_CHECKING, Iterable

import astropy.units as u
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from astropy.coordinates import Longitude
from scipy.interpolate import interp1d
from sklearn.model_selection import GridSearchCV, RepeatedKFold
from sklearn.neighbors import KernelDensity
from tensorflow.estimator import ModeKeys

from trans_finder.flares.exceptions import CalibFileMissing, InputOutOfRange
from trans_finder.flares.fit_mixture_model import create_bgmm_model
from trans_finder.flares.helpers import PValueFit
from trans_finder.flares.model_utils.utils import get_batch_size
from trans_finder.flares.state_manager import StateManager
from trans_finder.flares.utils import (
    get_index_in_array,
    safe_np_div,
    str_join,
    write_single_output_df,
)

if TYPE_CHECKING:
    from trans_finder.flares.log_parser import LogParser
    from trans_finder.flares.manager import Manager


class Calib:
    def __init__(self, mngr):
        self.mngr: Manager = mngr
        self.log: LogParser = mngr.log

    def derive_calib_single(self, calib_stage):
        """derive calibration of test statistics for anomaly detection"""
        self.log.info([["y", f"derive-calib-single({calib_stage}) "]])

        df_inputs_grp, encs_grp, decs_grp, wgts_grp, ids_grp = (
            dict(),
            dict(),
            dict(),
            dict(),
            dict(),
        )
        meta_bins_grp, df_preds = dict(), dict()

        # loop over all events and store the data needed as input for the
        # model by cut group
        self.get_calib_single_input_grps(
            df_inputs_grp=df_inputs_grp,
            encs_grp=encs_grp,
            decs_grp=decs_grp,
            wgts_grp=wgts_grp,
            ids_grp=ids_grp,
            meta_bins_grp=meta_bins_grp,
            calib_stage=calib_stage,
        )

        # for each cut group, evaluate multiple realisation of TS values
        # (either by shuffling the inputs, or by probability sampling)
        df_calib = self.get_calib_single_ts_realizations(
            df_inputs_grp=df_inputs_grp,
            encs_grp=encs_grp,
            decs_grp=decs_grp,
            wgts_grp=wgts_grp,
            ids_grp=ids_grp,
            meta_bins_grp=meta_bins_grp,
            df_preds=df_preds,
        )

        self.log.info([["b", " - fit-p-values ..."]])

        if self.mngr.is_enc_dec_anom_det or self.mngr.is_cluster_anom_det:
            fit_p_values_func = self.fit_p_values_log_linear
        elif self.mngr.is_enc_dec_clasif:
            fit_p_values_func = self.fit_p_values_kde
        else:
            raise Exception("problem with ts configuration")

        self.fit_p_values_by_flat_grp(
            df=df_calib,
            calib_stage=calib_stage,
            ts_tags=self.mngr.df_cols_ts_0,
            fit_func=fit_p_values_func,
        )

        output = {
            "df_calib": df_calib,
            "df_preds": df_preds,
        }

        return output

    def get_calib_single_input_grps(
        self,
        df_inputs_grp,
        encs_grp,
        decs_grp,
        wgts_grp,
        meta_bins_grp,
        calib_stage,
        ids_grp=None,
        *args,
        **kwargs,
    ):
        """
        loop over all events and store the data by cut group
        """
        assert ids_grp is not None

        keep_cols = None
        n_predict = -1
        n_filtered_grp = dict()

        while True:
            try:
                perdict_data = self.mngr.data.get_perdict_data(
                    n_predict=n_predict,
                    mode=ModeKeys.PREDICT,
                    batch_size=self.mngr.calib_batch_size,
                )
            except InputOutOfRange:
                break

            inputs = perdict_data["inputs"]
            pd_df = perdict_data["df"]
            # print(perdict_data['df'])

            if keep_cols is None:
                keep_cols = self.mngr.get_phase_cols_to_keep(pd_df)

                for calib_stage_now in self.mngr.calib_stages:
                    flat_key_tag = self.get_flat_key_tag(calib_stage_now)
                    keep_cols += [flat_key_tag]

            n_predict += len(pd_df)

            # add the flat representation of the metadata groups, needed for the rest
            # of the calibration process
            self.add_flat_meta_keys(df_calib=pd_df)
            # print(pd_df)

            # loop over all groups and derive the new TS
            grps_0, keys_0 = self.get_flat_meta_groups(
                df=pd_df, calib_stage=calib_stage
            )

            for n_grp_0 in range(len(keys_0)):
                grp_key = keys_0[n_grp_0]
                indx = grps_0.get_group(name=grp_key).index
                n_indx = len(indx)
                df_grp = pd_df.loc[indx]

                if grp_key not in n_filtered_grp:
                    n_filtered_grp[grp_key] = 0
                n_filtered_grp[grp_key] += n_indx

                self.log.info(
                    [
                        ["g", " - filter group: "],
                        ["b", grp_key],
                        ["g", " (now # "],
                        ["p", "{:>5d}".format(n_indx)],
                        ["g", ") --> tot # "],
                        ["y", "{:>5d}".format(n_filtered_grp[grp_key])],
                    ]
                )

                # filter selected events for this group. at this point the indx as
                # index is the same as the order of elements, and so can also be
                # used to slice the np arrays
                df_grp = df_grp[keep_cols].reset_index(drop=True)
                # print(df_grp)

                encs_now = inputs[self.mngr.tag_enc][indx, :]
                decs_now = inputs[self.mngr.tag_dec][indx, :]
                ids_now = inputs[self.mngr.tag_model_input_ids][indx]

                if len(inputs[self.mngr.tag_wgt].shape) > 1:
                    wgts_now = inputs[self.mngr.tag_wgt][indx, :]
                else:
                    wgts_now = inputs[self.mngr.tag_wgt][indx]

                # store the data by the group
                if grp_key not in df_inputs_grp:
                    # if this is the first time we have events in this group
                    df_inputs_grp[grp_key] = df_grp

                    encs_grp[grp_key] = encs_now
                    decs_grp[grp_key] = decs_now
                    wgts_grp[grp_key] = wgts_now
                    ids_grp[grp_key] = ids_now

                else:
                    # if adding to an existing group
                    df_inputs_grp[grp_key] = df_inputs_grp[grp_key].append(
                        df_grp,
                        ignore_index=True,
                    )

                    encs_grp[grp_key] = np.concatenate(
                        (encs_grp[grp_key], encs_now),
                        axis=0,
                    )
                    decs_grp[grp_key] = np.concatenate(
                        (decs_grp[grp_key], decs_now),
                        axis=0,
                    )
                    wgts_grp[grp_key] = np.concatenate(
                        (wgts_grp[grp_key], wgts_now),
                        axis=0,
                    )
                    ids_grp[grp_key] = np.concatenate(
                        (ids_grp[grp_key], ids_now),
                        axis=0,
                    )

    def get_calib_single_ts_realizations(
        self,
        df_inputs_grp,
        encs_grp,
        decs_grp,
        wgts_grp,
        meta_bins_grp,
        df_preds,
        ids_grp=None,
        *args,
        **kwargs,
    ):
        """
        for each cut group, evaluate multiple realisation of TS values (either by shuffling the
        inputs, or by probability sampling)

        change from rnn_blind Hermann 06.12.2022: add input_id to model_inputs
        """

        assert ids_grp is not None

        df_ts_grp = dict()
        df_calib = pd.DataFrame()

        for grp_key in encs_grp.keys():
            encs_now = encs_grp[grp_key]
            decs_now = decs_grp[grp_key]
            wgts_now = wgts_grp[grp_key]
            ids_now = ids_grp[grp_key]

            # determine how many loops to do (more than one will include data shuffling)
            n_evts = len(encs_now)
            n_extra_evals = self.mngr.n_calib_ts_events * (
                2 if self.mngr.is_enc_dec_clasif else 1
            )
            n_evals = int(1 + (n_extra_evals // n_evts))
            if self.mngr.use_tf_prob:
                n_evals = max(n_evals, self.mngr.min_n_calib_sample)

            for n_eval_now in range(n_evals):
                if (n_evals < 10) or (n_eval_now % (n_evals // 10) == 0):
                    self.log.info(
                        [
                            ["p", " - deriving ts: "],
                            ["g", grp_key, ""],
                            ["b", "{:>4d}".format(n_eval_now)],
                            ["p", " / "],
                            ["g", n_evals],
                            ["p", " ..."],
                        ]
                    )

                if grp_key not in df_ts_grp.keys():
                    df_ts_grp[grp_key] = pd.DataFrame()
                    df_preds[grp_key] = pd.DataFrame()

                # shuffle the data if needed (if using probabilities, we simply sample
                # each time on the same inputs without shufling)
                if n_eval_now > 0 and not self.mngr.use_tf_prob:
                    shuffled = self.shuffle_internal_time_step_order(encs_now, decs_now)
                    encs_now = shuffled[self.mngr.tag_enc]
                    decs_now = shuffled[self.mngr.tag_dec]

                # evaluate with the (possibly shuffled) inputs
                model_inputs = {
                    self.mngr.tag_enc: encs_now,
                    self.mngr.tag_dec: decs_now,
                    self.mngr.tag_wgt: wgts_now,
                    self.mngr.tag_model_input_ids: ids_now,
                }
                model_outputs = self.mngr.model[self.mngr.tag_model].predict(
                    model_inputs, batch_size=get_batch_size(self.mngr)
                )

                # add some bookkeeping info - we override this here, so that the info is caried
                # over as part of get_ts_from_preds()
                df_inputs_grp[grp_key].loc[:, self.mngr.n_realization_tag] = n_eval_now

                # derive the TS
                ts_output = self.get_ts_from_preds(
                    model_inputs=model_inputs,
                    model_outputs=model_outputs,
                    df_full=df_inputs_grp[grp_key],
                )

                # derive list of possible extra output colums (beyond the nominal set
                # of add_cols_0, add_cols_1) which should be propagated to the next stage
                add_cols_0 = [self.mngr.tag_wgt, self.mngr.tag_ts]
                add_cols_1 = [self.mngr.tag_pred, self.mngr.tag_dec]
                add_cols_extra = (
                    set(ts_output.keys()) - set(add_cols_0) - set(add_cols_1)
                )

                # store the results in a DataFrame
                for n_data in range(2):
                    if self.mngr.is_enc_dec_clasif and n_data > 0:
                        continue

                    if n_data == 0:
                        add_cols = [
                            [ts_output[self.mngr.tag_wgt], self.mngr.df_cols_wgt_0],
                            [ts_output[self.mngr.tag_ts], self.mngr.df_cols_ts_0],
                        ]
                        for add_extra in add_cols_extra:
                            add_cols += [
                                [ts_output[add_extra], [add_extra]],
                            ]
                        df_store = df_ts_grp

                    elif n_data == 1:
                        add_cols = [
                            [ts_output[self.mngr.tag_pred], self.mngr.df_cols_pred],
                            [ts_output[self.mngr.tag_dec], self.mngr.df_cols_dec],
                        ]
                        df_store = df_preds

                    df_now = pd.DataFrame()
                    for col in add_cols:
                        if len(df_now) == 0:
                            df_now = df_now.append(
                                pd.DataFrame(col[0], columns=col[1]),
                                ignore_index=True,
                            )
                        else:
                            df_now = df_now.merge(
                                pd.DataFrame(col[0], columns=col[1]),
                                left_index=True,
                                right_index=True,
                            )

                    df_store[grp_key] = df_store[grp_key].append(
                        df_now,
                        ignore_index=True,
                    )

                # choose the new iteration from the dataframe. reset the
                # index of df_ts_grp in order to properly merge with df_inputs_grp
                df_now = (
                    df_ts_grp[grp_key]
                    .iloc[-len(df_inputs_grp[grp_key]) :]
                    .reset_index(drop=True)
                )

                df_now = df_inputs_grp[grp_key].merge(
                    df_now,
                    left_index=True,
                    right_index=True,
                )

                df_calib = df_calib.append(df_now, ignore_index=True)

                # clean up the temporary bookkeeping data
                df_inputs_grp[grp_key].drop(
                    columns=[self.mngr.n_realization_tag], inplace=True
                )

        return df_calib

    def get_ts_step_combo(self, df_calib, calib_stage):
        """combine several p-values into a new test statistic

        Parameters
        ----------
        df_calib : pd.DataFrame
            input data including columns for TS-values and weights
        calib_fit_tag : tuple
            the unique key (cut combination) for the input data
        calib_stage : str
            tags for files to store p-value fit results

        Returns
        -------
        dict
            the derived input p-values and the corresponding combined TS-values
        """

        prev_calib_stage = self.get_prev_calib_stage(calib_stage)

        calib_tags = self.mngr.get_calib_tags(calib_stage=calib_stage)
        tag_ts_step_combo = calib_tags["ts"]
        tag_wgt_combo = calib_tags["wgt"]

        _ = self.get_p_values_by_stage(
            df_calib=df_calib,
            calib_stage=prev_calib_stage,
        )

        ts_tags = self.mngr.df_cols_ts_0
        log_p_val_tags = [self.mngr.ts_to_p_val_col[t] for t in ts_tags]
        wgt_tags = [self.mngr.ts_to_wgt_col[t] for t in ts_tags]

        tss = np.abs(
            df_calib[log_p_val_tags].values
        )  # - log of the p value for each ts
        # tss = np.abs(np.log10(np.abs(p_vals) + 1e-20)).values

        wgts = df_calib[wgt_tags].values

        # use maximum of individual ts values and pick the weight of that c
        max_idx = np.argmax(tss, axis=1)
        ts_combo = tss[np.arange(tss.shape[0]), max_idx].astype(np.float32)
        wgt_combo = wgts[np.arange(wgts.shape[0]), max_idx]

        df_calib.loc[:, tag_ts_step_combo] = ts_combo
        df_calib.loc[:, tag_wgt_combo] = wgt_combo

        # sanity check
        fill_mask = df_calib[tag_ts_step_combo].isnull()
        if len(df_calib[fill_mask]) > 0:
            raise Exception(
                f" - problem filling df_calib ({np.sum(fill_mask)})",
                fill_mask,
                df_calib,
            )

    def derive_calib(self):
        """derive the consecutive TS-calibration stages"""

        self.log.info([["g", " - derive-calib() "]])

        calib_stage_output = None
        for n_calib_stage in range(len(self.mngr.calib_stages)):
            calib_stage = self.mngr.calib_stages[n_calib_stage]

            if calib_stage == "calib_cluster_single":
                calib_stage_output = self.derive_calib_single(calib_stage=calib_stage)

            elif calib_stage == "calib_step_combo":
                calib_stage_output = self.derive_calib_single_step_combo(
                    calib_stage=calib_stage, calib_stage_input=calib_stage_output
                )

        # plot the distributions
        self.mngr.plots.plot_fit_p_values(
            data_ts=calib_stage_output["df_calib"],
            data_preds=calib_stage_output["df_preds"],
        )

        # plot the debugging plots
        self.mngr.diagnostics.plot_diagnostics()

    def get_p_values_by_stage(self, df_calib, calib_stage=None):
        if calib_stage is None:
            calib_stage = self.mngr.calib_stages[-1]

        if calib_stage == "calib_cluster_single":
            eval_output = self.get_p_values_singel_steps(
                df_calib=df_calib,
            )

        elif calib_stage == "calib_step_combo":
            eval_output = self.get_p_values_step_combo(
                df_calib=df_calib,
                calib_stage=calib_stage,
            )

        return eval_output

    def load_mcmc_states(self):
        # load bgmm and relevant metadata about the normalisation and quality of fit
        bgmm_quality_of_fit = StateManager(
            self.mngr, self.mngr.tag_bgmm_quality_of_fit
        ).load_state()

        # normalisation = StateManager(
        #     mngr=self.mngr, state_id="cluster_space_norm"
        # ).load_state()
        bgmm_state = StateManager(mngr=self.mngr, state_id="bgmm").load_state()

        self.mngr.mcmc_states = {
            self.mngr.tag_bgmm_quality_of_fit: bgmm_quality_of_fit,
            # "cluster_space_norm": normalisation,
            "bgmm": bgmm_state,
        }

    def get_ts_from_preds(self, model_inputs, model_outputs, df_full=None):
        # test statistic output with original inputs
        ts_output = self.get_ts_from_preds_clustering(
            model_inputs=model_inputs,
            model_outputs=model_outputs,
        )

        self.write_diagnostic_df(
            model_inputs=model_inputs,
            model_outputs=model_outputs,
            df_full=df_full,
            ts_output=ts_output,
        )

        return ts_output

    def write_diagnostic_df(self, model_inputs, model_outputs, df_full, ts_output):
        skip_diagnostics = (
            self.mngr.run_phase not in self.mngr.do_diagnostics.keys()
            or not self.mngr.do_diagnostics[self.mngr.run_phase]
        )
        if skip_diagnostics:
            return

        cluster_space_sample = model_outputs[self.mngr.tag_cluster_space_sample]
        cluster_space_mean = model_outputs[self.mngr.tag_cluster_space_mean]
        cluster_space_std = model_outputs[self.mngr.tag_cluster_space_std]
        auto_enc_reconstruction_error = model_outputs[
            self.mngr.tag_auto_enc_reconstruction_error
        ]
        forecasting_error = model_outputs[self.mngr.tag_model_forecasting_error]
        input_ids = model_inputs[self.mngr.tag_model_input_ids]

        df_diagnostics = df_full.merge(
            pd.DataFrame(ts_output["wgt"], columns=self.mngr.df_cols_wgt_0),
            left_index=True,
            right_index=True,
        )

        # add test statistic columns
        for ts_idx, ts_name in enumerate(self.mngr.df_cols_ts_0):
            df_diagnostics[f"{ts_name}"] = ts_output[self.mngr.tag_ts][:, ts_idx]

        df_diagnostics = df_diagnostics.assign(
            **{
                self.mngr.tag_auto_enc_reconstruction_error: auto_enc_reconstruction_error,
                self.mngr.tag_model_forecasting_error: forecasting_error,
                self.mngr.tag_model_input_ids: input_ids,
            }
        )

        # fix merge on existing columns
        for col in [c for c in df_diagnostics.columns if c.endswith("_x")]:
            df_diagnostics.drop(col, axis=1, inplace=True)

            new_col_name = col.replace("_x", "")
            orig_col_name = f"{new_col_name}_y"
            df_diagnostics.rename(columns={orig_col_name: new_col_name}, inplace=True)

        # df_diagnostics = df_diagnostics.merge(
        #     pd.DataFrame(
        #         cluster_space_sample,
        #         columns=[
        #             f"{self.mngr.tag_cluster_space_sample}_{i}"
        #             for i in range(self.mngr.model_learned_representation_n_dim)
        #         ],
        #     ),
        #     left_index=True,
        #     right_index=True,
        # )

        normed_observations = self.get_normed_mcmc_observations(
            cluster_space_mean, auto_enc_reconstruction_error
        )

        if self.mngr.bgmm_add_auto_enc_reconstruction_error_as_extra_dim:
            df_diagnostics = df_diagnostics.merge(
                pd.DataFrame(
                    normed_observations[:, -1],
                    columns=[self.mngr.tag_auto_enc_reconstruction_error_normed],
                ),
                left_index=True,
                right_index=True,
            )

            df_diagnostics = df_diagnostics.merge(
                pd.DataFrame(
                    normed_observations[:, :-1],
                    columns=[
                        f"{self.mngr.tag_cluster_space_mean_normed}_{i}"
                        for i in range(self.mngr.model_learned_representation_n_dim)
                    ],
                ),
                left_index=True,
                right_index=True,
            )
        else:
            df_diagnostics = df_diagnostics.merge(
                pd.DataFrame(
                    normed_observations[:, :],
                    columns=[
                        f"{self.mngr.tag_cluster_space_mean_normed}_{i}"
                        for i in range(self.mngr.model_learned_representation_n_dim)
                    ],
                ),
                left_index=True,
                right_index=True,
            )

        df_diagnostics = df_diagnostics.merge(
            pd.DataFrame(
                cluster_space_mean,
                columns=[
                    f"{self.mngr.tag_cluster_space_mean}_{i}"
                    for i in range(self.mngr.model_learned_representation_n_dim)
                ],
            ),
            left_index=True,
            right_index=True,
        )

        df_diagnostics = df_diagnostics.merge(
            pd.DataFrame(
                cluster_space_std,
                columns=[
                    f"{self.mngr.tag_cluster_space_std}_{i}"
                    for i in range(self.mngr.model_learned_representation_n_dim)
                ],
            ),
            left_index=True,
            right_index=True,
        )

        if self.mngr.run_phase == "calib":
            output_dir = self.mngr.data.get_dir_name("calib_diagnostics")
        elif self.mngr.run_phase == "predict":
            output_dir = self.mngr.data.get_dir_name("predict_diagnostics")
        else:
            raise Exception(" - unknown run_phase for diagnostic ...")

        write_single_output_df(
            df_in=df_diagnostics,
            output_dir=output_dir,
            in_file_filt_str="",
            out_file_patt=self.mngr.diagnostics_file_name_patt,
            tag_hdf=self.mngr.tag_hdf,
            log=self.log,
        )

    def get_normed_mcmc_observations(
        self, latent_representations, auto_enc_reconstruction_error
    ):
        if self.mngr.mcmc_states is None:
            self.load_mcmc_states()

        if self.mngr.bgmm_add_auto_enc_reconstruction_error_as_extra_dim:
            # adding the autoencoder reconstruction error as extra dimension
            observations = np.column_stack(
                [latent_representations, auto_enc_reconstruction_error]
            )
        else:
            observations = latent_representations

        # normed_observations = self.mngr.mcmc_states["cluster_space_norm"].transform(
        #     observations
        # )

        return observations

    def get_ts_from_preds_clustering(self, model_inputs, model_outputs):
        output_dec = model_inputs[self.mngr.tag_dec]
        output_wgt = model_inputs[self.mngr.tag_wgt]
        output_pred = model_outputs[self.mngr.tag_pred]
        latent_representations = model_outputs[self.mngr.tag_cluster_space_sample]
        auto_enc_reconstruction_error = model_outputs[
            self.mngr.tag_auto_enc_reconstruction_error
        ]

        output_dec = output_dec[:, :, self.mngr.data_col_indices]
        output_pred = output_pred[:, :, self.mngr.data_col_indices]
        output_wgt = output_wgt[:, :, self.mngr.data_col_indices]
        output_ids = model_outputs[self.mngr.tag_model_input_ids].astype(int)

        if self.mngr.do_train_scale:
            norms = self.mngr.data.train_feat_norms[self.mngr.tag_steps]
            shift = norms["shift"][0, self.mngr.data_col_indices]
            scale = norms["scale"][0, self.mngr.data_col_indices]

            output_dec = (output_dec / scale) - shift
            output_pred = (output_pred / scale) - shift

        # n_input_length x n_input_dims
        if latent_representations.ndim != 2:
            raise Exception(
                " - model_output latent_representation is not a matrix (n_samples x n_dims)"
            )

        n_input_length = latent_representations.shape[0]

        normed_observations = self.get_normed_mcmc_observations(
            latent_representations, auto_enc_reconstruction_error
        )

        n_selected_chain = np.argmin(
            self.mngr.mcmc_states[self.mngr.tag_bgmm_quality_of_fit]
        )

        bgmmodel = create_bgmm_model(self.mngr, n_input_length)
        bgmmodel.update_state(self.mngr.mcmc_states["bgmm"])

        outlier_score = bgmmodel.get_outlier_score(
            normed_observations,
            n_chain=n_selected_chain,
        )

        outlier_score = outlier_score[:, np.newaxis]

        if self.mngr.use_reco_err_ts_combo:
            auto_enc_reconstruction_error = auto_enc_reconstruction_error.T[
                :, np.newaxis
            ]
            output_ts = np.concatenate(
                (outlier_score, auto_enc_reconstruction_error), axis=1
            )
        else:
            output_ts = outlier_score

        # simple weighted sum
        sum_wgts = np.sum(output_wgt, axis=1)

        # average with weights over time steps for each feature
        output_dec = safe_np_div(
            np.sum((output_dec * output_wgt), axis=1),
            sum_wgts,
        )
        output_pred = safe_np_div(
            np.sum((output_pred * output_wgt), axis=1),
            sum_wgts,
        )

        # output_wgt is one scalar per event (sum over features and time steps)
        output_wgt_0 = np.sum(sum_wgts, axis=-1)

        # expand to the dimension of the ts list
        output_wgt_ts_0 = output_wgt_0[:, np.newaxis]

        if self.mngr.use_reco_err_ts_combo:
            output_wgt_ts_1 = output_wgt_0.T[:, np.newaxis]
            output_wgt = np.concatenate((output_wgt_ts_0, output_wgt_ts_1), axis=1)
        else:
            output_wgt = output_wgt_ts_0

        output = {
            self.mngr.tag_dec: output_dec,
            self.mngr.tag_pred: output_pred,
            self.mngr.tag_wgt: output_wgt,
            self.mngr.tag_ts: output_ts,
        }

        return output

    def fit_p_values_by_flat_grp(self, df, calib_stage, ts_tags, fit_func):
        # loop over all groups and derive the p-value fits
        grps_0, keys_0 = self.get_flat_meta_groups(df=df, calib_stage=calib_stage)

        for n_grp_0 in range(len(keys_0)):
            grp_key = keys_0[n_grp_0]
            indx = grps_0.get_group(name=grp_key).index
            df_grp = df.loc[indx]

            for ts_tag in ts_tags:
                mask = ~df_grp[ts_tag].isnull()
                df_grp = df_grp[mask]

            fit_func(
                df_calib=df_grp,
                ts_tags=ts_tags,
                calib_fit_tag=grp_key,
                calib_stage=calib_stage,
                is_verb=True,
            )

    def fit_p_values_log_linear(
        self, df_calib, ts_tags, calib_fit_tag, calib_stage, is_verb=True
    ):
        """fit TS distributions and parametrise as p-values

        each cumulative TS distribution is parametrised by a histogram
        covering the low range of TS values.
        the tail of the distribution is parametrised by a fit
        of an exponential curve.
        thre quantiles are defined for the TS distribution, denoted as
        quant-min/mid/max:
        - the histogram parametrisation extends up to quant-mid.
        - the parametrisation for TS values higher than that of quant-mid
        is taken from the value of the fitted function.
        - the fit of the TS-tail is performed between quant-min and quant-max (even
        though it is only used for evaluaition from quant-mid).
        there is thus is overlap between the histograms and the fitted range
        (between quant-min and quant-mid), which mitigates possible threshold
        jumps when evaluating the p-value around quant-mid.

        Parameters
        ----------
        df_calib : pd.DataFrame
            dataframe holding TS values
        ts_tags : list of str
            column names for the data DataFrame used for the fit
        calib_fit_tag : tuple
            the bin definition (e.g., (0,3,1,0) for a 4-dim cut combination)
        calib_stage : str
            tags for files to store p-value fit results
        is_verb : bool
            flag indicating is extra logging info should be done
        """

        # for _ in range(10):
        #     data = data.append(data)
        # print('\n - testing for large sample-', len(data), '\n\n')

        # minimal number of events at the edge of the distribution (highest values)
        # to keep as a margin, i.e., to exclude from the quantile fit
        n_evt_fit_margin = 5

        # how many orders of magnitude between quant-min and -max, used for the
        # fit of the cumilative distribution. for example:
        #   for fit_quant_range_scale = 2
        #   --> fit quants may be: min = (1 - 1e-03), max = (1 - 1e-05)
        fit_quant_range_scale = 2

        if is_verb:
            self.log.info(
                [["p", f" - fit-p-values-log-linear --> derive p_values({ts_tags})"]]
            )

        ts_write, log_pv_write, fit_pars = dict(), dict(), dict()
        for n_tag in range(len(ts_tags)):
            tag_ts = ts_tags[n_tag]
            wgt_tag = self.mngr.ts_to_wgt_col[tag_ts]

            out_vals = np.asarray(df_calib[tag_ts], dtype=np.float32)

            # copy the weights and renormalise
            if wgt_tag in df_calib.columns:
                out_wgts = np.asarray(df_calib[wgt_tag].copy(), dtype=np.float32)
            else:
                out_wgts = np.ones(df_calib[tag_ts].shape, dtype=np.float32)

            sum_wgts = np.sum(out_wgts, dtype=np.float32)

            # normalise the weights to the number of events
            n_evts = len(out_vals)
            out_wgts *= n_evts
            if sum_wgts > 0:
                out_wgts /= sum_wgts

            quant_max = n_evt_fit_margin / n_evts
            # if the sample is too small, set the quant range by hand
            if fit_quant_range_scale > -np.log10(quant_max):
                fit_quant_range_scale = -0.5 * np.log10(quant_max)

            # derive minimal and mid-point quantiles for the given maximum
            quant_min = np.power(10, (np.log10(quant_max) + fit_quant_range_scale))
            quant_mid = np.power(
                10, (np.log10(quant_max) + fit_quant_range_scale / 2.0)
            )

            if is_verb and (len(ts_write.keys()) == 0):
                quant_str = ", ".join(
                    [("%1.2e" % q) for q in [quant_min, quant_mid, quant_max]]
                )
                self.log.info([["b", " - fit-points[min/mid/max]: "], ["y", quant_str]])

            # derive the p-values up to the quantile mid-point
            n_cdf_bins = 50
            quant_range = [0, (1 - quant_mid)]

            cdf_range = self.get_quants(
                data=out_vals,
                quants=quant_range,
                weights=out_wgts,
            )

            cdf_bins_x = np.linspace(cdf_range[0], cdf_range[1], n_cdf_bins + 1)

            out_norm = 1.0 / n_evts
            ts_vals, sf_vals = [], []
            for ts_val in cdf_bins_x:
                wgts_in = out_wgts[out_vals >= ts_val]
                sf_val = np.sum(wgts_in, dtype=np.float32) * out_norm
                if sf_val > 0:
                    ts_vals += [ts_val]
                    sf_vals += [sf_val]

            if len(ts_vals) < 2:
                raise Exception("no positive-defined sf values ?!?", ts_vals, sf_vals)

            ts_write[tag_ts] = ts_vals
            log_pv_write[tag_ts] = np.log10(sf_vals)

            # derive the p-values between the quantile min- and max-points
            cdf_range = self.get_quants(
                data=out_vals,
                quants=[1 - quant_min, 1 - quant_max],
                weights=out_wgts,
            )

            # use a small number of bins, estimate the p-values, and use
            # those to interpolate a pline within the [min->max] quantile range
            n_cdf_bins = 10
            cdf_bins_x = np.linspace(cdf_range[0], cdf_range[1], n_cdf_bins + 1)

            fit_x, fit_y, fit_w = [], [], []
            for ts_val in cdf_bins_x:
                wgts_in = out_wgts[out_vals >= ts_val]
                n_ts_in = len(wgts_in)
                if n_ts_in < 2:
                    break

                sf_val = 1
                wgts_sum = np.sum(wgts_in, dtype=np.float32)
                if wgts_sum > 0:
                    sf_val = wgts_sum * out_norm

                fit_x += [ts_val]
                fit_y += [sf_val]
                # fit_w += [ np.sqrt(sf_val) ]
                fit_w += [1]

            p_value_fit = PValueFit(
                log=self.log,
                name=str_join(calib_fit_tag, tag_ts),
                fit_dist_type="log_linear",
                is_verb=is_verb,
            )

            # fill out of bounds values with fill_value=1
            spline = interp1d(
                ts_write[tag_ts],
                log_pv_write[tag_ts],
                kind="slinear",
                bounds_error=False,
                fill_value=0,
            )

            # set the spline for the fitter
            p_value_fit.set_spline(
                spline=spline,
                max_ts=ts_write[tag_ts][-1],
            )

            # perform the fit
            p_value_fit.fit(fit_x=fit_x, fit_y=fit_y, fit_w=fit_w)
            fit_pars[tag_ts] = p_value_fit.get_fit_pars()

        # store the results in file
        sf = {
            "sf_ts": ts_write,
            "sf_log_pv": log_pv_write,
            "fit_pars": fit_pars,
        }

        save_postfix = self.get_meta_grp_postfix(
            calib_stage=calib_stage, grp_key=calib_fit_tag
        )

        self.mngr.data.save_calib_file(
            tag_dir=self.mngr.tag_calib_dir,
            tag_file=save_postfix,
            data=sf,
        )

        # if overwrite_calib:
        #     # if we have the separate p-values for the features
        # else:
        #     # if we have do TS combination, first load the p-values for the
        #     # different features, then overwrite the file
        #     sf_feat = self.mngr.data.load_calib_file(
        #         tag_dir=self.mngr.tag_calib_dir,
        #         tag_file=save_postfix,
        #         is_verb=True,
        #     )

        #     for key in sf_feat.keys():
        #         print(key)
        #         print(sf[key])
        #         print()
        #         sf[key].update(sf_feat[key])

        #     self.mngr.data.save_calib_file(
        #         tag_dir=self.mngr.tag_calib_dir,
        #         tag_file=save_postfix,
        #         data=sf,
        #     )

    def fit_p_values_kde(
        self, df_calib, ts_tags, calib_fit_tag, calib_stage, is_verb=True
    ):
        """fit TS distributions and parametrise as p-values

        each cumulative TS distribution is parametrised by a histogram
        covering the low range of TS values.
        the tail of the distribution is parametrised by a fit
        of an exponential curve.
        thre quantiles are defined for the TS distribution, denoted as
        quant-min/mid/max:
        - the histogram parametrisation extends up to quant-mid.
        - the parametrisation for TS values higher than that of quant-mid
        is taken from the value of the fitted function.
        - the fit of the TS-tail is performed between quant-min and quant-max (even
        though it is only used for evaluaition from quant-mid).
        there is thus is overlap between the histograms and the fitted range
        (between quant-min and quant-mid), which mitigates possible threshold
        jumps when evaluating the p-value around quant-mid.

        Parameters
        ----------
        df_calib : pd.DataFrame
            dataframe holding TS values
        ts_tags : list of str
            column names for the data DataFrame used for the fit
        calib_fit_tag : tuple
            the bin definition (e.g., (0,3,1,0) for a 4-dim cut combination)
        calib_stage : str
            tags for files to store p-value fit results
        is_verb : bool
            flag indicating is extra logging info should be done

        Returns
        -------
        type
            description
        """

        if is_verb:
            self.log.info(
                [
                    ["p", " - derive p_values: "],
                    ["c", self.mngr.tag_calib_dir],
                ]
            )

        ts_write, log_pv_write, fit_pars = dict(), dict(), dict()
        for n_tag in range(len(ts_tags)):
            tag_ts = ts_tags[n_tag]
            wgt_tag = self.mngr.ts_to_wgt_col[tag_ts]

            out_vals = np.asarray(df_calib[tag_ts], dtype=np.float32)

            # copy the weights and renormalise
            if wgt_tag in df_calib.columns:
                out_wgts = np.asarray(df_calib[wgt_tag].copy(), dtype=np.float32)
            else:
                out_wgts = np.ones(df_calib[tag_ts].shape, dtype=np.float32)

            sum_wgts = np.sum(out_wgts, dtype=np.float32)
            if sum_wgts < 1e-10:
                self.log.error(
                    [
                        ["wr", " - sum_wgts = ", sum_wgts, " ?!?"],
                        ["p", "\n", tag_ts, "\n"],
                        ["r", df_calib[tag_ts], "\n"],
                        ["y", out_wgts, "\n"],
                    ]
                )
                raise Exception("sum_wgts too small ?!?")

            # check that the classification distribution is not exclusively peaked
            # around 0 (in which case, the kde will fail, but a simple histogram will work)
            quants = [1e-5, 0.5, 1 - 1e-5]
            quant_vals = np.quantile(out_vals, quants)

            has_kde = np.std(quant_vals) > 1e-5
            if not has_kde:
                raise Exception(
                    "variance of classification data too low ?!?!"
                    + " --> something wrong with training?"
                    + f"{np.std(quant_vals)} , for:\n {quant_vals}"
                )

            has_kde = has_kde and len(self.mngr.kde_bandwidths) > 0
            if not has_kde:
                raise Exception(
                    "histogram calibration not currently supported - use kde instead"
                )

            if has_kde:
                fit_data = out_vals[:, np.newaxis]

                if is_verb:
                    self.log.info(
                        [
                            ["b", f" - fitting kde({self.mngr.kde_kernel}) bandwidth "],
                            [
                                "c",
                                [
                                    float("{:1.5e}".format(b))
                                    for b in self.mngr.kde_bandwidths
                                ],
                            ],
                            ["b", " with n_evt "],
                            ["c", len(fit_data)],
                        ]
                    )

                # find the best bandwidth parameter and the corresponding kde estimator
                cv = RepeatedKFold()
                if self.mngr.kde_n_split_debug:
                    cv = RepeatedKFold(n_splits=2, n_repeats=1)

                grid = GridSearchCV(
                    estimator=KernelDensity(kernel=self.mngr.kde_kernel),
                    param_grid={"bandwidth": self.mngr.kde_bandwidths},
                    cv=cv,
                )
                grid_fit = grid.fit(
                    fit_data,
                    sample_weight=out_wgts,
                )

                kde = grid_fit.best_estimator_
                bandwidth = grid_fit.best_params_["bandwidth"]

                if is_verb:
                    self.log.info([["b", " - optimal bandwidth "], ["p", bandwidth]])

                # derive the log prabability densities for the given range of TS values
                kde_score_bins = self.mngr.kde_sample_bins
                kde_log_probs = kde.score_samples(
                    np.asarray(kde_score_bins)[:, np.newaxis]
                )

                # clip values which might cause overflow / underflow errors
                mask_max = kde_log_probs > np.log(self.mngr.kde_prob_limits["max"])
                mask_min = kde_log_probs < np.log(self.mngr.kde_prob_limits["min"])

                kde_log_probs[mask_max] = np.log(self.mngr.kde_prob_limits["max"])
                kde_log_probs[mask_min] = np.log(self.mngr.kde_prob_limits["min"])

                # derive the probability densities
                kde_prob_dens = np.exp(kde_log_probs)

                # check that the integral of the probabilities is consistent with unity
                kde_integral_prob = np.trapz(kde_prob_dens, kde_score_bins)
                if np.abs(1 - kde_integral_prob) > 1e-2:
                    self.log.warn(
                        [
                            ["wr", f" - got integrated KDE prob: {kde_integral_prob}"],
                            ["wr", " -> does kde_sample_bins have too few bins ?!?"],
                        ]
                    )

            # if not has_kde:
            #     if is_verb:
            #         self.log.info(
            #             [
            #                 ["b", " - skipping kde fit and using simple histogram"],
            #             ]
            #         )

            #     kde_prob_dens, bin_edges = np.histogram(
            #         out_vals,
            #         weights=out_wgts,
            #         bins=self.mngr.kde_sample_bins,
            #     )

            #     kde_score_bins = [
            #         (bin_edges[i] + 0.5 * (bin_edges[i + 1] - bin_edges[i]))
            #         for i in range(len(bin_edges) - 1)
            #     ]

            # derive the survival probability for each TS bin (p-values)
            kde_sf_norm = np.sum(kde_prob_dens, dtype=np.float32)
            ts_vals, kde_sf = [], []
            for n_bin in range(len(kde_score_bins)):
                sf_now = np.sum(kde_prob_dens[n_bin:], dtype=np.float32) / kde_sf_norm
                if sf_now > self.mngr.kde_prob_limits["min"]:
                    ts_vals += [kde_score_bins[n_bin]]
                    kde_sf += [sf_now]
                    # print(ts_vals[-1], kde_sf[-1])

            if len(ts_vals) < 2:
                raise Exception("no positive-defined sf values ?!?", ts_vals, kde_sf)

            ts_write[tag_ts] = ts_vals
            log_pv_write[tag_ts] = np.log10(kde_sf)

        # store the results in file
        sf = {
            "sf_ts": ts_write,
            "sf_log_pv": log_pv_write,
            "fit_pars": fit_pars,
        }

        save_postfix = self.get_meta_grp_postfix(
            calib_stage=calib_stage, grp_key=calib_fit_tag
        )

        self.mngr.data.save_calib_file(
            tag_dir=self.mngr.tag_calib_dir,
            tag_file=save_postfix,
            data=sf,
        )

        debug_kde_fit = False
        # debug_kde_fit = True
        if debug_kde_fit:
            eval_output = self.eval_p_values(
                df_calib=kde_score_bins,
                ts_tags=tag_ts,
                calib_stage=calib_stage,
                calib_fit_tag=calib_fit_tag,
            )
            fit_vals = eval_output["log_p_vals"]
            fit_vals = np.power(10, fit_vals)

            plt.figure(figsize=[4, 4])
            plt.plot(
                kde_score_bins,
                kde_prob_dens,
                color="orange",
                alpha=0.2,
                label="kde-prob-density",
            )
            if has_kde:
                kde_probs = kde_prob_dens * bandwidth
                plt.plot(
                    kde_score_bins,
                    kde_probs,
                    color="red",
                    alpha=0.2,
                    label="kde-prob",
                )
            plt.plot(
                kde_score_bins,
                kde_sf,
                color="green",
                alpha=0.2,
                linewidth=5,
                label="kde-sf",
            )
            plt.plot(
                kde_score_bins,
                fit_vals,
                color="blue",
                alpha=0.2,
                label="kde-sf-spline",
            )
            ax = plt.gca()
            ax.set_yscale("log")
            plt.legend(loc="upper right", fancybox=True, framealpha=0)
            plt.savefig("fig1.png", bbox_inches="tight", dpi=300)

        return

    def get_quants(self, data, quants, weights=None):
        """derive quantiles from an array of data

        Parameters
        ----------
        data : np.array
            array of input data
        quants : list of floats
            the quantiles to be calculated
        weights : np.array
            optional array of weights for each data element

        Returns
        -------
        type
            description
        """

        out = [0 for q in quants]

        try:
            if weights is None:
                out = np.quantile(data, quants)
            else:
                if not isinstance(weights, np.ndarray):
                    weights = np.asarray(weights, dtype=np.float32)

                if np.sum(weights, dtype=np.float32) > 0:
                    id_x = np.argsort(data)
                    data_sorted = data[id_x]

                    cdf = np.add.accumulate(weights[id_x], dtype=float)
                    cdf /= cdf[-1]

                    out = np.interp(quants, cdf, data_sorted).tolist()
        except Exception:
            self.log.info([["wr", " - cant get quantiles ... "], ["b", data]])

        return out

    def load_p_value_funcs(self, ts_tags, calib_fit_tag, calib_stage, is_verb=True):
        """setup functions that calculate p-values from their corresponding TS-values

        Parameters
        ----------
        ts_tags : list of str
            names of columns of TS-values
            (if not given, all available tags will be loaded)
        calib_fit_tag : tuple
            unique key (cut combination) for the input data
            (if given, only the corresponding function will be loaded)
        calib_stage : str
            tags for files to load p-value fit results
        is_verb : bool
            flag indicating is extra logging info should be done

        Returns
        -------
        dict of func
            the functions which have been loaded
        """

        meta_cut_keys = self.get_meta_cut_keys(calib_stage=calib_stage)

        # loop over all cut combinations
        for grp_key in meta_cut_keys:
            flat_key = self.get_flat_meta_grp(grp_key=grp_key, calib_stage=calib_stage)
            grp_name = self.get_meta_grp_name(flat_key)

            tag = self.get_meta_grp_postfix(calib_stage=calib_stage, grp_key=flat_key)

            if grp_name not in self.mngr.p_value_fits.keys():
                self.mngr.p_value_fits[grp_name] = dict()

            try:
                # if calib_fit_tag is given, exclude all others
                if (calib_fit_tag is not None) and (calib_fit_tag != flat_key):
                    continue

                try:
                    sf_now = self.mngr.data.load_calib_file(
                        tag_dir=self.mngr.tag_calib_dir,
                        tag_file=tag,
                        is_verb=False,
                        # is_verb=True,
                    )
                except CalibFileMissing:
                    continue
                except Exception as e:
                    raise e

                sf_ts = sf_now["sf_ts"]
                sf_log_pv = sf_now["sf_log_pv"]
                fit_pars = sf_now["fit_pars"]

                # if ts_tags is not given, load all available
                if ts_tags is None:
                    ts_tags = sf_ts.keys()

                # setup the functions given the information from the loaded file
                for tag_now in ts_tags:
                    if tag_now not in sf_ts.keys():
                        continue

                    if tag_now in self.mngr.p_value_fits[grp_name].keys():
                        continue

                    self.mngr.p_value_fits[grp_name][tag_now] = PValueFit(
                        log=self.log,
                        name=str_join(grp_name, tag_now),
                        fit_dist_type="log_linear",
                    )

                    # fill out of bounds values with fill_value=1
                    spline = interp1d(
                        sf_ts[tag_now],
                        sf_log_pv[tag_now],
                        kind="slinear",
                        bounds_error=False,
                        fill_value=0,
                    )
                    self.mngr.p_value_fits[grp_name][tag_now].set_spline(
                        spline=spline, max_ts=sf_ts[tag_now][-1]
                    )
                    self.mngr.p_value_fits[grp_name][tag_now].set_fit_pars(
                        fit_pars[tag_now]
                    )

            except Exception as e:
                raise e

    def check_load_p_value_funcs(
        self, ts_tags, calib_fit_tag, calib_stage, is_retry=False
    ):
        grp_name = self.get_meta_grp_name(calib_fit_tag)

        try:
            ts_tags_list = ts_tags
            if not isinstance(ts_tags, list):
                ts_tags_list = [ts_tags]

            if grp_name not in self.mngr.p_value_fits.keys():
                raise Exception

            any_missing = [
                t not in self.mngr.p_value_fits[grp_name].keys()
                for t in ts_tags_list
                if t is not None
            ]
            if any(any_missing):
                raise Exception

        except Exception:
            if is_retry:
                raise CalibFileMissing(
                    f" - missing calibration file for calib_fit_tag = {calib_fit_tag}, grp_name = {grp_name}"
                )

            self.load_p_value_funcs(
                ts_tags=ts_tags_list,
                calib_fit_tag=calib_fit_tag,
                calib_stage=calib_stage,
            )
            self.check_load_p_value_funcs(
                ts_tags=ts_tags,
                calib_fit_tag=calib_fit_tag,
                calib_stage=calib_stage,
                is_retry=True,
            )

    def try_load_calib(self, ts_tags, calib_stage, calib_fit_tag, n_evts, exps):
        try:
            self.check_load_p_value_funcs(
                ts_tags=ts_tags,
                calib_fit_tag=calib_fit_tag,
                calib_stage=calib_stage,
                is_retry=False,
            )
        except CalibFileMissing as e:
            exps += [
                {
                    "n_skipped": n_evts,
                    "exception": e,
                }
            ]
            self.log.warn([["r", e]])
            return False

        except Exception as e:
            raise e
        return True

    def eval_p_values(self, df_calib, ts_tags, calib_stage=None, calib_fit_tag=None):
        if not isinstance(ts_tags, list):
            ts_tags = [ts_tags]

        if calib_fit_tag is not None and isinstance(df_calib, np.ndarray):
            output = self.eval_p_values_single(
                df_calib=df_calib,
                ts_tags=ts_tags,
                calib_stage=calib_stage,
                calib_fit_tag=calib_fit_tag,
            )
        else:
            output = self.eval_p_values_df(
                df_calib=df_calib,
                ts_tags=ts_tags,
                calib_stage=calib_stage,
                calib_fit_tag=calib_fit_tag,
            )

        return output

    def eval_p_values_single(
        self, df_calib, ts_tags, calib_stage=None, calib_fit_tag=None
    ):
        log_p_vals, exceptions = [], []

        for tag_now in ts_tags:
            calib_stage = self.mngr.ts_to_calib_stage[tag_now]

            n_evts = len(df_calib)
            log_p_vals_now = np.ones(n_evts)
            grp_name = self.get_meta_grp_name(calib_fit_tag)

            has_loaded = self.try_load_calib(
                ts_tags=[tag_now],
                calib_fit_tag=calib_fit_tag,
                calib_stage=calib_stage,
                n_evts=n_evts,
                exps=exceptions,
            )
            if has_loaded:
                log_p_vals_now = self.mngr.p_value_fits[grp_name][
                    tag_now
                ].get_log_p_value(df_calib)
                log_p_vals += [log_p_vals_now]

            if len(ts_tags) == 1 and len(log_p_vals) > 0:
                log_p_vals = log_p_vals[0]

        output = {
            "log_p_vals": log_p_vals,
            "exceptions": exceptions,
        }
        return output

    def eval_p_values_df(self, df_calib, ts_tags, calib_stage=None, calib_fit_tag=None):
        log_p_val_tags, exceptions = [], []

        for tag_now in ts_tags:
            calib_stage = self.mngr.ts_to_calib_stage[tag_now]

            # loop over all groups and derive the new TS
            grps_0, keys_0 = self.get_flat_meta_groups(
                df=df_calib, calib_stage=calib_stage
            )

            for n_grp_0 in range(len(keys_0)):
                grp_key = keys_0[n_grp_0]
                indx = grps_0.get_group(name=grp_key).index
                df_grp = df_calib.loc[indx]

                grp_name = self.get_meta_grp_name(grp_key)

                has_loaded = self.try_load_calib(
                    ts_tags=[tag_now],
                    calib_fit_tag=grp_key,
                    calib_stage=calib_stage,
                    n_evts=len(indx),
                    exps=exceptions,
                )
                if not has_loaded:
                    continue

                p_vals_func = self.mngr.p_value_fits[grp_name][tag_now]

                col_name = self.mngr.ts_to_p_val_col[tag_now]
                log_p_val_tags += [col_name]

                df_calib.loc[indx, col_name] = p_vals_func.get_log_p_value(
                    df_grp[tag_now]
                )

        # log_p_vals.rename(columns=self.mngr.ts_to_p_val_col, inplace=True)
        output = {
            "df_calib": df_calib,
            "ts_tags": ts_tags,
            "log_p_val_tags": log_p_val_tags,
            "exceptions": exceptions,
        }
        return output

    def get_prev_calib_stage(self, calib_stage):
        try:
            calib_stage_index = self.mngr.calib_stages.index(calib_stage) - 1
            if calib_stage_index < 0:
                raise Exception("problem with index ?")
            prev_calib_stage = self.mngr.calib_stages[calib_stage_index]

        except Exception as e:
            msg = [
                f"""problem with calib_stage for get-prev-calib-stage({calib_stage})""",
                f"""[{','.join(self.mngr.calib_stages)}]\n""",
            ]
            raise Exception(" ".join(msg), e)

        # print('prev_calib_stageprev_calib_stageprev_calib_stage',prev_calib_stage)
        return prev_calib_stage

    def get_meta_cuts(self, df_in, meta_bins_dict, add_group_to_df, calib_stage):
        """
        Parameters
        ----------
        df_in
        meta_bins_dict
        add_group_to_df
        calib_stage

        Returns
        -------
        meta_cuts: list
            list with all meta cuts that are applied
        """

        if len(meta_bins_dict["meta_bin_cols"]) == 0:
            # inclusive group if no meta bins are defined
            meta_cuts = [
                pd.cut(
                    np.ones(len(df_in)).astype(int),
                    np.linspace(0, 2, 2),
                    include_lowest=True,
                    retbins=False,
                    labels=False,
                )
            ]

            return meta_cuts

        # group the data by metadata bins
        meta_cuts = []
        for meta_bin_key in meta_bins_dict["meta_bins"].keys():
            meta_bin_col = self.mngr.get_meta_bin_col_func(meta_bin_key, meta_bins_dict)

            if calib_stage is not None:
                if meta_bin_col not in self.mngr.meta_bin_calib_stage_cols[calib_stage]:
                    continue

            if self.mngr.prefix_meta_bins != "" and meta_bin_col.startswith(
                self.mngr.prefix_meta_bins
            ):
                # use result that was already stored.
                meta_cut = df_in[meta_bin_col]

            else:
                if meta_bins_dict["meta_types"][meta_bin_key] in ["linear"]:
                    meta_cut = pd.cut(
                        df_in[meta_bin_col],
                        meta_bins_dict["meta_bins"][meta_bin_key],
                        include_lowest=True,
                        retbins=False,
                        labels=False,
                    )

                elif meta_bins_dict["meta_types"][meta_bin_key] in ["fixed"]:
                    meta_cut = get_index_in_array(
                        np.array(meta_bins_dict["meta_bins"][meta_bin_key]),
                        df_in[meta_bin_col].to_numpy(),
                    )

                elif meta_bins_dict["meta_types"][meta_bin_key] in [
                    "range_longitude_deg"
                ]:
                    df_values = Longitude(
                        df_in[meta_bin_key].to_numpy() * u.rad
                    ).to_value("deg")
                    meta_cut = np.full(len(df_in), -1, dtype=float)

                    for n_bin, values in enumerate(
                        meta_bins_dict[meta_bin_key]["values"]
                    ):
                        values = Longitude(values * u.deg).to_value("deg")
                        if values[0] > values[1]:
                            # wrapping angle around 0
                            mask = (df_values > values[0]) | (df_values <= values[1])
                        else:
                            mask = (df_values > values[0]) & (df_values <= values[1])
                        meta_cut[mask] = n_bin
                    meta_cut[meta_cut < 0] = np.nan

                elif meta_bins_dict["meta_types"][meta_bin_key] in ["list"]:
                    meta_cut = np.full(len(df_in), -1, dtype=np.float)
                    for n_bin, values in enumerate(
                        meta_bins_dict["meta_bins"][meta_bin_key]
                    ):
                        mask = np.isin(df_in[meta_bin_col].to_numpy(), values)
                        meta_cut[mask] = n_bin
                    meta_cut[meta_cut < 0] = np.nan

                else:
                    NotImplementedError()

                if np.any(np.isnan(meta_cut)):
                    raise Exception(
                        " - should not have entries outside of the metadate selection"
                    )

                meta_cut = meta_cut.astype(int)

                if add_group_to_df:
                    df_in[meta_bin_key] = meta_cut

            meta_cuts.append(meta_cut)

        return meta_cuts

    def get_meta_groups(
        self, df_in, meta_bins_dict=None, add_group_to_df=False, calib_stage=None
    ):
        """
        group the data by their corresponding metadata bins

        Parameters
        ----------
        df_in: input dataframe
        meta_bins_dict: dict
            Optional give the `meta_bins_dict`. (e.g. required to deal with
            two sets of meta bins). If includes key `meta_col_dict`, will use
            this to map the meta columns to column names with its prefix.
        add_group_to_df: bool
            Add the meta group to df_in
        calib_stage : str
            tags for files to store p-value fit results

        Returns
        -------
        output: dict
            dictionary with the grouped input dataframe and the meta keys
        df_in: optional
            if add_group_to_df return the input data with the meta bins appended

        """
        meta_bins_dict = (
            self.mngr.meta_bins_dict if meta_bins_dict is None else meta_bins_dict
        )

        meta_cuts = self.get_meta_cuts(
            df_in=df_in,
            meta_bins_dict=meta_bins_dict,
            add_group_to_df=add_group_to_df,
            calib_stage=calib_stage,
        )

        pd_grps = df_in.groupby(meta_cuts)
        grp_keys = pd_grps.groups.keys()
        # print('count....'); print(pd_grps.agg('count').fillna(value=0)) # ; return
        # assert 0
        # print(pd_grps[metadata_col].agg('describe').fillna(value=0)) # ; return

        if self.mngr.do_meta_sanity_check:
            # sanity check, to make sure that self.mngr.meta_bins
            # covers all entries in the data
            for grp_key in grp_keys:
                if isinstance(grp_key, Iterable):
                    has_nan = any(x < 0 for x in grp_key)
                else:
                    has_nan = math.isnan(grp_key)
                if has_nan:
                    self.log.error(
                        [
                            ["r", " - problem with df_meta.groupby():\n"],
                            ["b", pd_grps.agg("count").fillna(value=0), "\n"],
                            ["g", grp_keys],
                        ]
                    )

                    raise Exception("problem with df_meta.groupby()")
        else:
            # get all keys of all groups that are valid meta bins
            grp_keys_meta = []
            for keys in grp_keys:
                if isinstance(keys, Iterable):
                    if not any([k < 0 for k in keys]):
                        grp_keys_meta.append(keys)
                elif not pd.isna(keys):
                    grp_keys_meta.append(keys)
            grp_keys = grp_keys_meta

        output = {
            "groups": pd_grps,
            "keys": grp_keys,
        }
        if add_group_to_df:
            return output, df_in
        else:
            return output

    def get_meta_cut_keys(self, calib_stage):
        """get short key representation of all unique combinations of cut variables

        Example:
        --------
        given self.mngr.meta_bins with 3 variables, having the following bins
        [[0, 1, 2], [10, 11], [20]]
        the unique bin combinations will be
        [(0, 10, 20), (0, 11, 20), (1, 10, 20), (1, 11, 20), (2, 10, 20), (2, 11, 20)]

        Returns
        -------
        list of tuples
            the list of group keys for filtering a dataframe
        """

        # if there are no meta bins, then there is only a single group with [key==0]
        if len(self.mngr.meta_bin_cols) == 0:
            return [0]

        meta_cut_keys = []
        for col in self.mngr.meta_bin_cols:
            if col not in self.mngr.meta_bin_calib_stage_cols[calib_stage]:
                continue

            if self.mngr.meta_types[col] in ("fixed", "list", "range_longitude_deg"):
                meta_cut_keys += [list(range(len(self.mngr.meta_bins[col])))]
            else:
                meta_cut_keys += [list(range(len(self.mngr.meta_bins[col]) - 1))]

        meta_cut_keys = list(itertools.product(*meta_cut_keys))

        # if there is only a single dimension, we need a list of
        # bins instead of the list of tuples of bins
        if len(self.mngr.meta_bin_cols) == 1:
            meta_cut_keys = [k[0] for k in meta_cut_keys]

        return meta_cut_keys

    def get_meta_grp_name(self, grp_key):
        """get the name of the cut bin

        Parameters
        ----------
        grp_key : tuple
            the bin definition (e.g., (0,3,1,0) for a 4-dim cut combination)
        """

        if isinstance(grp_key, Iterable):
            name = str_join("cuts", *grp_key)
        else:
            name = str_join("cuts", grp_key)

        return name

    def get_meta_grp_postfix(self, calib_stage, grp_key):
        meta_grp_postfix = str_join("p_value_fit", self.get_meta_grp_name(grp_key))
        if calib_stage is not None:
            meta_grp_postfix = str_join(calib_stage, meta_grp_postfix)
        return meta_grp_postfix

    def get_flat_meta_grp(self, grp_key, calib_stage):
        # construct a multi-index of the group-key, with an upper bound of max_bin_value

        if isinstance(grp_key, int):
            return grp_key

        dims = []
        for bin_name, bin_n in self.mngr.meta_n_bins.items():
            if bin_name in self.mngr.meta_bin_calib_stage_cols[calib_stage]:
                dims += [bin_n]

        flat_grp_key = np.ravel_multi_index(grp_key, dims=dims)

        return flat_grp_key

    def add_flat_meta_keys(self, df_calib):
        for calib_stage in self.mngr.calib_stages:
            flat_key_tag = self.get_flat_key_tag(calib_stage)

            meta_groups = self.get_meta_groups(df_calib, calib_stage=calib_stage)
            pd_grps = meta_groups["groups"]
            grp_keys = meta_groups["keys"]

            # loop over each bin in the cut dimensions
            for grp_key in grp_keys:
                # get the index of the selected events
                indx = pd_grps.get_group(name=grp_key).index

                flat_key = self.get_flat_meta_grp(
                    grp_key=grp_key, calib_stage=calib_stage
                )
                df_calib.loc[indx, flat_key_tag] = flat_key
                # print('-------', flat_key_tag, ' ------- ', flat_key, [indx[0:2], len(indx)])

    def get_flat_key_tag(self, calib_stage):
        return str_join(self.mngr.flat_meta_key_tag, calib_stage)

    def get_flat_meta_groups(self, df, calib_stage):
        # flat_key_tag = self.mngr.flat_meta_key_tag
        flat_key_tag = self.get_flat_key_tag(calib_stage)

        dims = []
        for bin_name, bin_n in self.mngr.meta_n_bins.items():
            if bin_name in self.mngr.meta_bin_calib_stage_cols[calib_stage]:
                dims += [bin_n]

        meta_bins = np.asarray(range(-1, np.prod(dims))) + 0.5

        meta_cuts = [
            pd.cut(
                df[flat_key_tag],
                meta_bins,
                include_lowest=True,
                retbins=False,
                labels=False,
            )
        ]

        grps_0 = df.groupby(meta_cuts)
        keys_0 = list(grps_0.groups.keys())

        return grps_0, keys_0

    def shuffle_internal_time_step_order(self, data_enc, data_dec):
        """shuffle data in the time-step axis
        where the time steps of the encoder and the time steps of the decoder
        are each independently shuffled for each complete time series

        Parameters
        ----------
        data_in : np.array
            array of input data, having shape
            (n_events, n_steps, n_features)

            example input:
            [[[  0   1   2   3]
            [ 10  11  12  13]
            [ 20  21  22  23]
            [ 30  31  32  33]]
            [[100 101 102 103]
            [110 111 112 113]
            [120 121 122 123]
            [130 131 132 133]]]

        Returns
        -------
        np.array
            shuffled array

            example output:
            [[[ 30  31  32  33]
            [ 20  21  22  23]
            [  0   1   2   3]
            [ 10  11  12  13]]
            [[120 121 122 123]
            [130 131 132 133]
            [110 111 112 113]
            [100 101 102 103]]]
        """

        def axis_shuffle(data, axis):
            idx = np.random.rand(*data.shape).argsort(axis=axis)
            output = {
                "data": np.take_along_axis(data, idx, axis=axis),
                "idx": idx,
            }
            return output

        def shuffle_data(data_in):
            n_evts = len(data_in)
            if n_evts == 0:
                return data_in
            n_steps = len(data_in[0])
            if n_steps == 0:
                return data_in
            n_feats = len(data_in[0][0])

            # generate a list of indices for slicing
            #  -> example: [ 0  1  2  3  4  5  6  7  8  9 10 11]
            indices = np.array(range(n_steps * n_evts))
            #  -> example: [[ 0  1  2] [ 3  4  5] [ 6  7  8] [ 9 10 11]]
            indices = indices.reshape(n_evts, n_steps)
            #  -> example: [[ 1  0  2] [ 3  5  4] [ 6  8  7] [11  9 10]]
            indices = axis_shuffle(indices, axis=1)["data"]
            #  -> example: [ 1  0  2  3  5  4  6  8  7 11  9 10]
            indices = indices.reshape(n_evts * n_steps)

            # randomise the time steps
            data_out = data_in.reshape(n_evts * n_steps, n_feats)
            data_out = data_out[indices]
            data_out = data_out.reshape(n_evts, n_steps, n_feats)

            return data_out

        out_enc = shuffle_data(data_enc)

        out_dec = data_dec
        if self.mngr.is_enc_dec_anom_det:
            out_dec = shuffle_data(data_dec)

        output = {self.mngr.tag_enc: out_enc, self.mngr.tag_dec: out_dec}

        return output

    def get_p_values_singel_steps(self, df_calib):
        """derive the p-values from individual steps

        Parameters
        ----------
        df_calib : pd.DataFrame
            input data including columns for TS-values and weights

        Returns
        -------
        dict
            the derived input p-values and the corresponding metadata
        """

        output = self.eval_p_values(
            df_calib=df_calib,
            ts_tags=self.mngr.df_cols_ts_0,
        )

        return output

    def get_ts_from_preds_enc_dec_anom_det(
        self, model_inputs, model_outputs, *args, **kwargs
    ):
        """derive the anomaly detection test statistic for a particular
        cut group for anomaly detection

         Parameters
         ----------
         model_outputs : dict
             the validation output of the network

         Returns
         -------
         dict
             the derived TS, in addition to the (averaged) inputs
        """

        output_dec = model_inputs[self.mngr.tag_dec]
        output_wgt = model_inputs[self.mngr.tag_wgt]
        output_pred = model_outputs[self.mngr.tag_pred]

        if self.mngr.avg_time_steps:
            output_dec = output_dec[:, :, self.mngr.data_col_indices]
            output_wgt = output_wgt[:, :, self.mngr.data_col_indices]

            output_dec = output_dec * output_wgt
            output_wgt = np.sum(output_wgt, axis=1)
            output_dec = safe_np_div(np.sum(output_dec, axis=1), output_wgt)

            output_pred = output_pred[:, self.mngr.data_col_indices]
        else:
            output_dec = output_dec[:, :, self.mngr.data_col_indices]
            output_pred = output_pred[:, :, self.mngr.data_col_indices]
            output_wgt = output_wgt[:, :, self.mngr.data_col_indices]

        if self.mngr.do_train_scale:
            norms = self.mngr.data.train_feat_norms[self.mngr.tag_steps]
            shift = norms["shift"][0, self.mngr.data_col_indices]
            scale = norms["scale"][0, self.mngr.data_col_indices]

            output_dec = (output_dec / scale) - shift
            output_pred = (output_pred / scale) - shift

        # print('---'*35)
        # print(' -- decs  --\n',output_dec,'\n'+'---'*35)
        # print(' -- pred --\n',output_pred,'\n'+'-+-'*35)
        # print(' -- wgts --\n',output_wgt,'\n'+'-+-'*35)
        # print('-+-'*35)

        # mask = output_pred == np.inf
        # if len(output_pred[mask]) > 0:
        #     print(' ????????? 0 ', len(output_pred[mask]))
        # output_pred[mask] = 1000

        ts_wgts = self.mngr.ts_weights_enc_dec_anom_det["func"]()
        ts_wgts = ts_wgts[np.newaxis, :]

        output_dec = np.multiply(output_dec, ts_wgts)
        output_pred = np.multiply(output_pred, ts_wgts)
        output_wgt = np.multiply(output_wgt, ts_wgts)

        delta_preds = np.sqrt(np.abs(output_pred) + 1)
        # mask = delta_preds == np.inf
        # if len(delta_preds[mask]) > 0:
        #     print(' ????????? 1 ', len(delta_preds[mask]))
        # delta_preds[mask] = 1

        output_ts = (output_dec - output_pred) / delta_preds
        # mask = output_ts == np.inf
        # if len(output_ts[mask]) > 0:
        #     print(' ????????? 2 ', len(output_ts[mask]))
        # output_ts[mask] = 10

        # events which fail the cuts are mapped to downward fluctuations
        # (around -delta_preds) and thus will pile-up at p-values close to unity
        evt_ts_mask = self.get_evt_ts_mask(
            dec_vals=output_dec,
            dec_wgts=output_wgt,
        )
        if evt_ts_mask is not None:
            output_ts[~evt_ts_mask] = -1

        # print(' -- delta_preds --\n',delta_preds,'\n'+'-+-'*35)
        # print(' -- 0-output_ts   --\n',output_ts,'\n'+'-+-'*35)
        # assert 0

        if not self.mngr.avg_time_steps:
            if self.mngr.sum_evt_lc_ts_func is not None:
                # if there is a specialised summation function
                output = self.mngr.sum_evt_lc_ts_func(
                    dec=output_dec,
                    pred=output_pred,
                    ts=output_ts,
                )
                output_dec = output[self.mngr.tag_dec]
                output_pred = output[self.mngr.tag_pred]
                output_ts = output[self.mngr.tag_ts]
            else:
                # simple weighted sum
                sum_wgts = np.sum(output_wgt, axis=1)
                # print(' -- sum_wgts --\n',sum_wgts,'\n'+'-+-'*35)

                # average with weights over time steps for each feature
                output_dec = safe_np_div(
                    np.sum((output_dec * output_wgt), axis=1),
                    sum_wgts,
                )
                output_pred = safe_np_div(
                    np.sum((output_pred * output_wgt), axis=1),
                    sum_wgts,
                )
                output_ts = safe_np_div(
                    np.sum((output_ts * output_wgt), axis=1),
                    sum_wgts,
                )

                output_wgt = sum_wgts

            # print(' -- 1-output_ts   --\n',output_ts,'\n'+'-+-'*35)

        output = {
            self.mngr.tag_dec: output_dec,
            self.mngr.tag_pred: output_pred,
            self.mngr.tag_wgt: output_wgt,
            self.mngr.tag_ts: output_ts,
        }
        if evt_ts_mask is not None:
            output[self.mngr.tag_evt_ts_mask] = evt_ts_mask.astype(bool)

        return output

    def get_ts_from_preds_enc_dec_clasif(
        self, model_inputs, model_outputs, *args, **kwargs
    ):
        """derive the classification test statistic for a particular
        cut group for classification

         Parameters
         ----------
         model_outputs : dict
             the validation output of the network

         Returns
         -------
         dict
             the derived TS, in addition to the (averaged) inputs
        """

        # scale/shift factors for TS values (in order to avoid log(0) errors)
        ts_scale = 10
        ts_log_shift = np.power(10.0, -ts_scale)

        output_dec = model_inputs[self.mngr.tag_dec]
        output_wgt = model_inputs[self.mngr.tag_wgt]
        output_pred = model_outputs[self.mngr.tag_pred]

        output_dec = output_dec[:, :, self.mngr.data_col_indices]

        if self.mngr.do_train_scale:
            norms = self.mngr.data.train_feat_norms[self.mngr.tag_steps]
            shift = norms["shift"][0, self.mngr.data_col_indices]
            scale = norms["scale"][0, self.mngr.data_col_indices]

            output_dec = (output_dec / scale) - shift

        # events which fail the cuts are mapped to downward fluctuations
        # (around -delta_preds) and thus will pile-up at p-values close to unity
        evt_ts_mask = self.get_evt_ts_mask(dec_vals=output_dec)

        pred_bck = output_pred[:, self.mngr.bck_cats]
        pred_bck = np.sum(pred_bck, axis=1)

        # just in case, though this should never happen ...
        pred_bck[pred_bck < 0] = 0
        pred_bck[pred_bck > 1] = 1

        # map the probabilities to the TS range [0-->1], where TS=0
        # is for sum( all background probabilities ) = 1
        output_ts = -np.log10(pred_bck + ts_log_shift) / ts_scale

        # add insignificant random noise in order to stabilize the kde fit later
        # output_ts += np.random.normal(0, 1e-3, len(output_ts))

        if evt_ts_mask is not None:
            output_ts[~evt_ts_mask] = 0

        output = {
            self.mngr.tag_wgt: output_wgt,
            self.mngr.tag_ts: output_ts,
        }
        if evt_ts_mask is not None:
            output[self.mngr.tag_evt_ts_mask] = evt_ts_mask.astype(bool)

        return output

    def derive_calib_single_step_combo(self, calib_stage, calib_stage_input):
        # derive the combined ts and fit the corresponding p-values
        self.log.info([["y", f"derive-calib-single-step-combo({calib_stage}) "]])

        calib_tags = self.mngr.get_calib_tags(calib_stage=calib_stage)
        tag_ts_step_combo = calib_tags["ts"]
        tag_wgt_combo = calib_tags["wgt"]

        df_calib = calib_stage_input["df_calib"]
        df_preds = calib_stage_input["df_preds"]

        # df_calib = df_calib.sort_values('n_realization')##.reset_index(drop=True)

        self.get_ts_step_combo(
            df_calib=df_calib,
            calib_stage=calib_stage,
        )

        # loop over all groups and derive the p-value fits
        self.fit_p_values_by_flat_grp(
            df=df_calib,
            calib_stage=calib_stage,
            ts_tags=[tag_ts_step_combo],
            fit_func=self.fit_p_values_log_linear,
        )

        output = {
            "df_calib": df_calib,
            "df_preds": df_preds,
        }

        return output

    def get_p_values_step_combo(self, df_calib, calib_stage):
        """derive the p-values from the combination of individual steps

        Parameters
        ----------
        data : pd.DataFrame
            input data including columns for TS-values and weights
        calib_stage : str
            tags for files to store p-value fit results

        Returns
        -------
        dict
            the derived input p-values and the corresponding metadata
        """

        tag_ts_step_combo = self.mngr.get_calib_tags(calib_stage=calib_stage, tag="ts")

        if tag_ts_step_combo not in df_calib:
            self.get_ts_step_combo(
                df_calib=df_calib,
                calib_stage=calib_stage,
            )

        eval_output = self.eval_p_values(
            df_calib=df_calib,
            ts_tags=[tag_ts_step_combo],
        )

        return eval_output
