from __future__ import annotations

from typing import TYPE_CHECKING

from tensorflow import keras

from trans_finder.flares.helpers_debugging import (
    denormalize_model_output,
    write_np_ndarray_to_csv,
)

if TYPE_CHECKING:
    from trans_finder.flares.manager import Manager


def write_model_outputs__and_errors_to_csv_for_debugging(
    mngr: Manager, model_outputs, data_desc=None
):
    if data_desc is None:
        data_desc = mngr.run_phase

    # write denormalized autoencoder input, output and error to csv
    auto_enc_input = model_outputs[mngr.tag_auto_enc_input]
    auto_enc_reconstruction = model_outputs[mngr.tag_auto_enc_reconstruction]
    denormed_input = denormalize_model_output(mngr, auto_enc_input)
    denormed_output = denormalize_model_output(mngr, auto_enc_reconstruction)

    write_np_ndarray_to_csv(
        mngr,
        f"{data_desc}_auto_enc_denormed_error",
        denormed_output - denormed_input,
    )

    write_np_ndarray_to_csv(
        mngr,
        f"{data_desc}_auto_enc_denormed_input",
        denormed_input,
    )
    write_np_ndarray_to_csv(
        mngr,
        f"{data_desc}_auto_enc_denormed_output",
        denormed_output,
    )

    # write denormalized forecasting error to csv
    d1 = model_outputs[mngr.tag_input_d1]
    bck_forecast = model_outputs[mngr.tag_bck_forecast_d1]
    denormed_forecast_input = denormalize_model_output(mngr, d1)
    denormed_forecast_output = denormalize_model_output(mngr, bck_forecast)

    write_np_ndarray_to_csv(
        mngr,
        f"{data_desc}_bck_forecast_denormed_error",
        denormed_forecast_output - denormed_forecast_input,
    )

    # write all model ouputs to csv
    for key, value in model_outputs.items():
        write_np_ndarray_to_csv(mngr, f"{data_desc}_{key}", value)
        pass

    # # write mcmc gauss mix state to csv, if exists
    # try:
    #     cluster_probs, loc, chol_precision = mngr.train.load_model_file(
    #         "mcmc_gauss_mix_state"
    #     )
    #     write_np_ndarray_to_csv(
    #         mngr,
    #         f"{data_desc}_model_mcmc_state_cluster_probs",
    #         cluster_probs.numpy(),
    #     )
    #     write_np_ndarray_to_csv(
    #         mngr,
    #         f"{data_desc}_model_mcmc_state_chol_precision",
    #         chol_precision.numpy(),
    #     )
    #     write_np_ndarray_to_csv(
    #         mngr, f"{data_desc}_model_mcmc_state_cluster_loc", loc
    #     )
    # except Exception as e:
    #     pass


class WriteModelOutputsToCSVCallback(keras.callbacks.Callback):
    """Writes the model ouputs of the first batch of training data to csv files"""

    def __init__(self, mngr, dataset_train, dataset_eval):
        self.dataset_train = dataset_train
        self.dataset_eval = dataset_eval
        self.mngr: Manager = mngr
        self.is_first_epoch = True

    def write_model_outputs(self, data_desc):
        model_outputs = self.model.predict(list(self.dataset_train)[0][0])

        write_model_outputs__and_errors_to_csv_for_debugging(
            self.mngr,
            model_outputs,
            data_desc,
        )

    def on_epoch_end(self, epoch, logs=None):
        self.write_model_outputs(f"e{epoch}_train")
        pass

    def on_epoch_begin(self, epoch, logs=None):
        if self.is_first_epoch:
            data_desc = "e_initial" if epoch == 0 else f"e{epoch - 1}_train"
            self.write_model_outputs(data_desc)
            self.is_first_epoch = False
