import os.path

import numpy as np
from matplotlib import pyplot as plt
from tensorflow.estimator import ModeKeys

from trans_finder.flares.checks.generic_checks import calculate_dist_overlap
from trans_finder.flares.helpers_debugging import get_debug_output_filepath
from trans_finder.flares.plot_list.util.plot_config import gridlines_config, tick_params
from trans_finder.flares.state_manager import StateManager
from trans_finder.flares.utilities.pdf_plot_collection import PDFPlotCollection

stat_names = ["median", "std"]
comparison_tags = ["forecast_input", "mixture_model_input", "bck_no_plateaus"]


def calculate_stats_deduplicated(mngr, pd_vals):
    # remove subsequent identical slices (appear due to a more dense other channel, while no new data
    # is available in current channel )

    stats = {}

    for channel_tag in mngr.data_prep_channel_tags:
        #  mask all network features for current channel
        channel_mask = [
            idx
            for idx, (ft_channel_tag, ft_name) in enumerate(mngr.data_step_col_names)
            if ft_channel_tag == channel_tag
        ]

        # mask all network features for current channel except temporal offsets
        ft_mask = [
            idx
            for idx, (ft_channel_tag, ft_name) in enumerate(mngr.data_step_col_names)
            if ft_channel_tag == channel_tag and ft_name != mngr.tag_temporal_offset
        ]

        channel_pd_vals = pd_vals[..., channel_mask]
        ft_pd_vals = pd_vals[..., ft_mask]

        # calculate difference between subsequent slices on the feature values only (e.g. fluxes)
        diff = np.diff(ft_pd_vals.reshape(len(ft_pd_vals), -1), axis=0)

        # absolute element-wise sum of difference between subsequent slices
        diff = np.absolute(diff).sum(axis=1)

        # keep the first slice, and from the second onwards, only those which changed compared to the predecessor
        channel_pd_vals = np.concatenate(
            [channel_pd_vals[[0]], channel_pd_vals[1:][diff.nonzero()]]
        )

        channel_stats = calculate_stats(channel_pd_vals)
        stats[channel_tag] = channel_stats

    return stats


def calculate_stats(pd_vals):
    medians = np.median(pd_vals, axis=1)
    stds = np.std(pd_vals, axis=1)

    return [medians, stds]


def calculate_stats_from_channel_lightcurve(mngr, pd_vals, input_df_meta):
    raise NotImplementedError()


def save_descriptive_stats(mngr, tag, input_df_meta, pd_vals):
    # evt_classes = input_df_meta.evt_class
    # unique_evt_classes = evt_classes.unique()
    # assert len(unique_evt_classes) == 1 and unique_evt_classes[0] == "bck_pure"

    # calculate histograms of mean / std per feature

    # create histogram of original input light curve
    stats_deduplicated_sclices = calculate_stats_deduplicated(mngr, pd_vals)
    stats_all_slices = calculate_stats(pd_vals)

    stats = [stats_all_slices, stats_deduplicated_sclices]

    # stats = calculate_stats(pd_vals)
    # stats = calculate_stats_from_channel_lightcurve(mngr, pd_vals, input_df_meta)
    StateManager(mngr, f"{tag}_descriptive_stats").save_state(stats)


def do_checks_on_background_network_inputs(mngr, mode, pd_vals, input_df_meta):
    if mode == ModeKeys.TRAIN and mngr.run_phase == "train" and mngr.train_phase == 0:
        tag = "forecast_input"
        save_descriptive_stats(mngr, tag, input_df_meta, pd_vals)

    elif mode == ModeKeys.TRAIN and mngr.run_phase == "bgmm_fit":
        tag = "mixture_model_input"
        save_descriptive_stats(mngr, tag, input_df_meta, pd_vals)

    elif (
        mode == ModeKeys.PREDICT
        and mngr.run_phase == "predict"
        and mngr.dir_input_predict == mngr.tag_custom_realisations
    ):
        selection = pd_vals[input_df_meta.realisation_id == "bck_no_plateaus"]
        save_descriptive_stats(mngr, "bck_no_plateaus", input_df_meta, selection)

        ref_pd_vals = pd_vals[input_df_meta.realisation_id == "bck_no_shuffling"]

        do_comparisons(mngr, ref_pd_vals)
    else:
        return


def do_comparisons(mngr, ref_pd_vals):
    reference_stats_all = calculate_stats(ref_pd_vals)
    reference_stats_deduplicated = calculate_stats_deduplicated(mngr, ref_pd_vals)

    for comparison_name in comparison_tags:
        [stats_all_slices, stats_deduplicated_sclices] = StateManager(
            mngr, f"{comparison_name}_descriptive_stats"
        ).load_state()

        calculate_stat_dist_overlap(
            mngr, comparison_name, reference_stats_all, stats_all_slices
        )

        for isDensity, subdir_name in [
            (False, os.path.join("counts", "all_slices")),
            (True, os.path.join("density", "all_slices")),
        ]:
            create_comparison_plot(
                mngr,
                subdir_name=subdir_name,
                reference_stats=reference_stats_all,
                stats=stats_all_slices,
                comparison_name=comparison_name,
                density=isDensity,
            )

        for isDensity, subdir_name in [
            (False, os.path.join("counts", "deduplicated_slices")),
            (True, os.path.join("density", "deduplicated_slices")),
        ]:
            create_comparison_plot_deduplicated(
                mngr,
                subdir_name=subdir_name,
                reference_stats=reference_stats_deduplicated,
                stats=stats_deduplicated_sclices,
                comparison_name=comparison_name,
                density=isDensity,
            )


def calculate_stat_dist_overlap(
    mngr, comparison_name, reference_stats_all, stats_all_slices
):
    for stat_name, ref_single_stat, comp_single_stat in zip(
        stat_names, reference_stats_all, stats_all_slices
    ):
        dist_overlap = calculate_dist_overlap(
            ref_single_stat, comp_single_stat, bins=50
        )

        fit_error_string = np.array2string(
            dist_overlap, formatter={"float_kind": lambda x: "%.2f" % x}
        )
        mngr.log.info(
            [
                [
                    "b",
                    f"{comparison_name} vs. original background: {stat_name} dist overlap is {np.mean(dist_overlap):.2%} | separately per feature: {fit_error_string}",
                ],
            ]
        )


def get_stats(mngr, comparison_name, slices_tag):
    [stats_all_slices, stats_deduplicated_sclices] = StateManager(
        mngr, f"{comparison_name}_descriptive_stats"
    ).load_state()

    if slices_tag == "all_slices":
        return stats_all_slices
    elif slices_tag == "deduplicated_slices":
        return stats_deduplicated_sclices
    else:
        raise ValueError(f"Unknown {slices_tag=}")


def create_comparison_plot(
    mngr, subdir_name, reference_stats, comparison_name, stats, density=False
):
    # custom realisations, get the background_no_shuffling input

    n_stats = len(stat_names)
    n_features = reference_stats[0].shape[1]

    fig, axes = plt.subplots(
        nrows=n_features,
        ncols=n_stats,
        figsize=(5 * n_stats, 3.5 * n_features),
        constrained_layout=True,
    )

    for network_ft_idx, (channel_tag, ft_name) in enumerate(mngr.data_step_col_names):
        # if ftname == mngr.tag_temporal_offset:
        #     continue

        selected_stats = [s[..., network_ft_idx] for s in stats]
        selected_reference_stats = [s[..., network_ft_idx] for s in reference_stats]

        plot_stat(
            axes,
            network_ft_idx,
            ft_name,
            channel_tag,
            comparison_name,
            selected_reference_stats,
            selected_stats,
            density,
        )

    PDFPlotCollection(mngr, figs=[fig]).save_to_file(
        get_debug_output_filepath(
            mngr,
            os.path.join("background", subdir_name),
            f"{comparison_name}.pdf",
        )
    )


def create_comparison_plot_deduplicated(
    mngr, reference_stats, subdir_name, comparison_name, stats, density=False
):
    # custom realisations, get the background_no_shuffling input

    n_stats = len(stat_names)
    n_features = len(mngr.data_step_col_names)

    fig, axes = plt.subplots(
        nrows=n_features,
        ncols=n_stats,
        figsize=(5 * n_stats, 3.5 * n_features),
        constrained_layout=True,
    )

    for channel_tag in stats:
        stat_ft_index = -1
        for network_ft_idx, (ft_channel_tag, ft_name) in enumerate(
            mngr.data_step_col_names
        ):
            if ft_channel_tag != channel_tag:
                continue

            stat_ft_index += 1
            selected_stats = [s[..., stat_ft_index] for s in stats[channel_tag]]
            selected_reference_stats = [
                s[..., stat_ft_index] for s in reference_stats[channel_tag]
            ]

            plot_stat(
                axes,
                network_ft_idx,
                ft_name,
                channel_tag,
                comparison_name,
                selected_reference_stats,
                selected_stats,
                density,
            )

    PDFPlotCollection(mngr, figs=[fig]).save_to_file(
        get_debug_output_filepath(
            mngr,
            os.path.join("background", subdir_name),
            f"{comparison_name}.pdf",
        )
    )


def plot_stat(
    axes,
    network_ft_idx,
    ft_name,
    channel_tag,
    comparison_name,
    selected_reference_stats,
    selected_stats,
    density,
):
    for stat_idx, (name, reference, undertest) in enumerate(
        zip(stat_names, selected_reference_stats, selected_stats)
    ):
        # test_result = kstest(undertest, reference)
        # null_holds = test_result.pvalue >= alpha
        # are_equivalent = null_holds

        # print(channel, ftname, tag, name, are_equivalent)

        ax = axes[network_ft_idx, stat_idx]

        bins = np.linspace(
            min(reference.min(), undertest.min()),
            max(reference.max(), undertest.max()),
            30,
        )
        ax.grid(**gridlines_config)

        ax.set_xlabel(
            f"{name}({channel_tag}_{ft_name})"  # - {'equal' if are_equivalent else 'different'}"
        )
        ax.tick_params(**tick_params)
        ax.hist(
            [reference, undertest],
            bins,
            label=["bck_no_shuffling", f"{comparison_name}"],
            density=density,
        )
        ax.get_yaxis().set_visible(False)
        ax.legend(loc="upper right")
