from pathlib import Path

import joblib
import numpy as np
from matplotlib import pyplot as plt

from trans_finder.flares.plot_list.util.plot_config import gridlines_config, tick_params
from trans_finder.flares.utilities.pdf_plot_collection import PDFPlotCollection


def check_in_range(mngr, data, interval, desc: str):
    """counts data points outside of given interval and warns if any exist."""

    n_total = np.product(np.array(data.shape))
    n_below_min = (data < interval[0]).sum()
    n_above_max = (data > interval[1]).sum()
    n_out_of_range = n_below_min + n_above_max

    percentage_above_max = n_above_max / n_total
    percentage_below_min = n_below_min / n_total
    percentage_out_of_range = n_out_of_range / n_total

    point_max = np.max(data)
    point_min = np.min(data)

    warning_strings = []
    if n_below_min > 0:
        warning_strings.append(
            f"{n_below_min} values ({percentage_below_min:.2%}) below {interval[0]} (min={point_min})"
        )
    if n_above_max > 0:
        warning_strings.append(
            f"{n_above_max} values ({percentage_above_max:.2%}) above {interval[1]} (max={point_max})"
        )

    if n_out_of_range > 0:
        mngr.log.warn(
            f" - Check range({desc}): {percentage_out_of_range:.2%} out of range | {warning_strings[0] if len(warning_strings) == 1 else ' | '.join(warning_strings)}"
        )
        return n_below_min + n_above_max
    else:
        return 0


def calculate_dist_overlap(
    a,
    b,
    bins="auto",
):
    """
    Args:
        a: 2D array (n_samples x n_dims )
        b: 2D array (n_samples x n_dims )
        bin_width: (optional) width of each bin, is prioritized over number of bins
        bins: number of bins

    Returns: the percentage of overlap per dimension as 1D numpy array (n_dim)

    """
    assert a.shape[1:] == b.shape[1:]
    assert len(a.shape) == 2
    assert a.shape[1] == b.shape[1]
    n_dims = a.shape[1]

    # precondition: overlap of numerical ranges should be there (e.g. >1%),
    # otherwise calculating common bins may not be robust

    union_width = np.min([a.max(axis=0), b.max(axis=0)], axis=0) - np.max(
        [a.min(axis=0), b.min(axis=0)], axis=0
    )
    intersection_width = np.max([a.max(axis=0), b.max(axis=0)], axis=0) - np.min(
        [a.min(axis=0), b.min(axis=0)], axis=0
    )
    numerical_range_overlap = np.divide(intersection_width, union_width)

    if (numerical_range_overlap < 0.01).any():
        # 0-1% overlap in numerical range, return just that overlap as an upper bound for the distribution overlap
        return numerical_range_overlap

    # if bins == "blocks":

    # else:
    #     # get min and max along first axis and calculate bin edges
    #     bin_min = np.min([a.min(axis=0), b.min(axis=0)], axis=0)
    #     bin_max = np.max([a.max(axis=0), b.max(axis=0)], axis=0)
    #
    #     bin_edges = np.linspace(bin_min, bin_max, num=int(bins))

    dim_overlap = np.zeros(n_dims)

    for i_dim in range(n_dims):
        bin_edges = np.histogram_bin_edges(
            np.concatenate([a[:, i_dim], b[:, i_dim]]), bins=bins
        )
        # create histogram of each dimension
        hist_a, _ = np.histogram(a[:, i_dim], bin_edges)
        normed_hist_a = hist_a / np.sum(hist_a)

        hist_b, _ = np.histogram(b[:, i_dim], bin_edges)
        normed_hist_b = hist_b / np.sum(hist_b)

        # calculate overlapping area of normed histograms
        dim_overlap[i_dim] = np.sum(np.min([normed_hist_a, normed_hist_b], axis=0))

    return dim_overlap


def describe_overlap(overlaps, desc: str, dim_name: str = "dim"):
    fit_error_string = np.array2string(
        np.array(overlaps),
        formatter={"float_kind": lambda x: "%.2f" % x},
    )

    return f"{desc} dist overlap = {np.mean(overlaps):.2%} | per {dim_name}: {fit_error_string}"


def temp_save(data, name):
    Path("temp").mkdir(parents=True, exist_ok=True)
    joblib.dump(data, f"temp/{name}.joblib")


def temp_load(name):
    return joblib.load(f"temp/{name}.joblib")


def get_hash_string(obj):
    import hashlib

    return hashlib.md5("".join(obj).encode("utf-8")).hexdigest()


def plot_dim_hists_of_ndarrays(
    mngr, ndarrays, labels, out_path, subplots_kwargs=None, hist_kwargs=None
):
    # assert same shape, except for first dim which are the number of samples
    if hist_kwargs is None:
        hist_kwargs = {}
    if subplots_kwargs is None:
        subplots_kwargs = {}

    flattened = ndarrays

    # add row dim if there is no row dim
    if len(flattened[0].shape) == 1:
        flattened = [np.expand_dims(a, -1) for a in flattened]

    assert all(a.shape[1:] == ndarrays[0].shape[1:] for a in ndarrays)

    # add col dim if there is no col dim
    if len(flattened[0].shape) == 2:
        flattened = [np.expand_dims(a, -1) for a in flattened]

    fig, axes = plt.subplots(
        nrows=flattened[0].shape[1],
        figsize=(max(8, flattened[0].shape[2] * 4), max(8, flattened[0].shape[1] * 4)),
        ncols=flattened[0].shape[2],
        squeeze=False,
        **subplots_kwargs,
    )
    for row_dim in range(flattened[0].shape[1]):
        for col_dim in range(flattened[0].shape[2]):
            ax = axes[row_dim, col_dim]
            axes[row_dim, col_dim].hist(
                [a[:, row_dim, col_dim] for a in flattened],
                bins="auto",
                histtype="step",
                label=labels,
                **hist_kwargs,
            )
            ax.grid(**gridlines_config)
            ax.tick_params(**tick_params)
            # ax.set_xlabel("ts: negative log prob")
            ax.set_ylabel(
                "Density"
                if "density" in hist_kwargs and hist_kwargs["density"] == True
                else "Counts"
            )
            ax.legend()
    PDFPlotCollection(mngr, figs=[fig]).save_to_file(out_path)
