from __future__ import annotations

from typing import Any, Dict, List, Literal, Optional

from pydantic import (
    BaseModel,
    NonNegativeFloat,
    NonNegativeInt,
    PositiveFloat,
    PositiveInt,
    confloat,
    conint,
    conlist,
)

from trans_finder.flares.core.config_components import (
    FlareTemplateConfig,
    FluctuationPreset,
    LongTermStatePreset,
)
from trans_finder.flares.core.constants import TAG_EVT_TYPE
from trans_finder.flares.core.types import (
    CalibStage,
    EventClass,
    MinMaxStr,
    RunPhase,
    TrainPhase,
    sized_list,
    to_list,
)


class OptionalConfigParams(
    BaseModel,
    protected_namespaces=(),  # disabling the model protected namespace of pydantic due to model_ attributes
    validate_default=True,
):
    # maximal number of events to load for different run types
    max_evt: Dict[RunPhase, int] = {k: -1 for k in to_list(RunPhase)}

    # fraction of evaluation events from the entire training
    # sample (used only during training)
    frac_train_eval: float = 0.8

    # threshold (epoch number) to start decreasing the learning rate for each phase
    # (this overrides learn_rate_epoch_thresh)
    learn_rate_epoch_thresh_phases: Dict[TrainPhase, int] = {
        k: 1000 for k in to_list(TrainPhase)
    }

    use_tf_prob: bool = True

    # configure the naming of metadata features (i.e. features with a single static value per light curve)
    meta_step_cols: list = []

    add_meta_bin_to_steps: bool = False

    batching_do_drop_remainder: bool = False

    # to enable/avoid using the reconstruction error as a TS
    use_reco_err_ts_combo: bool = True

    # the consecutive TS-calibration stages performed during the "calib" run_phase
    calib_stages: List[CalibStage] = ["calib_cluster_single", "calib_step_combo"]

    # lists of meta-data columns for each calibration stage (must be sub-sets of meta_bin_col_params)
    meta_bin_calib_stage_cols: Dict[str, List[str]] = {
        "calib_cluster_single": [TAG_EVT_TYPE],
        "calib_step_combo": [TAG_EVT_TYPE],
    }

    # using the entire length of the dataset as a buffer size to shuffle before batching the data,
    # or the current default value, which is 100 * batch_size
    # details: https://www.tensorflow.org/api_docs/python/tf/data/Dataset#shuffle
    shuffle_dataset_use_dataset_length_as_buffer_size: bool = True

    # A flag to set the shuffle parameter of keras model fit function
    # details: https://www.tensorflow.org/api_docs/python/tf/keras/Model#fit
    model_fit_do_shuffle_data_on_train_epoch_begin: bool = True

    # Use keras ModelCheckpointCallback and save tensorflow model weights only if val_loss has improved
    model_save_best_weights_only: bool = True

    # number of gaussian components for the mixture model
    model_gaus_mix_n_comps: int = 10

    # minimum number of tries before the mcmc mixture model fit will stop increasing
    # the complexity and run again in case the quality of fit has not improved
    mcmc_dynamic_fit_n_min_tries: PositiveInt = 2

    model_use_offset_weights: bool = True

    # >= 0, small value makes the decay of old signals slower, large values faster. 1 = linear decay
    # retrain the model when chaning this parameter
    model_offset_weight_power: confloat(ge=0) = 1.0

    # randomly uniform sample from this range during reconstruction training
    model_offset_weight_power_range: List[confloat(ge=0)] = [1, 2]

    model_n_samples_of_input_errors: conint(ge=0) = 20

    train_batch_size: PositiveInt = 128
    predict_batch_size: PositiveInt = 100000
    calib_batch_size: PositiveInt = 100000

    aim_track_val_metrics_every_n_steps: conint(ge=0) = 0

    # number of allowed parallel iterations for the fit
    model_hmc_max_parallel_iterations: PositiveInt = 32

    # number of randomized chains to run in parallel
    model_hmc_n_parallel_chains: PositiveInt = 3

    # number of steps to sample from the chain using Hamiltonian Monte Carlo
    model_hmc_n_steps: PositiveInt = 200

    model_hmc_input_normalisation_range: sized_list(float, 2) = [-1, 1]

    # number of hidden units in the encoder-decoder and in the forecasting LSTMs
    lstm_enc_sizes: conlist(PositiveInt, min_length=1) = [64, 32]
    lstm_dec_sizes: conlist(PositiveInt, min_length=1) = [32, 64]

    autoenc_enc_sizes: conlist(PositiveInt, min_length=1) = [256, 128]
    autoenc_dec_sizes: conlist(PositiveInt, min_length=1) = [128, 256]

    # flag to determine if we write out extra diagnostics info and corresponding plots
    do_diagnostics: Dict[RunPhase, bool] = {"calib": False, "predict": True}

    evt_type_ratios: Dict[str, Dict[EventClass, float]] = {
        "input_norm": {"bck_pure": 1},
        "train": {k: -1 for k in to_list(EventClass)},
        "bgmm_fit": {"bck_pure": 1},
        "calib": {"bck_pure": 1},
        "predict": {},
    }

    # model_cluster_space_regularisation_factor: 1.0
    model_cluster_space_regularisation_factor: float = 0.001

    # setting the number of latent dimensions (cluster space) equivalent to
    # this compression factor relative to the input dimensionality
    model_cluster_space_dimensions_compression_factor: confloat(gt=0) = 0.5

    # prep data will generate the following number of randomized realisations
    data_prep_train_n_realisations: Dict[
        Literal["background", "fluctuation"], NonNegativeInt
    ] = {"background": 100, "fluctuation": 100}

    data_prep_shuffling_sliding_window_scale: float = 2.0

    data_prep_imputation_sliding_window_scale: float = 3.0

    data_prep_fluctuation_presets: Dict[str, FluctuationPreset] = {
        "fluct_type_0": {
            "amplitude_confs": {
                "up": {"min": 2, "max": 100},
                "down": {"min": -2, "max": -10},
            },
            "duration": {"min": 1, "max": 8},
            "temporal_channel_shift": {"min": 0, "max": 5},
            "types": ["flat_plateau", "gaussian"],
        }
    }

    # is used like a stretch factor for time values when specifying durations and temporal shifts
    # of short term and long term state injections
    data_prep_fluctuation_time_unit: PositiveFloat = 1.0

    # if the most recent signal in a time series is older than
    # threshold * min_time_bin_width * n_steps_dec
    # it will be labeled bck_old_signal instead of signal
    data_prep_signal_to_background_time_threshold: PositiveFloat = 4.0

    # Background long-term plateaus added in data augmentation
    data_prep_long_term_state_presets: Dict[str, LongTermStatePreset] = {}
    #     {
    #     "upward": {
    #         "amplitude": {"min": 1, "max": 3},
    #         "duration": {"min": 100, "max": 200},
    #         "types": ["plateau", "linear_trend"],
    #     },
    #     "downward": {
    #         "amplitude": {"min": -1, "max": -3},
    #         "duration": {"min": 100, "max": 200},
    #         "types": ["plateau", "linear_trend"],
    #     },
    # }
    # todo: discuss and decide on what to do in paper and the default here

    # Data augmentation by globally adding a constant in units of
    # feature standard deviations to the entire light curve.
    # This is the range in units of feature standard deviation to
    # randomly draw values from
    data_prep_global_flat_plateaus: Dict[
        Literal["amplitude"], Dict[MinMaxStr, float]
    ] = {"amplitude": {"min": -6, "max": 6}}

    n_randomized_amplitudes_of_signal_dec_fluctuations: NonNegativeInt = 0

    tensorflow_run_eagerly: bool = False

    debug_do_overwrite_data_with_gaussians: bool = False
    debug_model_auto_enc_use_only_d1_as_input: bool = False
    debug_enable_comet: bool = False

    # create plots and csv files, e.g. for inspecting the learned hidden representations
    debug_create_model_output_plots: bool = False

    debug_create_model_output_csv_files: bool = False
    debug_create_data_prep_plots: bool = False
    debug_plot_of_data_prep_events_realisation_range: sized_list(int, 2) = [0, 0]
    debug_plot_realisation_by_id: dict = {}
    debug_data_prep_plots_per_realisation: NonNegativeInt = 0
    debug_data_prep_inspect_temporal_offset_dists: bool = False
    debug_data_prep_plot_reconstructed_lightcurves: bool = False
    debug_plot_predict_model_outputs: bool = False
    skip_flare_templates: bool = False

    do_error_analysis_test_statistic: bool = False
    do_error_analysis_reconstruction_error: bool = False

    error_analysis_custom_realisations: dict = {}
    # dictionary in structure:
    #  {realisation_name}:
    #    start_time: 450
    #    end_time: 550

    error_analysis_n_examples: NonNegativeInt = 1

    data_prep_reproduce_seed: Any = None
    data_prep_reproduce_indices: List[NonNegativeInt] = []

    plot_inputs_with_sliding_window_context_conf: dict = {}

    plot_flare_significance_over_template_flux: dict = {"recreate_input_data": False}

    plot_scenario_highlights: dict = {}
    infer_plot_highlights: dict = {}

    plot_cluster_space: bool = True

    # extra column names to keep in DataFrame of prediction results
    # e.g., extra metadata needed to be passed along for plotting
    df_cols_to_keep: List[dict] = [
        {
            "phases": [
                "train",
                "calib",
                "predict",
                "debug_keras_model",
                "inspect_inputs",
                "bgmm_fit",
            ],
            "cols": [
                "fluct_id",
                "fluct_type_name",
                "fluct_max_amplitude",
                "fluct_total_duration",
                "n_signal_steps_fermi_flux_0",
                "n_signal_steps_fermi_flux_1",
                "n_signal_steps_cta_flux_0",
                "n_signal_steps_cta_flux_1",
                "time_right_edge",
                "time_left_edge",
                "realisation_id",
                "realisation_seed",
                "realisation_row_index",
                "custom_realisation_name",
                "cta_n_flare_data_points",
                "flare_flux_normalisation",
                "flare_injection_start",
                "flare_injection_end",
                "split_rnd_number",
                "diagnostics_extra_col",
            ],
        }
    ]

    train_logging_n_epoch_msg: NonNegativeInt = 1

    # predict and do diagnostics plots on a separate test set. If set to false, use
    # entire output of the prep_train phase
    predict_diagnostics_on_test_split: bool = False

    # event class to category mapping, where category index is the
    # list index index.
    # for example, given:
    #     self.evt_type_cats[self.tag_sig_cats] = (
    #         [['sig_0'], ['sig_1', 'sig_2']]
    #     )
    # we have two categories, with cat-0 including
    # ['sig_0'], and cat-1 including ['sig_1', 'sig_2']
    # this is equivalent to:
    #     self.evt_type_to_cat = {
    #         'bck_0': 0, 'bck_1': 0, 'bck_2': 0,
    #         'sig_0': 1, 'sig_1': 2, 'sig_2': 2
    #     }
    # the total number of event classes is part of the model, and so
    # must be consistent for different self.run_type values
    evt_type_cats: dict = {
        "bck_cat": [["bck_0", "bck_1"]],
        "sig_cat": [["sig_0", "sig_1"]],
    }

    # -------------------> from rnn_blind -------------------->

    # flag for resetting the base directory (e.g., if training on a new model, any
    # previous output directory will be deleted)
    reset_tag_dir: bool = False

    # run phases during which reset_tag_dir can be performed
    reset_tag_dir_phases: List[RunPhase] = ["train", "prep_train"]

    # number of events to use for the calibration in each metadata bin
    n_calib_ts_events: int = 5.0e3
    # n_calib_ts_events: 100 # for debugging the calib stage

    # directory for prediction outputs
    dir_output_predict: str = "predict_out"

    # directory for simulated prediction outputs
    dir_output_predict_sim: str = "predict_sim_out"

    # minimal number of distribution sampling during calibration
    # if using use_tf_prob
    min_n_calib_sample: int = 3

    # sort the input dataframe after loading based on
    # calib_roi_neighbours_config.run_id_col and calib_roi_neighbours_config.run_time_col
    is_run_wise_sorted_inputs: bool = False

    # for development, can set the random seeds to a fixed value
    is_const_random_seed: bool = False

    # flag indicating that we integrate time steps for training,
    # predictions amd TS calculations
    avg_time_steps: bool = False

    # possible alternative function to sum TS values from the different
    # features for anomaly detection
    sum_evt_lc_ts_func: Any = None

    # maximal number of events to write to a given output file for prediction
    max_write_evt: int = 1.0e4

    # match these patterns to input data file names
    in_file_patt: List[str] = ["summary", ".hdf5"]

    # output data file names
    out_file_patt: List[str] = ["summary", ".hdf5"]

    # default value for a filter function on input files
    is_accepted_input_file: Any = None

    # dropout fraction for training
    rnn_dropout: NonNegativeFloat = 0.1

    # recurrent dropout fraction for training
    # if None, use rnn_dropout
    rnn_recurrent_dropout: Optional[NonNegativeFloat] = None

    # dropout for input at decoder phase
    # if None, use rnn_dropout
    input_dec_dropout: Optional[NonNegativeFloat] = None

    # regularization scale
    regularizer_l2_scale: NonNegativeFloat = 0.0001

    # learning rate
    learning_rate: NonNegativeFloat = 0.005

    # the number of epochs without improvement on the val_loss
    # before stopping the training. when 0: inactive, no early stopping
    use_early_stopping: int = 10

    # the amount of augmented / randomized realisations within one training epoch
    # rule of thumb: 2-3 realisations. If set too small - slow training due to frequent evaluation of
    # val split. If set too high: a single epoch may lead to overfitting already,
    # and it early stopping based on the validation loss does not work reliably.
    # The more repetetive / similar the data within a realisation is,
    # the lower this number should be.
    n_realisations_per_train_epoch: PositiveFloat = 2.5

    # factor to decrease learning rate at each epoch
    learn_rate_decrease_factor: float = -0.05

    # adjust learning rate in periods. Useful when reloading new data periodically.
    # Example: When loading new data each 10 epochs. The initial learning rate at each of these
    #          periods is decreased by -0.1. The learning rate of the 10 epochs och each period
    #          us adjusted according to `learn_rate_epoch_thresh` and `learn_rate_decrease_factor`.
    # learn_rate_new_data:
    #   period: 10
    #   decrease_factor: -0.1
    learn_rate_new_data: Optional[dict] = None

    # optional override for the initial learning rate, regardless of initial epoch
    # (this overrides the value of the rate which is
    # dynamically taken form the training history file)
    learn_rate_init: Optional[NonNegativeFloat] = None

    # threshold (epoch number) to start decreasing the learning rate
    learn_rate_epoch_thresh: NonNegativeInt = 5

    # number of training epochs. nominally this is an integer. in case train_phases
    # is provided, it can also be a list (see below)
    n_train_epochs: PositiveInt = 100

    # can also optionally override train_phase with a list of phases (eg [0, 1])
    # in case train_phases is given, one can also use n_train_epochs as
    # a list with corresponding values for each train_phase (eg n_train_epochs:[0, 1, 5])
    # For the flares project, to use multiple train_phases in sequence,
    # check that the data is reloaded in between
    train_phases: sized_list(TrainPhase, 1) = [0]

    # flag indicating if features are normalised for training
    do_train_scale: bool = True

    # only calculate the training scales, but then stop and e.g., do not train
    # for instance: load a sub sample of a large dataset (representative of different
    # metabins) and pre-calculate the scaling factors before the training phase
    only_do_train_scale: bool = False

    # reset the calibration, prediction directories
    reset_calib_dir: Any = None
    reset_predict_dir: Any = None

    # -------------------> classification -------------------->

    # kernel for the kde estimator used to calibrate the TS for classification
    # eg one of ('gaussian', 'exponential')
    kde_kernel: str = "gaussian"

    # range of bandwidth values to try and for the kde of the TS distribution
    kde_bandwidths: List[float] = [3.0e-4]
    # kde_bandwidths: [5.0e-5, 1.0e-4, 5.0e-4, 1.0e-3]

    # numerical limits on the kde results, to avoid
    # overflow / underflow error with np.exp()
    kde_prob_limits: Dict[MinMaxStr, float] = {"min": 1e-30, "max": 1e10}

    # range to perform the kde fit on, given that the classification TS
    # by construction is limited to the range [0, 1]. the kde_scale must
    # be large enough that probability does not leak out (eg the integrated
    # probability density in the kde fit range sums up to 1)
    kde_scale: float = 5.0e4
    kde_bins_min: float = -0.1
    kde_bins_max: float = 1.1

    # n_split parameter for the kde grid search
    kde_n_split_debug: bool = False

    # <------------------- classification <--------------------

    # creates a csv for normalisation shift and keras_back with one row per timestep and one column per feature
    # names: feature_norm_shift.csv and feature_norm_scale.csv
    # Moreover, this creates plots of the feature value distribution before and after normalisation
    # see files generated at debug/inspect_data_normalisation/...
    debug_inspect_data_normalisation: bool = False

    # a name for a version of the model, e.g. for comparing different model runs in tensorboard
    model_version_name: Optional[str] = None

    flare_template_conf: Optional[FlareTemplateConfig] = None

    tag: str = "flares"
    tag_dir: str = "flares"
    tag_input_dir: Optional[str] = None
    tag_base_output_dir: Optional[str] = None
    in_file_filt: int = -1
    in_file_filt_now: int = -1

    # directory with inputs for prediction / calibration
    dir_input_predict: str = "intermediate_input_shuffled"

    # The maximum number of concurrently running jobs, such as the number of processes or the size of a thread-pool.
    # If -1 (default), all CPUs are used.
    # If 1, no parallel computing code is used at all, which is useful for debugging.
    # See joblib.parallel.Parallel's `n_jobs` for details
    parallel_n_jobs: int = -1

    label_significance_combo: str = r"$\sigma_\mathrm{combo}}$"
    label_significance_mm: str = r"$\sigma_\mathrm{mm}}$"
    label_significance_rec: str = r"$\sigma_\mathrm{rec}}$"
