from __future__ import annotations

from typing import Any, Dict, List, Union

from pydantic import BaseModel, PositiveInt

from trans_finder.flares.core.types import ChannelName, FeatureName


class InputDataSpec(
    BaseModel,
    validate_default=True,
    extra="forbid",
):
    """Specification of the input dataset structure.
    Additional metdata in the provided yaml file is allowed but ignored.
    """

    # column name, input values assumed to be in units of MJD
    data_prep_day_column_name: str = "time"

    # columns to apply fluctuations to
    data_prep_fluctuation_cols: Dict[ChannelName, List[FeatureName]]

    plot_channel_y_normalisation: Dict[ChannelName, float] = {}

    data_prep_signal_noise_threshold_in_feature_stds: Dict[ChannelName, float]

    # the unit of the input uncertainty in normal standard deviations
    input_error_col_scale_in_stds: PositiveInt = 2

    # Specification of the input dataframe structure.
    # a dict with keys as the name of a channel
    # each value is a dict with keys for the name of a feature :
    # int, the number of bins of this feature, or None if there are no bins
    step_features_per_channel: Dict[
        ChannelName, Dict[FeatureName, Union[PositiveInt, None]]
    ]

    channel_labels: Dict[ChannelName, str]

    channel_y_labels: Dict[ChannelName, str]
    channel_units: Dict[ChannelName, str]

    channel_groups: Dict[str, List[ChannelName]]

    # groups of channels to shift jointly instead of independently in the data augmentation
    augmentation_shifting_groups: Dict[str, List[ChannelName]] = {}

    # extra information that is not used
    dataset_metadata: Any = None

    @classmethod
    def backwards_compatibility(cls, dictionary):
        """updates input spec data to ensure backwards compatibility"""

        # keys are the old and  values are the new name
        rename_mapping = {
            "data_prep_time_column_name": "data_prep_day_column_name",
        }

        for old_name, new_name in rename_mapping.items():
            if old_name in dictionary:
                dictionary[new_name] = dictionary[old_name]
                del dictionary[old_name]
