from __future__ import annotations

import functools
import itertools
import os
import shutil
from pathlib import Path
from typing import TYPE_CHECKING

# from line_profiler_pycharm import profile
from tqdm import tqdm

from trans_finder.flares.custom_realisations.background_scenarios import (
    get_background_scenarios,
)
from trans_finder.flares.custom_realisations.flare_scenarios import (
    custom_timeline_with_simulated_flare,
    get_flare_scenarios,
    physical_flare,
    systematic_all_flare_templates,
)
from trans_finder.flares.custom_realisations.fluctuation_scenarios import (
    get_fluctuation_scenarios,
)
from trans_finder.flares.custom_realisations.write_custom_scenario import (
    write_custom_scenario,
)
from trans_finder.flares.data_prep_helpers.fluctuation import FluctuationGenerator
from trans_finder.flares.data_prep_helpers.long_term_state import LongTermStateGenerator
from trans_finder.flares.data_prep_helpers.physical_flare import PhysicalFlareGenerator
from trans_finder.flares.plot_list.scenario_highlights import (
    do_plot_scenario_highlights,
)
from trans_finder.flares.utilities.df_batched_hdf_writer import (
    DataFrameBatchedHDFWriter,
)

if TYPE_CHECKING:
    from trans_finder.flares.manager import Manager


class CreateCustomRealisations:
    tag = ["b", f" - flares.{Path(__file__).stem} - "]

    def __init__(self, mngr):
        self.mngr: Manager = mngr

    def create_flare_significance_over_template_flux_scenarios(self, aggregated_data):
        mngr = self.mngr
        plot_conf = mngr.plot_flare_significance_over_template_flux

        if len(plot_conf) == 0:
            return []

        realisation_custom_name = "flare_significance_over_template_flux"
        # counter = 0

        if "realisation_seeds" in plot_conf:
            seeds = plot_conf["realisation_seeds"]
        else:
            seed_dict = find_sparsity_seeds(
                mngr, aggregated_data, initial_seed=plot_conf["find_seeds_start"]
            )
            seeds = list(itertools.chain(*seed_dict.values()))

        n_realisations = (
            len(seeds)
            * len(plot_conf["template_flux_scaling"])
            * plot_conf["n_template_seeds"]
        )

        progress_bar = tqdm(
            total=n_realisations,
            unit="realisations",
            desc="flare_significance_over_template_flux",
            position=0,
            leave=True,
        )

        for seed in seeds:
            for template_flux_scaling in plot_conf["template_flux_scaling"]:
                for template_seed in range(plot_conf["n_template_seeds"]):
                    realisation_id = f"seed_{seed}_scaling_{template_flux_scaling}_template_seed_{template_seed}"
                    scenario = create_flare_significance_over_template_flux_realisation(
                        mngr,
                        aggregated_data,
                        seed,
                        id=realisation_id,
                        custom_name=realisation_custom_name,
                        scaling=template_flux_scaling,
                        template_seed=template_seed,
                    )
                    yield scenario

                    progress_bar.update()
                    # counter += 1

    def flare_significance_over_flux_realisations(self, mngr, aggregated_data):
        if mngr.plot_flare_significance_over_template_flux["recreate_input_data"]:
            # reset output dir
            data_output_dir = os.path.join(
                mngr.tag_input_dir, "flare_template_realisations"
            )
            Path(data_output_dir).mkdir(parents=True, exist_ok=True)
            shutil.rmtree(data_output_dir)

            df_batched_hdf_writer = DataFrameBatchedHDFWriter(mngr, data_output_dir)

            for scenario in self.create_flare_significance_over_template_flux_scenarios(
                aggregated_data
            ):
                time_filter_range = None
                realisation, metadata_dict = scenario
                if realisation.custom_name == "flare_significance_over_template_flux":
                    time_filter_range = [
                        metadata_dict["flare_injection_start"],
                        metadata_dict["flare_injection_end"],
                    ]

                write_custom_scenario(
                    mngr,
                    scenario,
                    df_batched_hdf_writer,
                    do_plot_lightcurve=False,
                    time_filter_range=time_filter_range,
                )

            df_batched_hdf_writer.flush()

    # @profile
    def run(self):
        mngr = self.mngr
        mngr.log.info([self.tag, ["p", "CreateCustomRealisations.run()"]])

        data_output_dir = os.path.join(mngr.tag_input_dir, mngr.tag_custom_realisations)

        # clear output directory
        Path(data_output_dir).mkdir(parents=True, exist_ok=True)
        shutil.rmtree(data_output_dir)

        # load input data for each channel, ensuring an ordering in time
        input_data = mngr.data_prep.flares_load_input_data(
            mngr.dir_input_data_prep_train
        )

        # rebin channel data as configured and average high density data points
        aggregated_data = mngr.data_prep.data_aggregation(input_data)
        mngr.data_prep.set_aggregated_data_conf(aggregated_data)

        df_batched_hdf_writer = DataFrameBatchedHDFWriter(mngr, data_output_dir)

        if not mngr.skip_flare_templates and self.mngr.flare_template_conf is not None:
            mngr.flare_generator = PhysicalFlareGenerator(mngr=mngr)
        mngr.data_prep.fluctuation_generator = FluctuationGenerator()
        mngr.data_prep.long_term_state_generator = LongTermStateGenerator()

        scenario_functions = [
            (None, get_background_scenarios),
            (None, get_fluctuation_scenarios),
        ]

        if self.mngr.flare_template_conf is not None:
            scenario_functions = [
                (
                    None,
                    functools.partial(
                        custom_timeline_with_simulated_flare,
                        flare_scaling=1.0,
                        name="custom_flare_1.0",
                    ),
                ),
                (
                    None,
                    functools.partial(
                        custom_timeline_with_simulated_flare,
                        flare_scaling=0.1,
                        name="custom_flare_0.1",
                    ),
                ),
                (
                    None,
                    functools.partial(
                        custom_timeline_with_simulated_flare,
                        flare_scaling=-0.15,
                        name="custom_flare_downward",
                        # elevate_background=0,
                    ),
                ),
            ] + scenario_functions

            scenario_functions.extend(
                [
                    (None, get_flare_scenarios),
                    (None, systematic_all_flare_templates),
                ]
            )

        for output_dir, scenario_fn in scenario_functions:
            if output_dir is None:
                output_dir = "custom_realisations"

            scenarios = scenario_fn(mngr, aggregated_data)
            for scenario in scenarios:
                reconstructed_df = write_custom_scenario(
                    mngr,
                    scenario,
                    df_batched_hdf_writer,
                    do_plot_lightcurve=True,
                    plot_output_dir=output_dir,
                )
                realisation, metadata_dict = scenario
                do_plot_scenario_highlights(
                    mngr,
                    realisation,
                    reconstructed_df,
                    output_dir=output_dir,
                )

        df_batched_hdf_writer.flush()

        # try:
        #     self.flare_significance_over_flux_realisations(mngr, aggregated_data)
        # except Exception as e:
        #     print(e)


def create_flare_significance_over_template_flux_realisation(
    mngr,
    aggregated_data,
    seed,
    scaling,
    id,
    custom_name,
    template_seed,
):
    return physical_flare(
        mngr,
        aggregated_data,
        id,
        custom_name=custom_name,
        realisation_seed=seed,
        flare_fermi=True,
        flare_cta=True,
        fermi_flare_offset=0,
        scaling=scaling,
        flare_seed=template_seed,
    )


def find_sparsity_seeds(mngr, aggregated_data, initial_seed):
    """generates and prints a fixed number of seeds that give a number of flare data points in cta"""

    cta_sparsity_values = mngr.plot_flare_significance_over_template_flux[
        "cta_sparsity_values"
    ]

    cta_sparsity_seeds = {v: [] for v in cta_sparsity_values}

    n_examples_target = mngr.plot_flare_significance_over_template_flux[
        "n_realisations_per_cta_sparsity_value"
    ]

    seed = initial_seed

    while any(len(seeds) < n_examples_target for seeds in cta_sparsity_seeds.values()):
        seed += 1
        scenario = create_flare_significance_over_template_flux_realisation(
            mngr,
            aggregated_data,
            seed,
            id="find_sparsity_seeds",
            custom_name="find_sparsity_seeds",
            scaling=1,
            template_seed=0,
        )
        realisation, metadata_dict = scenario
        cta_n_flare_data_points = metadata_dict["cta_n_flare_data_points"]
        if (
            cta_n_flare_data_points in cta_sparsity_seeds
            and len(cta_sparsity_seeds[cta_n_flare_data_points]) < n_examples_target
        ):
            cta_sparsity_seeds[cta_n_flare_data_points].append(realisation.seed)

            print(f"- {realisation.seed} # {repr(metadata_dict)}")

    """
     
    all_full = False
    while not all_full:
        realisation = None # new random
        # sliding window of flare_duation size, count cta data points

        for value in cta_sparsity_values:
            # sample one position from the sliding window pass with sparsity == value
            # generate realisation with filtered aggregated data
            
    clip aggregated data 15 data points before + right after the injection  
    """

    return cta_sparsity_seeds
