from typing import Dict, List, Optional

import numpy as np
import pandas as pd

from trans_finder.flares.data_prep_helpers.channel_padding import (
    create_channel_padding_infer,
)
from trans_finder.flares.data_prep_helpers.physical_flare import (
    PhysicalFlare,
    PhysicalFlareGenerator,
)
from trans_finder.flares.data_prep_helpers.realisation import Realisation
from trans_finder.flares.state_manager import StateManager


def get_flare_scenarios(mngr, aggregated_data):
    scenarios = []

    return scenarios


def create_flare(
    mngr,
    flare_start,
    flare_intensity,
    flare_seed,
    fermi_flare_offset=0,
    flare_fermi=True,
    flare_cta=True,
    duration=None,
):
    # flare_type = {"fermi": "simple", "cta": "gauss"}

    flare_type = {channel_tag: "box" for channel_tag in flare_intensity}

    channel_template_names = {}
    if duration is None:
        duration = mngr.plot_flare_significance_over_template_flux["flare_duration"]
    for channel_tag in flare_intensity:
        channel_template_names[channel_tag] = (
            f"{channel_tag}_type_{flare_type[channel_tag]}_intensity_{flare_intensity[channel_tag]}_duration_{duration}_realisation_{flare_seed}.hdf5"
        )

    channel_templates = {
        channel_tag: mngr.flare_generator.read_flare_template(
            channel_template_names[channel_tag], channel_tag
        )
        for channel_tag in flare_intensity
    }

    flare_config = {
        "type_name": "physical_flare",
        "id": -1,
        "start_time": flare_start,
        "end_time": flare_start + duration,
        "channel_start_times": {c: flare_start for c in channel_templates},
        "channel_end_times": {c: flare_start + duration for c in channel_templates},
        "channel_templates": channel_templates,
    }

    phys_flare = PhysicalFlare(mngr, **flare_config)
    return phys_flare


def flare_intensity_comparison(mngr, aggregated_data):
    name = "flare_intensity_comparison"

    realisation = mngr.data_prep.create_realisation(
        aggregated_data,
        shuffling_strategy="step_based",
        do_randomize_temporal_offsets=True,
        do_shift_input_data_in_time=True,
        add_fluctuations=False,
        add_flares=False,
        add_background_states=False,
        realisation_id=name,
        realisation_seed=0,
        channel_padding=create_channel_padding_infer(mngr),
    )
    realisation.custom_name = name

    scalings = mngr.plot_flare_significance_over_template_flux["template_flux_scaling"]
    flare_seed = 0

    flare_start = 350.2

    for idx, scaling in enumerate(scalings):
        flare_start += 20
        flare = create_flare(
            mngr,
            flare_start=flare_start,
            flare_intensity={
                channel_name: scaling
                for channel_name in mngr.flare_template_conf["flare_channels"]
            },
            flare_seed=flare_seed,
        )
        realisation.flare_template_injections.append(flare)

    return realisation, {}


def physical_flare(
    mngr,
    aggregated_data,
    id,
    custom_name=None,
    realisation_seed=0,
    flare_fermi=True,
    flare_cta=True,
    fermi_flare_offset=0,
    scaling=1.0,
    flare_seed=0,
):
    if custom_name is None:
        custom_name = id

    realisation = mngr.data_prep.create_realisation(
        aggregated_data,
        shuffling_strategy="step_based",
        do_randomize_temporal_offsets=False,
        do_shift_input_data_in_time=False,
        add_fluctuations=False,
        add_flares=False,
        add_background_states=False,
        realisation_id=id,
        realisation_seed=realisation_seed,
        channel_padding=create_channel_padding_infer(mngr),
    )

    realisation.custom_name = custom_name

    flare_start = 350.2
    flare = create_flare(
        mngr,
        flare_start=flare_start,
        flare_intensity={"fermi": scaling, "cta": scaling},
        flare_seed=flare_seed,
    )

    cta_n_flare_data_points = (
        realisation.real_times["cta"]
        .between(
            flare.channel_start_times["cta"],
            flare.channel_end_times["cta"],
            inclusive="both",
        )
        .sum()
    )

    fermi_n_flare_data_points = (
        realisation.real_times["fermi"]
        .between(
            flare.channel_start_times["fermi"],
            flare.channel_end_times["fermi"],
            inclusive="both",
        )
        .sum()
    )

    realisation.signal_injections.append(flare)

    metadata_dict = {
        "flare_injection_start": flare.start_time,
        "flare_injection_end": flare.end_time,
        "cta_n_flare_data_points": cta_n_flare_data_points,
        "fermi_n_flare_data_points": fermi_n_flare_data_points,
        "flare_flux_normalisation": flare.input_flux,
    }

    return realisation, metadata_dict


def custom_timeline_with_simulated_flare(
    mngr,
    aggregated_data: Dict[str, pd.DataFrame],
    name: str,
    flare_scaling: float,
    elevate_background: Optional[float] = None,
):
    flare_generator = PhysicalFlareGenerator(mngr)
    channel_tags = mngr.data_prep_channel_tags
    inference_channel_padding = create_channel_padding_infer(mngr)
    template_conf = mngr.flare_template_conf
    time_col_name = mngr.data_prep_day_column_name

    flares = []

    for flare_start, scaling, duration in [
        [1, flare_scaling, 3],
    ]:
        if scaling < 0:
            make_negative = True
            scaling *= -1
        else:
            make_negative = False
        flare = flare_generator.create_custom_flare(
            template_names={
                channel_tag: flare_generator.get_template_name(
                    channel_tag,
                    "box",
                    scaling,
                    template_conf["flare_duration"],
                    0,
                )
                for channel_tag in channel_tags
            },
            start_time=flare_start,
            duration=duration,
            make_negative=make_negative,
        )

        flare_start += duration

        flares.append(flare)

    #  use the template dataframes and make a time series out of them by concatenation
    new_data = concatenate_flare_templates_into_time_series(
        mngr,
        flares,
        aggregated_data,
        inference_channel_padding,
        channel_tags,
        do_fill_up_with_padding=False,
    )

    if elevate_background is None:
        background = aggregated_data
        padding = inference_channel_padding
    else:
        background = {}
        padding = {}
        for channel_tag in channel_tags:
            # get a dataframe of the global std for each data point (rows) and channel feature (columns)
            relevant_cols = mngr.data_prep_fluctuation_cols[channel_tag]

            # background data
            stds = aggregated_data[channel_tag][relevant_cols].std()
            df = aggregated_data[channel_tag][relevant_cols] + stds * elevate_background
            background[channel_tag] = aggregated_data[channel_tag].copy()
            background[channel_tag][relevant_cols] = df

            # padding data
            df = (
                inference_channel_padding[channel_tag][relevant_cols]
                + stds * elevate_background
            )
            padding[channel_tag] = inference_channel_padding[channel_tag].copy()
            padding[channel_tag][relevant_cols] = df

    segments = [
        (background, 0, 80),
        (new_data, None, None),
        (background, 80, 120),
    ]

    full_data = {}

    pivot = 0

    for i_segment, (segment, t_min, t_max) in enumerate(segments):
        for channel_tag in channel_tags:
            actual_segment = (
                segment[channel_tag]
                if t_min is None
                else segment[channel_tag][
                    segment[channel_tag][time_col_name].between(t_min, t_max)
                ]
            )
            new_df = actual_segment.assign(
                **{
                    time_col_name: actual_segment[time_col_name]
                    + (pivot + 1 - (t_min if t_min is not None else 0))
                }
            )

            if channel_tag in full_data:
                full_data[channel_tag] = pd.concat(
                    [full_data[channel_tag], new_df[new_df[time_col_name].gt(0)]]
                )
            else:
                full_data[channel_tag] = new_df

        for channel_tag in channel_tags:
            pivot = max(pivot, full_data[channel_tag][time_col_name].max())

    for channel_tag in channel_tags:
        full_data[channel_tag] = (
            full_data[channel_tag]
            .reset_index(drop=True)
            .sort_values(
                axis=0,
                by=mngr.data_prep_day_column_name,
                ascending=True,
                ignore_index=True,
            )
        )

    realisation: Realisation = mngr.data_prep.create_realisation(
        full_data,
        shuffling_strategy=None,
        do_randomize_temporal_offsets=False,
        do_shift_input_data_in_time=False,
        add_fluctuations=False,
        add_flares=False,
        add_background_states=False,
        realisation_id=name,
        realisation_seed=0,
        channel_padding=padding,
    )
    realisation.custom_name = name

    return [
        (realisation, {}),
    ]


def concatenate_flare_templates_into_time_series(
    mngr,
    flares: List[PhysicalFlare],
    aggregated_data: Dict[str, pd.DataFrame],
    padding: Dict[str, pd.DataFrame],
    channel_tags_to_edit: List[str],
    do_fill_up_with_padding=True,
):
    """use the template dataframes and make a time series out of it by concatenating them,
    adding a single point from the padding to the in front and remove upper limits using the padding data points
    """

    new_data = {}

    for channel_tag in channel_tags_to_edit:
        time_col_name = mngr.data_prep_day_column_name

        relevant_template_cols = [col for col in aggregated_data[channel_tag].columns]

        relevant_output_cols = [col for col in aggregated_data[channel_tag].columns]

        # concatenate channel dataframes
        template_df_list = [
            flare.channel_templates[channel_tag][
                flare.channel_templates[channel_tag][time_col_name].between(
                    0,
                    flare.channel_end_times[channel_tag]
                    - flare.channel_start_times[channel_tag],
                    inclusive="left",
                )
            ][relevant_template_cols].assign(
                **{time_col_name: lambda x: x.time + flare.start_time}
            )
            for flare in flares
            if channel_tag in flare.channel_templates
        ]

        template_concat_df = (
            pd.concat(template_df_list)
            .reset_index(drop=True)
            .sort_values(
                axis=0,
                by=mngr.data_prep_day_column_name,
                ascending=True,
                ignore_index=True,
            )
        )

        # Replace negative values with zero in the concatenated templates
        # template_concat_df[relevant_template_cols] = template_concat_df[
        #     relevant_template_cols
        # ].clip(lower=0)

        # add flare template data points  on top of background from channel padding
        template_times = template_concat_df[mngr.data_prep_day_column_name]

        additive_background_df = (
            padding[channel_tag]
            .sample(n=len(template_times), replace=True)
            .reset_index(drop=True)
            .assign(**{time_col_name: template_times})
        )

        error_cols = (
            mngr.data_prep_error_up_cols[channel_tag]
            + mngr.data_prep_error_down_cols[channel_tag]
        )

        # exclude time column from values being summed for flare template and background
        additive_cols = [
            col
            for col in relevant_output_cols
            if col != time_col_name and col not in error_cols
        ]

        # # columns that need to be added in log space
        # log_add_columns = []

        # columns that can be added directly
        direct_add_columns = [
            col for col in additive_cols  # if col not in log_add_columns
        ]

        channel_df = additive_background_df
        # (1) do inverse log transform , (2) add together, (3) to log transform
        # only on feature (flux) columns
        # channel_df[log_add_columns] = data_prep_simple_log_transform(
        #     data_prep_inverse_simple_log_transform(channel_df[log_add_columns]).add(
        #         data_prep_inverse_simple_log_transform(
        #             template_concat_df[log_add_columns]
        #         ).fillna(0)
        #     )
        # )

        # If errors in templates are zero or Nan/Inf, use background errors
        condition = (template_concat_df[error_cols] == 0) | ~np.isfinite(
            template_concat_df[error_cols]
        )

        # Otherwise, use smaller uncertainty from either template or background
        channel_df[error_cols] = np.where(
            condition,
            additive_background_df[error_cols],
            np.minimum(
                additive_background_df[error_cols], template_concat_df[error_cols]
            ),
        )

        # adding directly
        channel_df[direct_add_columns] = (channel_df[direct_add_columns]).add(
            (template_concat_df[direct_add_columns])
        )

        if do_fill_up_with_padding:
            # fill up empty time with channel padding data points
            background_times = pd.Series(np.arange(0, max(channel_df["time"])))
            background_times = background_times[
                ~background_times.isin(channel_df["time"])
            ].reset_index(drop=True)

            background_df = (
                padding[channel_tag]
                .sample(n=len(background_times), replace=True)
                .reset_index(drop=True)
                .assign(**{time_col_name: background_times})
            )

            # combine background and concatenated flare templates
            channel_df = (
                pd.concat([channel_df, background_df])
                .reset_index(drop=True)
                .sort_values(
                    axis=0,
                    by=mngr.data_prep_day_column_name,
                    ascending=True,
                    ignore_index=True,
                )
            )

        new_data[channel_tag] = channel_df

    for channel_tag in aggregated_data:
        if channel_tag not in new_data:
            new_data[channel_tag] = (
                pd.concat(
                    [
                        padding[channel_tag],
                        padding[channel_tag].iloc[0:1]
                        # adding another data point in the past to complete this channel
                        .assign(
                            **{
                                time_col_name: lambda x: padding[channel_tag][
                                    time_col_name
                                ].max()
                                + mngr.data_prep_n_days_per_step[channel_tag]
                            }
                        ),
                    ]
                )
                .reset_index(drop=True)
                .sort_values(
                    axis=0,
                    by=mngr.data_prep_day_column_name,
                    ascending=True,
                    ignore_index=True,
                )
            )

    return new_data


def systematic_all_flare_templates(mngr, aggregated_data: Dict[str, pd.DataFrame]):
    if mngr.skip_flare_templates:
        return []

    if not hasattr(mngr, "flare_template_conf"):
        mngr.log.info(
            "Skipping flare templates as config entry 'flare_template_conf' does not exist."
        )
        return []

    systematic_flare_evaluation_metadata = {}

    scenarios = []

    base_name = "systematic_all_flare_templates"
    template_conf = mngr.flare_template_conf
    flare_generator = PhysicalFlareGenerator(mngr)
    inference_channel_padding = create_channel_padding_infer(mngr)

    for flare_duration in template_conf["systematic_evaluation_durations"]:
        for combination_name, channel_tags in template_conf[
            "systematic_evaluation_combinations"
        ].items():
            full_name = f"{base_name}_{combination_name}_{flare_duration}"

            flares = []
            flare_start = 10

            combination_metadata = []

            relevant_template_cols = []
            for channel_tag in channel_tags:
                relevant_template_cols.extend(
                    [col for col in aggregated_data[channel_tag].columns]
                )

            for scaling in template_conf["template_flux_scaling"]:
                for template_seed in range(template_conf["n_template_seeds"]):
                    if scaling == 0:
                        flare = flare_generator.create_custom_flare(
                            template_names={
                                channel_tag: flare_generator.get_template_name(
                                    channel_tag,
                                    "box",
                                    1.0,
                                    template_conf["flare_duration"],
                                    template_seed,
                                )
                                for channel_tag in channel_tags
                            },
                            start_time=flare_start,
                            duration=flare_duration,
                        )

                        for channel_tag in flare.channel_templates:
                            flare.channel_templates[channel_tag] = (
                                flare.channel_templates[channel_tag].copy()
                            )
                            cols = (
                                mngr.data_prep_error_up_cols[channel_tag]
                                + mngr.data_prep_error_down_cols[channel_tag]
                                + mngr.data_prep_error_feature_cols[channel_tag]
                            )
                            flare.channel_templates[channel_tag][cols] = (
                                flare.channel_templates[channel_tag][cols] * 0
                            )

                    else:
                        flare = flare_generator.create_custom_flare(
                            template_names={
                                channel_tag: flare_generator.get_template_name(
                                    channel_tag,
                                    "box",
                                    scaling,
                                    template_conf["flare_duration"],
                                    template_seed,
                                )
                                for channel_tag in channel_tags
                            },
                            start_time=flare_start,
                            duration=flare_duration,
                        )
                    flares.append(flare)

                    combination_metadata.append(
                        [flare.start_time, flare.end_time, flare.input_flux]
                    )

                    flare_start += flare_duration + (
                        2
                        * mngr.n_steps_tot
                        * mngr.data_prep.data_prep_analysis_min_time_bin_width
                    )

            systematic_flare_evaluation_metadata[full_name] = combination_metadata

            #  use the template dataframes and make a time series out of them by concatenation
            new_data = concatenate_flare_templates_into_time_series(
                mngr,
                flares,
                aggregated_data,
                inference_channel_padding,
                channel_tags,
            )

            realisation: Realisation = mngr.data_prep.create_realisation(
                new_data,
                shuffling_strategy=None,
                do_randomize_temporal_offsets=False,
                do_shift_input_data_in_time=False,
                add_fluctuations=False,
                add_flares=False,
                add_background_states=False,
                realisation_id=full_name,
                realisation_seed=0,
                channel_padding=inference_channel_padding,
            )
            realisation.custom_name = full_name

            scenarios.append((realisation, {}))

    StateManager(
        mngr,
        "systematic_flare_evaluation_metadata",
        dir_tag="custom_realisations_input",
    ).save_state(systematic_flare_evaluation_metadata)

    return scenarios
