from __future__ import annotations

import gc
import os
import shutil
import sys
from math import ceil, floor
from typing import TYPE_CHECKING, List

import joblib
import numpy as np
import pandas as pd
import tensorflow as tf
from numpy.random import default_rng
from tensorflow.estimator import ModeKeys

from trans_finder.flares.checks.background_preprocessing import (
    do_checks_on_background_network_inputs,
)
from trans_finder.flares.exceptions import (
    CalibFileMissing,
    DataScalingMissing,
    InputOutOfRange,
)
from trans_finder.flares.plot_list.data_normalisation import (
    make_feature_distribution_plots,
)
from trans_finder.flares.utils import create_dir, get_columns_wildcard, str_join

if TYPE_CHECKING:
    from trans_finder.flares.log_parser import LogParser
    from trans_finder.flares.manager import Manager


class Data:
    """class for custom setting for processing input data"""

    def __init__(self, mngr):
        self.mngr: Manager = mngr
        self.log: LogParser = mngr.log

    def get_all_step_col_names(self, col_function):
        lc_col_names = []
        for n_step in range(self.mngr.n_steps_tot):
            # main features come first
            for n_feat in range(self.mngr.n_main_cols):
                lc_col_names += [col_function(n_step=n_step, n_feat=n_feat)]
        return lc_col_names

    # def get_custom_step_values(self, df, col_function):
    #     """reshape the pd dataset columns into an array
    #
    #     transform e.g.,
    #     ['lc_0_0', 'lc_0_1', 'lc_0_2', 'lc_0_3', ...
    #     'lc_17_2', 'lc_17_3', 'lc_17_4', 'lc_17_5']
    #     into an array, e.g.,
    #     [[[0 1 0 0 0 1]
    #     [0 0 2 0 0 0]
    #     ...
    #     [1 1 1 0 0 0]
    #     [0 3 0 2 0 0]]]
    #
    #     Parameters
    #     -----------
    #     df : pd.DataFrame()
    #         input dataframe with light curve columns, one per bin/time
    #
    #     Returns
    #     -------
    #     np.array
    #         the data
    #     """
    #     lc_col_names = self.get_all_step_col_names(col_function)
    #
    #     n_cols = self.mngr.n_main_cols
    #
    #     custom_values = df[lc_col_names]
    #     n_evts = len(custom_values)
    #
    #     custom_values = np.asarray(custom_values.values)
    #     custom_values = custom_values.reshape(n_evts, self.mngr.n_steps_tot, n_cols)
    #
    #     return custom_values

    def get_step_values(self, df):
        """reshape the pd dataset columns into an array

        transform e.g.,
        ['lc_0_0', 'lc_0_1', 'lc_0_2', 'lc_0_3', ...
        'lc_17_2', 'lc_17_3', 'lc_17_4', 'lc_17_5']
        into an array, e.g.,
        [[[0 1 0 0 0 1]
        [0 0 2 0 0 0]
        ...
        [1 1 1 0 0 0]
        [0 3 0 2 0 0]]]

        Parameters
        -----------
        df : pd.DataFrame()
            input dataframe with light curve columns, one per bin/time

        Returns
        -------
        np.array
            the data
        """

        # the main key prefix of the light curve / time-series columns (e.g., lc_)
        lc_tag = self.mngr.lc_col_func

        lc_col_names = []
        for n_step in range(self.mngr.n_steps_tot):
            # main features come first
            for n_feat in range(self.mngr.n_main_cols):
                lc_col_names += [self.mngr.lc_col_func(n_step=n_step, n_feat=n_feat)]

            # add additional optional columns for step-wise metadata
            for meta_col in self.mngr.meta_step_cols:
                lc_col_names += [
                    self.mngr.meta_step_col_func(col_name=meta_col, n_step=n_step)
                ]

        n_cols = self.mngr.n_main_cols + len(self.mngr.meta_step_cols)

        step_values = df[lc_col_names]
        n_evts = len(step_values)

        step_values = np.asarray(step_values.values)
        step_values = step_values.reshape(n_evts, self.mngr.n_steps_tot, n_cols)

        # print(df) ;print(step_values) ; print('+-'*30)

        return step_values

    def set_train_scales(self):
        """calculate the scaling parameters of the inputs, and store them for later"""

        self.log.info([["r", " - set-train-scales()"]])

        mode = ModeKeys.TRAIN

        input_df = self.mngr.input_data[self.mngr.tag_input_df][mode]
        masks = self.mngr.train_val_splitter.get_masks_train_val_split(input_df)
        train_mask, _ = masks

        tag_types = [self.mngr.tag_steps]
        for tag_type in tag_types:
            is_data = tag_type == self.mngr.tag_steps
            tag_file = self.mngr.tag_input_scalings[tag_type]

            self.log.info(
                [
                    ["b", " - processing "],
                    ["o", ("data" if is_data else "metadata"), ""],
                    ["g", len(self.mngr.input_data[self.mngr.tag_steps][mode])],
                    ["b", " events ..."],
                ]
            )

            if is_data:
                df_values = np.asarray(self.mngr.input_data[self.mngr.tag_steps][mode])
                # scale on train split only
                df_values = df_values[train_mask]
            else:
                df_values = self.mngr.input_data[self.mngr.tag_input_df][mode][
                    self.mngr.meta_bin_cols
                ]
                # scale on train split only
                df_values = df_values[train_mask]

                df_values = np.asarray(df_values)

            # for classification, scale based on the background categories only
            if self.mngr.is_enc_dec_clasif:
                mask = self.mngr.input_data[self.mngr.tag_input_df][mode][
                    self.mngr.tag_is_bck_cat
                ]
                df_values = df_values[mask]

            if is_data:
                df_values = np.concatenate(df_values, axis=0)

            quant_locs = [
                self.mngr.input_scaling_quant_cut,
                0.5,
                1 - self.mngr.input_scaling_quant_cut,
            ]
            min_quant_range = sys.float_info.min * 100
            # min_quant_scaling = 1e-10

            quants = []
            for n_feat in range(df_values.shape[-1]):
                df_feat = df_values[:, n_feat]

                feat_range = np.quantile(df_feat, quant_locs)

                if (feat_range[2] - feat_range[0]) < min_quant_range:
                    # feat_range[0] = -1 * min_quant_scaling
                    feat_range[0] = -1
                    feat_range[1] = 0
                    feat_range[2] = 1
                if (
                    np.abs(feat_range[1] - feat_range[0]) < min_quant_range
                    or np.abs(feat_range[2] - feat_range[1]) < min_quant_range
                ):
                    feat_range[1] = 0.5 * (feat_range[2] - feat_range[0])

                quants += [feat_range]

                self.log.info(
                    [
                        ["g", " - n_feat = "],
                        ["y", "{:>3d}".format(n_feat)],
                        ["g", "  -> "],
                        ["p", " , ".join([("{:>6.3g}".format(f)) for f in feat_range])],
                    ]
                )

            self.save_calib_file(
                tag_dir=self.mngr.tag_model_dir,
                tag_file=tag_file,
                data=quants,
            )

            if self.mngr.debug_inspect_data_normalisation:
                from trans_finder.flares.helpers_debugging import (
                    get_debug_output_filepath,
                )

                file_name = get_debug_output_filepath(
                    self.mngr,
                    "inspect_data_normalisation",
                    f"complete_dist.csv",
                )
                data_distribution = pd.DataFrame(df_values).describe()
                pd.DataFrame(data_distribution).to_csv(file_name, float_format="%.3f")

                file_name = get_debug_output_filepath(
                    self.mngr,
                    "inspect_data_normalisation",
                    f"complete_dist.pdf",
                )
                make_feature_distribution_plots(self.mngr, df_values, file_name)

        return

    def apply_user_sample_cuts(self, df_in):
        if self.mngr.prep_train_scales:
            return df_in

        elif (
            self.mngr.run_phase in ["train", "debug_keras_model", "inspect_inputs"]
            and len(self.mngr.train_phases) == 1
            and self.mngr.train_phases[-1] == 0
        ):
            # train_phase is forecasting,
            # hence drop events with fluctuations in decoder for training the background forecasting
            signal_classes = self.mngr.train_phase_0_evt_class_blacklist
            indices_to_drop = df_in[self.mngr.tag_evt_type].isin(signal_classes)

            df_out = df_in.loc[~indices_to_drop].reset_index(drop=True)

            return df_out

        elif (
            self.mngr.run_phase in ["train", "debug_keras_model", "inspect_inputs"]
            and len(self.mngr.train_phases) == 1
            and self.mngr.train_phases[-1] == 1
        ):
            # drop realisations without injected signals
            meta_df = pd.read_csv(self.mngr.path_realisation_metadata, index_col=0)
            ids_with_injections = meta_df[meta_df["has_injections"]]["id"]
            mask = df_in["realisation_id"].isin(ids_with_injections)
            df_out = df_in.loc[mask].reset_index(drop=True)
            return df_out

        elif (
            self.mngr.run_phase == "predict"
            and self.mngr.dir_output_prep_train == self.mngr.dir_input_predict
            and self.mngr.predict_diagnostics_on_test_split
        ):
            # we are predicting on the outputs of prep_train
            # hence, apply the cut to only use the test set

            # get test split mask
            raise NotImplementedError(
                "Switch everywhere from get_masks_train_val_split to 3-split."
            )
            masks = self.mngr.train_val_splitter.get_masks_train_val_test_split(df_in)
            _, _, test_mask = masks

            # filter by mask
            df_out = df_in[test_mask]

            return df_out

        else:
            return df_in

    def overwrite_data_with_gaussians(self, df, outputs, seed=42):
        train_gaussians = []

        means = [100, 200]

        for mean in means:
            train_gaussians.append((mean, 15))
            # train_gaussians.append((mean, 10))
            train_gaussians.append((mean, 1))

        predict_gaussians = train_gaussians + [
            # (10, 5),
            # (10, 10),
            (250, 1),
            (150, 1),
            # (10, 20),
        ]

        if self.mngr.run_phase in ["train", "bgmm_fit", "calib"]:
            gaussians = train_gaussians
        elif self.mngr.run_phase == "predict":
            gaussians = predict_gaussians
        else:
            raise ValueError("Unsupported run_phase in transform_intputs()")

        # All features (data and metadata) for one light curve (a row in the dataset) are randomly sampled
        # from the same Gaussian. The number of samples per Gaussian dist is evenly distributed.
        rng = default_rng(seed=seed)

        # the number of time bins to overwrite, counted from the end of the time series
        step_mask = self.mngr.n_steps_tot

        n_total_samples = outputs.shape[0]
        n_segments = len(gaussians)
        segment_start_idx = 0
        for index, (mu, sig) in enumerate(gaussians):
            segment_end_idx = int(((index + 1) / n_segments) * n_total_samples)

            # to leave the metadata columns unchanged, swap the following with self.mngr.n_main_cols
            # n_features = outputs.shape[-1]
            n_features = self.mngr.n_main_cols

            shape = ((segment_end_idx - segment_start_idx), step_mask, n_features)

            newdata = outputs[
                segment_start_idx:segment_end_idx, -step_mask:, :n_features
            ] * 0 + rng.normal(mu, sig, shape)

            outputs[segment_start_idx:segment_end_idx, -step_mask:, :n_features] = (
                newdata
            )

            isTrainGauss = (mu, sig) in train_gaussians
            label = f"mean={mu}, std={sig} {'(training)' if isTrainGauss else ''}"
            df.loc[segment_start_idx:segment_end_idx, self.mngr.tag_evt_type] = label

            segment_start_idx = segment_end_idx

        if self.mngr.debug_inspect_data_normalisation:
            from trans_finder.flares.helpers_debugging import get_debug_output_filepath

            file_name = get_debug_output_filepath(
                self.mngr,
                "inspect_data_normalisation",
                "overwrite_data_with_gaussians()_data_distribution.csv",
            )
            import pandas as pd

            data_distribution = pd.DataFrame(outputs.flatten()).describe()
            pd.DataFrame(data_distribution).to_csv(file_name, float_format="%.3f")

        return outputs

    def transform_inputs(self, df, inputs, keep_orig):
        outputs = inputs
        if keep_orig:
            outputs = np.copy(inputs)

        if self.mngr.debug_do_overwrite_data_with_gaussians:
            self.log.info(
                [
                    ["g", " - transform_inputs("],
                    ["y", self.mngr.run_phase],
                    ["g", ") flares:  overwriting data with gaussianss "],
                ]
            )
            outputs = self.overwrite_data_with_gaussians(df, outputs)

        return outputs

    def add_evt_cut_mask(self, df, inputs):
        """add a general event cut based on the input data

        Parameters
        ----------
        df : pd.DataFrame
            the full DataFrame of the input data
        inputs : np.array
            the reshaped step-data which is the input to the rnn
        """
        df.loc[:, self.mngr.tag_evt_cut_mask] = True
        return

    def dataset_model_map(self, dataset):
        """mapping function for tf.data.Dataset, which provides the inputs
        and outputs for the model

        note that we do not add the third optional weight output, since weights
        are explicitly accounted for as part of the custom loss calculation

         Parameters
         ----------
         dataset : dict of tensors with following elements
             self.mngr.tag_steps : tf.Tensor
                 input data corresponding to the time series
             self.mngr.tag_wgt : tf.Tensor
                 input weights
             self.mngr.tag_evt_cats : tf.Tensor
                 the event class corresponding to the input data (e.g., background or
                 signal of a given category)

         Returns
         -------
         dict, dict
             dicts of the input and the output tensors for the model
        """

        # split the time series data according to the length of the encoder
        encs = dataset[self.mngr.tag_steps][:, : -self.mngr.n_steps_dec, :]
        decs = dataset[self.mngr.tag_steps][:, -self.mngr.n_steps_dec :, :]
        wgts = dataset[self.mngr.tag_wgt]
        model_input_ids = dataset[self.mngr.tag_model_input_ids]

        model_inputs = {
            self.mngr.tag_enc: encs,
            self.mngr.tag_dec: decs,
            self.mngr.tag_wgt: wgts,
            self.mngr.tag_model_input_ids: model_input_ids,
        }

        # since the loss is computed as part of the model itself, the ground-truth
        # should just be a tensor of zeros (with the correct batch size), which
        # is then used with tf.keras.losses.MeanAbsoluteError(), which therefore
        # simply gives the sum (over batch) of abs(loss)
        model_outputs = {
            # self.mngr.tag_loss: tf.math.multiply(encs[:, 0, 0], 0, name="null_loss"),
        }

        data_out = (model_inputs, model_outputs)

        return data_out

    def split_and_set_inputs(self, inputs, mode_keys):
        # split the data into separate parts by mode_key and structure it by the type of data
        df_cut, data_vals, meta_vals, wgt_vals = dict(), dict(), dict(), dict()
        for mode in mode_keys:
            if mode in [ModeKeys.TRAIN, ModeKeys.PREDICT]:
                df_cut[mode] = inputs[self.mngr.tag_input_df]
                data_vals[mode] = inputs[self.mngr.tag_steps]
                wgt_vals[mode] = inputs[self.mngr.tag_wgt]

                # extract the metadata columns
                if self.mngr.has_meta_bins:
                    meta_vals[mode] = df_cut[mode][self.mngr.meta_bin_cols].values

            # split the dataset into two samples, training and evaluation
            # (used also as part of training for checking the performance)
            # use the full dataset for prediction

            mode_tag = ModeKeys.TRAIN if (mode == ModeKeys.EVAL) else mode

            if mode_tag in [ModeKeys.TRAIN, ModeKeys.EVAL]:
                masks = self.mngr.train_val_splitter.get_masks_train_val_split(
                    df_cut[mode_tag]
                )
                train_mask, eval_mask = masks

            if mode == ModeKeys.TRAIN:
                self.mngr.input_data[self.mngr.tag_input_df][mode] = df_cut[mode_tag][
                    train_mask
                ].reset_index()
                vals_now = data_vals[mode_tag][train_mask]
                wgts_now = wgt_vals[mode_tag][train_mask]
                if self.mngr.has_meta_bins:
                    meta_now = meta_vals[mode_tag][train_mask]

            elif mode == ModeKeys.EVAL:
                self.mngr.input_data[self.mngr.tag_input_df][mode] = df_cut[mode_tag][
                    eval_mask
                ].reset_index()
                vals_now = data_vals[mode_tag][eval_mask]
                wgts_now = wgt_vals[mode_tag][eval_mask]
                if self.mngr.has_meta_bins:
                    meta_now = meta_vals[mode_tag][eval_mask]

            elif mode == ModeKeys.PREDICT:
                self.mngr.input_data[self.mngr.tag_input_df][mode] = df_cut[mode_tag]
                vals_now = data_vals[mode_tag]
                wgts_now = wgt_vals[mode_tag]
                if self.mngr.has_meta_bins:
                    meta_now = meta_vals[mode_tag]

            else:
                raise Exception("unexpected mode=", mode)

            vals_merged = vals_now

            # add the metadata bin columns on the right of the main features
            # (duplicated values for each step)
            if self.mngr.add_meta_bin_to_steps:

                def merge_meta(meta_in, data_in, n_evt):
                    data = data_in[n_evt]
                    meta = np.tile(meta_in[n_evt], (self.mngr.n_steps_tot, 1))
                    merged = np.concatenate((data, meta), axis=1)
                    return merged

                vals_merged = [
                    merge_meta(
                        meta_in=meta_now,
                        data_in=vals_now,
                        n_evt=n,
                    )
                    for n in range(len(vals_now))
                ]

            # store the derived dataset for global access
            self.mngr.input_data[self.mngr.tag_steps][mode] = vals_merged
            self.mngr.input_data[self.mngr.tag_wgt][mode] = wgts_now

            self.log.info(
                [
                    ["b", " - loaded data(" + mode + "):\t"],
                    ["p", str(len(vals_merged)), "events"],
                ]
            )

            self.after_split_and_set_inputs(mode)

    def after_split_and_set_inputs(self, mode):
        pd_vals = self.mngr.input_data[self.mngr.tag_steps][mode]
        input_df_meta = self.mngr.input_data[self.mngr.tag_input_df][mode]

        self.log.debug(f" - after_split_and_set_inputs({mode=}, {pd_vals.shape=})")

        do_checks_on_background_network_inputs(self.mngr, mode, pd_vals, input_df_meta)

    def get_offset_weights(self, data, power_param):
        if power_param is None:
            power_param = self.mngr.model_offset_weight_power

        # self.mngr.model_offset_weight_power_range
        n_evts = len(data)
        step_vals = self.get_step_values(data)[:, -self.mngr.n_steps_dec :, :]
        wgts = np.ones((n_evts, self.mngr.n_steps_dec, self.mngr.n_features))

        col_ranges = []
        start_col = 0

        for offset_col in self.mngr.temporal_offset_column_indices:
            col_ranges.append((start_col, offset_col))

            start_col = offset_col + 1
            if start_col >= self.mngr.n_features:
                break

        for idx, (start_col, offset_col) in enumerate(col_ranges):
            offsets = np.abs(step_vals[..., offset_col])

            sensitivity_window = self.mngr.n_steps_dec

            offsets = np.maximum(np.zeros_like(offsets), offsets - sensitivity_window)

            wgt = np.divide(1, 1 + offsets)
            wgt = np.power(wgt, power_param)
            for idx in range(start_col, offset_col):
                wgts[..., idx] = wgt

        return wgts

    def weight_train_func(self, data):
        """function to define weights for training
        by default, a simple scheme with weights of 1 everywhere

        Parameters
        ----------
        df_in : pd.Dataframe()
            input dataframe

        Returns
        -------
        np.array
            array of weights
        """

        n_evts = len(data)

        if self.mngr.is_enc_dec_anom_det or self.mngr.is_cluster_anom_det:
            wgts = np.ones((n_evts, self.mngr.n_steps_dec, self.mngr.n_features))

            if self.mngr.run_phase in ["bgmm_fit", "calib", "train"]:
                return wgts

            if self.mngr.model_use_offset_weights:
                offset_weight_power_param = None
                if self.mngr.run_phase == "train":
                    # randomly sample offset weights power param
                    offset_weight_power_param = np.random.uniform(
                        *self.mngr.model_offset_weight_power_range, size=n_evts
                    )
                    offset_weight_power_param = np.repeat(
                        offset_weight_power_param, self.mngr.n_steps_dec
                    ).reshape(-1, self.mngr.n_steps_dec)

                offset_wgts = self.get_offset_weights(
                    data, power_param=offset_weight_power_param
                )

                if self.mngr.run_phase == "train":
                    # apply weights randomly half of the time
                    n_half = int(0.5 * n_evts)
                    random_indices = np.random.choice(
                        wgts.shape[0], size=n_half, replace=False
                    )
                    wgts[random_indices] = (
                        wgts[random_indices] * offset_wgts[random_indices]
                    )
                else:
                    wgts = wgts * offset_wgts

            # print(wgts)

        elif self.mngr.is_enc_dec_clasif:
            wgts = np.ones(n_evts)

        # debug_wgts = False
        # # debug_wgts = True
        # if debug_wgts:
        #     for x in range(self.mngr.n_steps_dec):
        #         wgts[:, x, :] = round(1 / (1.0 + x), 2)
        #     for x in range(self.mngr.n_features):
        #         wgts[:, :, x] += 0.1 * x
        #     for x in range(n_evts):
        #         wgts[x, :, :] += 1 * x
        #     # print(wgts)

        return wgts

    # def after_set_train_scales(self, df_values):
    #
    #     if self.mngr.debug_inspect_data_normalisation:
    #         from trans_finder.flares.helpers_debugging import (
    #             get_debug_output_filepath,
    #         )
    #
    #         file_name = get_debug_output_filepath(
    #             self.mngr,
    #             "inspect_data_normalisation",
    #             f"complete_dist.csv",
    #         )
    #         data_distribution = pd.DataFrame(df_values).describe()
    #         pd.DataFrame(data_distribution).to_csv(file_name, float_format="%.3f")
    #
    #         file_name = get_debug_output_filepath(
    #             self.mngr,
    #             "inspect_data_normalisation",
    #             f"complete_dist.pdf",
    #         )
    #         make_feature_distribution_plots(self.mngr, df_values, file_name)

    def smear_df_by_errors(
        self,
        df_in: pd.DataFrame,
        error_up_cols: List[str],
        error_down_cols: List[str],
        value_cols: List[str],
        drop_error_cols=False,
    ):
        """create a version of df_in with randomized errors added to the data"""

        # add randomized errors to value columns
        randomness = np.random.normal(
            0,
            1.0 / self.mngr.input_error_col_scale_in_stds,
            size=df_in[value_cols].shape,
        )

        random_errors = np.where(
            randomness < 0,
            randomness * df_in[error_down_cols].to_numpy(),
            randomness * df_in[error_up_cols].to_numpy(),
        )

        if drop_error_cols:
            df_out = df_in.drop(columns=error_up_cols + error_down_cols).copy()
        else:
            df_out = df_in.copy()

        df_out[value_cols] += random_errors

        return df_out

    def before_prep_inputs(self, df_in):
        # add copies with smearing by input errors, but not for the prep_train dataset
        if (
            self.mngr.run_phase != "predict"
            or self.mngr.dir_output_prep_train == self.mngr.dir_input_predict
        ):
            return df_in

        self.log.info(
            [
                ["g", " - before_prep_inputs("],
                ["y", self.mngr.run_phase],
                [
                    "g",
                    f") flares:  sampling input errors {self.mngr.model_n_samples_of_input_errors} times",
                ],
            ]
        )

        all_dfs = [df_in]

        error_up_cols = self.get_all_step_col_names(self.mngr.lc_error_up_col_func)
        error_down_cols = self.get_all_step_col_names(self.mngr.lc_error_down_col_func)
        step_cols = self.get_all_step_col_names(self.mngr.lc_col_func)

        # set zero errors for time colum names
        time_error_cols = [
            n
            for n in error_up_cols + error_down_cols
            if self.mngr.tag_temporal_offset in n
        ]
        df_in = pd.concat(
            [
                df_in,
                pd.DataFrame({col: 0 for col in time_error_cols}, index=df_in.index),
            ],
            axis=1,
        )

        # check for nans or np.inf in step and error columns and raise an error
        if not np.isfinite(
            df_in[step_cols + error_up_cols + error_down_cols].to_numpy()
        ).all():
            self.log.error(
                [self.tag, ["r", "NaN or np.inf encountered in input dataframe."]]
            )

        for _ in range(self.mngr.model_n_samples_of_input_errors):
            smeared_df = self.smear_df_by_errors(
                df_in, error_up_cols, error_down_cols, step_cols, drop_error_cols=True
            )
            all_dfs.append(smeared_df)

        output_df = pd.concat(all_dfs).sort_index().reset_index(drop=True)
        return output_df

    def read_dataset(self, mode, pd_range=None, full_data=None):
        """return function which reads data for tf.estimator

        Parameters
        -----------
        mode : ModeKeys
            one of [ModeKeys.TRAIN, ModeKeys.EVAL, ModeKeys.PREDICT]
        pd_range : list
            range of values (index range) to read in
        """

        pd_vals = self.mngr.input_data[self.mngr.tag_steps][mode]
        input_wgt = self.mngr.input_data[self.mngr.tag_wgt][mode]
        input_df_meta = self.mngr.input_data[self.mngr.tag_input_df][mode]
        input_ids = self.mngr.input_data[self.mngr.tag_input_df][mode][
            self.mngr.tag_event_id
        ]

        df_in = self.mngr.input_data[self.mngr.tag_input_df][mode]

        if self.mngr.is_enc_dec_clasif:
            if self.mngr.tag_evt_cats in input_df_meta.columns:
                input_evt_class = input_df_meta[self.mngr.tag_evt_cats]
            else:
                input_evt_class = np.zeros(len(input_df_meta))

        batch_size_now = (
            self.mngr.train_batch_size
            if mode == ModeKeys.TRAIN
            else self.mngr.predict_batch_size
        )

        if pd_range is not None:
            if len(pd_range) != 2:
                raise Exception("expected len(pd_range) == 2")

            batch_size_now = pd_range[1] - pd_range[0]
            pd_vals = pd_vals[pd_range[0] : pd_range[1]]
            input_wgt = input_wgt[pd_range[0] : pd_range[1]]
            input_ids = input_ids[pd_range[0] : pd_range[1]]

            if self.mngr.is_enc_dec_clasif:
                input_evt_class = input_evt_class[pd_range[0] : pd_range[1]]

        if full_data is not None:
            if not isinstance(full_data, dict):
                raise Exception("if used, full_data should be a dict")
            full_data["pd_vals"] = pd_vals
            full_data["input_wgt"] = input_wgt
            full_data["input_df_meta"] = input_df_meta

        # if we are out of events, raise custom
        # exception to indicate an evt-loop break
        self.mngr.n_dataset_samples[mode] = len(pd_vals)

        if self.mngr.n_dataset_samples[mode] == 0:
            raise InputOutOfRange(" - pd_vals empty ...")

        dataset = tf.data.Dataset.from_tensor_slices(
            tf.convert_to_tensor(pd_vals, dtype=tf.float32)
        )

        # add weights
        dataset_weight = tf.data.Dataset.from_tensor_slices(
            tf.convert_to_tensor(input_wgt, dtype=tf.float32)
        )

        dataset_input_ids = tf.data.Dataset.from_tensor_slices(
            tf.convert_to_tensor(input_ids, dtype=tf.float32)
        )

        zips = {
            self.mngr.tag_steps: dataset,
            self.mngr.tag_wgt: dataset_weight,
            self.mngr.tag_model_input_ids: dataset_input_ids,
        }

        dataset = tf.data.Dataset.zip(zips)

        # shuffle if enabled
        if self.mngr.do_shuffle_dataset(mode):
            if not self.mngr.shuffle_dataset_use_dataset_length_as_buffer_size:
                raise NotImplementedError(
                    f"{self.mngr.shuffle_dataset_use_dataset_length_as_buffer_size=}"
                )
            else:
                shuffle_buffer_size = len(dataset)

            dataset = dataset.shuffle(
                buffer_size=shuffle_buffer_size,
                reshuffle_each_iteration=self.mngr.model_fit_do_shuffle_data_on_train_epoch_begin,
            )

        # do batching
        dataset = self.do_batching(dataset, batch_size_now, mode)

        # iterator transformation function for variable scaling
        if self.mngr.do_train_scale:
            if not hasattr(self.mngr, "train_feat_norms"):
                self.train_feat_norms = self.get_train_scales()

            tag_types = [self.mngr.tag_steps]
            for tag_type in tag_types:
                shift = tf.convert_to_tensor(
                    value=self.train_feat_norms[tag_type]["shift"],
                    dtype=tf.float32,
                    name=str_join("norms", tag_type, "shift"),
                )
                scale = tf.convert_to_tensor(
                    value=self.train_feat_norms[tag_type]["scale"],
                    dtype=tf.float32,
                    name=str_join("norms", tag_type, "scale"),
                )

                def trans(input):
                    input[tag_type] = tf.math.add(input[tag_type], shift)
                    input[tag_type] = tf.math.multiply(input[tag_type], scale)
                    return input

                dataset = dataset.map(trans)

        # map inputs and outputs for the model
        dataset_out = dataset.map(self.dataset_model_map)

        # input_df_meta

        # dataset_evt_ids = np.concatenate(
        #     [d[0][self.mngr.tag_model_input_ids] for d in dataset_out.as_numpy_iterator()]
        # )

        self.after_read_dataset(mode, dataset_out)

        return dataset_out

    def after_read_dataset(self, mode, dataset_out):
        if self.mngr.debug_inspect_data_normalisation:
            from trans_finder.flares.helpers_debugging import get_debug_output_filepath
            from trans_finder.flares.plot_list.data_normalisation import (
                make_superimposed_feature_dist_timesteps_plot,
            )

            file_name = get_debug_output_filepath(
                self.mngr,
                "inspect_data_normalisation",
                f"normed_{mode}_dist.csv",
            )
            df_values = tf.concat(
                [e[0]["dec"] for e in list(dataset_out)], axis=0
            ).numpy()
            # merge the first dimension with the time step dimension
            df_values = np.reshape(df_values, (-1, df_values.shape[-1]))
            data_distribution = pd.DataFrame(df_values).describe()
            data_distribution.to_csv(file_name, float_format="%.3f")

            file_name = get_debug_output_filepath(
                self.mngr,
                "inspect_data_normalisation",
                f"normed_{mode}_dist.pdf",
            )
            make_feature_distribution_plots(self.mngr, df_values, file_name)

            if hasattr(self.mngr, "train_feat_norms"):
                file_name = get_debug_output_filepath(
                    self.mngr, "inspect_data_normalisation", f"normalisation_shift.csv"
                )
                pd.DataFrame(self.train_feat_norms["steps"]["shift"]).to_csv(
                    file_name, float_format="%.3f"
                )

                file_name = get_debug_output_filepath(
                    self.mngr, "inspect_data_normalisation", f"normalisation_scale.csv"
                )
                pd.DataFrame(self.train_feat_norms["steps"]["scale"]).to_csv(
                    file_name, float_format="%.3f"
                )

            file_name = get_debug_output_filepath(
                self.mngr,
                "inspect_data_normalisation",
                f"normed_{mode}_d1_timesteps.pdf",
            )
            d1 = tf.concat([e[0]["dec"] for e in list(dataset_out)], axis=0).numpy()
            make_superimposed_feature_dist_timesteps_plot(self.mngr, d1, file_name)

            quants = self.load_calib_file(
                tag_dir=self.mngr.tag_model_dir,
                tag_file=self.mngr.tag_input_scalings[self.mngr.tag_steps],
            )

            file_name = get_debug_output_filepath(
                self.mngr,
                "inspect_data_normalisation",
                f"norm_{self.mngr.tag_steps}_data_quantiles.csv",
            )
            pd.DataFrame(quants).to_csv(file_name, float_format="%.3f")

    def clip_df_by_evt_type_ratios(self, df_evt_type, mode_name):
        """the flares version adds the options to set all evt types to -1, which disables the clipping"""
        max_evt_types = dict(
            (k, len(v)) for (k, v) in df_evt_type.items() if len(v) > 0
        )

        type_ratios = dict(
            (k, self.mngr.evt_type_ratios[mode_name][k])
            for (k, v) in max_evt_types.items()
            if k in self.mngr.evt_type_ratios[mode_name]
        )

        if len(type_ratios) == 0:
            return

        original_n_events = dict(
            (k, max_evt_types[k] * v) for (k, v) in type_ratios.items()
        )

        if all(v == -1 for v in type_ratios.values()):
            scaled_n_evts_1 = {k: len(v) for k, v in df_evt_type.items()}
        else:
            # for k in original_n_events.keys():
            #     if type_ratios[k] == -1:
            #         n_evts_0[k] = np.inf

            n_evts_0 = dict(original_n_events)
            min_n_evts_0 = min(
                n_evts_0.items(),
                key=lambda x: np.inf if type_ratios[x[0]] == -1 else x[1],
            )[0]

            ratio_n_evts_0 = max_evt_types[min_n_evts_0] / type_ratios[min_n_evts_0]

            scaled_n_evts_0 = dict(
                (k, ratio_n_evts_0 * v) for (k, v) in type_ratios.items()
            )

            ratio_n_evts_1 = max(
                v / max_evt_types[k] for (k, v) in scaled_n_evts_0.items()
            )

            scaled_n_evts_1 = dict(
                (k, max(1, int(floor(v / ratio_n_evts_1))))
                for (k, v) in scaled_n_evts_0.items()
            )

            for evt_type, n_evts in scaled_n_evts_1.items():
                df_evt_type[evt_type] = df_evt_type[evt_type].iloc[:n_evts]

        sum_n_evt = sum(len(v) for v in df_evt_type.values())
        for evt_type, n_evts in scaled_n_evts_1.items():
            frac_evt = 100 * n_evts / sum_n_evt

            self.log.info(
                [
                    ["c", " - final evt-class filters: "],
                    ["p", evt_type],
                    ["c", " -> n-evt-tot: "],
                    ["o", len(df_evt_type[evt_type])],
                    ["c", " -> "],
                    ["b", "{:>1.1f} %".format(frac_evt)],
                ]
            )

    def get_dir_name(self, dir_type):
        dir_output_base = os.path.join(
            self.mngr.tag_output_dir, self.mngr.tag_model_base
        )

        if dir_type == "predict_out":
            prefix = "predict"
            prefix = f"""{prefix}_{self.mngr.ts_weights_enc_dec_anom_det["tag"]}"""

            predict_out = os.path.join(
                dir_output_base, prefix, self.mngr.dir_output_predict
            )
            out = predict_out

        elif dir_type == "predict_sim_out":
            prefix = "predict"
            prefix = f"""{prefix}_{self.mngr.ts_weights_enc_dec_anom_det["tag"]}"""

            predict_out = os.path.join(
                dir_output_base, prefix, self.mngr.dir_output_predict_sim
            )
            out = predict_out

        elif dir_type == "train_in":
            train_in = os.path.join(self.mngr.tag_input_dir, self.mngr.dir_input_train)
            out = train_in

        elif dir_type == "predict_in":
            predict_in = os.path.join(
                self.mngr.tag_input_dir, self.mngr.dir_input_predict
            )
            out = predict_in

        elif dir_type == self.mngr.tag_model_dir:
            out = os.path.join(dir_output_base, self.mngr.tag_model_dir)

        elif dir_type in ["model_export", "tensor_board", "model_history"]:
            out = os.path.join(
                self.get_dir_name(self.mngr.tag_model_dir),
                dir_type,
                self.mngr.tag_model,
            )

        elif dir_type == "calib":
            postfix = f"""calib_{self.mngr.ts_weights_enc_dec_anom_det["tag"]}"""
            out = os.path.join(dir_output_base, postfix)

        elif dir_type == "calib_fits":
            predict_out = self.get_dir_name(dir_type.replace("_fits", ""))
            out = os.path.join(predict_out, "fits")

        elif dir_type == "calib_plots":
            predict_out = self.get_dir_name(dir_type.replace("_plots", ""))
            out = os.path.join(predict_out, "plots")

        elif dir_type in "calib_diagnostics":
            out = os.path.join(self.get_dir_name("calib"), f"diagnostics")

        elif dir_type in "predict_diagnostics":
            out = os.path.join(self.get_dir_name("predict_out"), f"diagnostics")

        elif dir_type == "predict_plots":
            postfix = (
                f"""predict_plots_{self.mngr.ts_weights_enc_dec_anom_det["tag"]}"""
            )
            out = os.path.join(dir_output_base, postfix)

        elif dir_type == "debug":
            out = os.path.join(dir_output_base, "debug")

        elif dir_type == "custom_realisations_input":
            out = os.path.join(
                self.mngr.tag_input_dir, self.mngr.tag_custom_realisations
            )

        else:
            raise Exception(f"unsupported option for get_dir_name({dir_type})")

        return out

    def create_out_dirs(self):
        """create or reset output directories"""

        if self.mngr.reset_tag_dir:
            self.log.info([["o", " - reset-dir "], ["b", self.mngr.tag_model_base]])

        if self.mngr.reset_calib_dir is not None:
            do_reset_calib = self.mngr.reset_calib_dir
        else:
            do_reset_calib = self.mngr.run_phase == "calib"

        if self.mngr.reset_predict_dir is not None:
            do_reset_pred = self.mngr.reset_predict_dir
        else:
            do_reset_pred = self.mngr.run_phase == "predict"

        create_dir(
            log=self.log,
            dir_name=os.path.join(self.mngr.tag_output_dir, self.mngr.tag_model_base),
            reset=self.mngr.reset_tag_dir,
            verb=True,
            keep_dirs=self.mngr.keep_dirs,
        )

        all_dirs = [
            [self.get_dir_name(self.mngr.tag_model_dir), False],
            [self.get_dir_name("predict_plots"), False],
        ]

        all_dirs += [
            [self.get_dir_name("model_export"), False],
            [self.get_dir_name("tensor_board"), False],
            [self.get_dir_name("model_history"), False],
            [self.get_dir_name("calib_fits"), do_reset_calib],
            [self.get_dir_name("calib_plots"), do_reset_calib],
            [self.get_dir_name("calib_diagnostics"), do_reset_calib],
            [self.get_dir_name("predict_out"), do_reset_pred],
            [self.get_dir_name("predict_sim_out"), do_reset_pred],
            [self.get_dir_name("predict_diagnostics"), do_reset_pred],
        ]

        for dir_now, reset in all_dirs:
            if reset and (self.mngr.in_file_filt is not None):
                self.clean_filt_files(dir_now)
                reset = False

            create_dir(log=self.log, dir_name=dir_now, reset=reset, verb=False)

        return

    def get_file_name(self, dir_type, file_name=None):
        if dir_type in ["model_history"]:
            if file_name is None:
                file_name = "history.csv"
            out = os.path.join(self.get_dir_name("model_history"), file_name)

        return out

    def load_data(self):
        """load data"""

        self.log.info(
            [
                ["g", " - loading input-data from file("],
                ["y", self.mngr.run_phase],
                ["g", ") ..."],
            ]
        )

        # setup dict if this is the first time the function is called
        input_keys = [
            self.mngr.tag_steps,
            self.mngr.tag_wgt,
            self.mngr.tag_input_df,
        ]
        for key in input_keys:
            if key not in self.mngr.input_data:
                self.mngr.input_data[key] = dict()

        if self.mngr.run_phase == "predict_plots":
            dir_name = self.get_dir_name("predict_out")

            if isinstance(dir_name, dict):
                self.mngr.input_data[self.mngr.tag_input_df][
                    self.mngr.run_phase
                ] = dict()
                for tag in dir_name.keys():
                    input_df = self.get_input_df(
                        dir_name=dir_name[tag],
                        mode_name="predict",
                    )
                    self.mngr.input_data[self.mngr.tag_input_df][self.mngr.run_phase][
                        tag
                    ] = input_df
            else:
                input_df = self.get_input_df(
                    dir_name=dir_name,
                    mode_name="predict",
                )
                self.mngr.input_data[self.mngr.tag_input_df][
                    self.mngr.run_phase
                ] = input_df

        elif self.mngr.run_phase == "create_custom_realisations":
            self.mngr.create_custom_realisations()

        elif self.mngr.run_phase == "reproduce_data_prep":
            self.mngr.reproduce_flares_data_prep()

        elif self.mngr.run_phase == "plot_diagnostics":
            self.mngr.plot_diagnostics_runphase()

        elif self.mngr.run_phase == "prep_train":
            self.mngr.data_prep.prep_train()

        elif self.mngr.run_phase == "prep_calib":
            self.mngr.data_prep.prep_calib()

        elif self.mngr.run_phase == "prep_bck":
            self.prep_bck()

        elif self.mngr.run_phase == "prep_predict":
            self.mngr.data_prep.prep_predict()

        elif self.mngr.run_phase == "prep_sim":
            self.prep_sim()

        elif self.mngr.run_phase == "inject_sim":
            self.inject_sim()

        else:
            self.prep_dataset_from_pd()
            self.extra_data_prep()

        return

    def extra_data_prep(self):
        """
        Method can be overwritten by subprojects to perform special data preparation
        tasks e.g. for a specific run_phase.
        """
        pass

    def get_train_scales(self, force_check=False):
        """load scaling factors for input variables

        Parameters
        -----------
        force_check : bool
            force force_check if the file can be loaded
        """

        self.log.info([["p", " - get-train-scales()"]])

        norms = dict()

        tag_types = [self.mngr.tag_steps]
        for tag_type in tag_types:
            is_data = tag_type == self.mngr.tag_steps
            tag_file = self.mngr.tag_input_scalings[tag_type]

            try:
                quants = self.load_calib_file(
                    tag_dir=self.mngr.tag_model_dir,
                    tag_file=tag_file,
                )
                if len(quants) == 0:
                    raise Exception("problem with load-calib-file()")

            except Exception as e:
                # if the file dow not exist and this is the first try, create it
                if force_check:
                    raise Exception("set_train_scales() problem", e)

                self.log.info(
                    [
                        ["r", " - will run set_train_scales() ..."],
                    ]
                )
                raise DataScalingMissing

            # once the file has been loaded, store the scaling parameters
            shift, scale = [], []
            for qunat in quants:
                base_dim = np.ones(self.mngr.n_steps_tot) if is_data else 1

                # shift += [base_dim * (-1 * qunat[0])]
                shift += [base_dim * (-1 * qunat[1])]
                scale += [base_dim * 2 / (qunat[2] - qunat[0])]

                # shift += [np.ones(self.mngr.n_steps_tot) * (-1 * qunat[0])]
                # if qunat[0] < 0:
                #     scale += [np.ones(self.mngr.n_steps_tot) / (qunat[2] - qunat[0])]
                # else:
                #     scale += [np.ones(self.mngr.n_steps_tot) / qunat[2]]

            norms[tag_type] = {
                "shift": np.asarray(shift).T,
                "scale": np.asarray(scale).T,
            }

        # print('norms \n',norms)
        # print('norms \n',norms[self.mngr.tag_steps]['shift'][0],'\n', norms[self.mngr.tag_steps]['scale'][0])
        # print('norms \n',norms[self.mngr.tag_meta_bins]['shift'],'\n', norms[self.mngr.tag_meta_bins]['scale'])
        # assert 0

        return norms

    def do_batching(self, tf_dataset, batch_size, mode):
        """splits a tensorflow dataset into batches and returns it"""

        tf_dataset_batched = tf_dataset.batch(
            batch_size,
            drop_remainder=self.mngr.batching_do_drop_remainder,
        )

        if self.mngr.batching_do_drop_remainder:
            count_before_batching = len(tf_dataset)
            count_after_batching = len(tf_dataset_batched) * batch_size

            self.log.info(
                [
                    ["b", " - do_batching(" + mode + "): dropped "],
                    ["y", count_before_batching - count_after_batching],
                    ["b", " of "],
                    ["y", count_before_batching],
                    ["b", " samples to ensure static train batch size "],
                ]
            )
        return tf_dataset_batched

    def apply_sample_cuts(self, df_in, is_verb=True):
        """aplly optional sample cuts on inputs

        Parameters
        -----------
        df_in : pd.Dataframe()
            input data
        is_verb : bool
            flag to get verbosity for logging
        """

        df_out = df_in

        # example cut scheme for debugging (choose all even rows)
        debug_cut = False
        # debug_cut = True
        if debug_cut:
            indx = [i for i in range(len(df_out)) if (i % 2 == 0)]
            df_out = df_out.iloc[indx]

        # with pd.option_context('display.max_rows', None, 'display.max_columns', None):
        #     if len(df_out) > 0:
        #         print(df_out.iloc[0])

        # clip by the allowed ranges of the metadata columns
        cut = np.ones(len(df_out)).astype(bool)
        for key in self.mngr.meta_bins.keys():
            if self.mngr.meta_bin_col_params[key]["type"] == "list":
                all_bins = np.asarray(
                    self.mngr.meta_bin_col_params[key]["values"]
                ).flatten()
                cut_str = np.zeros(len(df_out)).astype(bool)
                for bin_now in all_bins:
                    cut_str = cut_str | (df_out[key] == bin_now)
                cut = cut & cut_str
            else:
                cut = cut & df_out[key].between(
                    self.mngr.meta_bins[key][0],
                    self.mngr.meta_bins[key][-1],
                    inclusive=True,
                )
        df_out = df_out.loc[cut].reset_index(drop=True)

        # reset the df.index and filter the data/weights arrays
        if len(df_out) < len(df_in):
            self.log.info(
                [
                    ["b", " - apply-sample-cuts(): clipping by meta_bins: "],
                    ["y", len(df_in)],
                    ["b", "  -->  survived: "],
                    ["p", len(df_out)],
                ]
            )

        # apply all cut functions
        for cut_func in self.mngr.input_data_cut_funcs:
            df_now = cut_func(df_in=df_out)

            if is_verb and len(df_now) < len(df_out):
                self.log.info(
                    [
                        ["g", " - apply-sample-cuts(): cut-func(): "],
                        ["y", " # evt from/to: "],
                        ["p", len(df_in)],
                        ["y", " --> "],
                        ["p", len(df_now)],
                    ]
                )

            df_out = df_now

        return df_out

    def prep_dataset_from_pd(self):
        """parse the pandas dataframe and transform it into the structure of the model input dictionaries
        with separate dictionaries for train and validation splits, if the run_phase is train
        """

        self.log.info([["r", " - prep-dataset-from-pd()"]])

        # set input directory and dataset mode_keys based on the run_phase
        if self.mngr.run_phase in [
            "train",
            "bgmm_fit",
            "debug_keras_model",
            "inspect_inputs",
        ]:
            mode_keys = [ModeKeys.TRAIN, ModeKeys.EVAL]
            dir_name = self.get_dir_name("train_in")

        else:
            mode_keys = [ModeKeys.PREDICT]
            dir_name = self.get_dir_name("predict_in")

        # read input files from disk and compile dataframe
        df_full = self.get_input_df(dir_name=dir_name)

        # raise an exception if there is no data but we require data
        if len(df_full) == 0 and ModeKeys.PREDICT not in mode_keys:
            raise Exception("empty input dataset for:", dir_name)

        for trans_func in self.mngr.transform_input_feature_funcs:
            df_full = trans_func(df_in=df_full)

        # apply data transformations
        inputs = self.prep_inputs(df_full)

        # optional sort inputs
        inputs = self.sort_inputs_for_calib(inputs)

        self.split_and_set_inputs(inputs, mode_keys)

    def get_input_df(
        self,
        dir_name,
        mode_name=None,
        is_accepted_input_file=None,
        is_verb=True,
    ):
        """load data as pd.DataFrame()

        Parameters
        -----------
        dir_name : str
            path for directory containing the data
        mode : ModeKeys
            one of [ModeKeys.TRAIN, ModeKeys.EVAL, ModeKeys.PREDICT]
        is_accepted_input_file : func
            checks by filtering input file names if an input is accepted
        is_verb: bool
            verbosity flag for logging
        """

        if is_verb:
            self.log.info(
                [
                    ["c", " - get-input-df("],
                    ["b", dir_name.replace(self.mngr.tag_input_dir, "")],
                    ["g", ")"],
                ]
            )

        if (
            self.mngr.prep_train_scales
            and "input_norm" in self.mngr.evt_type_ratios.keys()
        ):
            mode_name = "input_norm"

        if mode_name is None:
            mode_name = self.mngr.run_phase

        if is_accepted_input_file is None:
            is_accepted_input_file = self.mngr.is_accepted_input_file

        if is_accepted_input_file is None:

            def is_accepted_input_file(file_name):
                return all((patt in file_name) for patt in self.mngr.in_file_patt)

        n_files = 0
        try:
            # get all files matching the filter patterns
            file_list = sorted(os.listdir(dir_name))
            file_list = [x for x in file_list if is_accepted_input_file(x)]
            n_files = len(file_list)

            if (self.mngr.in_file_filt is not None) and (
                n_files < self.mngr.in_file_filt
            ):
                raise Exception
        except Exception:
            raise Exception(
                f"problem with in_file_filt({dir_name}) setup {self.mngr.in_file_filt}"
                + f"/{n_files} -> in_file_filt too large ... ",
            )

        # get the number of events to accept for each event class
        evt_type_ratio = dict(
            (t, 1) for t in self.mngr.evt_type_ratios[mode_name].keys()
        )
        for evt_type in self.mngr.evt_type_ratios[mode_name].keys():
            if evt_type in self.mngr.evt_type_ratios[mode_name].keys():
                evt_type_ratio[evt_type] = self.mngr.evt_type_ratios[mode_name][
                    evt_type
                ]

        sum_ratios = np.sum(list(evt_type_ratio.values()))
        for evt_type in self.mngr.evt_type_ratios[mode_name].keys():
            evt_type_ratio[evt_type] /= sum_ratios

        n_evt_max = self.get_max_evt()

        if self.mngr.run_phase == "predict":
            self.log.info(
                [["g", " - loading all data without any event class cuts for predict"]]
            )
            n_evt_max_type = {
                None: n_evt_max,
            }

            is_done = {None: False}

            df_evt_type = {
                None: pd.DataFrame(),
            }
        else:
            n_evt_max_type = dict(
                (k, int(ceil(n_evt_max * v))) for (k, v) in evt_type_ratio.items()
            )

            is_done = dict(
                (t, False) for t in self.mngr.evt_type_ratios[mode_name].keys()
            )

            df_evt_type = dict(
                (t, pd.DataFrame()) for t in self.mngr.evt_type_ratios[mode_name].keys()
            )

        try:
            # add input files until reaching the maximal number of events needed
            self.load_data_files(
                dir_name=dir_name,
                file_list=file_list,
                n_evt_max_type=n_evt_max_type,
                is_done=is_done,
                df_evt_type=df_evt_type,
                is_verb=is_verb,
            )

            self.clip_df_by_evt_type_ratios(
                df_evt_type=df_evt_type,
                mode_name=mode_name,
            )

        except Exception as e:
            sub_dirs = [d for d in dir_name.split("/") if d != ""]
            if len(sub_dirs) > 0 and sub_dirs[-1] == "None":
                return pd.DataFrame()
            else:
                raise e

        # merge all together
        df_out = pd.DataFrame()
        for evt_type in df_evt_type.keys():
            df_out = df_out.append(df_evt_type[evt_type], ignore_index=False, sort=True)

            # try to prevent some memory overloads, by forcing garbage collection
            df_evt_type[evt_type] = None
            gc.collect()

        if len(df_out) == 0:
            raise Exception(" - no input events accepted ...")

        # recover the original order and reset the index (there will be missing entries)
        df_out = df_out.sort_index().reset_index(drop=True)

        # # shuffle the event classes
        # df_out = df_out.sample(frac=1)

        dir_name_print = dir_name.replace(self.mngr.tag_input_dir, "")

        if is_verb:
            self.log.info([["g", " - "], ["p", dir_name_print]])
            self.log.info(
                [
                    ["y", " - all / input files: "],
                    ["o", len(file_list), "/"],
                    ["g", " ---> accepted events: "],
                    [("p" if len(df_out) > 0 else "wr"), len(df_out)],
                ]
            )

        return df_out

    def load_data_files(
        self,
        dir_name,
        file_list,
        n_evt_max_type,
        is_done,
        df_evt_type,
        is_verb=False,
    ):
        """load data from input files

        Parameters
        -----------
        dir_name : str
            path for directory containing the data
        file_list : list
            list of input files within dir_name
        is_done : dict
            dictionary for keeping track of number of loaded events by their type
        n_evt_max_type : dict
            dictionary for keeping track of maximum needed number of events by their type
        df_evt_type : dict
            pd.DataFrame objects for each event typr
        is_verb: bool
            verbosity flag for logging
        """

        def input_msg(n_file, evt_type):
            if not is_verb:
                return
            if n_evt_max_type[evt_type] == 0:
                evt_frac = 0
            else:
                evt_frac = 100 * len(df_evt_type[evt_type]) / n_evt_max_type[evt_type]

            self.log.info(
                [
                    ["c", " - input file "],
                    ["b", n_file + 1, "/", n_files],
                    ["c", " -> evt-class: "],
                    ["p", evt_type],
                    ["c", " -> n-evt-needed: "],
                    ["y", n_evt_max_type[evt_type]],
                    ["c", " -> n-evt-tot: "],
                    ["o", len(df_evt_type[evt_type])],
                    ["c", " -> "],
                    ["b", "{:>1.1f} %".format(evt_frac)],
                ]
            )

        n_files = len(file_list)

        for n_file_now in range(n_files):
            # filter inputs if splitting the sample
            if self.mngr.in_file_filt is not None:
                if n_file_now % self.mngr.in_file_filt != self.mngr.in_file_filt_now:
                    continue

            file_name = file_list[n_file_now]
            file_path = os.path.join(dir_name, file_name)

            # read in the data
            df_in = pd.read_hdf(file_path, self.mngr.tag_hdf)

            # has_evt_type = self.mngr.tag_evt_type in df_in.columns
            # if self.mngr.run_phase == "predict" and has_evt_type:
            #     df_in = df_in.drop(self.mngr.tag_evt_type, axis=1)

            if all(list(is_done.values())):
                break

            df_in = self.apply_user_sample_cuts(df_in=df_in)

            for evt_type in df_evt_type.keys():
                if is_done[evt_type]:
                    continue

                if self.mngr.run_phase == "predict":
                    df_now = df_in
                    if evt_type is not None:
                        continue

                else:
                    # filter on event classes for this mode
                    evt_type_filt = f'{self.mngr.tag_evt_type} == "{evt_type}"'
                    df_now = df_in[df_in.eval(evt_type_filt)]
                    # df_now = df_in.copy()

                    # add event classes before applying sample cuts
                    if self.mngr.is_enc_dec_clasif:
                        evt_types = df_now[self.mngr.tag_evt_type]

                        input_evt_class = evt_types.apply(
                            lambda x: self.mngr.evt_type_to_cat[x]
                        )
                        df_now.loc[:, self.mngr.tag_evt_cats] = input_evt_class

                        is_bck_cat = input_evt_class.isin(self.mngr.bck_cats)
                        df_now.loc[:, self.mngr.tag_is_bck_cat] = is_bck_cat

                # apply optional sample cuts
                df_now = self.apply_sample_cuts(df_in=df_now)

                n_evt_now = len(df_now)
                if n_evt_now == 0:
                    continue

                df_evt_type[evt_type] = df_evt_type[evt_type].append(
                    df_now,
                    ignore_index=True,
                    sort=False,
                )

                # try to prevent some memory overloads, by forcing garbage collection
                del df_now
                gc.collect()

                is_done[evt_type] = (
                    len(df_evt_type[evt_type]) >= n_evt_max_type[evt_type]
                )
                if is_done[evt_type]:
                    df_evt_type[evt_type] = df_evt_type[evt_type].iloc[
                        : n_evt_max_type[evt_type]
                    ]

                if evt_type is None:
                    break

            for evt_type in df_evt_type.keys():
                if n_file_now % 5 == 0 or is_done[evt_type]:
                    input_msg(n_file=n_file_now, evt_type=evt_type)

            if n_files > 1 and n_file_now == n_files - 1:
                self.log.info([["c", " - final count of input files:"]])
                for evt_type in df_evt_type.keys():
                    input_msg(n_file=n_file_now, evt_type=evt_type)

    #
    # def get_max_evt(self, tag, *args, **kwargs):
    #     """wrapper function for the maximal number of events to load, which may be overriden

    #         Parameters
    #         ----------
    #         tag : str
    #             a key for the general self.n_max_evt dict

    #         Returns
    #         -------
    #         int
    #             the maximal number of events
    #     """

    #     max_evt = self.n_max_evt[tag]

    #     return max_evt

    def get_max_evt(self):
        n_evt_max = np.iinfo("int32").max

        # for self.mngr.prep_train_scales we keep the full dataset in order to get
        # the full range of values. For predict_plots run_phase we also want the
        # inclusive sample for full representation of the distribution
        if self.mngr.prep_train_scales:
            if self.mngr.max_evt["train_scales"] > 0:
                n_evt_max = self.mngr.max_evt["train_scales"]
        else:
            if self.mngr.max_evt[self.mngr.run_phase] > 0:
                n_evt_max = self.mngr.max_evt[self.mngr.run_phase]

        n_evt_max = int(n_evt_max)

        self.log.info(
            [
                ["p", " - get-evt-slice-indices()"],
                ["g", " --> range: "],
                ["y", [0, n_evt_max]],
            ]
        )

        return n_evt_max

    def get_perdict_data(self, n_predict, mode, batch_size=None):
        """get predictions for a given event range

        Parameters
        -----------
        n_predict : int
            the number of events to get
        mode : ModeKeys
            one of [ModeKeys.TRAIN, ModeKeys.EVAL, ModeKeys.PREDICT]
        batch_size : int
            the batch size
        """

        if batch_size is None:
            batch_size = (
                self.mngr.predict_batch_size
                if mode == ModeKeys.PREDICT
                else self.mngr.train_batch_size
            )

        pd_range = [n_predict + 1, n_predict + batch_size + 1]

        pd_range[1] = min(
            pd_range[1], len(self.mngr.input_data[self.mngr.tag_input_df][mode])
        )

        # if we are out of events, raise custom exception to indicate an evt-loop break
        if (pd_range[1] - pd_range[0]) <= 0:
            raise InputOutOfRange("problem with pd_range:", pd_range)

        # get the data from the estimator
        dataset = self.read_dataset(mode=mode, pd_range=pd_range)

        if self.mngr.tag_model not in self.mngr.model.keys():
            self.mngr.train.load_model(dataset)
            self.mngr.model[self.mngr.tag_model] = self.mngr.model[self.mngr.tag_model]

        # iterate over the dataset, expecting a tuple such as (inputs, outputs, weights)
        # where each element is a dict
        inputs = dict()
        for data_0 in list(dataset.as_numpy_iterator()):
            for data_1 in data_0:
                if not isinstance(data_1, dict):
                    raise Exception("unsupported data type", data_1)
                inputs.update(data_1)

        # match the full input data with the corresponding event range
        pd_df = self.mngr.input_data[self.mngr.tag_input_df][mode].iloc[
            pd_range[0] : pd_range[1]
        ]
        pd_df = pd_df.reset_index(drop=True)

        # message
        n_evt_now = len(inputs[self.mngr.tag_enc])
        n_evt_tot = len(self.mngr.input_data[self.mngr.tag_input_df][mode])
        n_evt_frac = int((100 * (pd_range[1] - n_evt_now)) // n_evt_tot)

        self.log.info(
            [
                ["c", " - inputs - "],
                ["y", n_evt_now, pd_range],
                ["c", " (tot: "],
                ["y", n_evt_tot],
                ["c", " -> "],
                ["y", str(n_evt_frac) + "%"],
                ["c", ") events ..."],
            ]
        )

        out = {
            "inputs": inputs,
            "df": pd_df,
        }

        return out

    def prep_inputs(self, df_in):
        """derive the input data according to user-defined functions

        Parameters
        ----------
        df_in : pd.Dataframe()
            input Dataframe

        Returns
        -------
        dict
            the input data and corresponding weights
        """

        # optional operations on the dataframe
        df_in = self.before_prep_inputs(df_in)

        # reshape the light curve df columns into an array
        data_vals = self.get_step_values(df_in)

        # before any smearing, add a general event acceptance column
        self.add_evt_cut_mask(df_in, data_vals)

        # apply a transformation to the inputs (eg smearing)
        data_vals = self.transform_inputs(df=df_in, inputs=data_vals, keep_orig=False)

        # weighting scheme
        wgt_vals = self.weight_train_func(df_in)

        # filter the subset of columns needed for later
        cols_to_keep = []
        if self.mngr.ts_compare is not None:
            cols_to_keep += [self.mngr.ts_compare]

        extra_cols = [
            self.mngr.meta_bin_cols,
            self.mngr.clas_cols_to_keep,
            self.mngr.aux_cols_to_keep,
        ]
        extra_cols += [self.mngr.get_phase_cols_to_keep(df_in)]

        if self.mngr.run_phase in ["predict_sim"]:
            # extra columns only for simulations (used to identify individual sims)
            # Currently only relevant for VERITAS subproject
            extra_cols += [self.unique_sim_index]
            extra_cols += [get_columns_wildcard(df_in, f"{self.sim_timestep_col}*")]

        for cols in extra_cols:
            cols_to_keep += [c for c in cols if c not in cols_to_keep]

        cols_to_keep = [c for c in cols_to_keep if c in df_in.columns]
        df_out = df_in[cols_to_keep]
        # df_out = df_out.copy()

        # filter data
        inputs = {
            self.mngr.tag_input_df: df_out,
            self.mngr.tag_steps: data_vals,
            self.mngr.tag_wgt: wgt_vals,
        }

        return inputs

    def sort_inputs_for_calib(self, inputs):
        run_phases = ["calib", "predict", "predict_sim"]
        if self.mngr.is_run_wise_sorted_inputs and self.mngr.run_phase in run_phases:
            # relevant columns required for grouping same events
            run_id_col = self.calib_roi_neighbours_config["run_id_col"]
            run_time_col = self.calib_roi_neighbours_config["run_time_col"]
            cut_cols = [run_time_col, run_id_col]

            # sort array by cut_cols
            cut_data = inputs[self.mngr.tag_input_df][cut_cols].to_numpy()
            args_sort = np.lexsort(cut_data.T, axis=0)

            inputs[self.mngr.tag_steps] = inputs[self.mngr.tag_steps][args_sort]
            inputs[self.mngr.tag_wgt] = inputs[self.mngr.tag_wgt][args_sort]
            inputs[self.mngr.tag_input_df] = inputs[self.mngr.tag_input_df].iloc[
                args_sort
            ]

            inputs[self.mngr.tag_input_df].reset_index(inplace=True, drop=True)

        return inputs

    def get_n_file_filt(self, in_file_filt_now=None):
        if in_file_filt_now is None:
            in_file_filt_now = self.mngr.in_file_filt_now

        return self.mngr.tag_filt_file + str("%05d" % in_file_filt_now) + "_"

    def clean_filt_files(self, dir_now):
        if not os.path.isdir(dir_now):
            return

        dir_files = sorted(os.listdir(dir_now))

        # get all files from the same in_file_filt_now or
        # that do not have tag_filt_file
        files_to_del = [
            f
            for f in dir_files
            if ((self.mngr.tag_filt_file not in f) or (self.get_n_file_filt() in f))
        ]

        # keep all files with pattern if not above this in_file_filt
        # does not protect if in_file_filt had been changed before ....
        filt_files = [f for f in dir_files if self.mngr.tag_filt_file in f]
        try:
            for filt_file_now in filt_files:
                filt_index = filt_file_now.split(self.mngr.tag_filt_file)
                if len(filt_index) < 2:
                    raise

                filt_index = filt_index[1].split("_")
                if len(filt_index) < 2:
                    raise

                filt_index = int(filt_index[0])
                if filt_index >= self.mngr.in_file_filt:
                    files_to_del += [filt_file_now]

        except Exception:
            raise Exception(
                f"""problem with filter setup:
                filt_files({filt_files}),
                filt_file_now({filt_file_now}),
                filt_index({filt_index})"""
            )

        self.log.info(
            [
                [
                    "g",
                    " - clean-filt(",
                    dir_now.replace(self.mngr.tag_input_dir, ""),
                    "): ",
                ],
                ["y", files_to_del],
            ]
        )
        for file_now in files_to_del:
            file_to_del = os.path.join(dir_now, file_now)

            if os.path.exists(file_to_del):
                try:
                    os.remove(file_to_del)
                except Exception as e:
                    shutil.rmtree(file_to_del)
                    if os.path.exists(file_to_del):
                        raise e

        return

    def save_calib_file(self, tag_dir, tag_file, data, tag_prefix=None):
        """save file with calibration info

        Parameters
        -----------
        tag_dir : str
            tag for self.get_dir_name() directory path
        tag_file: str
            tag for the name of the stored file
        tag_prefix : str
            optional prefix for the file name
        """

        if tag_prefix is None:
            tag_prefix = "calib"

        out_file = os.path.join(
            self.get_dir_name(tag_dir), f"{tag_prefix}_{tag_file}.joblib"
        )
        joblib.dump(data, out_file)

        self.log.info(
            [
                ["c", f" --- exported {tag_prefix} to "],
                ["g", out_file.replace(self.mngr.tag_base_output_dir, "")],
            ]
        )
        return

    def load_calib_file(self, tag_dir, tag_file=None, tag_prefix=None, is_verb=True):
        """load file with calibration info

        Parameters
        -----------
        tag_dir : str
            tag for self.get_dir_name() directory path
        tag_file : str
            tag for the name of the stored file
        tag_prefix : str
            optional prefix for the file name
        is_verb : bool
            flag to get verbosity for logging
        """

        if tag_prefix is None:
            tag_prefix = "calib"

        tag_name = "" if tag_file is None else str(tag_file)
        type_tag = f"{tag_prefix}_{tag_name}.joblib"
        load_file = os.path.join(self.get_dir_name(tag_dir), type_tag)

        try:
            output = joblib.load(load_file)
        except Exception as e:
            raise CalibFileMissing(
                " - could not load: ",
                load_file.replace(self.mngr.tag_base_output_dir, ""),
                e,
            )

        if is_verb:
            self.log.info(
                [
                    ["c", f" --- imported {tag_prefix} from: "],
                    ["g", load_file.replace(self.mngr.tag_base_output_dir, "")],
                ]
            )
        return output
