from __future__ import annotations

import os
from collections import Counter
from pathlib import Path
from typing import TYPE_CHECKING, Dict, List, Optional

import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from tqdm import tqdm

from trans_finder.flares.core.types import EventClass
from trans_finder.flares.data_prep_helpers.channel_padding import (
    create_channel_padding_infer,
    create_channel_padding_train,
)
from trans_finder.flares.data_prep_helpers.fluctuation import (
    Fluctuation,
    FluctuationGenerator,
)
from trans_finder.flares.data_prep_helpers.long_term_state import LongTermStateGenerator
from trans_finder.flares.data_prep_helpers.physical_flare import PhysicalFlare
from trans_finder.flares.data_prep_helpers.realisation import Realisation
from trans_finder.flares.data_prep_helpers.recreate_aggregated_data import (
    recreate_aggregated_data_from_output_df,
)
from trans_finder.flares.data_prep_helpers.sliding_window_shuffling import (
    do_no_shuffling,
    do_step_based_shuffling,
)
from trans_finder.flares.data_prep_helpers.temporal_injection import TemporalInjection
from trans_finder.flares.helpers_debugging import clear_directory
from trans_finder.flares.plot_list.data_prep_event import (
    do_debug_plot_of_data_prep_events,
)
from trans_finder.flares.plot_list.lightcurve import plot_lightcurve
from trans_finder.flares.plot_list.scenario_highlights import (
    do_plot_scenario_highlights,
)
from trans_finder.flares.utilities.df_batched_hdf_writer import (
    DataFrameBatchedHDFWriter,
)

if TYPE_CHECKING:
    from trans_finder.flares.manager import Manager
    from trans_finder.flares.log_parser import LogParser

from itertools import islice


def batch(iterable, size):
    iterator = iter(iterable)
    for first in iterator:  # stops when iterator is depleted
        yield tuple(islice(iterator, size - 1, size)) + (first,)


class DataPrep:
    tag = ["b", f" - flares.{Path(__file__).stem} - "]
    data_prep_output_file_id = 0

    def __init__(self, mngr):
        self.log: LogParser = mngr.log
        self.mngr: Manager = mngr

    def set_aggregated_data_conf(self, aggregated_data):
        self.set_data_prep_conf()

        (
            self.output_data_col_names,
            self.output_metadata_col_names,
            self.channel_column_names,
        ) = self.get_output_df_columns(aggregated_data)

    def set_data_prep_conf(self):
        self.data_prep_n_steps_ref_mean_and_std = int(self.mngr.n_steps_tot)

        self.n_steps_shuffling_window = int(
            self.mngr.n_steps_tot * self.mngr.data_prep_shuffling_sliding_window_scale
        )

        self.n_steps_feature_imputation_window = int(
            self.mngr.n_steps_tot * self.mngr.data_prep_imputation_sliding_window_scale
        )

        self.old_signal_threshold = -(
            self.data_prep_analysis_min_time_bin_width
            * self.mngr.n_steps_dec
            * self.mngr.data_prep_signal_to_background_time_threshold
        )

    # @profile
    def prep_train(self):
        """
        main function of the data_prep stage, combining and transforming raw input light curves into a dataset for
        training and calibration
        """
        self.log.info([self.tag, ["p", "prep-train()"]])

        data_prep_output_dir = os.path.join(
            self.mngr.tag_input_dir, self.mngr.dir_output_prep_train
        )

        # clear output directory
        clear_directory(self.mngr, "data_prep")
        clear_directory(self.mngr, data_prep_output_dir)

        # load input data for each channel, ensuring an ordering in time
        input_data = self.flares_load_input_data(self.mngr.dir_input_data_prep_train)

        # rebin channel data as configured and average high density data points
        aggregated_data = self.data_aggregation(input_data)

        self.set_aggregated_data_conf(aggregated_data)

        df_batched_hdf_writer = DataFrameBatchedHDFWriter(
            self.mngr, data_prep_output_dir
        )

        # create randomized realisations of the input light curves in the structure of the network and write them to disk
        # self.flare_generator = PhysicalFlareGenerator(mngr=self.mngr)
        self.fluctuation_generator = FluctuationGenerator()
        self.long_term_state_generator = LongTermStateGenerator()

        # realisations = self.generate_random_realisations(
        self.parallel_generate_realisations(
            aggregated_data,
            df_batched_hdf_writer,
            n_background=self.mngr.data_prep_train_n_realisations["background"],
            n_fluctuation=self.mngr.data_prep_train_n_realisations["fluctuation"],
        )
        # realisations = self.generate_random_realisations_by_type_ratios(aggregated_data)

        df_batched_hdf_writer.flush()

        # logging and analysis
        # self.data_prep_log_result_summary(realisations)

    def log_background_dataset_statistics(self, realisations):
        # events by event class
        counts_by_type = pd.DataFrame(
            [r.n_events_per_event_class for r in realisations]
        ).sum()

        target = "bck_pure"

        excess = counts_by_type.drop(target).sum() / counts_by_type.sum()

        self.log.info(
            [
                self.tag,
                [
                    "b",
                    f"Percentage of signal event classes with target being pure background: {excess} )",
                ],
            ]
        )

    def produce_prep_train_realisation(
        self,
        aggregated_data,
        realisation_id,
        realisation_seed,
        background_mode: bool,
        plot_output_dir,
    ):
        if background_mode:
            realisation = self.create_realisation(
                aggregated_data,
                shuffling_strategy="step_based",
                do_randomize_temporal_offsets=True,
                do_shift_input_data_in_time=True,
                do_smear_uncertainty=True,
                add_fluctuations=False,
                add_flares=False,
                add_background_states=True,
                realisation_id=realisation_id,
                realisation_seed=realisation_seed,
            )

            network_df = self.write_realisation(
                realisation,
                plot_output_dir=plot_output_dir,
                plot_uncertainty=False,
            )
        else:
            realisation = self.create_realisation(
                aggregated_data,
                shuffling_strategy="step_based",
                do_randomize_temporal_offsets=True,
                do_shift_input_data_in_time=True,
                do_smear_uncertainty=True,
                add_fluctuations=True,
                add_flares=False,
                add_background_states=True,
                realisation_id=realisation_id,
                realisation_seed=realisation_seed,
            )
            network_df = self.write_realisation(
                realisation,
                plot_output_dir=plot_output_dir,
                plot_uncertainty=False,
            )

        return realisation, network_df

    def parallel_generate_realisations(
        self,
        aggregated_data,
        df_batched_hdf_writer,
        n_background: int,
        n_fluctuation: int,
    ):
        realisations = []
        plot_output_dir = (
            "prep_train"
            if self.mngr.debug_data_prep_plot_reconstructed_lightcurves
            else None
        )

        realisation_id = 0
        realisation_seed = 0

        args_list = []
        for _ in range(n_background):
            realisation_id += 1
            realisation_seed += 1
            args_list.append(
                (
                    aggregated_data,
                    realisation_id,
                    realisation_seed,
                    True,
                    plot_output_dir,
                )
            )

        for _ in range(n_fluctuation):
            realisation_id += 1
            realisation_seed += 1
            args_list.append(
                (
                    aggregated_data,
                    realisation_id,
                    realisation_seed,
                    False,
                    plot_output_dir,
                )
            )

        def worker(*args):
            return self.produce_prep_train_realisation(*args)

        with Parallel(n_jobs=self.mngr.parallel_n_jobs, verbose=10) as parallel:
            batch_size = 30
            for i in range(0, len(args_list), batch_size):
                batch = args_list[i : i + batch_size]

                batch_results = parallel(delayed(worker)(*args) for args in batch)

                for realisation, network_df in batch_results:
                    df_batched_hdf_writer.add(network_df)

                    if self.mngr.run_phase == "prep_train":
                        df_batched_hdf_writer.add_realisation_metadata(realisation)

                    del realisation
                    del network_df

    def generate_random_realisations(
        self, aggregated_data, df_batched_hdf_writer, n_background, n_fluctuation
    ):
        realisation_id = 0
        realisation_seed = 0
        realisations = []

        progress_bar = tqdm(
            total=n_background + n_fluctuation,
            unit="realisation",
            position=0,
            leave=True,
        )

        plot_output_dir = (
            "prep_train"
            if self.mngr.debug_data_prep_plot_reconstructed_lightcurves
            else None
        )

        for _ in range(n_background):
            realisation_id += 1
            realisation_seed += 1

            realisation = self.create_realisation(
                aggregated_data,
                shuffling_strategy="step_based",
                do_randomize_temporal_offsets=True,
                do_shift_input_data_in_time=True,
                do_smear_uncertainty=True,
                add_fluctuations=False,
                add_flares=False,
                add_background_states=True,
                realisation_id=realisation_id,
                realisation_seed=realisation_seed,
            )

            network_df = self.write_realisation(
                realisation, plot_output_dir=plot_output_dir
            )
            df_batched_hdf_writer.add(network_df)

            if self.mngr.run_phase == "prep_train":
                df_batched_hdf_writer.add_realisation_metadata(realisation)

            realisations.append(realisation)

            progress_bar.update()

        if n_background > 0:
            self.log_background_dataset_statistics(realisations)

        for _ in range(n_fluctuation):
            realisation_id += 1
            realisation_seed += 1

            realisation = self.create_realisation(
                aggregated_data,
                shuffling_strategy="step_based",
                do_randomize_temporal_offsets=True,
                do_shift_input_data_in_time=True,
                do_smear_uncertainty=True,
                add_fluctuations=True,
                add_flares=False,
                add_background_states=True,
                realisation_id=realisation_id,
                realisation_seed=realisation_seed,
            )
            network_df = self.write_realisation(
                realisation, plot_output_dir=plot_output_dir
            )
            df_batched_hdf_writer.add(network_df)

            if self.mngr.run_phase == "prep_train":
                df_batched_hdf_writer.add_realisation_metadata(realisation)

            realisations.append(realisation)

            progress_bar.update()

        return realisations

    # @profile
    def create_realisation(
        self,
        aggregated_data,
        shuffling_strategy: Optional[str],
        do_randomize_temporal_offsets: bool,
        do_shift_input_data_in_time: bool,
        add_fluctuations: bool,
        add_flares: bool,
        add_background_states: bool,
        realisation_id,
        realisation_seed,
        do_smear_uncertainty: bool = False,
        channel_padding=None,
    ) -> Realisation:
        """:returns a dict containing the number of generated events per event class"""
        channel_tags = self.mngr.data_prep_channel_tags

        realisation = Realisation(
            mngr=self.mngr,
            channel_tags=channel_tags,
            realisation_id=realisation_id,
            has_randomized_real_times=do_randomize_temporal_offsets,
            has_shifted_real_times=do_shift_input_data_in_time,
            shuffling_strategy=shuffling_strategy,
            input_data=aggregated_data,
            realisation_seed=realisation_seed,
        )

        # randomize temporal offsets
        if do_randomize_temporal_offsets:
            realisation.randomize_temporal_offsets(mngr=self.mngr)

        # apply temporal shifts
        if do_shift_input_data_in_time:
            realisation.shift_real_times(mngr=self.mngr)

        if do_smear_uncertainty:
            realisation.smear_values_by_uncertainty(self.mngr)
        else:
            realisation.use_copy_of_input_data()

        realisation.replace_nan_feature_values(self.mngr)

        # add padding data points to the beginning of channels with too few data points
        if channel_padding is None:
            channel_padding = create_channel_padding_train(
                self.mngr, realisation.input_data
            )

        if channel_padding is not False:
            realisation.add_channel_padding(channel_padding)

        # sliding window shuffling
        if shuffling_strategy == "step_based":
            do_step_based_shuffling(self.mngr, realisation)

        elif shuffling_strategy is None:
            do_no_shuffling(self.mngr, realisation)

        else:
            raise NotImplementedError(f"Unknown shuffle type, {shuffling_strategy}")

        realisation.assert_length_of_index_slices()

        # keep the previous n-1 time steps the same for consecutive events:
        realisation.correlate_time_window_steps_individually_per_channel()

        realisation.assert_length_of_index_slices()

        if add_background_states:
            realisation.apply_global_flat_plateaus(self.mngr)

            if len(self.mngr.data_prep_long_term_state_presets.keys()) > 0:
                realisation.add_long_term_states(
                    self.mngr, self.long_term_state_generator
                )

        # add fluctuations
        if add_fluctuations and len(self.mngr.data_prep_fluctuation_presets.keys()) > 0:
            realisation.add_short_term_states(self.mngr, self.fluctuation_generator)

        # add physical flares
        if add_flares:
            raise NotImplementedError()

        return realisation

    # @profile
    def write_realisation(
        self,
        realisation: Realisation,
        fixed_evt_class=None,
        yet_desired_evt_classes=None,
        extra_metadata_dict=None,
        time_filter_range=None,
        extra_postprocessing_fn=None,
        plot_output_dir=None,
        always_apply_all_injections=True,
        plot_uncertainty=True,
    ):
        if yet_desired_evt_classes is None:
            yet_desired_evt_classes = []

        realisation.set_seed()
        realisation.apply_longterm_background_states(self.mngr)
        realisation.determine_ref_background_mean_and_std(self.mngr)

        # construct the combined dataframe based on the channel slices
        network_data: pd.DataFrame = self.produce_network_structure(
            realisation,
            fixed_evt_class,
            time_filter_range,
            always_apply_all_injections=always_apply_all_injections,
        )

        realisation.n_events_per_event_class = Counter(
            network_data[self.mngr.tag_evt_type]
        )

        # print("got", realisation.n_events_per_event_class)

        if plot_output_dir is not None:
            reconstructed_df = recreate_aggregated_data_from_output_df(
                self.mngr, network_data
            )

            event_classes = pd.Series(
                network_data[self.mngr.tag_evt_type].to_numpy(),
                index=network_data["time_right_edge"],
            )
            plot_lightcurve(
                self.mngr,
                reconstructed_df,
                realisation.id,
                plot_output_dir,
                event_classes=event_classes,
                # show uncertainty only for non-pure-background realisations
                plot_uncertainty=plot_uncertainty,
            )
            do_plot_scenario_highlights(
                self.mngr, realisation, reconstructed_df, plot_output_dir
            )

        if extra_metadata_dict is not None:
            network_data = network_data.assign(**extra_metadata_dict)

        if extra_postprocessing_fn is not None:
            extra_postprocessing_fn(self.mngr, network_data)

        return network_data

    def get_output_df_columns(self, aggregated_data):
        channel_tags = self.mngr.data_prep_channel_tags
        # (1) produce a list of column names for the output dataframe

        output_data_col_names = []
        output_metadata_col_names = []

        for channel_tag in channel_tags:
            channel_column_names = list(aggregated_data[channel_tag].columns)
            channel_column_names.append(self.mngr.tag_temporal_offset)
            # channel_column_names.append("absolute_offset")
            for i_time_step in range(self.mngr.n_steps_tot):
                output_data_col_names.extend(
                    f"{channel_tag}_{self.mngr.data_prep_time_step_prefix}_{i_time_step}_{col_name}"
                    for col_name in channel_column_names
                )
        for channel_tag in channel_tags:
            output_metadata_col_names.extend(
                (
                    f"{channel_tag}_first_signal_offset",
                    f"{channel_tag}_signal_integral_dec",
                )
            )
        output_metadata_col_names.extend(
            (
                self.mngr.tag_evt_type,
                "realisation_id",
                "realisation_row_index",
                "realisation_seed",
                "custom_realisation_name",
                "time_left_edge",
                "time_right_edge",
            )
        )

        output_metadata_col_names.extend(self.mngr.data_prep_static_df_columns.keys())

        return output_data_col_names, output_metadata_col_names, channel_column_names

    # @profile
    def create_output_df_row(
        self,
        real_time_values_per_channel,
        time_at_time_series_left_edge,
        time_at_time_series_right_edge,
        realisation,
        realisation_index,
        output_row_index,
        data_values,
        signal_injections,
        signal_injection_amplitudes,
        fixed_evt_class,
    ):
        # create one row for the final output dataframe

        time_col_name = self.mngr.data_prep_day_column_name

        # a dict of present temporal injections and its number of affected time steps in the encoder and decoder part
        # only temporal injections that actually appear in this event will be stored here
        temporal_injection_signal_steps = {}

        # a pandas dataframe with a boolean per step, True for signal, False for background. columns are channels, rows steps
        old_signal_steps = pd.DataFrame(index=range(self.mngr.n_steps_tot))
        new_signal_steps = pd.DataFrame(index=range(self.mngr.n_steps_tot))

        # signal_steps_by_feature = pd.DataFrame(index=range(self.mngr.n_steps_tot))

        signal_std_and_mean_per_channel = {}

        first_signal_offset = -np.inf

        event_data_values = []
        event_metadata_values = []
        for (
            channel_tag,
            real_time_values,
        ) in real_time_values_per_channel.items():
            data_indices = realisation.data_indices[channel_tag][realisation_index][
                -self.mngr.n_steps_tot :
            ]

            channel_values = (
                data_values[channel_tag]
                .loc[data_indices]
                .reset_index(drop=True)
                .assign(
                    **{
                        time_col_name: real_time_values_per_channel[channel_tag],
                    }
                )
            )

            signal_relevant_columns = self.mngr.data_prep_fluctuation_cols[channel_tag]

            df_bck_mean = realisation.ref_background_mean[channel_tag][
                realisation_index
            ]

            df_bck_std = realisation.ref_background_std[channel_tag][realisation_index]

            injected_channel_values = channel_values

            if signal_injections is not None:
                for injection, amplitude in zip(
                    signal_injections, signal_injection_amplitudes
                ):
                    if (
                        injection.start_time > real_time_values.iloc[-1]
                        or injection.end_time < real_time_values.iloc[0]
                    ):
                        continue

                    injected_channel_values = injection.apply_to_time_series(
                        channel_tag, injected_channel_values, amplitude, df_bck_std
                    )

            # check how many steps count as signal
            threshold = self.mngr.data_prep_signal_noise_threshold_in_feature_stds[
                channel_tag
            ]

            signal_std_and_mean_per_channel[channel_tag] = (
                df_bck_mean,
                df_bck_std,
            )

            # in units of time series feature standard deviations from the mean
            signal_strength = (
                injected_channel_values[signal_relevant_columns].subtract(df_bck_mean)
                # .subtract(df_bck_median.loc[i_event])
                .abs()
                # .divide(df_bck_std.loc[i_event])
                .divide(df_bck_std)
            )

            # derive a relative measure of the real times as additional column
            absolute_offsets = real_time_values - time_at_time_series_right_edge

            injected_channel_values = injected_channel_values.assign(
                **{self.mngr.tag_temporal_offset: absolute_offsets}
            )

            event_data_values.append(injected_channel_values.values.flatten())

            channel_signal_steps = signal_strength.ge(threshold)

            old_signal_steps[channel_tag] = channel_signal_steps.any(axis=1) & (
                absolute_offsets < self.old_signal_threshold
            )

            new_signal_steps[channel_tag] = channel_signal_steps.any(axis=1) & (
                absolute_offsets >= self.old_signal_threshold
            )

            true_signal_steps = channel_signal_steps.any(axis=1)[lambda x: x == True]
            if len(true_signal_steps) == 0:
                first_signal_channel_offset = np.nan
            else:
                first_signal_index = true_signal_steps.index[-1]
                first_signal_channel_offset = absolute_offsets[first_signal_index]

            event_metadata_values.append(first_signal_channel_offset)
            channel_signal_strength_dec_sum = (
                signal_strength[self.mngr.n_steps_enc :].sum().sum()
            )
            event_metadata_values.append(channel_signal_strength_dec_sum)

            first_signal_offset = max(first_signal_offset, first_signal_channel_offset)

        # get extra metadata columns describing a potentially present fluctuation or physical flare,
        # including evt_class

        has_signal_injection = (
            signal_injections is not None and len(signal_injections) > 0
        )
        if fixed_evt_class is not None:
            evt_class = fixed_evt_class
        else:
            evt_class = self.determine_event_class(
                old_signal_steps, new_signal_steps, has_signal_injection
            )

        event_metadata_values.append(evt_class)
        event_metadata_values.append(realisation.id)
        event_metadata_values.append(output_row_index)
        event_metadata_values.append(realisation.seed)
        event_metadata_values.append(realisation.custom_name)
        event_metadata_values.append(time_at_time_series_left_edge)
        event_metadata_values.append(time_at_time_series_right_edge)

        event_metadata_values.extend(self.mngr.data_prep_static_df_columns.values())

        do_debug_plot_of_data_prep_events(
            self.mngr,
            event_data_values,
            event_metadata_values,
            self.output_data_col_names,
            self.output_metadata_col_names,
            realisation,
            output_row_index,
            self.channel_column_names,
            old_signal_steps,
            new_signal_steps,
            signal_std_and_mean_per_channel,
        )

        return event_data_values, event_metadata_values, evt_class, first_signal_offset

    # @profile
    def produce_network_structure(
        self,
        realisation: Realisation,
        fixed_evt_class,
        time_filter_range=None,
        always_apply_all_injections=True,
    ) -> pd.DataFrame:
        """
        @param realisation: contains rows of indices on the data_values (possibly shuffled),
        indices on pd.Series of real times and the said pd.Series to derive real times for a time series

        @param data_values: dictionary, keys are channel_tags, values are data frames with binned data points

        @return: pandas df containing all channels and time series created previously
        """

        data_values = realisation.input_data
        channel_tags = self.mngr.data_prep_channel_tags

        # (2) Now, create and fill a list of output data frame rows
        output_data_values = []
        output_metadata_values = []

        # loop over all lists of indices
        n_events = realisation.get_n_events()

        signal_injection_iterator = iter(realisation.signal_injections)
        injection = None

        if always_apply_all_injections:
            row_amplitudes = [
                [i.get_random_amplitude() for i in realisation.signal_injections]
            ]
            row_injections = [realisation.signal_injections]
        else:
            row_amplitudes = None
            row_injections = None

        output_row_index = -1
        for realisation_index in range(n_events):
            # get real_time_values of this event per channel
            real_time_values_per_channel = {}

            for channel_tag in channel_tags:
                final_time_index = realisation.final_step_index[channel_tag][
                    realisation_index
                ]
                first_time_index = final_time_index - self.mngr.n_steps_tot + 1

                real_times = realisation.real_times[channel_tag]

                timesteps_range_start = first_time_index
                timesteps_range_end = final_time_index + 1

                real_time_values_per_channel[channel_tag] = real_times.iloc[
                    timesteps_range_start:timesteps_range_end
                ].reset_index(drop=True)

            real_time_values_per_channel = pd.DataFrame(real_time_values_per_channel)

            time_at_time_series_right_edge = np.max(
                real_time_values_per_channel.iloc[-1]
            )

            time_at_time_series_left_edge = np.min(real_time_values_per_channel.iloc[0])

            if time_filter_range is not None:
                if (
                    time_at_time_series_right_edge < time_filter_range[0]
                    or time_at_time_series_right_edge > time_filter_range[1]
                ):
                    continue

            if not always_apply_all_injections:
                if injection is not None and (
                    injection.end_time >= time_at_time_series_left_edge
                ):
                    pass

                else:
                    injection: TemporalInjection = next(signal_injection_iterator, None)

                    if injection is not None and injection.has_randomized_amplitude:
                        signal_injection_amplitudes = [injection.get_random_amplitude()]
                        for _ in range(
                            self.mngr.n_randomized_amplitudes_of_signal_dec_fluctuations
                        ):
                            signal_injection_amplitudes.append(
                                injection.get_random_amplitude()
                            )

                        row_injections = [[injection]] * len(
                            signal_injection_amplitudes
                        )
                        row_amplitudes = [[a] for a in signal_injection_amplitudes]
                    else:
                        row_injections = [None]
                        row_amplitudes = [None]

            for injections, amplitudes in zip(row_injections, row_amplitudes):
                output_row_index += 1

                (
                    event_data_values,
                    event_metadata_values,
                    event_class,
                    first_signal_offset,
                ) = self.create_output_df_row(
                    real_time_values_per_channel,
                    time_at_time_series_left_edge,
                    time_at_time_series_right_edge,
                    realisation,
                    realisation_index,
                    output_row_index,
                    data_values,
                    injections,
                    amplitudes,
                    fixed_evt_class,
                )

                output_metadata_values.append(event_metadata_values)
                output_data_values.append(np.concatenate(event_data_values, axis=None))
                assert len(self.output_data_col_names) == len(
                    np.concatenate(event_data_values, axis=None)
                )

                if (
                    # event_class not in yet_desired_evt_classes
                    fixed_evt_class is not None
                    or injection is None
                    or not injection.has_randomized_amplitude
                    or injection.start_time > time_at_time_series_right_edge
                ):
                    # more than one random injection amplitude in this example is either not desired
                    # or will not make a difference, as the injection is not contained in this time series
                    break

        data_df = pd.DataFrame(
            data=output_data_values, columns=self.output_data_col_names
        )
        meta_df = pd.DataFrame(
            data=output_metadata_values, columns=self.output_metadata_col_names
        )
        output_df = pd.concat((meta_df, data_df), axis=1)

        # sanity checks
        if not np.isfinite(data_df.to_numpy()).all():
            # Identify columns with non-finite values
            non_finite_cols = data_df.columns[
                ~np.isfinite(data_df).all(axis=0)
            ].tolist()

            # Log error with list of affected columns
            self.log.error(
                [
                    self.tag,
                    [
                        "r",
                        f"NaN or np.inf encountered in output dataframe in columns: {non_finite_cols}",
                    ],
                ]
            )

        # add random number for train/test split

        output_df["split_rnd_number"] = np.random.uniform(0, 1, (len(output_df)))

        return output_df

    def determine_event_class(
        self, old_signal_steps, new_signal_steps, has_signal_injection: bool
    ) -> EventClass:
        n_old_signal_steps = {
            self.mngr.tag_enc: old_signal_steps.iloc[: self.mngr.n_steps_enc]
            .any(axis=1)
            .sum(),
            self.mngr.tag_dec: old_signal_steps.iloc[self.mngr.n_steps_enc :]
            .any(axis=1)
            .sum(),
        }
        n_new_signal_steps = {
            self.mngr.tag_enc: new_signal_steps.iloc[: self.mngr.n_steps_enc]
            .any(axis=1)
            .sum(),
            self.mngr.tag_dec: new_signal_steps.iloc[self.mngr.n_steps_enc :]
            .any(axis=1)
            .sum(),
        }

        dec_old = n_old_signal_steps[self.mngr.tag_dec] >= 1
        dec_new = n_new_signal_steps[self.mngr.tag_dec] >= 1
        enc_old = n_old_signal_steps[self.mngr.tag_enc] >= 1
        enc_new = n_new_signal_steps[self.mngr.tag_enc] >= 1

        if enc_new and not enc_old:
            if not dec_new and not dec_old:
                # enc: new | dec: -
                return "signal_enc"
            if not dec_new and dec_old:
                # enc: new | dec: old
                return "signal_enc"
            if dec_new and not dec_old:
                # enc: new | dec: new
                return "signal_enc_dec"
            if dec_new and dec_old:
                # enc: new | dec: new+old
                return "signal_enc_dec"

        elif not enc_new and enc_old:
            if not dec_new and not dec_old:
                # enc: old | dec: -
                return "signal_old_enc"
            if not dec_new and dec_old:
                # enc: old | dec: old
                return "signal_old_enc_dec"
            if dec_new and not dec_old:
                # enc: old | dec: new
                return "signal_dec_signal_old_enc"
            if dec_new and dec_old:
                # enc: old | dec: new+old
                return "signal_dec_signal_old_enc"

        elif not enc_new and not enc_old:
            if not dec_new and not dec_old:
                # enc: - | dec: -
                if has_signal_injection:
                    return "bck_fluct"
                else:
                    return "bck_pure"
                # return "train_simple_bck"
            if not dec_new and dec_old:
                # enc: - | dec: old
                return "signal_old_dec"
            if dec_new and not dec_old:
                # enc: - | dec: new
                return "signal_dec"
            if dec_new and dec_old:
                # enc: - | dec: new+old
                return "signal_dec"

        elif enc_new and enc_old:
            if not dec_new and not dec_old:
                # enc: new+old | dec: -
                return "signal_enc"
            if not dec_new and dec_old:
                # enc: new+old | dec: old
                return "signal_enc"
            if dec_new and not dec_old:
                # enc: new+old | dec: new
                return "signal_enc_dec"
            if dec_new and dec_old:
                # enc: new+old | dec: new+old
                return "signal_enc_dec"

        raise Exception("Unexpected state in determine_event_class()")

    def flares_load_input_data(self, input_dir_name) -> Dict[str, pd.DataFrame]:
        """
        channel names are listed in data_prep_channel_tags.
        For each channel, there is one input file loaded, named like the channel
        from directory as specified in dir_input_data_prep

        This function ensures that the DataFrames are sorted ascending in time

        input dataframe format:
            - columns:
                - all features, e.g., flux in different energy bands
                - time (absolute real time)

        @return: a dict of dataframes with one entry per channel
        """
        input_data = {}
        input_dir = os.path.join(self.mngr.tag_input_dir, input_dir_name)

        self.log.info(
            [self.tag, ["b", "reading input dataframes from "], ["c", input_dir]]
        )

        for channel_tag in self.mngr.data_prep_channel_tags:
            channel_input_file_path = os.path.join(input_dir, f"{channel_tag}.hdf5")
            # noinspection PyTypeChecker
            df: pd.DataFrame = pd.read_hdf(channel_input_file_path)

            df = df.sort_values(
                axis=0,
                by=self.mngr.data_prep_day_column_name,
                ascending=True,
                ignore_index=True,
            )

            # assign zero as errors for columns that do not exist
            missing_error_cols = [
                col
                for col in (
                    self.mngr.data_prep_error_up_cols[channel_tag]
                    + self.mngr.data_prep_error_down_cols[channel_tag]
                )
                if col not in df.columns
            ]

            for col in missing_error_cols:
                df[col] = 0

            if len(missing_error_cols) > 0:
                self.log.info(
                    [
                        self.tag,
                        ["b", f"channel '{channel_tag}': assuming zero errors: "],
                        ["c", " ".join(missing_error_cols)],
                    ]
                )

            if (
                df[
                    self.mngr.data_prep_error_up_cols[channel_tag]
                    + self.mngr.data_prep_error_down_cols[channel_tag]
                ]
                < 0
            ).any(axis=None):
                raise ValueError(
                    f"channel '{channel_tag}': error columns have to be >= 0."
                )

            cols_to_keep = (
                [self.mngr.data_prep_day_column_name]
                + self.mngr.data_prep_fluctuation_cols[channel_tag]
                + self.mngr.data_prep_error_up_cols[channel_tag]
                + self.mngr.data_prep_error_down_cols[channel_tag]
            )

            dropped_cols = [col for col in df.columns if col not in cols_to_keep]

            # # print input df column names for debugging
            self.log.debug(
                [
                    self.tag,
                    ["b", f"channel '{channel_tag}': dropped columns: "],
                    ["y", " ".join(dropped_cols)],
                ]
            )

            cols_to_keep = [col for col in cols_to_keep if col in df.columns]
            input_data[channel_tag] = df[cols_to_keep]

        type_desc_str = " + ".join(
            [
                f"{channel_tag} ({len(input_data[channel_tag])})"
                for channel_tag in self.mngr.data_prep_channel_tags
            ]
        )

        self.log.info(
            [
                self.tag,
                [
                    "b",
                    f"input data summary: {len(self.mngr.data_prep_channel_tags)} channels: ",
                ],
                ["b", type_desc_str],
            ]
        )

        return input_data

    def ensure_fixed_time_intervals(self, real_times):
        """round values up to a multiple of self.data_prep_analysis_min_time_bin_width"""
        return (
            np.ceil(real_times / self.data_prep_analysis_min_time_bin_width)
            * self.data_prep_analysis_min_time_bin_width
        )

    def data_aggregation(
        self,
        input_data: Dict[str, pd.DataFrame],
    ) -> Dict[str, pd.DataFrame]:
        """
        for each element in input_data
        creare new dataframe with a single entry per self.mngr.data_prep_n_days_per_step
        store in aggregated_data

        - derive min time of anaylsys by max(min(all_times)) or some other way
        - fill dataframe with agg data, where each row is an average (or sample) within data_prep_n_days_per_step
        - filter out None values and reindex the dataframe

        ---> result in aggregated_data is dataframe with:
          - colums
            - all features, e.g., flux in different energy bands
            - other metadata that is propagated to the analysis
            - time_step number (not continuous between rowes)
             - real time corresponding to the [time-step number]
        """

        # get min and max time of analysis

        self.data_prep_analysis_min_time_bin_width = min(
            self.mngr.data_prep_n_days_per_step[channel_tag]
            for channel_tag in self.mngr.data_prep_channel_tags
        )

        analysis_min_time = min(
            input_data[channel_tag][self.mngr.data_prep_day_column_name].iat[0]
            for channel_tag in self.mngr.data_prep_channel_tags
        )
        self.log.info(f"absolute min time in data: {analysis_min_time}")

        self.data_prep_analysis_min_time = np.floor(analysis_min_time)

        analysis_time_offset = self.data_prep_analysis_min_time

        # for each channel individually, aggregate high density data by random sampling
        aggregated_data = {}
        for channel_tag in self.mngr.data_prep_channel_tags:
            df_channel = input_data[channel_tag]

            channel_time_col = df_channel[self.mngr.data_prep_day_column_name]
            channel_time_col -= analysis_time_offset
            channel_time_col = self.ensure_fixed_time_intervals(channel_time_col)

            df_channel[self.mngr.data_prep_day_column_name] = channel_time_col

            df_channel_binned = (
                df_channel.groupby(self.mngr.data_prep_day_column_name)
                .sample(n=1, random_state=1)
                .dropna(how="all")  # remove time bins with no data points
                .reset_index(drop=True)
            )

            n_nan_rows = df_channel_binned.isna().sum().sum()

            df_channel_binned = df_channel_binned.dropna(
                how="any"
            ).reset_index(  # remove time bins with no data points
                drop=True
            )

            if n_nan_rows > 0:
                self.log.warn(
                    [
                        self.tag,
                        ["r", f"dropping {n_nan_rows} input data rows from "],
                        ["r", channel_tag],
                        ["r", " because they have NaN values."],
                    ]
                )

            aggregated_data[channel_tag] = df_channel_binned

        type_desc_str = " + ".join(
            [
                f"{channel_tag} ({len(aggregated_data[channel_tag])})"
                for channel_tag in self.mngr.data_prep_channel_tags
            ]
        )

        self.log.info(
            [
                self.tag,
                ["b", "aggregating high-density data points, remaining: "],
                ["b", type_desc_str],
            ]
        )

        return aggregated_data

    def data_prep_log_result_summary(self, realisations: List[Realisation]):
        all_fluctuations = []
        all_flares = []
        n_realisations = len(realisations)
        n_randomized_offsets = 0
        n_background_states = 0
        n_shifted_real_times = 0

        for realisation in realisations:
            for temporal_injection in realisation.signal_injections:
                if isinstance(temporal_injection, PhysicalFlare):
                    all_flares.append(temporal_injection)
                elif isinstance(temporal_injection, Fluctuation):
                    all_fluctuations.append(temporal_injection)

            if realisation.has_randomized_real_times:
                n_randomized_offsets += 1

            if realisation.has_shifted_real_times:
                n_shifted_real_times += 1

            n_background_states += len(realisation.long_term_states)

        #  shuffling strategies
        type_desc_str = " + ".join(
            [
                f"{count} {shuffle_type}"
                for shuffle_type, count in Counter(
                    r.shuffling_strategy for r in realisations
                ).items()
            ]
        )
        self.log.info(
            [
                self.tag,
                [
                    "b",
                    f"Created {n_realisations} realisations with shuffling strategies: ( {type_desc_str} )",
                ],
            ]
        )

        # shifted real times
        self.log.info(
            [
                self.tag,
                [
                    "b",
                    f" -> shifted real times in {n_shifted_real_times} of {n_realisations} realisations",
                ],
            ]
        )

        # temporal offsets
        self.log.info(
            [
                self.tag,
                [
                    "b",
                    f" -> randomized temporal offsets in {n_randomized_offsets} of {n_realisations} realisations",
                ],
            ]
        )

        # summary of longterm background states
        self.log.info(
            [
                self.tag,
                [
                    "b",
                    f"generated {n_background_states} long-term background states",
                ],
            ]
        )

        # summary of fluctuations
        counts_by_type = Counter(fluct.type_name for fluct in all_fluctuations)
        type_desc_str = " + ".join(
            [f"{count} {fluct_type}" for fluct_type, count in counts_by_type.items()]
        )
        self.log.info(
            [
                self.tag,
                [
                    "b",
                    f"generated {len(all_fluctuations)} randomized fluctuations ( {type_desc_str} )",
                ],
            ]
        )

        # summary of physical flares
        counts_by_type = Counter(flare.type_name for flare in all_flares)
        type_desc_str = " + ".join(
            [f"{count} {fluct_type}" for fluct_type, count in counts_by_type.items()]
        )
        self.log.info(
            [
                self.tag,
                [
                    "b",
                    f"generated {len(all_flares)} physical flares ( {type_desc_str} )",
                ],
            ]
        )

        # events by event class
        counts_by_type = pd.DataFrame(
            [r.n_events_per_event_class for r in realisations]
        ).sum()

        type_desc_str = " + ".join(
            [f"{int(count)} {evt_class}" for evt_class, count in counts_by_type.items()]
        )
        self.log.info(
            [
                self.tag,
                [
                    "b",
                    f"Result: {int(sum(counts_by_type))} events: ( {type_desc_str} )",
                ],
            ]
        )

    def prep_calib(self):
        """
        data for calibration is generated as part of prep_train and can be selected using evt_type ratios in the config
        """
        raise NotImplementedError()

    def profile_datasets_temp(self):
        """print distribution of features for input_train, input_predict and background realisations"""
        data_predict = self.data_aggregation(
            self.flares_load_input_data(self.mngr.dir_input_data_prep_predict)
        )
        data_train = self.data_aggregation(
            self.flares_load_input_data(self.mngr.dir_input_data_prep_train)
        )

        channels = ["veritas"]  # data_train

        for channel_tag in channels:
            cols = self.mngr.data_prep_fluctuation_cols[channel_tag]
            print(channel_tag, "train:", data_train[channel_tag][cols].describe())
            print(channel_tag, "predict:", data_predict[channel_tag][cols].describe())

            print("---")

        self.long_term_state_generator = LongTermStateGenerator()

        for i in range(10):
            self.set_aggregated_data_conf(data_train)

            realisation = self.create_realisation(
                data_train,
                shuffling_strategy="step_based",
                do_randomize_temporal_offsets=True,
                do_shift_input_data_in_time=True,
                add_fluctuations=False,
                add_flares=False,
                add_background_states=True,
                realisation_id=f"{i}",
                realisation_seed=i,
            )

            network_df = self.write_realisation(
                realisation,
                plot_output_dir="inference",
                fixed_evt_class="custom",
            )
            reconstructed_df = recreate_aggregated_data_from_output_df(
                self.mngr, network_df
            )

            for channel_tag in channels:
                cols = self.mngr.data_prep_fluctuation_cols[channel_tag]
                print(
                    channel_tag,
                    "background realization:",
                    reconstructed_df[channel_tag][cols].describe(),
                )

                print("---")

        exit(0)  # terminating here

    def prep_predict(self):
        # profile_input_data(self.mngr)
        self.log.info([["p", " - prep-predict()"]])

        output_dir = os.path.join(
            self.mngr.tag_input_dir, self.mngr.dir_output_data_prep_predict
        )

        clear_directory(self.mngr, output_dir)

        # load input data for each channel, ensuring an ordering in time
        input_data = self.flares_load_input_data(self.mngr.dir_input_data_prep_predict)

        # rebin channel data as configured and average high density data points
        aggregated_data = self.data_aggregation(input_data)

        self.set_aggregated_data_conf(aggregated_data)

        df_batched_hdf_writer = DataFrameBatchedHDFWriter(self.mngr, output_dir)

        inference_channel_padding = create_channel_padding_infer(self.mngr)

        # create network structure from aggregated data
        realisation = self.create_realisation(
            aggregated_data,
            shuffling_strategy=None,
            do_randomize_temporal_offsets=False,
            do_shift_input_data_in_time=False,
            add_fluctuations=False,
            add_flares=False,
            add_background_states=False,
            realisation_id="inference_realisation",
            realisation_seed=0,
            channel_padding=inference_channel_padding,
        )

        realisation.custom_name = "inference_realisation"
        network_df = self.write_realisation(
            realisation,
            plot_output_dir="inference",
            fixed_evt_class="custom",
        )

        df_batched_hdf_writer.add(network_df)
        df_batched_hdf_writer.flush()
