from __future__ import annotations

from typing import TYPE_CHECKING, Dict

import numpy as np
import pandas as pd

from trans_finder.flares.state_manager import StateManager

if TYPE_CHECKING:
    from trans_finder.flares.manager import Manager


def large_negative_offsets(mngr: Manager, channel_tag, n_steps_padding):
    step_size = mngr.data_prep_n_days_per_step[channel_tag]
    time_offset = 10 * mngr.n_steps_tot * step_size
    times = (
        np.arange(start=0, stop=step_size * n_steps_padding, step=step_size)
        - time_offset
    )
    return times


def smear_errors_on_channel_df(
    mngr: Manager, df_padding: pd.DataFrame, channel_tag: str
) -> pd.DataFrame:
    df_padding = mngr.data.smear_df_by_errors(
        df_padding,
        error_up_cols=mngr.data_prep_error_up_cols[channel_tag],
        error_down_cols=mngr.data_prep_error_down_cols[channel_tag],
        value_cols=mngr.data_prep_error_feature_cols[channel_tag],
    )

    return df_padding


def create_channel_padding_train(
    mngr: Manager, aggregated_data: Dict[str, pd.DataFrame]
):
    channel_padding = {}

    for channel_tag in aggregated_data:
        channel_df = aggregated_data[channel_tag]

        if len(channel_df) < 3:
            raise Exception(
                f"Input channel '{channel_df}' has less than 3 data points. At least 3 data points per channel are required."
            )

        else:
            # add padding s.t. the final step in the decoder will be the first real data point, if it is at time 0, otherwise
            # the final decoder step will still be padding
            n_steps_padding = mngr.data_prep.n_steps_shuffling_window - (
                1 if channel_df[mngr.data_prep_day_column_name].iloc[0] == 0 else 0
            )

            reference_df = channel_df.iloc[: mngr.data_prep.n_steps_shuffling_window, :]

            has_errors = (
                reference_df[
                    mngr.data_prep_error_up_cols[channel_tag]
                    + mngr.data_prep_error_down_cols[channel_tag]
                ]
                > 0
            ).any(axis=None)

            if not has_errors:
                # randomly generate data points from median and 1sig quantiles
                # relevant_cols = mngr.data_prep_fluctuation_cols[channel_tag]
                median = reference_df.median()
                delta = reference_df.quantile(0.841) - channel_df.quantile(0.159)

                # temporal offset column will be set manually below
                delta[mngr.data_prep_day_column_name] = np.nan

                padding = np.random.normal(
                    median, delta, (n_steps_padding, len(median))
                )
                df_padding = pd.DataFrame(padding, columns=channel_df.columns)

            else:
                # if errors are nonzero,  and randomly sample from and smear available data points,
                #  and sample fluxes from them, keep original error
                df_padding = reference_df.sample(
                    n=n_steps_padding, replace=n_steps_padding > len(reference_df)
                ).reset_index(drop=True)
                df_padding = smear_errors_on_channel_df(mngr, df_padding, channel_tag)

            # in any case, use some large offset,
            # s.t. it will be treated as background in the autoencoder through the offset weights
            times = large_negative_offsets(mngr, channel_tag, n_steps_padding)
            df_padding[mngr.data_prep_day_column_name] = times

            channel_padding[channel_tag] = df_padding

    return channel_padding


def create_channel_padding_infer(mngr: Manager):
    channel_padding_state = StateManager(mngr, mngr.tag_inference_channel_padding)

    if channel_padding_state.has_state():
        return channel_padding_state.load_state()

    # load input data for each channel, ensuring an ordering in time
    input_data = mngr.data_prep.flares_load_input_data(mngr.dir_input_data_prep_train)

    # rebin channel data as configured and average high density data points
    aggregated_data = mngr.data_prep.data_aggregation(input_data)

    mngr.data_prep.set_aggregated_data_conf(aggregated_data)

    realisation = mngr.data_prep.create_realisation(
        aggregated_data,
        shuffling_strategy="step_based",
        do_randomize_temporal_offsets=False,
        do_shift_input_data_in_time=False,
        add_fluctuations=False,
        add_flares=False,
        add_background_states=False,
        realisation_id="channel_padding",
        realisation_seed=0,
    )

    # We do not need to create network inputs and reconstruct the original light curve here.
    # We do need the logic in create_realisation to add the train_padding to sparse channels though.
    background_lc_data = realisation.input_data

    channel_padding = {}

    for channel_tag in background_lc_data:
        channel_df = background_lc_data[channel_tag]

        # add padding s.t. the final step in the decoder will be the first real data point, if it is at time 0, otherwise
        # the final decoder step will still be padding
        n_steps_padding = mngr.data_prep.n_steps_shuffling_window - (
            1 if channel_df[mngr.data_prep_day_column_name].iloc[0] == 0 else 0
        )

        df_padding = channel_df.iloc[-n_steps_padding:].copy()

        times = large_negative_offsets(mngr, channel_tag, n_steps_padding)
        df_padding[mngr.data_prep_day_column_name] = times

        channel_padding[channel_tag] = df_padding

    channel_padding_state.save_state(channel_padding)

    return channel_padding
