from __future__ import annotations

import random
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Dict

import numpy as np
import pandas as pd

from trans_finder.flares.data_prep_helpers.temporal_injection import (
    TemporalInjection,
    TemporalInjectionGenerator,
)

if TYPE_CHECKING:
    from trans_finder.flares.manager import Manager


class Fluctuation(TemporalInjection, ABC):
    id: int
    amplitude: Dict[str, float]
    has_randomized_amplitude = True

    def __init__(
        self,
        mngr,
        type_name,
        id,
        start_time,
        end_time,
        amplitude_conf,
        channel_start_times,
        channel_end_times,
        randomize_feature_amplitudes,
    ):
        self.type_name = mngr.tag_fluctuation
        self.mngr: Manager = mngr
        self.type_name = type_name
        self.id = id
        self.start_time = start_time
        self.end_time = end_time
        self.amplitude_conf = amplitude_conf
        self.channel_start_times = channel_start_times
        self.channel_end_times = channel_end_times
        self.randomize_feature_amplitudes = randomize_feature_amplitudes

    def get_random_amplitude(self):
        if self.randomize_feature_amplitudes:
            return self.get_amplitudes_with_randomized_feature_amplitudes()
        else:
            return self.get_amplitudes_with_all_features_identical()

    @abstractmethod
    def get_intensity_at(
        self,
        real_times: pd.Series,
        channel_tag: str,
        amplitude,
    ) -> pd.Series:
        """returns a scalar at each time in real_times,
        that is the positive or negative value added to each feature in units of feature standard deviations
        """
        pass

    def apply_to_time_series(
        self,
        channel_tag,
        channel_values: pd.DataFrame,
        amplitude,
        stddev,
    ) -> pd.DataFrame:
        relevant_cols = self.mngr.data_prep_fluctuation_cols[channel_tag]
        time_col_name = self.mngr.data_prep_day_column_name

        # affected_rows = channel_values[time_col_name].between(
        #     self.channel_start_times[channel_tag], self.channel_end_times[channel_tag]
        # )

        scaling_factor = self.get_intensity_at(
            channel_values[time_col_name],
            channel_tag,
            amplitude,
        )

        # if self.amplitude[channel_tag] < 0:
        #     channel_values.loc[affected_rows, relevant_cols] = channel_values.loc[
        #         affected_rows, relevant_cols
        #     ] / abs(scaling_factor)
        # else:
        fluct_diff = (
            scaling_factor
            * stddev
            # * self.mngr.data_prep_feature_stds[channel_tag][relevant_cols]
        )

        injected_relevant_cols = channel_values[relevant_cols] + fluct_diff

        injected_channel_values = channel_values.assign(
            **{col: injected_relevant_cols[col] for col in relevant_cols}
        )
        #
        # injected_channel_values = channel_values
        # injected_channel_values.loc[affected_rows, relevant_cols] += fluct_diff

        assert all(channel_values.columns == injected_channel_values.columns)
        assert all(channel_values.index == injected_channel_values.index)

        if injected_channel_values[relevant_cols].isnull().values.any():
            raise ValueError("NaN encountered")

        return injected_channel_values


class FluctuationMetaDataWriter:
    def __init__(self, mngr):
        self.mngr: Manager = mngr

    def get_description_column_names(self):
        col_names = [
            f"fluct_{attribute}"
            for attribute in [
                "id",
                "type_name",
                # "intensity_enc",
                # "intensity_dec",
                # "total_duration",
            ]
        ]
        # col_names.extend(
        #     f"amplitude_{tag}" for tag in self.mngr.data_prep_channel_tags
        # )

        return col_names

    def get_description_default_values(self):
        col_values = [
            -1,
            "no fluctuation",
            # -1,
        ]
        # # amplitude
        # col_values.extend(-1 for tag in self.mngr.data_prep_channel_tags)
        return col_values

    def get_description_values(self, fluctuation: Fluctuation):
        col_values = [
            fluctuation.id,
            fluctuation.type_name,
            # fluctuation.end_time - fluctuation.start_time,
        ]
        # col_values.extend(
        #     fluctuation.amplitude[tag] for tag in self.mngr.data_prep_channel_tags
        # )
        return col_values


class PlateauMetaDataWriter(FluctuationMetaDataWriter):
    def get_description_column_names(self):
        pass

    def get_description_default_values(self):
        pass

    def get_description_values(self, fluctuation: Fluctuation):
        pass


class ShortTermStateFluctuation(Fluctuation, ABC):
    pass


class LinearTrendFluctuation(Fluctuation, ABC):
    def get_intensity_at(
        self, real_times: pd.Series, channel_tag: str, amplitude
    ) -> pd.DataFrame:
        """return 0 for real times where nothing should be added, otherwise the added score in units of feature stds"""

        real_time_mask = real_times.ge(
            self.channel_start_times[channel_tag]
        ) & real_times.le(self.channel_end_times[channel_tag])

        intensities = pd.DataFrame(
            0,
            index=real_times.index,
            columns=self.mngr.data_prep_fluctuation_cols[channel_tag],
        )

        linear_interpolation = np.interp(
            real_times[real_time_mask],
            [
                self.channel_start_times[channel_tag],
                self.channel_end_times[channel_tag],
            ],
            [0, amplitude[channel_tag]],
            left=0,
            right=0,
        )

        for col in self.mngr.data_prep_fluctuation_cols[channel_tag]:
            intensities.loc[real_time_mask, col] = linear_interpolation

        return intensities


class FlatPlateauScalingFluctuation(Fluctuation, ABC):
    def get_intensity_at(
        self, real_times: pd.Series, channel_tag: str, amplitude
    ) -> pd.DataFrame:
        """return 0 for real times where nothing should be added, otherwise the added score in units of feature stds"""

        real_time_mask = real_times.ge(
            self.channel_start_times[channel_tag]
        ) & real_times.le(self.channel_end_times[channel_tag])

        intensities = pd.DataFrame(
            0,
            index=real_times.index,
            columns=self.mngr.data_prep_fluctuation_cols[channel_tag],
        )

        if len(amplitude[channel_tag]) == 1:
            intensities.loc[real_time_mask, :] = amplitude[channel_tag][0]
        else:
            intensities.loc[real_time_mask, :] = amplitude[channel_tag]

        return intensities


class GaussianScalingFluctuation(Fluctuation):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.mu = (
            pd.Series(self.channel_end_times) + pd.Series(self.channel_start_times)
        ) / 2
        self.sig = (
            pd.Series(self.channel_end_times) - pd.Series(self.channel_start_times)
        ) / 4

    def get_intensity_at(
        self,
        real_times: pd.Series,
        channel_tag: str,
        amplitude,
    ) -> pd.DataFrame:
        real_time_mask = real_times.ge(
            self.channel_start_times[channel_tag]
        ) & real_times.le(self.channel_end_times[channel_tag])

        intensities = pd.DataFrame(
            0,
            index=real_times.index,
            columns=self.mngr.data_prep_fluctuation_cols[channel_tag],
        )

        if real_time_mask.sum() < 1:
            return intensities

        time_profile = np.exp(
            -np.power(real_times[real_time_mask] - self.mu[channel_tag], 2.0)
            / (2 * np.power(self.sig[channel_tag], 2.0))
        )

        time_mask_intensities = np.outer(time_profile, amplitude[channel_tag])

        intensities.loc[real_time_mask, :] = time_mask_intensities

        return intensities


class ExponentialDecayFluctuation(Fluctuation):
    def get_intensity_at(
        self,
        real_times: pd.Series,
        channel_tag: str,
        amplitude,
    ) -> pd.DataFrame:
        raise NotImplementedError()


def create_fluctuation(mngr, fluct_type, fluctuation_config):
    if fluct_type == "flat_plateau":
        return FlatPlateauScalingFluctuation(mngr, **fluctuation_config)
    elif fluct_type == "gaussian":
        return GaussianScalingFluctuation(mngr, **fluctuation_config)
    elif fluct_type == "exp_decay":
        return ExponentialDecayFluctuation(mngr, **fluctuation_config)


def create_custom_fluctuation_from_conf(mngr, conf):
    conf["start_time"] = min(conf["channel_start_times"].values())
    conf["end_time"] = max(conf["channel_end_times"].values())

    # for unconfigured channels, specify no activity
    for channel_tag in mngr.data_prep_channel_tags:
        if channel_tag not in conf["channel_start_times"]:
            conf["channel_start_times"][channel_tag] = conf["start_time"]
            conf["channel_end_times"][channel_tag] = conf["start_time"]
            conf["amplitude_conf"][channel_tag] = {"min": 0, "max": 0}

    return create_fluctuation(mngr, conf["type_name"], conf)


class FluctuationGenerator(TemporalInjectionGenerator):
    id_counter = -1

    def generate_randomized_temporal_injection(self, mngr, start_time) -> Fluctuation:
        self.id_counter += 1

        channel_tags = mngr.data_prep_channel_tags

        preset_name = random.choice(list(mngr.data_prep_fluctuation_presets.keys()))

        fluct_conf = mngr.data_prep_fluctuation_presets[preset_name]
        fluct_type = random.choice(fluct_conf["types"])

        time_unit = mngr.data_prep_fluctuation_time_unit

        durations = {
            tag: np.random.uniform(
                fluct_conf["duration"]["min"], fluct_conf["duration"]["max"]
            )
            * time_unit
            for tag in channel_tags
        }

        shifts = {
            tag: np.random.uniform(
                fluct_conf["temporal_channel_shift"]["min"],
                fluct_conf["temporal_channel_shift"]["max"],
            )
            * time_unit
            for tag in channel_tags
        }

        min_shift = min(shifts.values())
        shifts = {tag: shifts[tag] - min_shift for tag in channel_tags}

        channel_start_times = {tag: start_time + shifts[tag] for tag in channel_tags}

        channel_end_times = {
            tag: channel_start_times[tag] + durations[tag] for tag in channel_tags
        }

        fluctuation_config = {
            "type_name": fluct_type,
            "id": self.id_counter,
            "start_time": start_time,
            "end_time": max(channel_end_times.values()),
            "amplitude_conf": {
                tag: random.choice(list(fluct_conf["amplitude_confs"].values()))
                for tag in channel_tags
            },
            "channel_start_times": channel_start_times,
            "channel_end_times": channel_end_times,
            "randomize_feature_amplitudes": True,
        }

        return create_fluctuation(mngr, fluct_type, fluctuation_config)
