import random
from abc import ABC

import numpy as np
import pandas as pd

from trans_finder.flares.data_prep_helpers.fluctuation import (
    FlatPlateauScalingFluctuation,
    Fluctuation,
    LinearTrendFluctuation,
)
from trans_finder.flares.data_prep_helpers.temporal_injection import (
    TemporalInjectionGenerator,
)


class LongTermLinearTrendFluctuation(LinearTrendFluctuation, ABC):
    def get_intensity_at(
        self, real_times: pd.Series, channel_tag: str, amplitude
    ) -> pd.DataFrame:
        """return 0 for real times where nothing should be added, otherwise the added score in units of feature stds"""

        intensities = pd.DataFrame(
            0,
            index=real_times.index,
            columns=self.mngr.data_prep_fluctuation_cols[channel_tag],
        )

        for i_col, col in enumerate(self.mngr.data_prep_fluctuation_cols[channel_tag]):
            linear_interpolation = np.interp(
                real_times,
                [
                    self.channel_start_times[channel_tag],
                    self.channel_end_times[channel_tag],
                ],
                [0, amplitude[channel_tag][i_col]],
                left=0,
                right=amplitude[channel_tag][i_col],
            )
            intensities[col] = linear_interpolation

        return intensities


class LongTermPlateauFluctuation(FlatPlateauScalingFluctuation, ABC):
    """return 0 for real times where nothing should be added, otherwise the added score in units of feature stds"""

    pass
    # def get_intensity_at(
    #     self,
    #     real_times: pd.Series,
    #     channel_tag: str,
    #     channel_values,
    #     amplitude,
    # ) -> pd.Series:
    #     real_time_mask = real_times.ge(
    #         self.channel_start_times[channel_tag]
    #     ) & real_times.le(self.channel_end_times[channel_tag])
    #
    #     intensities = pd.DataFrame(
    #         0,
    #         index=real_times.index,
    #         columns=self.mngr.data_prep_fluctuation_cols[channel_tag],
    #     )
    #
    #     intensities.loc[real_time_mask, :] = np.full(
    #         len(self.mngr.data_prep_fluctuation_cols[channel_tag]),
    #         self.amplitude,
    #     )
    #
    #     return intensities


class LongTermStateGenerator(TemporalInjectionGenerator):
    id_counter = -1

    def generate_randomized_temporal_injection(self, mngr, start_time) -> Fluctuation:
        self.id_counter += 1

        channel_tags = mngr.data_prep_channel_tags
        time_unit = mngr.data_prep_fluctuation_time_unit
        preset_names_list = list(mngr.data_prep_long_term_state_presets.keys())

        preset_name = random.choice(preset_names_list)

        long_term_state_conf = mngr.data_prep_long_term_state_presets[preset_name]

        long_term_state_type = random.choice(long_term_state_conf["types"])

        durations = {
            tag: np.random.uniform(
                long_term_state_conf["duration"]["min"],
                long_term_state_conf["duration"]["max"],
            )
            * time_unit
            for tag in channel_tags
        }

        channel_start_times = {tag: start_time for tag in channel_tags}

        channel_end_times = {
            tag: channel_start_times[tag] + durations[tag] for tag in channel_tags
        }

        long_term_state_config = {
            "type_name": "long_term_state",
            "id": self.id_counter,
            "start_time": start_time,
            "end_time": max(channel_end_times.values()),
            "amplitude_conf": {
                tag: long_term_state_conf["amplitude"] for tag in channel_tags
            },
            "channel_start_times": channel_start_times,
            "channel_end_times": channel_end_times,
            "randomize_feature_amplitudes": False,
        }

        if long_term_state_type == "plateau":
            return LongTermPlateauFluctuation(mngr, **long_term_state_config)
        elif long_term_state_type == "linear_trend":
            return LongTermLinearTrendFluctuation(mngr, **long_term_state_config)
        else:
            raise NotImplementedError
