from __future__ import annotations

import os
from typing import TYPE_CHECKING, Dict, List, Optional

import numpy as np
import pandas as pd

from trans_finder.flares.data_prep_helpers.temporal_injection import (
    TemporalInjection,
    TemporalInjectionGenerator,
)
from trans_finder.flares.helpers import (
    data_prep_inverse_simple_log_transform,
    data_prep_simple_log_transform,
)

if TYPE_CHECKING:
    from trans_finder.flares.manager import Manager


class PhysicalFlare(TemporalInjection):
    """injections based on a template dataframe"""

    def __init__(
        self,
        mngr,
        type_name,
        id,
        start_time,
        end_time,
        channel_start_times,
        channel_end_times,
        channel_templates,
    ):
        self.mngr: Manager = mngr
        self.type_name = type_name
        self.id = id
        self.start_time = start_time
        self.end_time = end_time
        self.channel_start_times = channel_start_times
        self.channel_end_times = channel_end_times
        self.channel_templates = channel_templates
        self.input_flux = max(
            channel_templates[tag].loc[0, "sim_flux_input"] for tag in channel_templates
        )

    def apply_to_time_series(
        self,
        channel_tag,
        channel_values: pd.DataFrame,
        _,
        stddev,
    ):
        relevant_cols = self.mngr.data_prep_fluctuation_cols[channel_tag]
        time_col_name = self.mngr.data_prep_day_column_name

        if channel_tag not in self.channel_start_times.keys():
            return channel_values

        affected_rows = channel_values[time_col_name].between(
            self.channel_start_times[channel_tag], self.channel_end_times[channel_tag]
        )

        if not affected_rows.any():
            return channel_values

        real_times = channel_values[time_col_name]
        relative_real_times = real_times - self.channel_start_times[channel_tag]
        template_times = self.channel_templates[channel_tag][
            self.mngr.data_prep_day_column_name
        ]

        template_times -= template_times.iat[0]

        template_fluxes = self.channel_templates[channel_tag][relevant_cols]

        delinearized_template_fluxes = data_prep_inverse_simple_log_transform(
            template_fluxes
        )

        delinearized_template_fluxes.loc[:, relevant_cols] = (
            delinearized_template_fluxes[relevant_cols].fillna(value=0)
        )

        delinearized_fluxes = data_prep_inverse_simple_log_transform(
            channel_values[relevant_cols]
        )

        new_fluxes = pd.DataFrame()

        for flux_feature in delinearized_template_fluxes:
            interpolated_flux_feature = np.interp(
                relative_real_times,
                template_times,
                delinearized_template_fluxes[flux_feature],
                left=0,
                right=0,
            )

            new_flux_feature = (
                delinearized_fluxes[flux_feature] + interpolated_flux_feature
            )

            log_new_flux_feature = pd.Series(
                data_prep_simple_log_transform(new_flux_feature),
                index=real_times.index,
            )

            new_fluxes[flux_feature] = log_new_flux_feature

        injected_channel_values = channel_values.assign(
            **{col: new_fluxes[col] for col in relevant_cols}
        )

        assert all(channel_values.columns == injected_channel_values.columns)
        assert all(channel_values.index == injected_channel_values.index)

        # if injected_channel_values[relevant_cols].isnull().values.any():
        #     raise ValueError("NaN encountered")

        return injected_channel_values

    def get_random_amplitude(self):
        return {tag: None for tag in self.mngr.data_prep_channel_tags}

    @classmethod
    def get_description_column_names(cls):
        col_names = [
            f"flare_{attribute}"
            for attribute in [
                "id",
                "type_name",
                "max_amplitude",
                "total_duration",
            ]
        ]
        return col_names

    @classmethod
    def get_description_default_values(cls):
        return [-1, "no flare", -1, -1]

    def get_description_values(self):
        return [
            self.id,
            self.type_name,
            0,
            # max(self.amplitude.values()),
            self.end_time - self.start_time,
        ]


class PhysicalFlareGenerator(TemporalInjectionGenerator):
    id_counter = -1

    def __init__(self, mngr):
        self.mngr: Manager = mngr
        template_file_names = self.find_flare_templates()
        self.channel_tags = self.mngr.data_prep_channel_tags
        self.templates = {
            channel_tag: {
                name: self.read_flare_template(name, channel_tag)
                for name in template_file_names
                if name.startswith(channel_tag)
            }
            for channel_tag in self.channel_tags
        }

        self.normalize_times_in_templates()

        for channel_tag, template_list in self.templates.items():
            if len(template_list) == 0:
                self.mngr.log.warn(
                    [
                        [
                            "y",
                            f" - PhysicalFlareGenerator - No physical flare templates found for channel {channel_tag}",
                        ]
                    ]
                )

    def get_template_name(
        self, channel_tag: str, type: str, scaling: float, duration: float, seed: int
    ):
        return f"{channel_tag}_type_{type}_intensity_{scaling}_duration_{duration}_realisation_{seed}.hdf5"

    def generate_randomized_temporal_injection(self, mngr, start_time) -> PhysicalFlare:
        raise NotImplementedError()

    def normalize_times_in_templates(self):
        def round_time(times, time_bin_width):
            return np.round(times / time_bin_width) * time_bin_width

        for channel_tag in self.channel_tags:
            time_bin_width = self.mngr.data_prep_n_days_per_step[channel_tag]

            for template_df in self.templates[channel_tag].values():
                time_col = template_df[self.mngr.data_prep_day_column_name]
                time_col -= time_col.iat[0]

                # todo: rounding the times here is a fix to proceed now, time bin with is always 1, cleanup templates ...
                assert time_bin_width == 1

                time_col = round_time(time_col, time_bin_width)
                template_df[self.mngr.data_prep_day_column_name] = time_col

                assert time_col.is_unique

    def find_flare_templates(self):
        def is_accepted_input_file(file_name):
            return all(
                (patt in file_name) for patt in self.mngr.physical_flare_file_name_patt
            )

        input_dir = os.path.join(
            self.mngr.tag_input_dir, self.mngr.dir_input_flare_templates
        )

        if not os.path.isdir(input_dir):
            return []

        file_list = sorted(os.listdir(input_dir))
        file_list = [x for x in file_list if is_accepted_input_file(x)]
        return file_list

    def read_flare_template(self, template_name, channel_tag):
        template_path = os.path.join(
            self.mngr.tag_input_dir,
            self.mngr.dir_input_flare_templates,
            template_name,
        )
        template_df = pd.read_hdf(template_path)
        template_df: pd.DataFrame

        # check for nans or np.inf in step and error columns and raise an error
        relevant_cols = [
            self.mngr.data_prep_day_column_name
        ] + self.mngr.data_prep_fluctuation_cols[channel_tag]

        value_cols = self.mngr.data_prep_fluctuation_cols[channel_tag]
        time_col = self.mngr.data_prep_day_column_name

        if not np.isfinite(template_df[value_cols].to_numpy()).all():
            self.mngr.log.warn(
                [
                    self.mngr.tag,
                    [
                        "r",
                        f"NaN or np.inf encountered in template values. ({template_name}). Replaced with zero. ",
                    ],
                ]
            )

        # Replace NaN and infinite values with zero
        template_df[value_cols] = template_df[value_cols].where(
            np.isfinite(template_df[value_cols]), 0
        )

        if not np.isfinite(template_df[time_col].to_numpy()).all():
            raise ValueError(
                f"NaN or np.inf encountered in template time col. ({template_name})."
            )

        return template_df

    def create_custom_flare(
        self,
        template_names: Dict[str, str],
        start_time: float,
        shifts: Optional[Dict[str, float]] = None,
        duration=None,
        make_negative=False,
    ):
        channel_tags = list(template_names.keys())

        if shifts is None:
            shifts = {tag: 0 for tag in channel_tags}

        min_shift = min(shifts.values())
        shifts = {tag: shifts[tag] - min_shift for tag in channel_tags}

        channel_start_times = {tag: start_time + shifts[tag] for tag in channel_tags}

        if duration is None:
            duration = self.mngr.flare_template_conf["flare_duration"]

        durations = {channel_tag: duration for channel_tag in channel_tags}

        channel_end_times = {
            tag: channel_start_times[tag] + durations[tag] for tag in channel_tags
        }

        channel_templates = {
            channel_tag: self.templates[channel_tag][template_names[channel_tag]]
            for channel_tag in channel_tags
        }

        if make_negative:
            # use channel templates with negated value columns
            negated_templates = {
                k: df.assign(
                    **{col: -df[col] for col in self.mngr.data_prep_fluctuation_cols[k]}
                )
                for k, df in channel_templates.items()
            }

            channel_templates = negated_templates

        flare_config = {
            "type_name": "custom_flare",
            "id": self.id_counter,
            "start_time": start_time,
            "end_time": max(channel_end_times.values()),
            "channel_start_times": channel_start_times,
            "channel_end_times": channel_end_times,
            "channel_templates": channel_templates,
        }

        return PhysicalFlare(self.mngr, **flare_config)


def inject_flares_into_timeseries(
    mngr,
    flares: List[PhysicalFlare],
    data: Dict[str, pd.DataFrame],
):
    for flare in flares:
        for channel_tag in flare.channel_templates:
            channel_df = data[channel_tag]

            # drop any data points within the interval of the flare injection
            time_col_name = mngr.data_prep_day_column_name
            real_time_mask = channel_df[time_col_name].ge(
                flare.channel_start_times[channel_tag]
            ) & channel_df[time_col_name].le(flare.channel_end_times[channel_tag])

            channel_df.drop(channel_df[real_time_mask].index).reset_index(drop=True)

            # generate zero data points
            n_data_points = 3

            time_bin_width = mngr.data_prep_n_days_per_step[channel_tag]

            def round_time(times, time_bin_width):
                return np.ceil(times / time_bin_width) * time_bin_width

            start = round_time(flare.start_time, time_bin_width)
            end = round_time(flare.end_time, time_bin_width)
            real_times = np.random.choice(
                np.arange(start, end, time_bin_width), n_data_points, replace=False
            )

            for time in real_times:
                channel_df.loc[len(channel_df)] = 0
                channel_df.loc[len(channel_df) - 1][time_col_name] = time

            channel_df = channel_df.sort_values(
                axis=0,
                by=mngr.data_prep_day_column_name,
                ascending=True,
                ignore_index=True,
            )

            injected_channel_df = flare.apply_to_time_series(
                channel_tag, channel_df, None, None
            )
            print("test")
            channel_df = injected_channel_df

        data[channel_tag] = channel_df

        # for each channel
        # randomly generate zero data points accor
        pass

    pass

    pass


def create_dense_input_for_flare_injections(
    mngr,
    flares: List[PhysicalFlare],
    data: Dict[str, pd.DataFrame],
):
    for channel_tag in data:
        pass
    pass

    return data
