from __future__ import annotations

import random
from typing import TYPE_CHECKING, Dict, List

import numpy as np
import pandas as pd

# from line_profiler_pycharm import profile
from trans_finder.flares.data_prep_helpers.physical_flare import PhysicalFlare
from trans_finder.flares.data_prep_helpers.temporal_injection import (
    TemporalInjection,
    TemporalInjectionGenerator,
)
from trans_finder.flares.inspections.data_prep_temporal_offsets import (
    inspect_temporal_offsets,
)

if TYPE_CHECKING:
    from trans_finder.flares.manager import Manager


class Realisation:
    """the raw input data is randomized into different realisations. This class represents a single realisation.

    Attributes:
        data_indices        for each channel, rows of indices on data points
        final_step_index    for each channel, the index of the final step in a channel, used to derive real times
    """

    random_time_shift_fraction_of_total_duration = 0.5

    def __init__(
        self,
        mngr,
        realisation_id: int,
        channel_tags: List[str],
        has_randomized_real_times: bool,
        has_shifted_real_times: bool,
        shuffling_strategy: str,
        input_data: Dict[str, pd.DataFrame],
        realisation_seed,
        custom_name="",
    ):
        self.mngr: Manager = mngr
        self.id = realisation_id
        self.has_randomized_real_times = has_randomized_real_times
        self.has_shifted_real_times = has_shifted_real_times
        self.data_indices = {channel_tag: [] for channel_tag in channel_tags}
        self.final_step_index = {channel_tag: [] for channel_tag in channel_tags}
        self.signal_injections: List[TemporalInjection] = []
        self.flare_template_injections: List[PhysicalFlare] = []
        self.long_term_states: List[TemporalInjection] = []
        self.channel_tags = channel_tags
        self.n_events_per_event_class = {}
        self.shuffling_strategy = shuffling_strategy
        self.ref_background_mean = {}
        self.ref_background_std = {}
        self.input_data = {tag: input_data[tag] for tag in channel_tags}
        self.real_times = {
            tag: input_data[tag][mngr.data_prep_day_column_name] for tag in channel_tags
        }
        self.seed = realisation_seed
        self.custom_name = custom_name

        self.set_seed()

    def set_seed(self):
        np.random.seed(self.seed)
        random.seed(self.seed)

    def get_n_events(self):
        return len(next(iter(self.data_indices.values())))

    def get_first_temporal_injection_start_time(self, mngr, randomize=False):
        right_window_index = {
            channel_tag: mngr.data_prep.n_steps_shuffling_window
            - 1
            + (random.randint(0, int(0.25 * mngr.n_steps_tot)) if randomize else 0)
            for channel_tag in self.channel_tags
        }

        real_times_at_right_window_index = [
            self.real_times[channel_tag][right_window_index[channel_tag]]
            for channel_tag in self.channel_tags
        ]

        return max(real_times_at_right_window_index)

    def correlate_time_window_steps_individually_per_channel(self):
        """
        mimic the time structure of mutliple time series,
        where only one new time step is added to the time series and the oldest one is removed, the rest remains
        For this: use n-1 steps from the previous time series if it has moved in the current channel,
        otherwise use the entire previous time series for that channel
        """
        n_slices = self.get_n_events()
        for i_slice in range(1, n_slices):
            for channel_tag in self.channel_tags:
                # use the same previous values, if the final step in time is the same.

                previous_final_step = self.final_step_index[channel_tag][i_slice - 1]
                current_final_step = self.final_step_index[channel_tag][i_slice]

                previous_time_series = self.data_indices[channel_tag][i_slice - 1]

                if previous_final_step == current_final_step:
                    self.data_indices[channel_tag][i_slice] = previous_time_series

                else:
                    # final step has increased by one
                    assert current_final_step == previous_final_step + 1

                    # use n-1 steps from previous and one of the new window

                    self.data_indices[channel_tag][i_slice][:-1] = previous_time_series[
                        1:
                    ]

    def assert_length_of_index_slices(self):
        for channel_tag in self.channel_tags:
            if any(
                len(x) != self.mngr.data_prep.n_steps_shuffling_window
                for x in self.data_indices[channel_tag]
            ):
                raise ValueError(
                    f"Expected uniform length in index slices. Not the case in {channel_tag}."
                )

    def get_global_time_structure_indices(self):
        # start with first slice
        indices = {k: self.data_indices[k][0] for k in self.channel_tags}

        n_slices = self.get_n_events()
        for i_slice in range(1, n_slices):
            for channel_tag in self.channel_tags:
                # use the same previous values, if the final step in time is the same.

                previous_final_step = self.final_step_index[channel_tag][i_slice - 1]
                current_final_step = self.final_step_index[channel_tag][i_slice]

                if previous_final_step != current_final_step:
                    # final step has increased by one
                    assert current_final_step == previous_final_step + 1

                    # add one of the new window
                    indices[channel_tag].append(
                        self.data_indices[channel_tag][i_slice][-1]
                    )

        self.global_data_indices = indices

    # @profile
    def determine_ref_background_mean_and_std(self, mngr):
        n_last_ref_steps = mngr.data_prep.data_prep_n_steps_ref_mean_and_std

        if (
            n_last_ref_steps + mngr.n_steps_tot
            > mngr.data_prep.n_steps_shuffling_window
        ):
            raise ValueError(
                "n_steps_shuffling_window needs to be at least of size data_prep_n_steps_ref_mean_and_std + n_steps_tot."
            )

        for channel_tag in self.channel_tags:
            relevant_cols = mngr.data_prep_fluctuation_cols[channel_tag]
            channel_ref_mean = []
            channel_ref_std = []
            for i_sliding_window, channel_data_indices in enumerate(
                self.data_indices[channel_tag]
            ):
                df = self.input_data[channel_tag].loc[
                    channel_data_indices, relevant_cols
                ]

                rolling = df.reset_index(drop=True).rolling(
                    n_last_ref_steps, closed="left"
                )

                channel_ref_std.append(
                    rolling.std().iloc[-mngr.n_steps_tot :].reset_index(drop=True)
                )
                channel_ref_mean.append(
                    rolling.mean().iloc[-mngr.n_steps_tot :].reset_index(drop=True)
                )

            assert all(
                np.isfinite(slice_df.to_numpy()).all() for slice_df in channel_ref_mean
            )
            assert all(
                np.isfinite(slice_df.to_numpy()).all() for slice_df in channel_ref_std
            )

            self.ref_background_mean[channel_tag] = channel_ref_mean
            self.ref_background_std[channel_tag] = channel_ref_std

    def add_long_term_states(
        self, mngr, temporal_injection_generator: TemporalInjectionGenerator
    ):
        temporal_injections = []

        first_temporal_injection_start_time = (
            self.get_first_temporal_injection_start_time(mngr=mngr, randomize=False)
        )

        input_data_end_time = max(
            (values.iloc[-1] for values in self.real_times.values())
        )

        fluct_start_time = first_temporal_injection_start_time

        while True:
            # randomize start time, adding a random offet of up to 5 * the min time bin width
            random_diff = np.random.uniform(
                mngr.data_prep.data_prep_analysis_min_time_bin_width,
                mngr.data_prep.data_prep_analysis_min_time_bin_width * 5,
            )
            fluct_start_time = fluct_start_time + random_diff

            temporal_injection = (
                temporal_injection_generator.generate_randomized_temporal_injection(
                    mngr=mngr, start_time=fluct_start_time
                )
            )

            if temporal_injection.end_time > input_data_end_time:
                break

            temporal_injections.append(temporal_injection)

            fluct_start_time = temporal_injection.end_time + np.random.uniform(
                0, temporal_injection.end_time - temporal_injection.start_time
            )

        for temporal_injection in temporal_injections:
            self.long_term_states.append(temporal_injection)

    def add_short_term_states(
        self, mngr, temporal_injection_generator: TemporalInjectionGenerator
    ):
        temporal_injections = []

        first_temporal_injection_start_time = (
            self.get_first_temporal_injection_start_time(mngr=mngr, randomize=False)
        )

        input_data_end_time = max(
            (values.iloc[-1] for values in self.real_times.values())
        )

        fluct_start_time = first_temporal_injection_start_time

        while True:
            temporal_injection = (
                temporal_injection_generator.generate_randomized_temporal_injection(
                    mngr=mngr, start_time=fluct_start_time
                )
            )

            if temporal_injection.end_time > input_data_end_time:
                break

            temporal_injections.append(temporal_injection)

            distance = np.random.uniform(
                0,
                mngr.n_steps_dec * mngr.data_prep.data_prep_analysis_min_time_bin_width,
            )

            fluct_start_time = temporal_injection.end_time + distance

        for temporal_injection in temporal_injections:
            self.signal_injections.append(temporal_injection)

    def shift_real_times(self, mngr):
        shifted_real_times = {}

        min_real_time_across_channels = min(
            self.real_times[channel_tag].iat[0] for channel_tag in self.channel_tags
        )
        max_real_time_across_channels = max(
            self.real_times[channel_tag].iat[-1] for channel_tag in self.channel_tags
        )

        total_time_range = max_real_time_across_channels - min_real_time_across_channels
        maximum_shift_value = int(
            total_time_range * self.random_time_shift_fraction_of_total_duration
        )

        channel_shifts = {
            channel_tag: random.randrange(
                -maximum_shift_value,
                maximum_shift_value,
                mngr.data_prep_n_days_per_step[channel_tag],
            )
            for channel_tag in self.channel_tags
        }

        # if configured, shift groups of channels always in the same way
        for _, channels in self.mngr.augmentation_shifting_groups.items():
            shift = None
            for i, channel in enumerate(channels):
                if i == 0:
                    shift = channel_shifts[channel]
                else:
                    channel_shifts[channel] = shift

        for channel_tag in self.channel_tags:
            channel_shift = channel_shifts[channel_tag]
            shifted_times_in_channel = pd.Series(
                (t + channel_shift) % (max_real_time_across_channels + 1)
                for t in self.real_times[channel_tag]
            )
            shifted_real_times[channel_tag] = shifted_times_in_channel.sort_values(
                ignore_index=True
            )

            delta_t_in_shifted_times = shifted_real_times[channel_tag].diff()
            delta_t_in_shifted_times.at[0] = mngr.data_prep_n_days_per_step[channel_tag]

            assert not any(
                delta_t_in_shifted_times < mngr.data_prep_n_days_per_step[channel_tag]
            )

        self.real_times = shifted_real_times

    def randomize_temporal_offsets(self, mngr):
        """change the real_times per channel, where the first and last time remain unchanged,
        and the gaps (here "offsets") between points i and i-1 are randomised"""

        randomized_real_times = {}

        for channel_tag in self.channel_tags:
            time_bin_width = mngr.data_prep_n_days_per_step[channel_tag]

            channel_times = self.real_times[channel_tag].to_numpy() / time_bin_width

            times_0 = np.asarray(channel_times[:-1])
            times_1 = np.asarray(channel_times[1:])
            time_diffs = times_1 - times_0
            rnd_diff_scales = np.random.rand(len(time_diffs))
            rnd_diffs = np.floor(rnd_diff_scales * time_diffs).astype(int)

            times_final = rnd_diffs + times_0
            times_final = np.append(times_final, channel_times[-1])

            randomized_real_times[channel_tag] = pd.Series(
                times_final * time_bin_width, index=self.real_times[channel_tag].index
            )

        if mngr.debug_data_prep_inspect_temporal_offset_dists:
            inspect_temporal_offsets(
                self.mngr, self.real_times, randomized_real_times, self.id
            )

        self.real_times = randomized_real_times

    def use_copy_of_input_data(self):
        # copy the input data so that these plateaus do not affect other realisations
        for channel_tag in self.channel_tags:
            self.input_data[channel_tag] = self.input_data[channel_tag].copy()

    def smear_values_by_uncertainty(self, mngr):
        for channel_tag in self.channel_tags:
            df_smeared = mngr.data.smear_df_by_errors(
                self.input_data[channel_tag],
                error_up_cols=mngr.data_prep_error_up_cols[channel_tag],
                error_down_cols=mngr.data_prep_error_down_cols[channel_tag],
                value_cols=mngr.data_prep_error_feature_cols[channel_tag],
            )
            self.input_data[channel_tag] = df_smeared

    def replace_nan_feature_values(self, mngr):
        for channel_tag in self.channel_tags:
            df = self.input_data[channel_tag]
            # imputed_df = df.copy()

            # loop in sliding window
            window_size = mngr.data_prep.n_steps_feature_imputation_window
            for right_edge in range(min(len(df), window_size) - 1, len(df)):
                left_edge = max(0, right_edge - window_size)
                window_df = df.loc[left_edge:right_edge]

                for ft_name in mngr.data_prep_fluctuation_cols[channel_tag]:
                    # for each feature, replace all nan values with any random values from the non nan values
                    is_null = window_df[ft_name].isnull()
                    n_null_values = is_null.sum()

                    if n_null_values == right_edge - left_edge + 1:
                        raise Exception(
                            f"""Encountered a sliding window of only NaN values in {channel_tag}_{ft_name}. 
                        You may want to double check the data or increase the window size for random replacement of missing values in the config: 
                        data_prep_imputation_sliding_window_scale"""
                        )

                    if n_null_values > 0:
                        given_values = window_df[ft_name].loc[~is_null]

                        # sample with replacement only if necessary
                        df[ft_name].loc[is_null[is_null == True].index] = (
                            given_values.sample(
                                n_null_values,
                                replace=n_null_values > len(given_values),
                                ignore_index=True,
                            ).to_numpy()
                        )

            assert (
                not self.input_data[channel_tag][
                    mngr.data_prep_fluctuation_cols[channel_tag]
                ]
                .isna()
                .any(axis=None)
            )

    def apply_longterm_background_states(
        self,
        mngr,
    ):
        if len(self.long_term_states) == 0:
            return

        before_plateau_stds = {}
        for channel_tag in self.channel_tags:
            # get a dataframe of the global std for each data point (rows) and channel feature (columns)
            relevant_cols = self.mngr.data_prep_fluctuation_cols[channel_tag]
            stds = self.input_data[channel_tag][relevant_cols].std()
            before_plateau_stds[channel_tag] = stds

        for long_term_state in self.long_term_states:
            plateau_amplitude = long_term_state.get_random_amplitude()

            for channel_tag in self.channel_tags:
                input_data_with_long_term_state = long_term_state.apply_to_time_series(
                    channel_tag,
                    self.input_data[channel_tag],
                    plateau_amplitude,
                    before_plateau_stds[channel_tag],
                )

                self.input_data[channel_tag] = input_data_with_long_term_state

    def add_channel_padding(self, channel_padding):
        # add padding to realisation.real_times and to realisation.aggregated_data

        for channel_tag in self.channel_tags:
            if channel_padding[channel_tag] is not None:
                self.input_data[channel_tag] = pd.concat(
                    [
                        channel_padding[channel_tag],
                        self.input_data[channel_tag],
                    ]
                ).reset_index(drop=True)[self.input_data[channel_tag].columns]

                self.real_times[channel_tag] = pd.concat(
                    [
                        channel_padding[channel_tag][
                            self.mngr.data_prep_day_column_name
                        ],
                        self.real_times[channel_tag],
                    ]
                ).reset_index(drop=True)

    def apply_flare_templates(self, mngr):
        raise NotImplementedError()

    def apply_global_flat_plateaus(self, mngr):
        """multiply all data points of each channel by a randomized scalar,
        if interval is configured in data_prep_global_flat_plateaus"""
        if hasattr(mngr, "data_prep_global_flat_plateaus"):
            for channel_tag in self.channel_tags:
                # get a dataframe of the global std for each data point (rows) and channel feature (columns)
                relevant_cols = self.mngr.data_prep_fluctuation_cols[channel_tag]
                conf_ampiltude = mngr.data_prep_global_flat_plateaus["amplitude"]
                channel_amplitude = np.random.uniform(
                    conf_ampiltude["min"], conf_ampiltude["max"]
                )
                stds = self.input_data[channel_tag][relevant_cols].std()
                self.input_data[channel_tag][relevant_cols] += stds * channel_amplitude
