from __future__ import annotations

import bisect
import random
from typing import TYPE_CHECKING

import pandas as pd

if TYPE_CHECKING:
    from trans_finder.flares.manager import Manager


def sliding_window(mngr: Manager, realisation, n_steps_sliding_window):
    # (A) Find the maximum real time across channels (max_real_time_across_channels) for the current time series
    # where there are enough data points in each channel

    right_window_index = {
        channel_tag: n_steps_sliding_window - 1
        for channel_tag in realisation.channel_tags
    }

    real_times_at_right_window_index = [
        realisation.real_times[channel_tag][right_window_index[channel_tag]]
        for channel_tag in realisation.channel_tags
    ]

    max_real_time_at_window_end = max(real_times_at_right_window_index)

    # Find all possible end times of an ouput time series

    all_real_times = pd.concat(
        [
            realisation.real_times[channel_tag]
            for channel_tag in realisation.channel_tags
        ],
        ignore_index=True,
    )

    max_real_times_of_timeseries = (
        all_real_times.where(all_real_times >= max_real_time_at_window_end)
        .dropna()
        .sort_values()
    )

    # filter max_real_times_of_time series by entries that are closer to the previous value
    # than the analysis_min_time_bin_width
    # and: remove duplicated times
    max_real_times_of_timeseries = mngr.data_prep.ensure_fixed_time_intervals(
        max_real_times_of_timeseries
    ).unique()

    # time_count, bin_edges = np.histogram(
    #     max_real_times_of_timeseries,
    #     np.arange(
    #         max_real_times_of_timeseries.iat[0],
    #         max_real_times_of_timeseries.iat[-1],
    #         mngr.data_prep.data_prep_analysis_min_time_bin_width,
    #     ),
    # )
    #
    # # take the right interval edge (corresponding to larger time) as max_real_time
    # max_real_times_of_timeseries = np.where(time_count > 0, bin_edges[1:], np.nan)

    for time in max_real_times_of_timeseries:
        yield time


def sliding_window_indices(
    mngr: Manager, realisation, channel_tag, n_steps_sliding_window
):
    for max_real_time_of_timeseries in sliding_window(
        mngr, realisation, n_steps_sliding_window
    ):
        # right_range_end = the index of the last real time that is <= max_real_time_of_timeseries
        right_range_end = bisect.bisect_right(
            realisation.real_times[channel_tag],
            max_real_time_of_timeseries,
        )
        left_index = right_range_end - n_steps_sliding_window
        window_index_list = list(range(left_index, right_range_end))
        yield window_index_list


def do_step_based_shuffling(mngr: Manager, realisation):
    for channel_tag in realisation.channel_tags:
        for window_index_list in sliding_window_indices(
            mngr,
            realisation,
            channel_tag,
            mngr.data_prep.n_steps_shuffling_window,
        ):
            realisation.final_step_index[channel_tag].append(window_index_list[-1])

            shuffled_indices = random.sample(
                window_index_list, mngr.data_prep.n_steps_shuffling_window
            )

            realisation.data_indices[channel_tag].append(shuffled_indices)


def do_no_shuffling(mngr: Manager, realisation):
    for channel_tag in realisation.channel_tags:
        for window_index_list in sliding_window_indices(
            mngr,
            realisation,
            channel_tag,
            mngr.data_prep.n_steps_shuffling_window,
        ):
            realisation.final_step_index[channel_tag].append(window_index_list[-1])
            realisation.data_indices[channel_tag].append(window_index_list)
