from abc import ABC, abstractmethod
from typing import Dict

import numpy as np
import pandas as pd


class TemporalInjection(ABC):
    start_time: float
    end_time: float
    channel_start_times: Dict[str, float]
    channel_end_times: Dict[str, float]
    type_name: str
    mngr: None
    has_randomized_amplitude: bool = False
    amplitude_conf = None

    @abstractmethod
    def apply_to_time_series(
        self, channel_tag, channel_values: pd.DataFrame, amplitude, stddev
    ):
        """
        Args:
            amplitude (np.array): an array with one element per channel feature
        """
        pass

    def get_amplitudes_with_all_features_identical(self):
        assert self.amplitude_conf is not None

        amplitudes = {}

        for channel_tag in self.mngr.data_prep_channel_tags:
            channel_amplitude = np.random.uniform(
                self.amplitude_conf[channel_tag]["min"],
                self.amplitude_conf[channel_tag]["max"],
            )
            n_features = len(self.mngr.data_prep_fluctuation_cols[channel_tag])
            amplitudes[channel_tag] = np.repeat(channel_amplitude, repeats=n_features)

        return amplitudes

    def get_amplitudes_with_randomized_feature_amplitudes(self):
        assert self.amplitude_conf is not None

        amplitudes = {}

        for channel_tag in self.mngr.data_prep_channel_tags:
            max_channel_amplitude = np.random.uniform(
                self.amplitude_conf[channel_tag]["min"],
                self.amplitude_conf[channel_tag]["max"],
            )
            n_features = len(self.mngr.data_prep_fluctuation_cols[channel_tag])
            feature_amplitudes = np.random.uniform(
                0, max_channel_amplitude, size=n_features
            )
            amplitudes[channel_tag] = feature_amplitudes

        return amplitudes

    @abstractmethod
    def get_random_amplitude(self):
        pass

    def get_temporal_overlap(self, window_start, window_end, channel_tag):
        if channel_tag not in self.channel_start_times:
            return 0

        overlap = max(
            0,
            min(window_end, self.channel_end_times[channel_tag])
            - max(window_start, self.channel_start_times[channel_tag]),
        )
        return overlap

    def may_affect_channel_values(
        self, channel_tag, channel_values: pd.DataFrame
    ) -> bool:
        time_col_name = self.mngr.data_prep_day_column_name
        window_start = min(channel_values[time_col_name])
        window_end = max(channel_values[time_col_name])

        if self.start_time == window_end:
            return True

        overlap = self.get_temporal_overlap(window_start, window_end, channel_tag)

        is_there_temporal_overlap = overlap > 0

        return is_there_temporal_overlap

    def get_n_affected_steps(self, channel_tag, channel_values):
        if channel_tag not in self.channel_start_times:
            return {
                self.mngr.tag_enc: 0,
                self.mngr.tag_dec: 0,
            }

        time_col = channel_values[self.mngr.data_prep_day_column_name]

        temporal_injection_time_start = self.channel_start_times[channel_tag]
        temporal_injection_time_end = self.channel_end_times[channel_tag]

        encoder_times = time_col.iloc[: self.mngr.n_steps_enc]
        decoder_times = time_col.iloc[self.mngr.n_steps_enc :]

        n_affected_data_points_enc = encoder_times[
            (encoder_times >= temporal_injection_time_start)
            & (encoder_times <= temporal_injection_time_end)
        ].count()
        n_affected_data_points_dec = decoder_times[
            (decoder_times >= temporal_injection_time_start)
            & (decoder_times <= temporal_injection_time_end)
        ].count()

        return {
            self.mngr.tag_enc: n_affected_data_points_enc,
            self.mngr.tag_dec: n_affected_data_points_dec,
        }

    # def get_n_steps_with_signal_over_threshold(self, channel_tag, channel_values):
    #     time_col = channel_values[self.mngr.data_prep_day_column_name]
    #     # relevant_cols = self.mngr.data_prep_fluctuation_cols[channel_tag]
    #
    #     # feature_stds = self.mngr.data_prep_feature_stds[channel_tag][relevant_cols]
    #     threshold = self.mngr.data_prep_signal_noise_ratio_in_feature_stds
    #     # feature_means = self.mngr.data_prep_feature_means[channel_tag][relevant_cols]
    #
    #     encoder_times = time_col.iloc[: self.mngr.n_steps_enc]
    #     decoder_times = time_col.iloc[self.mngr.n_steps_enc :]
    #
    #     intensities_enc = self.get_intensity_at(
    #         encoder_times, channel_tag, channel_values
    #     )
    #     intensities_dec = self.get_intensity_at(
    #         decoder_times, channel_tag, channel_values
    #     )
    #
    #     encoder_signal_steps = intensities_enc.ge(threshold).any(axis=1).sum()
    #     decoder_signal_steps = intensities_dec.ge(threshold).any(axis=1).sum()
    #
    #     # encoder_signal_steps = (
    #     #     channel_values.loc[: self.mngr.n_steps_enc, relevant_cols]
    #     #     .subtract(feature_means)
    #     #     .abs()
    #     #     .ge(feature_stds * threshold)
    #     #     .any(axis=1)
    #     #     .sum()
    #     # )
    #     #
    #     # decoder_signal_steps = (
    #     #     channel_values.loc[self.mngr.n_steps_enc :, relevant_cols]
    #     #     .subtract(feature_means)
    #     #     .abs()
    #     #     .ge(feature_stds * threshold)
    #     #     .any(axis=1)
    #     #     .sum()
    #     # )
    #
    #     result = {
    #         self.mngr.tag_enc: encoder_signal_steps,
    #         self.mngr.tag_dec: decoder_signal_steps,
    #     }
    #
    #     return result

    # @classmethod
    # @abstractmethod
    # def get_description_column_names(cls):
    #     pass
    #
    # @classmethod
    # @abstractmethod
    # def get_description_default_values(cls):
    #     pass
    #
    # @abstractmethod
    # def get_description_values(self):
    #     pass


class TemporalInjectionGenerator(ABC):
    @abstractmethod
    def generate_randomized_temporal_injection(
        self, mngr, start_time
    ) -> TemporalInjection:
        pass
