from __future__ import annotations

import os
from pathlib import Path

import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt
from tensorflow.estimator import ModeKeys

from trans_finder.flares.core.tags import TAG_AUTO_ENC_INPUT_BEFORENORM
from trans_finder.flares.helpers_debugging import (
    clear_directory,
    get_debug_output_filepath,
)
from trans_finder.flares.model_utils.remove_temporal_offsets import (
    remove_temporal_offsets,
)
from trans_finder.flares.model_utils.utils import get_batch_size
from trans_finder.flares.plot_list.heatmap_prediction_vs_truth_1d import (
    plot_2dhistogram_of_ndarray_per_step_and_feature,
)
from trans_finder.flares.plot_list.residual_violin_plot_grid import violin_plot_grid
from trans_finder.flares.plot_list.util.ecdf import plot_ecdf
from trans_finder.flares.plot_list.util.plot_config import gridlines_config, tick_params
from trans_finder.flares.utilities.optional_aim import log_metrics_in_aim
from trans_finder.flares.utilities.pdf_plot_collection import PDFPlotCollection
from trans_finder.flares.utils import log_table


def calculate_mae_per_feature(y_true, y_pred, ft_names):
    assert y_true.shape == y_pred.shape
    assert len(ft_names) == y_true.shape[-1]

    scores = {}

    for ft_idx, ft_name in enumerate(ft_names):
        ft_mae = np.mean(np.abs(y_true[..., ft_idx] - y_pred[..., ft_idx]))
        scores[ft_name] = ft_mae

    mean_ft_mae = np.mean(list(scores.values()))
    scores["* mean"] = mean_ft_mae

    return scores


def calculate_r2_per_feature(y_true, y_pred, ft_names):
    assert y_true.shape == y_pred.shape
    assert len(ft_names) == y_true.shape[-1]

    all_ft_sum_y_true_ssd = 0
    all_ft_sum_y_pred_ssd = 0

    scores = {}

    for ft_idx, ft_name in enumerate(ft_names):
        y_true_mean = np.mean(y_true[..., ft_idx])
        y_true_sum_squared_diff = np.sum(np.square(y_true[..., ft_idx] - y_true_mean))
        all_ft_sum_y_true_ssd += y_true_sum_squared_diff

        y_pred_sum_squared_diff = np.sum(
            np.square(y_true[..., ft_idx] - y_pred[..., ft_idx])
        )
        all_ft_sum_y_pred_ssd += y_pred_sum_squared_diff

        scores[ft_name] = 1 - (y_pred_sum_squared_diff / y_true_sum_squared_diff)

    scores["* overall"] = 1 - (all_ft_sum_y_pred_ssd / all_ft_sum_y_true_ssd)

    return scores


from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from trans_finder.flares.manager import Manager


class KerasModelDebugger:
    tag = ["b", f" - flares.{Path(__file__).stem} - "]

    def __init__(self, mngr):
        self.mngr: Manager = mngr
        self.root_dir = None

    def run(self):
        mngr = self.mngr
        mngr.log.info([self.tag, ["p", "run()"]])

        # mode = ModeKeys.TRAIN
        # mode = ModeKeys.EVAL
        for mode in [ModeKeys.TRAIN, ModeKeys.EVAL]:
            data_split_tag = mode
            input_dict = {}
            dataset = mngr.data.read_dataset(mode=mode, full_data=input_dict)
            # dataset_eval = self.mngr.data.read_dataset(mode=ModeKeys.EVAL)

            # input_df_meta = self.mngr.input_data[self.mngr.tag_input_df][mode]

            # input_df_meta = input_dict["input_df_meta"]
            mngr.train.load_model(dataset)

            model = mngr.model["keras_model"]

            # assert len(model_outputs[mngr.tag_input_d1]) == len(input_df_meta)

            for n_phase in range(len(mngr.train_phases)):
                mngr.train_phase = mngr.train_phases[n_phase]

                if mngr.train_phase == 0:
                    self.do_plots_forecasting(model, dataset, data_split_tag)

                elif mngr.train_phase == 1:
                    self.do_plots_reconstruction(model, dataset, data_split_tag)

    def calculate_metrics_per_feature_forecasting(self, model_outputs, data_split_tag):
        input_d1 = model_outputs[self.mngr.tag_input_d1]
        forecast = model_outputs[self.mngr.tag_bck_forecast_d1]

        for metric_name, metric_fn in [
            # ("r2", calculate_r2_per_feature),
            ("MAE", calculate_mae_per_feature),
        ]:
            metric_per_feature = metric_fn(
                y_true=remove_temporal_offsets(self.mngr, input_d1),
                y_pred=remove_temporal_offsets(self.mngr, forecast),
                ft_names=[
                    n
                    for n in self.mngr.step_feature_names
                    if self.mngr.tag_temporal_offset not in n
                ],
            )

            log_metrics_in_aim(
                self.mngr,
                metric_per_feature,
                data_split_tag,
                prefix=f"forecasting_{metric_name}",
            )

            log_table(
                self.mngr.log.info,
                metric_per_feature,
                title=f" - forecasting {metric_name} per feature ({data_split_tag} dataset, train_phase={self.mngr.train_phase}):",
                add_space=True,
            )

    def calculate_metrics_per_feature_reconstruction(
        self, model_outputs, data_split_tag
    ):
        # input_d1 = model_outputs[self.mngr.tag_input_d1]
        # forecast = model_outputs[self.mngr.tag_bck_forecast_d1]

        auto_enc_input = model_outputs[self.mngr.tag_auto_enc_input]
        reconstruction = model_outputs[self.mngr.tag_auto_enc_reconstruction]

        for metric_name, metric_fn in [
            # ("r2", calculate_r2_per_feature),
            ("MAE", calculate_mae_per_feature),
        ]:
            metric_per_feature = metric_fn(
                y_true=auto_enc_input,
                y_pred=reconstruction,
                ft_names=[
                    n
                    for n in self.mngr.step_feature_names
                    if self.mngr.tag_temporal_offset not in n
                ],
            )

            log_metrics_in_aim(
                self.mngr,
                metric_per_feature,
                data_split_tag,
                prefix=f"reconstruction_{metric_name}",
            )

            log_table(
                self.mngr.log.info,
                metric_per_feature,
                title=f" - reconstruction {metric_name} per feature ({data_split_tag} dataset, train_phase={self.mngr.train_phase}):",
                add_space=True,
            )

    def log_autoenc_input_dists(self, model_outputs, data_split_tag):
        diff = model_outputs[TAG_AUTO_ENC_INPUT_BEFORENORM]

        ft_names = [
            n
            for n in self.mngr.step_feature_names
            if self.mngr.tag_temporal_offset not in n
        ]

        assert len(ft_names) == diff.shape[-1]
        descriptions = {}
        for ft_idx, ft_name in enumerate(ft_names):
            descriptions[ft_name] = pd.DataFrame(
                diff[..., ft_idx].reshape(-1)
            ).describe()

        file_name = get_debug_output_filepath(
            self.mngr,
            os.path.join("autoencoder", data_split_tag),
            "d1_minus_forecast_dist_per_feature.csv",
        )
        pd.DataFrame().assign(**descriptions).to_csv(
            file_name
        )  # , float_format="%.3e")

    def do_calculate_detailed_metrics(
        self, model, dataset, data_split_tag, model_outputs=None
    ):
        pass

    def do_residual_violin_plot(self, model_outputs, data_split_tag):
        input_d1 = model_outputs[self.mngr.tag_input_d1]
        forecast = model_outputs[self.mngr.tag_bck_forecast_d1]

        residuals = input_d1 - forecast

        mngr = self.mngr
        ft_names = [
            n for n in mngr.step_feature_names if mngr.tag_temporal_offset not in n
        ]

        residuals_features = residuals[
            :,
            :,
            [
                i
                for i, n in enumerate(mngr.step_feature_names)
                if mngr.tag_temporal_offset not in n
            ],
        ]

        fig, axes = violin_plot_grid(
            residuals_features, ft_names, "Residuals (forecast)"
        )

        PDFPlotCollection(mngr, [fig]).save_to_file(
            get_debug_output_filepath(
                mngr,
                os.path.join("forecasting", data_split_tag),
                f"trainphase_{mngr.train_phase}_residuals.pdf",
            )
        )

    def do_plots_forecasting(self, model, dataset, data_split_tag):
        model_outputs = model.predict(
            dataset, batch_size=get_batch_size(self.mngr, force_predict_size=True)
        )

        # assert len(model_outputs[mngr.tag_input_d1]) == len(input_df_meta)

        assert self.mngr.train_phase == 0

        self.calculate_metrics_per_feature_forecasting(model_outputs, data_split_tag)

        input_d1 = model_outputs[self.mngr.tag_input_d1]
        input_d0 = model_outputs[self.mngr.tag_input_d0]
        forecast = model_outputs[self.mngr.tag_bck_forecast_d1]

        diff_forecast_d1 = model_outputs[TAG_AUTO_ENC_INPUT_BEFORENORM]

        variants = [("", 0.00), ("_clipped_outliers", 0.01)]

        ft_names = [
            n
            for n in self.mngr.step_feature_names
            if self.mngr.tag_temporal_offset not in n
        ]

        for postfix, outlier_clip in variants:
            # Forecasting 2dhistogram: truth vs prediction
            fig, _ = plot_2dhistogram_of_ndarray_per_step_and_feature(
                self.mngr,
                input_d1,
                forecast,
                self.mngr.step_feature_names,
                title=f"Forecasting truth vs prediction {postfix}",
                outlier_clip=outlier_clip,
                skip_offset_indices=True,
            )
            PDFPlotCollection(self.mngr, [fig]).save_to_file(
                get_debug_output_filepath(
                    self.mngr,
                    os.path.join("forecasting", data_split_tag),
                    f"trainphase_{self.mngr.train_phase}_2dhist_truth_vs_pred{postfix}.pdf",
                )
            )

        # Forecasting inputs
        fig, _ = self.plot_hists_of_ndarray_per_step_and_feature(
            self.mngr,
            input_d1,
            title="Forecasting inputs",
            skip_offset_indices=True,
        )
        PDFPlotCollection(self.mngr, [fig]).save_to_file(
            get_debug_output_filepath(
                self.mngr,
                os.path.join("forecasting", data_split_tag),
                f"trainphase_{self.mngr.train_phase}_inputs_d1.pdf",
            )
        )

        fig, _ = self.plot_hists_of_ndarray_per_step_and_feature(
            self.mngr, input_d0, title="Forecasting inputs", skip_offset_indices=True
        )
        PDFPlotCollection(self.mngr, [fig]).save_to_file(
            get_debug_output_filepath(
                self.mngr,
                os.path.join("forecasting", data_split_tag),
                f"trainphase_{self.mngr.train_phase}_inputs_d0.pdf",
            )
        )

        # Relative forecasting error
        x = forecast - input_d1
        y = input_d1

        # mask = np.abs(y) < 1e-10

        # x[mask] = 5
        # y[mask] = 1
        #
        # relative_error = x / y

        with np.errstate(divide="ignore", invalid="ignore"):
            c = np.true_divide(forecast - input_d1, input_d1)
            c[c == np.inf] = 5
            c = np.nan_to_num(c, nan=5, posinf=5, neginf=5)

        clip_value = 10
        c[c > clip_value] = clip_value
        c[c < -clip_value] = -clip_value

        fig, _ = self.plot_hists_of_ndarray_per_step_and_feature(
            self.mngr,
            c,
            title=f"Forecasting relative error, clipped at +/- {clip_value}, inf and None set to 5",
            hist_kwargs=dict(bins=50),
            ylog=False,
            skip_offset_indices=True,
        )
        PDFPlotCollection(self.mngr, [fig]).save_to_file(
            get_debug_output_filepath(
                self.mngr,
                os.path.join("forecasting", data_split_tag),
                f"trainphase_{self.mngr.train_phase}_relative_error.pdf",
            )
        )

        # Forecasting residuals
        fig, _ = self.plot_hists_of_ndarray_per_step_and_feature(
            self.mngr,
            diff_forecast_d1,
            title="Forecasting residuals",
        )
        PDFPlotCollection(self.mngr, [fig]).save_to_file(
            get_debug_output_filepath(
                self.mngr,
                os.path.join("forecasting", data_split_tag),
                f"trainphase_{self.mngr.train_phase}_residuals.pdf",
            )
        )

        # self.plot_per_dimension_forecasting_residual_histograms(
        #     self.mngr, model_outputs, "residuals of forecasting", data_split_tag
        # )

        error = np.mean(np.abs(diff_forecast_d1))
        title = f"mean absolute forecasting error: {error:.3f}"
        # self.plot_forecast_histogram_diff(mngr, model_outputs)
        self.plot_forecast_error_baseline_comparison(
            self.mngr, model_outputs, data_split_tag
        )
        self.plot_per_dimension_forecasting_histograms(
            self.mngr, model_outputs, title, data_split_tag
        )
        self.plot_per_dimension_forecasting_error(
            self.mngr, model_outputs, title, data_split_tag
        )

    def do_plots_reconstruction(self, model, dataset, data_split_tag):
        model_outputs = model.predict(
            dataset, batch_size=get_batch_size(self.mngr, force_predict_size=True)
        )

        assert self.mngr.train_phase == 1

        self.log_autoenc_input_dists(model_outputs, data_split_tag)

        # self.plot_r2_per_feature_forecasting(model_outputs, data_split_tag)
        self.calculate_metrics_per_feature_reconstruction(model_outputs, data_split_tag)

        # Reconstruction 2dhistogram per time step and feature: truth vs prediction
        ft_names = [
            n
            for n in self.mngr.step_feature_names
            if self.mngr.tag_temporal_offset not in n
        ]
        for postfix, outlier_clip, zeroclip in [
            ("_plain", 0.00, None),
            ("_masked_zero", 0.00, 0.3),
            ("_masked_zero_clipped_outliers", 0.01, 0.015),
            ("_clipped_outliers", 0.01, None),
        ]:
            fig, _ = plot_2dhistogram_of_ndarray_per_step_and_feature(
                self.mngr,
                truth=model_outputs[self.mngr.tag_auto_enc_input],
                prediction=model_outputs[self.mngr.tag_auto_enc_reconstruction],
                ft_names=ft_names,
                title=f"Reconstruction truth vs prediction {postfix}",
                outlier_clip=outlier_clip,
                clip_around_zero=zeroclip,
            )
            PDFPlotCollection(self.mngr, [fig]).save_to_file(
                get_debug_output_filepath(
                    self.mngr,
                    os.path.join("autoencoder", data_split_tag),
                    f"{self.mngr.train_phase}_2D_projection{postfix}.pdf",
                )
            )

        # auto_enc_input = model_outputs[self.mngr.tag_auto_enc_input]
        # reconstruction = model_outputs[self.mngr.tag_auto_enc_reconstruction]

        # error = self.mean_absolute_error(auto_enc_input, reconstruction)
        # title = f"mean absolute reconstruction error: {error:.3f}"
        #
        # self.plot_per_dimension_autoenc_residual_histograms(
        #     self.mngr, model_outputs, "residuals of reconstruction", data_split_tag
        # )
        auto_enc_input = model_outputs[self.mngr.tag_auto_enc_input]
        reconstruction = model_outputs[self.mngr.tag_auto_enc_reconstruction]
        fig, _ = self.plot_hists_of_ndarray_per_step_and_feature(
            self.mngr,
            auto_enc_input - reconstruction,
            title="Reconstruction residuals",
            ylog=True,
            skip_offset_labels=True,
        )
        PDFPlotCollection(self.mngr, [fig]).save_to_file(
            get_debug_output_filepath(
                self.mngr,
                os.path.join("autoencoder", data_split_tag),
                f"trainphase_{self.mngr.train_phase}_residuals.pdf",
            )
        )

        self.plot_per_dimension_autoenc_input_and_reconstruction_histograms(
            self.mngr, model_outputs, "input vs. reconstruction", data_split_tag
        )

        self.plot_d1_minus_forecast(self.mngr, model_outputs, data_split_tag)

        # self.do_error_analysis_on_reconstruction_error(
        #     mngr, model_outputs, input_df_meta
        # )

    def plot_forecast_histogram_diff(
        self,
        mngr,
        model_outputs,
        data_split_tag,
        plot_axis_range=None,
    ):
        input_d1 = model_outputs[mngr.tag_input_d1]
        forecast = model_outputs[mngr.tag_bck_forecast_d1]

        n_features = input_d1.shape[2]
        n_time_steps = input_d1.shape[1]

        input_norm = input_d1.copy()
        input_norm[np.abs(input_norm) < 1e-10] = 1

        # features = [f"feature {i}" for i in range(n_features)]

        diff = pd.DataFrame(
            np.mean((forecast - input_d1), axis=1),
            columns=[f"feature {i}" for i in range(n_features)],
        )
        normed_diff = (forecast - input_d1) / input_norm

        fig, axes = plt.subplots(
            nrows=n_time_steps,
            ncols=n_features,
            sharex=True,
            figsize=(5 * n_features, 6 * n_time_steps),
        )
        for axes_row, time_step in zip(axes, range(n_time_steps)):
            for dim, ax in zip(range(n_features), axes_row):
                if dim in mngr.temporal_offset_column_indices:
                    continue

                g = sns.histplot(
                    ax=ax,
                    data=normed_diff[:, time_step, dim],
                    legend=dim == 0,
                    common_norm=False,
                    element="step",
                )
                ax.set_xlabel(f"dimension {dim}")
                # ax.set_xlim(-1, 1)
                plt.xlabel(f"dimension {dim}")

        filename = f"trainphase_{mngr.train_phase}_forecasting_histogram_diff.pdf"
        output_path = get_debug_output_filepath(
            mngr, os.path.join("forecasting", data_split_tag), filename
        )
        PDFPlotCollection(mngr, [fig]).save_to_file(output_path)

    def mean_absolute_error(self, truth, prediction):
        return np.mean(
            np.abs(
                truth.reshape(truth.shape[0], -1)
                - prediction.reshape(truth.shape[0], -1)
            )
        )

    def plot_forecast_error_baseline_comparison(
        self,
        mngr,
        model_outputs,
        data_split_tag,
        plot_axis_range=None,
    ):
        input_d0 = model_outputs[mngr.tag_input_d0]
        input_d1 = model_outputs[mngr.tag_input_d1]
        forecast = model_outputs[mngr.tag_bck_forecast_d1]

        input_d1 = remove_temporal_offsets(self.mngr, input_d1)
        forecast = remove_temporal_offsets(self.mngr, forecast)

        n_time_steps = input_d1.shape[1]
        n_features = input_d1.shape[2]
        assert n_time_steps == forecast.shape[1]

        assert n_features == forecast.shape[2]

        def fc_error(predictions):
            return np.mean(
                np.abs(
                    input_d1.reshape(input_d1.shape[0], -1)
                    - predictions.reshape(input_d1.shape[0], -1)
                ),
                axis=1,
            )

        # get baseline forecast

        last_3_timesteps = remove_temporal_offsets(self.mngr, input_d0)[:, -3:, :]
        mean_of_last_3 = np.mean(last_3_timesteps, axis=1)
        l3_baseline = np.tile(mean_of_last_3, mngr.n_steps_dec).reshape(input_d1.shape)

        last_one = remove_temporal_offsets(self.mngr, input_d0)[:, -1, :]
        l1_baseline = np.tile(last_one, mngr.n_steps_dec).reshape(input_d1.shape)

        fig, ax = plt.subplots(
            nrows=1,
            ncols=1,
            figsize=(5, 6),
        )

        for label, predictions in {
            "forecast": forecast,
            "l3_baseline": l3_baseline,
            "l1_baseline": l1_baseline,
        }.items():
            plot_ecdf(ax, fc_error(predictions), label=label)
            # ax.hist(
            #     ,
            #     bins="auto",
            #     # alpha=0.5,
            #     label=label,
            #     # edgecolor="black",
            #     histtype="step",
            #     # linewidth=1.2,
            #     density=False,
            # )

        # Legend
        ax.legend()
        ax.grid(**gridlines_config)
        ax.tick_params(**tick_params)
        ax.set_xlabel("MAE")

        # g = sns.histplot(
        #     ax=ax,
        #     data={
        #         "forecast": forecasting_error,
        #         "baseline": baseline_error,
        #     },
        #     legend=True,
        #     # stat="density",
        #     common_norm=False,
        #     element="step",
        #     binrange=plot_axis_range,
        #     binwidth=0.05,
        # )

        filename = f"trainphase_{mngr.train_phase}_forecasting_error_vs_baseline.pdf"
        output_path = get_debug_output_filepath(
            mngr, os.path.join("forecasting", data_split_tag), filename
        )
        PDFPlotCollection(mngr, [fig]).save_to_file(output_path)

    def plot_hists_of_ndarray_per_step_and_feature(
        self,
        mngr,
        data,
        title=None,
        hist_kwargs=None,
        skip_offset_indices=False,
        skip_offset_labels=False,
        ylog=True,
    ):
        if hist_kwargs is None:
            hist_kwargs = {}

        if skip_offset_indices:
            data = remove_temporal_offsets(mngr, data)

        if skip_offset_indices or skip_offset_labels:
            ft_names = [
                n for n in mngr.step_feature_names if mngr.tag_temporal_offset not in n
            ]
        else:
            ft_names = mngr.step_feature_names

        assert data.ndim == 3
        n_time_steps = data.shape[1]
        n_features = data.shape[2]

        fig, axes = plt.subplots(
            nrows=n_time_steps,
            ncols=n_features,
            sharex=True,
            # sharey=True,
            figsize=(4.5 * n_features, 3.5 * n_time_steps),
        )
        if title:
            fig.suptitle(title, fontsize=24)
        plt.tight_layout(pad=2)

        for axes_row, time_step in zip(axes, range(n_time_steps)):
            for idx, (dim, ax) in enumerate(zip(range(n_features), axes_row)):
                # if (
                #     skip_offset_indices
                #     and dim in mngr.temporal_offset_column_indices
                # ):
                #     continue

                # g = sns.histplot(
                #     ax=ax,
                #     data={"residual": data[:, time_step, dim]},
                #     legend=dim == 0,
                #     common_norm=False,
                #     element="step",
                #     binwidth=0.05,
                # )

                if "bins" not in hist_kwargs:
                    hist_kwargs["bins"] = "auto"

                ax.hist(
                    data[:, time_step, dim],
                    histtype="step",
                    # label=data_labels,
                    **hist_kwargs,
                )
                ax.set_xlabel(ft_names[idx])
                if ylog:
                    ax.set_yscale("log")
                ax.grid(**gridlines_config)
                ax.tick_params(**tick_params)
                plt.setp(ax.get_xticklabels(), visible=True)
                # ax.set_xlim(-1, 1)
                # plt.xlabel(ft_name)

        return fig, axes

    def plot_per_dimension_autoenc_input_and_reconstruction_histograms(
        self,
        mngr,
        model_outputs,
        title,
        data_split_tag,
        plot_axis_range=None,
    ):
        auto_enc_input = model_outputs[mngr.tag_auto_enc_input]
        reconstruction = model_outputs[mngr.tag_auto_enc_reconstruction]

        n_time_steps = auto_enc_input.shape[1]
        n_features = auto_enc_input.shape[2]
        assert n_time_steps == reconstruction.shape[1]
        assert n_features == reconstruction.shape[2]

        fig, axes = plt.subplots(
            nrows=n_time_steps,
            ncols=n_features,
            sharex=True,
            figsize=(4 * n_features, 3 * n_time_steps),
        )
        fig.suptitle(title, fontsize=24)
        plt.tight_layout(pad=2)

        ft_names = [
            n for n in mngr.step_feature_names if mngr.tag_temporal_offset not in n
        ]

        for axes_row, time_step in zip(axes, range(n_time_steps)):
            for dim, ax, ft_name in zip(range(n_features), axes_row, ft_names):
                g = sns.histplot(
                    ax=ax,
                    data={
                        "reconstruction": reconstruction[:, time_step, dim],
                        "autoenc_input": auto_enc_input[:, time_step, dim],
                    },
                    legend=dim == 0,
                    common_norm=False,
                    element="step",
                    binrange=plot_axis_range,
                    binwidth=0.05,
                )
                ax.set_xlabel(ft_name)
                # ax.set_yscale("log")
                ax.grid(**gridlines_config)
                ax.tick_params(**tick_params)
                plt.setp(ax.get_xticklabels(), visible=True)
                # ax.set_xlim(-1, 1)
                # plt.xlabel(ft_name)

        filename = f"input_vs_reconstruction.pdf"
        output_path = get_debug_output_filepath(
            mngr, os.path.join("autoencoder", data_split_tag), filename
        )
        PDFPlotCollection(mngr, [fig]).save_to_file(output_path)

    def plot_per_dimension_forecasting_error(
        self,
        mngr,
        model_outputs,
        title,
        data_split_tag,
        plot_axis_range=None,
    ):
        input_d1 = model_outputs[mngr.tag_input_d1]
        forecast = model_outputs[mngr.tag_bck_forecast_d1]

        n_time_steps = input_d1.shape[1]
        n_features = input_d1.shape[2]
        assert n_time_steps == forecast.shape[1]
        assert n_features == forecast.shape[2]

        fig, axes = plt.subplots(
            nrows=n_time_steps,
            ncols=n_features,
            # sharey=True,
            sharex=True,
            figsize=(4 * n_features, 5 * n_time_steps),
        )
        fig.suptitle(title, fontsize=24)
        plt.tight_layout(pad=3)

        for axes_row, time_step in zip(axes, range(n_time_steps)):
            for dim, ax in zip(range(n_features), axes_row):
                error = forecast[:, time_step, dim] - input_d1[:, time_step, dim]
                # error = np.abs(error)

                if np.min(error) == np.max(error):
                    continue

                g = sns.histplot(
                    ax=ax,
                    data={
                        "forecasting_error": error,
                    },
                    # legend=dim == 0,
                    legend=False,
                    # stat="density",
                    common_norm=False,
                    element="step",
                    binrange=plot_axis_range,
                    binwidth=0.05,
                )
                ax.set_xlabel(mngr.step_feature_names[dim])
                # ax.set_xlim(-1, 1)
                # plt.xlabel(f"dimension {dim}")
                ax.grid(**gridlines_config)
                ax.set_yscale("log")

        filename = f"trainphase_{mngr.train_phase}_forecasting_error.pdf"
        output_path = get_debug_output_filepath(
            mngr, os.path.join("forecasting", data_split_tag), filename
        )
        PDFPlotCollection(mngr, [fig]).save_to_file(output_path)

    def plot_d1_minus_forecast(self, mngr, model_outputs, data_split_tag):
        ft_names = [
            n
            for n in self.mngr.step_feature_names
            if self.mngr.tag_temporal_offset not in n
        ]

        diff = model_outputs[TAG_AUTO_ENC_INPUT_BEFORENORM]
        n_time_steps = diff.shape[1]
        n_features = diff.shape[2]

        fig, axes = plt.subplots(
            nrows=n_time_steps,
            ncols=n_features,
            # sharey=True,
            sharex=True,
            figsize=(4 * n_features, 5 * n_time_steps),
        )
        fig.suptitle("Decoder data minus forecast", fontsize=24)
        plt.tight_layout(pad=3)

        for axes_row, time_step in zip(axes, range(n_time_steps)):
            for dim, ax in zip(range(n_features), axes_row):
                dimension_diff = diff[:, time_step, dim]
                # error = np.abs(error)

                if np.min(dimension_diff) == np.max(dimension_diff):
                    continue

                g = sns.histplot(
                    ax=ax,
                    data={
                        "d1 - forecast": dimension_diff,
                    },
                    # legend=dim == 0,
                    legend=False,
                    # stat="density",
                    common_norm=False,
                    element="step",
                    # binrange=plot_axis_range,
                    binwidth=0.05,
                )
                ax.set_xlabel(ft_names[dim])
                # ax.set_xlim(-1, 1)
                # plt.xlabel(f"dimension {dim}")
                ax.grid(**gridlines_config)

        filename = "d1_minus_forecast.pdf"
        output_path = get_debug_output_filepath(
            mngr, os.path.join("autoencoder", data_split_tag), filename
        )
        PDFPlotCollection(mngr, [fig]).save_to_file(output_path)

    def plot_per_dimension_forecasting_histograms(
        self,
        mngr,
        model_outputs,
        title,
        data_split_tag,
        plot_axis_range=None,
    ):
        input_d1 = model_outputs[mngr.tag_input_d1]
        forecast = model_outputs[mngr.tag_bck_forecast_d1]

        # input_d1 = remove_temporal_offsets(self.mngr, input_d1)
        # forecast = remove_temporal_offsets(self.mngr, forecast)

        n_time_steps = input_d1.shape[1]
        n_features = input_d1.shape[2]
        assert n_time_steps == forecast.shape[1]
        assert n_features == forecast.shape[2]

        fig, axes = plt.subplots(
            nrows=n_time_steps,
            ncols=n_features,
            # sharey=True,
            sharex=True,
            figsize=(4 * n_features, 5 * n_time_steps),
        )
        fig.suptitle(title, fontsize=24)
        plt.tight_layout(pad=3)

        for axes_row, time_step in zip(axes, range(n_time_steps)):
            for dim, ax in zip(range(n_features), axes_row):
                g = sns.histplot(
                    ax=ax,
                    data={
                        "forecast": forecast[:, time_step, dim],
                        "input_d1": input_d1[:, time_step, dim],
                    },
                    legend=dim == 0,
                    # stat="density",
                    common_norm=False,
                    element="step",
                    binrange=plot_axis_range,
                    binwidth=0.05,
                )
                ax.set_yscale("log")
                ax.set_xlabel(mngr.step_feature_names[dim])
                # ax.set_xlim(-1, 1)
                # plt.xlabel(f"dimension {dim}")
                ax.grid(**gridlines_config)

        filename = f"trainphase_{mngr.train_phase}_input_vs_forecast.pdf"
        output_path = get_debug_output_filepath(
            mngr, os.path.join("forecasting", data_split_tag), filename
        )
        PDFPlotCollection(mngr, [fig]).save_to_file(output_path)

    def do_error_analysis_on_reconstruction_error(
        self, mngr, model_outputs, input_df_meta
    ):
        mngr.log.info([self.tag, ["y", "do_error_analysis_on_reconstruction_error()"]])
        n_examples = 100

        base_output_dir = "autoenc_error_analysis"

        clear_directory(mngr, base_output_dir)

        # highest and lowest reconstruction error
        label_high = "high_reconstruction_error"
        label_low = "low_reconstruction_error"
        for category in [label_high, label_low]:
            reco_error = pd.Series(
                model_outputs[mngr.tag_auto_enc_reconstruction_error]
            )

            if category == label_high:
                rows = reco_error.nlargest(n_examples).index
            elif category == label_low:
                rows = reco_error.nsmallest(n_examples).index

            for index in rows:
                output_dir = os.path.join(
                    base_output_dir,
                    category,
                    input_df_meta.loc[index][mngr.tag_evt_type],
                )
                id = f"seed_{input_df_meta.loc[index]['realisation_seed']}_index_{input_df_meta.loc[index]['realisation_row_index']}"
                filename = f"re_{reco_error[index]:.2f}_{id}.pdf"
                model_output = {k: model_outputs[k][index] for k in model_outputs}
                mngr.predictions.do_error_analysis_plot_of_event(
                    model_output,
                    input_df_meta.loc[index],
                    id,
                    get_debug_output_filepath(mngr, output_dir, filename),
                )
