from __future__ import annotations

import copy
import os
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from scipy import stats

from trans_finder.flares.core.tags import (
    TAG_LOWERQUANT,
    TAG_MEDIAN,
    TAG_SIGNIFICANCE_CLUSTER_SPACE,
    TAG_SIGNIFICANCE_COMBINED,
    TAG_SIGNIFICANCE_RECO_ERROR,
    TAG_UPPERQUANT,
)
from trans_finder.flares.helpers_debugging import get_debug_output_filepath
from trans_finder.flares.inspections.pure_bck_with_high_ts import (
    plot_pure_bck_with_high_ts_relisation_diagnostics,
)
from trans_finder.flares.plot_list.bgmm import (
    bgmm_plot_outlier_score_ecdfplot,
    bgmm_plot_outlier_score_histograms,
)
from trans_finder.flares.plot_list.cluster_space import do_cluster_space_plots
from trans_finder.flares.plot_list.performance_by_time_sliding_window import (
    do_performance_by_time_sliding_window_plots,
)
from trans_finder.flares.plot_list.significance_distribution import (
    plot_significance_distribution,
    plot_significance_distribution_per_realisation,
)
from trans_finder.flares.plot_list.significance_squared_distribution import (
    plot_significance_squared_distribution,
)
from trans_finder.flares.plot_list.systematic_all_flare_templates_plot import (
    plot_systematic_all_flare_templates,
)
from trans_finder.flares.utilities.result_table import make_result_table

if TYPE_CHECKING:
    from trans_finder.flares.log_parser import LogParser
    from trans_finder.flares.manager import Manager


significance_clips = [10, 25]


class DiagnosticsPlots:
    def __init__(self, mngr):
        self.mngr: Manager = mngr
        self.log: LogParser = mngr.log

    def get_signal_step_sum(self, df_diagnostics):
        col_names = []

        for channel_tag in self.mngr.data_prep_channel_tags:
            for ft_name in self.mngr.data_prep_fluctuation_cols[channel_tag]:
                col_name = f"n_signal_steps_{channel_tag}_{ft_name}"
                col_names.append(col_name)
                # signal_step_plot_df = signal_step_plot_df.assign(
                #     **{col_name: df_diagnostics[col_name]}
                # )

        # df_diagnostics = df_diagnostics.assign(
        #     signal_step_sum=df_diagnostics[col_names].sum(axis=1)
        # )

        signal_step_sum = df_diagnostics[col_names].sum(axis=1)

        signal_step_sum = pd.cut(
            signal_step_sum, bins=[0, 1, 2, 3, 4, 5, 10, 15, np.inf], right=False
        )

        # df_diagnostics = df_diagnostics.assign(
        #     signal_step_sum=pd.cut(
        #         df_diagnostics["signal_step_sum"],
        #         bins=[0, 1, 2, 10, np.inf],
        #         right=False,
        #     )
        # )

        return signal_step_sum

    def calculate_significance(self, log_p_val, p_value_clip):
        log_p_val = log_p_val.astype(float)

        p_value_ten_sigma = stats.norm.sf(p_value_clip)
        log_p_val_10_sigma = np.log10(p_value_ten_sigma)
        log_p_val[log_p_val < log_p_val_10_sigma] = log_p_val_10_sigma

        p_val = np.power(10, log_p_val)

        signif = np.asarray(stats.norm.isf(p_val))
        signif[(signif < 0) | (np.isneginf(signif))] = 0

        return signif

    def do_predict_diagnostics_plots(self, df_diagnostics, data_desc):
        calib_tags = self.mngr.get_calib_tags(calib_stage=self.mngr.calib_stages[-1])
        test_statistics = df_diagnostics[calib_tags["ts"]].to_numpy()

        print(
            f"------------------ PLOT ts dist and representations for entire dataset of {self.mngr.run_phase}"
        )

        test_statistics = test_statistics.flatten()

        def p_value_to_ts(p_value):
            calib_tags = self.mngr.get_calib_tags(
                calib_stage=self.mngr.calib_stages[-1]
            )
            log_p_val = df_diagnostics[calib_tags["log_p_val"]].astype(float)

            log_p_val_10_sigma = np.log10(stats.norm.sf(p_value))

            ts_above_p_value_thresholds = test_statistics[
                log_p_val < log_p_val_10_sigma
            ]

            if len(ts_above_p_value_thresholds) == 0:
                return None

            else:
                return np.min(ts_above_p_value_thresholds)

        def get_ts_clip_data(ts_clip_p_val):
            ts_clip_ts = p_value_to_ts(ts_clip_p_val)

            if ts_clip_ts is not None:
                ts_clip = copy.deepcopy(test_statistics)
                ts_clip[ts_clip > ts_clip_ts] = ts_clip_ts
                ts_clip_label = (
                    f"test statistic clipped at p_value of {ts_clip_p_val:.2f}"
                    + f" -> ts of {ts_clip_ts:.2f}"
                )

                return ts_clip, ts_clip_label
            else:
                return (
                    test_statistics,
                    "test statistic",
                )

        ts_clip_4, ts_clip_label_4 = get_ts_clip_data(4)
        ts_clip_10, ts_clip_label_10 = get_ts_clip_data(10)

        df_diagnostics = df_diagnostics.assign(
            ts_clip_10=ts_clip_10, ts_clip_4=ts_clip_4
        )

        ts_vertical_lines = {f"{i}": p_value_to_ts(i) for i in [1, 2, 3]}

        scores = []

        for clip in significance_clips:
            scores.extend(
                v + f"_clip_{clip}"
                for v in self.mngr.p_val_to_significance_col.values()
            )
        scores.extend(self.mngr.ts_to_p_val_col.keys())
        scores.extend([self.mngr.tag_model_forecasting_error, "ts_clip_4"])

        groups = ["signal_strength_categorical", self.mngr.tag_evt_type]

        for score in scores:
            if score not in df_diagnostics.columns:
                continue

            for group in groups:
                plot_name = f"{score}_by_{group}"
                bgmm_plot_outlier_score_histograms(
                    mngr=self.mngr,
                    out_file_name=get_debug_output_filepath(
                        self.mngr,
                        os.path.join(f"by_{group}", score),
                        f"hist_{score}_by_{group}.pdf",
                    ),
                    data=df_diagnostics[score],
                    labels=df_diagnostics[group],
                    x_label=score,
                    vertical_lines={},
                )

                bgmm_plot_outlier_score_ecdfplot(
                    mngr=self.mngr,
                    out_file_name=get_debug_output_filepath(
                        self.mngr,
                        os.path.join(f"by_{group}", score),
                        f"ecdf_{score}_by_{group}_ecdf.pdf",
                    ),
                    data=df_diagnostics[score],
                    labels=df_diagnostics[group],
                    x_label=score,
                    group_label_col_name=group,
                    vertical_lines={},
                )

    def plots_on_inference_realisation(self, df_diagnostics):
        plot_significance_squared_distribution(
            self.mngr, df_diagnostics, data_desc="inference"
        )
        plot_significance_distribution_per_realisation(
            self.mngr, df_diagnostics, realisation_desc="inference"
        )

        do_performance_by_time_sliding_window_plots(
            self.mngr,
            df_diagnostics,
            plot_output_dir="inference",
        )

    def plots_on_full_training_dataset(self, df_diagnostics):
        # plots on realisations without injected signals

        print(df_diagnostics.realisation_id.unique())
        # flare scenarios and other custom scenarios
        do_performance_by_time_sliding_window_plots(
            self.mngr, df_diagnostics, "prep_train"
        )

        try:
            meta_df = pd.read_csv(self.mngr.path_realisation_metadata, index_col=0)
            bck_realisation_ids = meta_df[~meta_df["has_injections"]]["id"].to_list()
            bck_realisation_df = df_diagnostics[
                df_diagnostics.realisation_id.isin(bck_realisation_ids)
            ]

            plot_significance_distribution(
                self.mngr, bck_realisation_df, data_desc="all_background_realisations"
            )

            plot_significance_squared_distribution(
                self.mngr, bck_realisation_df, "all_background_realisations"
            )

        except pd.errors.EmptyDataError as e:
            print(e)

        # get realisation ids without signal injections
        os.path.join(
            self.mngr.data.get_dir_name("predict_in"),
        )
        for df, desc in [
            [df_diagnostics[df_diagnostics.evt_class == "bck_pure"], "all_bck_pure"],
            [
                df_diagnostics[df_diagnostics.evt_class == "signal_dec"],
                "all_signal_dec",
            ],
        ]:
            plot_significance_distribution(self.mngr, df, data_desc=desc)

            plot_significance_squared_distribution(self.mngr, df, desc)

        plot_pure_bck_with_high_ts_relisation_diagnostics(self.mngr, df_diagnostics)

        self.do_predict_diagnostics_plots(df_diagnostics, "all")

    def aggregate_input_uncertainty_samples(self, df):
        aggregation_cols = [
            TAG_SIGNIFICANCE_CLUSTER_SPACE,
            TAG_SIGNIFICANCE_RECO_ERROR,
            TAG_SIGNIFICANCE_COMBINED,
        ]

        id_col = self.mngr.tag_model_input_ids

        grouped = df.groupby(["realisation_id", "input_mask", id_col])
        output = grouped.first().reset_index().drop(columns=aggregation_cols)

        aggr_functions = {
            TAG_MEDIAN: lambda x: x.quantile(q=0.5, interpolation="linear"),
            TAG_UPPERQUANT: lambda x: x.quantile(q=0.84, interpolation="linear"),
            TAG_LOWERQUANT: lambda x: x.quantile(q=0.16, interpolation="linear"),
        }

        for agg_tag, agg_fn in aggr_functions.items():
            output[["_".join([col, agg_tag]) for col in aggregation_cols]] = agg_fn(
                grouped[aggregation_cols]
            ).reset_index(drop=True)

        return output.reset_index(drop=True)

    def plot_diagnostics(self, df_diagnostics=None):
        self.log.info([["g", " - plot-diagnostics() for flares"]])

        if df_diagnostics is None:
            if (
                self.mngr.run_phase not in self.mngr.do_diagnostics.keys()
                or not self.mngr.do_diagnostics[self.mngr.run_phase]
            ):
                return  # skip diagnostic plots

            # obtain data for diagnostics plots
            df_diagnostics = get_diagnostics_data(self.mngr)

        if self.mngr.plot_cluster_space:
            do_cluster_space_plots(self.mngr, df_diagnostics)

        # this plot is done with all samples of input uncertainties, not based on the aggregated median/quantiles
        if self.mngr.dir_input_predict == self.mngr.tag_custom_realisations:
            if self.mngr.flare_template_conf is not None:
                plot_systematic_all_flare_templates(self.mngr, df_diagnostics)
            else:
                self.log.info(
                    "Skipping flare templates as config entry 'flare_template_conf' is None."
                )

        # aggregate relevant output columns into median and quantiles
        df_diagnostics = self.aggregate_input_uncertainty_samples(df_diagnostics)

        if self.mngr.dir_input_predict == self.mngr.tag_custom_realisations:
            # flare scenarios and other custom scenarios
            do_performance_by_time_sliding_window_plots(
                self.mngr, df_diagnostics, "custom_realisations"
            )

            return

        if self.mngr.dir_input_predict == self.mngr.dir_output_data_prep_predict:
            table_df = make_result_table(self.mngr, df_diagnostics)
            table_df.to_csv(
                get_debug_output_filepath(self.mngr, "inference", f"result.csv"),
                index=False,
                float_format="%.2f",
            )

            self.plots_on_inference_realisation(df_diagnostics)
            return

        self.plots_on_full_training_dataset(df_diagnostics)


def get_diagnostics_data(mngr, manual_runphase=None):
    if mngr.run_phase == "calib" or manual_runphase == "calib":
        dir_path = mngr.data.get_dir_name("calib_diagnostics")
    elif mngr.run_phase == "predict" or manual_runphase == "predict":
        dir_path = mngr.data.get_dir_name("predict_diagnostics")
    else:
        raise Exception(" - unknown run_phase for diagnostic ...")

    dir_files = sorted(os.listdir(dir_path))
    dir_files = [
        f for f in dir_files if all(o in f for o in mngr.diagnostics_file_name_patt)
    ]

    df_full = pd.DataFrame()
    for file_name in dir_files:
        file_path = os.path.join(dir_path, file_name)

        # read in the data
        df_in = pd.read_hdf(file_path, mngr.tag_hdf)

        df_full = df_full.append(
            df_in,
            ignore_index=True,
            sort=False,
        )

    eval_output = mngr.calib.get_p_values_by_stage(
        df_calib=df_full,
        calib_stage=mngr.calib_stages[-1],
    )
    df_diagnostics = eval_output["df_calib"]

    signal_integral_col_names = [
        f"{channel_tag}_signal_integral_dec"
        for channel_tag in mngr.data_prep_channel_tags
    ]

    signal_integral = df_diagnostics[signal_integral_col_names].sum(axis=1)

    signal_integral_norm = mngr.n_steps_dec * sum(
        len(features) for features in mngr.data_prep_fluctuation_cols.values()
    )
    signal_integral = signal_integral / signal_integral_norm

    for p_val_col, significance_col in mngr.p_val_to_significance_col.items():
        if p_val_col not in df_diagnostics.columns:
            continue

        for p_val_clip in significance_clips:
            df_diagnostics = df_diagnostics.assign(
                **{
                    f"{significance_col}_clip_{p_val_clip}": mngr.diagnostics.calculate_significance(
                        df_diagnostics[p_val_col].to_numpy(), p_value_clip=p_val_clip
                    ),
                }
            )

    df_diagnostics = df_diagnostics.assign(
        signal_strength=signal_integral,
        signal_strength_categorical=pd.cut(
            signal_integral, bins=[0, 2, 3, 4, np.inf], right=False
        ),
        # cta_signal_strength=df_diagnostics["cta_signal_integral_dec"]
        # / (mngr.n_steps_dec * len(mngr.data_prep_fluctuation_cols["cta"])),
        # fermi_signal_strength=df_diagnostics["fermi_signal_integral_dec"]
        # / (mngr.n_steps_dec * len(mngr.data_prep_fluctuation_cols["fermi"])),
    )

    return df_diagnostics
