from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt
from tensorflow.estimator import ModeKeys

from trans_finder.flares.checks.generic_checks import (
    calculate_dist_overlap,
    describe_overlap,
)
from trans_finder.flares.helpers_debugging import get_debug_output_filepath
from trans_finder.flares.mcmc_fit.dynamic_complexity_increase import (
    do_dynamic_mcmc_fit_with_increasing_complexity,
)
from trans_finder.flares.mcmc_fit.evaluate_fit import evaluate_mixture_model_fit
from trans_finder.flares.model_utils.bayesian_gaussian_mixture_model import (
    BayesianGaussianMixtureModel,
)
from trans_finder.flares.model_utils.utils import get_batch_size
from trans_finder.flares.plot_list.bgmm import (
    plot_per_dimension_input_and_posterior_histograms,
)
from trans_finder.flares.state_manager import StateManager
from trans_finder.flares.utilities.pdf_plot_collection import PDFPlotCollection

if TYPE_CHECKING:
    from trans_finder.flares.manager import Manager


def create_bgmm_model(mngr: Manager, n_input_length):
    n_input_dims = mngr.model_learned_representation_n_dim + (
        1 if mngr.bgmm_add_auto_enc_reconstruction_error_as_extra_dim else 0
    )

    hmc_n_burnin_steps = 700
    hmc_n_step_size_adaptation_steps = 600

    bgmmodel = BayesianGaussianMixtureModel(
        n_input_dims=n_input_dims,
        n_input_length=n_input_length,
        gmm_n_components=mngr.model_gaus_mix_n_comps,
        gmm_wishard_degrees_of_freedom=max(200, n_input_dims),
        hmc_max_parallel_iterations=mngr.model_hmc_max_parallel_iterations,
        hmc_n_parallel_chains=mngr.model_hmc_n_parallel_chains,
        hmc_n_steps=mngr.model_hmc_n_steps,
        hmc_n_burnin_steps=hmc_n_burnin_steps,
        hmc_n_step_size_adaptation_steps=hmc_n_step_size_adaptation_steps,
        hmc_step_size=0.01,
        hmc_n_leapfrog_integrator_steps=30,
    )

    return bgmmodel


def check_model_params(mngr: Manager, input_data):
    large_bgmm = create_bgmm_model(mngr, len(input_data))
    large_bgmm.adapt(input_data)

    # how much overlap is there with a single multivariate gaussian (uni-modal)?
    [
        sample_mean,
        sample_cov,
        sample_scale,
    ] = large_bgmm.compute_sample_stats(input_data)

    input_normal_samples = np.random.multivariate_normal(
        sample_mean.numpy(), sample_cov.numpy(), len(input_data)
    )

    mngr.log.info(
        describe_overlap(
            calculate_dist_overlap(input_data, input_normal_samples),
            desc="manual multivariate normal from mean & cov matrix",
        )
    )
    # plot histograms
    fig = plot_per_dimension_input_and_posterior_histograms(
        input_data, input_normal_samples
    )
    filename = get_debug_output_filepath(
        mngr, "bgmmodel", "normal_vs_input_projections.pdf"
    )
    # plt.savefig("inputoutputfit.png")
    PDFPlotCollection(mngr, [fig]).save_to_file(filename)

    # check the convergence of the mcmc fit on a small sample of the input data

    n_small = 5000
    if len(input_data) <= n_small:
        return

    small_input_data = input_data[
        np.random.choice(input_data.shape[0], n_small, replace=False)
    ]

    small_bgmm = create_bgmm_model(mngr, len(small_input_data))
    small_bgmm.adapt(small_input_data)

    # how good is the model before doing any fit?
    mngr.log.info(" - Evaluating mixture model with prior parameters: ")
    small_bgmm.compare_input_and_posterior_sample_dist_overlap(mngr, small_input_data)

    # mngr.log.info(" - [small data subset] Fitting model...")
    # chain_results = small_bgmm.fit(small_input_data)
    # mngr.log.info(" - [small data subset] Evaluating mixture model fit: ")
    # small_bgmm.compare_input_and_posterior_sample_dist_overlap(
    #     mngr, small_input_data
    # )

    bgmmodel = create_bgmm_model(mngr, len(small_input_data))
    bgmmodel.adapt(small_input_data)
    chain_results = bgmmodel.fit(small_input_data)
    mngr.log.info("Fit done.")

    quality_of_fit = bgmmodel.get_quality_of_fit_ks_2samp(input_data)
    evaluate_mixture_model_fit(
        mngr,
        bgmmodel,
        small_input_data,
        quality_of_fit,
        chain_results,
        "input_subset",
    )
    # exit(0)


class FitMixtureModelStep:
    tag = ["b", f" - flares.{Path(__file__).stem} - "]

    def __init__(self, mngr):
        self.mngr: Manager = mngr
        self.gmm_n_components = self.mngr.model_gaus_mix_n_comps

        self.n_samples_from_posterior = 1000
        self.input_normalisation_range = mngr.model_hmc_input_normalisation_range
        self.plot_axis_range = (
            mngr.model_hmc_input_normalisation_range[0] - 0.5,
            mngr.model_hmc_input_normalisation_range[1] + 0.5,
        )

    def fit_bayesian_gaussian_mixture_model(self, input_train, input_val):
        mngr = self.mngr
        assert input_train.ndim == 2  # n_input_length x n_input_dims

        # fit_data = input_train[:1000]
        #
        # bgmmodel = BayesianGaussianMixtureModel(
        #     n_input_dims=fit_data.shape[1],
        #     n_input_length=fit_data.shape[0],
        #     gmm_n_components=3,
        #     gmm_wishard_degrees_of_freedom=max(200, fit_data.shape[1]),
        #     hmc_max_parallel_iterations=multiprocessing.cpu_count(),
        #     hmc_n_parallel_chains=3,
        #     hmc_n_steps=50,
        #     hmc_n_burnin_steps=200,
        #     hmc_n_step_size_adaptation_steps=50,
        #     hmc_step_size=0.01,
        #     hmc_n_leapfrog_integrator_steps=30,
        # )
        #
        # mngr.log.info("MCMC fit starting")
        # st = time.time()
        # chain_results = bgmmodel.fit(fit_data)
        # et = time.time()
        # StateManager(mngr=self.mngr, state_id="bgmm").save_state(bgmmodel._state)
        # mngr.log.info(f"MCMC fit done. runtime: {et - st:.2f} seconds")
        #
        # StateManager(self.mngr, "bgmm_chain_results").save_state(chain_results)
        #
        # quality_of_fit_fit_data = bgmmodel.get_quality_of_fit_ks_2samp(fit_data)
        # quality_of_fit_input_train = bgmmodel.get_quality_of_fit_ks_2samp(fit_data)
        # quality_of_fit_input_val = bgmmodel.get_quality_of_fit_ks_2samp(input_val)
        #
        # def get_mean_prob(data):
        #     n_chains = chain_results.trace.shape[1]
        #     mean_prob_per_chain = [
        #         np.mean(bgmmodel.get_posterior_dist(i_chain).log_prob(data))
        #         for i_chain in range(n_chains)
        #     ]
        #     best_prob = np.max(mean_prob_per_chain)
        #     return best_prob
        #
        # mngr.log.info(f"MCMC fit - {quality_of_fit_fit_data=}")
        # mngr.log.info(f"MCMC fit - {quality_of_fit_input_train=}")
        # mngr.log.info(f"MCMC fit - {quality_of_fit_input_val=}")
        # mngr.log.info("-----")
        # mngr.log.info(f"MCMC fit - {get_mean_prob(fit_data)=}")
        # mngr.log.info(f"MCMC fit - {get_mean_prob(input_train)=}")
        # mngr.log.info(f"MCMC fit - {get_mean_prob(input_val)=}")
        #
        # StateManager(self.mngr, self.mngr.tag_bgmm_quality_of_fit).save_state(
        #     quality_of_fit_input_val
        # )

        # # evaluate
        # evaluate_mixture_model_fit(
        #     self.mngr,
        #     bgmmodel,
        #     input_train,
        #     quality_of_fit_input_train,
        #     chain_results,
        #     "train",
        # )

        # evaluate_mixture_model_fit(
        #     self.mngr,
        #     bgmmodel,
        #     input_data_val,
        #     quality_of_fit,
        #     None,
        #     "val",
        # )

        do_dynamic_mcmc_fit_with_increasing_complexity(mngr, input_train, input_val)

    def plot_inputs(self, observations):
        input_plot_filename = get_debug_output_filepath(
            self.mngr, "bgmmodel", f"input_plot.pdf"
        )

        figs_input = []

        plot_df = pd.DataFrame(
            observations, columns=[f"dim {i}" for i in range(observations.shape[1])]
        )

        g = sns.PairGrid(plot_df, diag_sharey=False, height=5)
        g.map_upper(sns.scatterplot, marker="x")
        g.map_lower(sns.kdeplot)
        g.map_diag(sns.kdeplot)
        g.add_legend()

        # g = sns.jointplot(
        #     x=observations[:, 0],
        #     y=observations[:, 1],
        #     xlim=self.plot_axis_range,
        #     ylim=self.plot_axis_range,
        #     kind="scatter",
        #     marginal_kws=dict(bins=30, fill=False),
        #     marker="x",
        # )

        plt.suptitle(f"bgmm fit input data")
        figs_input.append(g.fig)
        PDFPlotCollection(self.mngr, figs_input).save_to_file(input_plot_filename)

    def run(self):
        dataset_train = self.mngr.data.read_dataset(mode=ModeKeys.TRAIN)
        dataset_eval = self.mngr.data.read_dataset(mode=ModeKeys.EVAL)

        self.mngr.train.load_model(dataset_train)
        if self.mngr.tensorflow_run_eagerly:
            self.mngr.model[self.mngr.tag_model].compile(run_eagerly=True)

        model = self.mngr.model["keras_model"]

        model_outputs_train = model.predict(
            dataset_train, batch_size=get_batch_size(self.mngr)
        )

        # # todo: cleanup
        # print(
        #     "Temporarily for debugging, saving model outputs of this run_phase to compare them later with calib."
        # )
        # temp_save(model_outputs_train, "bgmm_fit_outputs_train")
        # temp_save(
        #     model.predict(dataset_eval, batch_size=get_batch_size(self.mngr)),
        #     "bgmm_fit_outputs_val",
        # )

        input_data_train = model_outputs_train[self.mngr.tag_cluster_space_mean]
        input_data_val = model.predict(
            dataset_eval, batch_size=get_batch_size(self.mngr)
        )[self.mngr.tag_cluster_space_mean]

        if self.mngr.bgmm_add_auto_enc_reconstruction_error_as_extra_dim:
            # adding the autoencoder reconstruction error as extra dimension
            auto_enc_reconstruction_error = model_outputs_train[
                self.mngr.tag_auto_enc_reconstruction_error
            ]
            input_data_train = np.column_stack(
                [input_data_train, auto_enc_reconstruction_error]
            )

        # check_in_range(self.mngr, input_data_train, [-1, 1], desc="input data")
        # check_in_range(self.mngr, input_data_val, [-1, 1], desc="input data")

        # evaluate stored fit if it already exists
        # if StateManager(self.mngr, "bgmm_chain_results").has_state():
        #     evalute_stored_mixture_model(
        #         self.mngr, input_data_train, data_desc="train"
        #     )
        #     evalute_stored_mixture_model(self.mngr, input_data_val, data_desc="val")
        #

        # fit mixture model from input data
        self.fit_bayesian_gaussian_mixture_model(input_data_train, input_data_val)

        # StateManager(self.mngr, "bgmm_chain_results").save_state(chain_results)
        #
        # quality_of_fit = bgmmodel.get_quality_of_fit_ks_2samp(input_data_train)
        #
        #
        # StateManager(self.mngr, self.mngr.tag_bgmm_quality_of_fit).save_state(
        #     quality_of_fit
        # )
        #
        # # evaluate
        # evaluate_mixture_model_fit(
        #     self.mngr,
        #     bgmmodel,
        #     input_data_train,
        #     quality_of_fit,
        #     chain_results,
        #     "train",
        # )
        #
        # evaluate_mixture_model_fit(
        #     self.mngr,
        #     bgmmodel,
        #     input_data_val,
        #     quality_of_fit,
        #     None,
        #     "val",
        # )


def evalute_stored_mixture_model(mngr, input_data, data_desc):
    mngr.log.info(f"Evaluating stored mixture model on {data_desc}...")

    bgmm = create_bgmm_model(mngr, len(input_data))

    # load state
    loaded_quality_of_fit = StateManager(
        mngr, mngr.tag_bgmm_quality_of_fit
    ).load_state()
    loaded_chain_results = StateManager(mngr, "bgmm_chain_results").load_state()
    bgmm.update_state_from_chain_results(loaded_chain_results)

    # evaluate
    evaluate_mixture_model_fit(
        mngr,
        bgmm,
        input_data,
        loaded_quality_of_fit,
        loaded_chain_results,
        data_desc,
    )
