import copy
import os
from typing import Iterable

import matplotlib.pyplot as plt

# import math
# import scipy
import numpy as np
import pandas as pd
from scipy import optimize
from scipy.signal import lombscargle

from trans_finder.flares.helpers_debugging import get_debug_output_filepath
from trans_finder.flares.utilities.pdf_plot_collection import PDFPlotCollection

# allow running on batch, where a display is not available
if "DISPLAY" not in os.environ:
    plt.switch_backend("agg")


class PValueFit(object):
    def __init__(self, log, fit_dist_type, name=None, is_verb=True):
        self.log = log
        self.fit_params = None
        self.spline = None
        self.max_spline_ts = 0
        # self.min_p_value = 1e-15
        # self.max_sig = 1e4

        self.fit_dist_type = fit_dist_type
        self.is_verb = is_verb

        self.name = "" if name is None else (" - " + str(name))

        # an ugly hack for now, as the other distributions need to be validated!!!!
        assert fit_dist_type == "log_linear"

        if fit_dist_type == "log_linear":
            self.fit_func = self.log_linear
        elif fit_dist_type == "generalized_extreme_value":
            self.fit_func = self.generalized_extreme_value
        elif fit_dist_type == "generalized_pareto":
            self.fit_func = self.generalized_pareto
        elif fit_dist_type == "pareto":
            self.fit_func = self.pareto
        else:
            raise Exception(
                ' - fit_dist_type = "' + str(fit_dist_type) + '" not recognised ...'
            )

        if fit_dist_type in ["generalized_pareto", "generalized_extreme_value"]:
            self.fit_opts = {
                "p0": [1e-2, 1, 0],
                "bounds": ((1e-6, -1e6, -1e6), (1e6, 1e6, 1e6)),
            }
        elif fit_dist_type in ["log_linear"]:
            self.fit_opts = {
                "p0": [1, 1],
                "bounds": ((-1e6, -1e6), (1e6, 1e6)),
            }
        else:
            self.fit_opts = {
                "p0": [1e-2, 1],
                "bounds": (1e-5, 30),
            }

        return

    # see: https://en.wikipedia.org/wiki/Generalized_extreme_value_distribution

    def generalized_extreme_value(self, x, sig, chi, mu):
        try:
            out_0 = np.power((1 + chi * (x - mu) / sig), (-1 / chi))
            out = (1 / sig) * np.power(out_0, (chi + 1)) * np.exp(-out_0)
            # out = (1 / sig) * np.power(
            #     (1 + chi * (x - mu) / sig),
            #     (-1 - 1/chi)
            # )
        except Exception:
            if isinstance(x, Iterable):
                try:
                    out = np.zeros(len(x))
                # can be numpy.ndarray of size one ...
                except Exception:
                    out = 0
        return out

    # see: https://en.wikipedia.org/wiki/Generalized_Pareto_distribution

    def generalized_pareto(self, x, sig, chi, mu):
        try:
            out = (1 / sig) * np.power((1 + chi * (x - mu) / sig), (-1 - 1 / chi))
        except Exception:
            if isinstance(x, Iterable):
                try:
                    out = np.zeros(len(x))
                # can be numpy.ndarray of size one ...
                except Exception:
                    out = 0
            else:
                out = 0
        return out

    # see: https://en.wikipedia.org/wiki/Pareto_distribution

    def pareto(self, x, a, b):
        try:
            out = np.power(a, b) * b / np.power(x, b + 1)
        except Exception:
            out = 0
        return out
        # return a + b * x

    def log_linear(self, x, a, b):
        try:
            out = a + b * x
        except Exception:
            out = 1
        return out

    def fit(self, fit_x, fit_y, fit_w=None):
        if len(fit_x) < 2:
            self.log.info(
                [
                    ["wr", self.name],
                    ["wr", " - nothing to fit for", self.fit_dist_type, "..."],
                ]
            )
            self.fit_params = copy.deepcopy(self.fit_opts["p0"])
        else:
            fit_opts = copy.deepcopy(self.fit_opts)
            if fit_w is not None:
                fit_opts["sigma"] = fit_w

            if self.fit_dist_type == "log_linear":
                fit_y = np.log10(fit_y)

            # print('-=-'*30)
            # print(fit_x)
            # print('-+-'*30)
            # print(fit_y)
            # print('-+-'*30)
            # print(fit_w)
            # print('-+-'*30)

            self.fit_params, _ = optimize.curve_fit(
                self.fit_func, fit_x, fit_y, **fit_opts
            )

            if self.is_verb:
                self.log.info(
                    [
                        ["b", self.name],
                        ["b", " - fitted", self.fit_dist_type, "dist "],
                        [
                            "g",
                            ", ".join(["{:>6.3f}".format(p) for p in self.fit_params]),
                        ],
                    ]
                )

        return

    def get_fit_pars(self):
        return self.fit_params

    def set_fit_pars(self, fit_params):
        self.fit_params = copy.deepcopy(fit_params)
        return

    def get_log_p_value(self, ts_in):
        ts = np.asarray(ts_in)
        mask = ts < self.max_spline_ts
        anti_mask = ~mask

        out_0 = self.spline(ts) * mask

        if self.fit_params is None:
            out = out_0
        else:
            out_1 = self.fit_func(ts, *self.fit_params) * anti_mask
            if self.fit_dist_type == "log_linear":
                out_1 = np.where(out_1 >= 0, 0, out_1)
            out_1 *= anti_mask
            out = out_0 + out_1

        return out

    def set_spline(self, spline, max_ts):
        self.spline = spline
        self.max_spline_ts = max_ts
        return

    # def get_sig(self, ts_in):
    #     ts = np.asarray(ts_in)
    #     p_value = self.get_p_value(ts)
    #     sig = stats.chi2.ppf(1 - p_value, df=1)
    #     sig = np.nan_to_num(np.sqrt(sig))
    #     sig[p_value < self.min_p_value] = np.sqrt(self.max_sig)
    #     return sig
    # best with p-value directly through isf:
    # p_values = data[p_val_combo_col].astype(float)
    # data.loc[:, sig_combo_col] = stats.chi2.isf(p_values, df=ndf)

    # def get_ts_from_sig(self, sig_in):
    #     sig = np.asarray(sig_in)
    #     sig[sig > np.sqrt(self.max_sig)] = self.max_sig

    #     try:
    #         return np.power(sig, 2)
    #     except Exception:
    #         return sig


#
# class PoissonFit(object):

#
#     def __init__(self, log):
#         self.log = log

#
#     # poisson pdf, parameter lamb is the fit parameter
#
#     def poiss(self, k, lamb):
#         # by using np.seterr(all='raise'), an overflow will raise an exception...
#         try:
#             fact = [max(1, math.gamma(x + 1)) for x in k]

#             # fact = [ max(f,1) for f in factorial(k) ]

#             # result = (lamb**k / fact) * np.exp(-lamb)
#             result = (np.power(lamb, k) / fact) * np.exp(-lamb)
#         except Exception:
#             result = 0

#         return result

#
#     # the negative log-Likelohood-Function
#
#     def neg_log_likelihood(self, params, data):
#         return -np.sum(np.log(self.poiss(data, params[0])))

#
#     # minimize the negative log-Likelihood
#
#     def poiss_minimize(self, data):
#         try:
#             if not isinstance(data, Iterable):
#                 data = [data]
#             x0 = np.ones(1) * max(1., np.average(data))

#             result = scipy.optimize.minimize(
#                 # function to minimize
#                 self.neg_log_likelihood,
#                 # start value
#                 x0=x0,
#                 # additional arguments for function
#                 args=(data, ),
#                 # minimization method, see docs: 'L-BFGS-B','SLSQP','TNC'
#                 method='SLSQP',
#                 bounds=scipy.optimize.Bounds(1e-6, 1e6),
#                 # options={'maxiter':2},
#             )
#             result = result.x[0]

#         except Exception:
#             result = np.average(data)
#             # result =  np.nan # how it would have resulted without the exception...


#         return result
def get_col_name_filter(filter_string: str):
    def is_flux_col(col_name: str):
        parts = col_name.split("_")
        is_flux = parts[0] == filter_string
        if not is_flux:
            return False
        if len(parts) == 1:
            return True
        elif len(parts) == 2:
            return parts[1].isdigit()
        else:
            return False

    return is_flux_col


def lomb_scargle_psd(times, values):
    t_range = max(times) - min(times)
    f_min = 1 / t_range
    f_max = 2.0
    steps = 10_000
    frequencies = np.linspace(f_min, f_max, int(steps))
    periodogram = lombscargle(times, values, frequencies, normalize=True)

    return frequencies, periodogram


def plot_psd(mngr, frequencies, periodograms, labels, suffix=""):
    fig = plt.figure(figsize=(10, 7))
    if not isinstance(periodograms, list):
        plt.plot(frequencies, periodograms, label=labels)
    else:
        for periodogram, label in zip(periodograms, labels):
            plt.plot(frequencies, periodogram, label=label)

    plt.tick_params(direction="in", which="both", top=True, right=True)
    plt.legend(loc="best", fontsize="small")
    plt.xlabel("f")
    # plt.yscale("log")
    out_file_name = get_debug_output_filepath(
        mngr, "data_prep", f"lomb_scargle_periodogram{suffix}.pdf"
    )
    PDFPlotCollection(mngr, [fig]).save_to_file(out_file_name)


def data_prep_simple_log_transform(inputs):
    transform = np.log10(inputs)
    return transform


def data_prep_inverse_simple_log_transform(inputs):
    inverse = np.power(10, inputs)
    return inverse


def data_prep_flux_log_transform(inputs):
    transform = -1 / np.log10(inputs)
    return transform


def data_prep_flux_inverse_transform(inputs):
    inverse = np.power(10, (-1 / inputs))
    return inverse


def shuffle_input_dict_of_dataframes_or_ndarrays(inputs, seed=42):
    """
    shuffle all dataframes and numpy ndarrays in the inputs dictionary in the same way
    All elements of the input dict have to be either a dataframe or nparray of the same length
    (that is for numpy arrays: along the first dimension)

    :param seed: the fixed random seed for repeatable outcomes
    :param inputs: a dict of dataframes
    :return: a dict of the same dataframes, shuffled in the same way
    """

    # create a shuffled index from the first dataframe in the input dict
    first_df_in_inputs = None
    for key in inputs:
        if isinstance(inputs[key], pd.DataFrame):
            first_df_in_inputs = inputs[key]
            break

    if first_df_in_inputs is None:
        raise ValueError("input must contain at least one dataframe")

    shuffle_idx = np.random.RandomState(seed=seed).permutation(first_df_in_inputs.index)

    # apply the shuffled index to all dataframes and ndarrays in the dict
    outputs = {}
    for key in inputs:
        if isinstance(inputs[key], pd.DataFrame):
            outputs[key] = inputs[key].reindex(shuffle_idx)
            outputs[key].reset_index(inplace=True, drop=True)

        elif isinstance(inputs[key], np.ndarray):
            # apply shuffled index along the first dimension
            outputs[key] = inputs[key][shuffle_idx, ...]

        else:
            raise ValueError(
                "elements in input dict must be either a pd.DataFrame or a numpy.ndarray"
            )

    return outputs
