import os
import shutil
from pathlib import Path

import numpy as np
import tensorflow as tf


def get_debug_output_filepath(mngr, subfolder_name: str, filename: str):
    """
    :param mngr:
    :param subfolder_name: the name of the folder under debug/
    :param filename:
    :return: a file path in output/.../debug/subfolder_name/filename
    """

    """returns """
    dir = os.path.join(mngr.data.get_dir_name("debug"), subfolder_name)
    Path(dir).mkdir(parents=True, exist_ok=True)
    out_file_name = os.path.join(dir, filename)
    return out_file_name


def clear_directory(mngr, subfolder_name: str):
    """
    :param mngr:
    :param subfolder_name: the name of the folder under debug/
    """

    dir = os.path.join(mngr.data.get_dir_name("debug"), subfolder_name)
    Path(dir).mkdir(parents=True, exist_ok=True)
    shutil.rmtree(dir)


def get_tfdataset_input_as_numpy_array(dataset, key: str):
    """
    :param dataset: tensorflow dataset
    :param key: a key of the input dict to the model, such as enc or dec
    :return: the input converted to a numpy array
    """

    data = tf.concat([e[0][key] for e in list(dataset)], axis=0).numpy()
    return data


def denormalize_model_output(mngr, model_output, data_part=None):
    """
    :param data_part:
    tag_dec (default) for the decoder part
    tag_enc for the encoder part
    :param mngr:
    :param model_output: the input or output of the autoencoder
    :return: the denormalized input or output of the autoencoder
    """

    if data_part is None:
        data_part = mngr.tag_dec

    if mngr.do_train_scale:
        norms = mngr.data.train_feat_norms[mngr.tag_steps]

        # if necessary, repeat scale and shift features to match the shape of the autoencoder input/output
        # (i.e. twice the number of features of d1 as the autoencoder input/output)
        n_features_in_norms = norms["shift"].shape[-1]
        n_features_in_X = model_output.shape[-1]
        n_times_all_data_columns = int(n_features_in_X / n_features_in_norms)

        shift = np.tile(norms["shift"], n_times_all_data_columns)
        scale = np.tile(norms["scale"], n_times_all_data_columns)

        # if necessary, use only the relevant time steps of the encoder/decoder data parts
        if data_part == mngr.tag_enc:
            shift = shift[: mngr.n_steps_enc]
            scale = scale[: mngr.n_steps_enc]
        elif data_part == mngr.tag_dec:
            shift = shift[-mngr.n_steps_dec :]
            scale = scale[-mngr.n_steps_dec :]
        else:
            raise ValueError(f"Unknown {data_part=}")

        return (model_output / scale) - shift
    else:
        return model_output


def write_np_ndarray_to_csv(
    mngr,
    name,
    ndarray,
    print_format="%.3f",
    n_rows=None,
    debug_subdir_name="model_csv",
):
    file_name = get_debug_output_filepath(mngr, debug_subdir_name, f"{name}.csv")

    data = ndarray.reshape(ndarray.shape[0], -1)
    if n_rows:
        data = data[:n_rows, :]

    np.savetxt(
        file_name,
        data,
        delimiter=",",
        fmt=print_format,
    )


def mock_object():
    class MockObject:
        def __getattr__(self, name):
            def catchall(*args):
                pass

            return catchall

    return MockObject()
