from __future__ import annotations

import os
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt
from tensorflow.estimator import ModeKeys

from trans_finder.flares.helpers_debugging import (
    clear_directory,
    denormalize_model_output,
    get_debug_output_filepath,
)
from trans_finder.flares.model_utils.utils import get_batch_size
from trans_finder.flares.plot_list.util.text_figure import create_text_figure
from trans_finder.flares.utilities.pdf_plot_collection import PDFPlotCollection

if TYPE_CHECKING:
    from trans_finder.flares.manager import Manager


def inspect_evt_class_proportion(mngr: Manager):
    train_df = mngr.input_data[mngr.tag_input_df][ModeKeys.TRAIN]
    eval_df = mngr.input_data[mngr.tag_input_df][ModeKeys.EVAL]

    train_fig = (
        train_df[mngr.tag_evt_type].value_counts().sort_values().plot(kind="bar").figure
    )
    eval_fig = (
        eval_df[mngr.tag_evt_type].value_counts().sort_values().plot(kind="bar").figure
    )

    PDFPlotCollection(mngr, [train_fig, eval_fig]).save_to_file(
        get_debug_output_filepath(mngr, "inspect_inputs", "evt_class_proportions.pdf")
    )


class InspectInputs:
    tag = ["b", f" - flares.{Path(__file__).stem} - "]

    def __init__(self, mngr):
        self.mngr: Manager = mngr

    def run(self):
        mngr = self.mngr
        mngr.log.info([self.tag, ["p", "run()"]])

        input_data_inspections = [
            inspect_evt_class_proportion,
        ]

        dataset_train = mngr.data.read_dataset(mode=ModeKeys.TRAIN)
        dataset_eval = self.mngr.data.read_dataset(mode=ModeKeys.EVAL)

        for inspection in input_data_inspections:
            inspection(mngr)

        # input_df_meta = input_dict["input_df_meta"]
        mngr.train.load_model(dataset_train)
        model = mngr.model["keras_model"]
        model_outputs = model.predict(dataset_train, batch_size=get_batch_size(self))

        # assert len(model_outputs[mngr.tag_input_d1]) == len(input_df_meta)

        assert len(mngr.train_phases) == 1
        train_phase = mngr.train_phases[0]

        # self.plot_inputs_by_evt_class(mngr, model_outputs, input_df_meta)

    def plot_inputs_by_evt_class(self, mngr, model_outputs, input_df_meta):
        mngr.log.info([self.tag, ["y", "do_error_analysis_on_reconstruction_error()"]])
        n_examples = 100

        base_output_dir = f"inspect_inputs_trainphase_{mngr.train_phase}"

        clear_directory(mngr, base_output_dir)

        # for signal_step_sum, group in df_out.groupby("signal_step_sum"):
        for evt_class, group in input_df_meta.groupby(mngr.tag_evt_type):
            rows = group.head(10).index

            for index in rows:
                output_dir = os.path.join(
                    base_output_dir,
                    evt_class,
                )
                id = f"seed_{input_df_meta.loc[index]['realisation_seed']}_index_{input_df_meta.loc[index]['realisation_row_index']}"
                filename = f"{id}.pdf"
                model_output = {k: model_outputs[k][index] for k in model_outputs}
                self.do_input_inspection_plot(
                    mngr,
                    model_output,
                    input_df_meta.loc[index],
                    id,
                    get_debug_output_filepath(mngr, output_dir, filename),
                )

    def do_input_inspection_plot(self, mngr, model_output, metadata, id, output_path):
        col_name_input_part = "input part"

        column_names = ["_".join(t) for t in mngr.data_step_col_names]

        temporal_offset_available = any(
            ftname == mngr.tag_temporal_offset
            for (channel, ftname) in mngr.data_step_col_names
        )

        denormed_event_df = pd.DataFrame(
            np.vstack(
                [
                    denormalize_model_output(
                        mngr,
                        model_output[mngr.tag_input_d0],
                        data_part=mngr.tag_enc,
                    ),
                    denormalize_model_output(
                        mngr,
                        model_output[mngr.tag_input_d1],
                        data_part=mngr.tag_dec,
                    ),
                ]
            ).T,
            column_names,
        ).T
        denormed_event_df[col_name_input_part] = ["d0"] * mngr.n_steps_enc + [
            "d1"
        ] * mngr.n_steps_dec

        event_df = pd.DataFrame(
            np.vstack(
                [model_output[mngr.tag_input_d0], model_output[mngr.tag_input_d1]]
            ).T,
            column_names,
        ).T
        event_df[col_name_input_part] = ["d0"] * mngr.n_steps_enc + [
            "d1"
        ] * mngr.n_steps_dec

        feature_names_by_channel_tag = {}
        for channel_tag, ft_name in mngr.data_step_col_names:
            if channel_tag not in feature_names_by_channel_tag:
                feature_names_by_channel_tag[channel_tag] = []
            feature_names_by_channel_tag[channel_tag].append(ft_name)

        def get_plot_df(data_df):
            plot_df_list = []
            for channel_tag, ft_names in feature_names_by_channel_tag.items():
                for ft_name in ft_names:
                    if ft_name == mngr.tag_temporal_offset:
                        continue

                    full_ft_name = f"{channel_tag}_{ft_name}"
                    full_offset_name = f"{channel_tag}_{mngr.tag_temporal_offset}"

                    data = {
                        "input step": data_df.index,
                        "channel name": channel_tag,
                        "feature name": ft_name,
                        "channel and feature name": full_ft_name,
                        "feature value": data_df[full_ft_name],
                    }
                    if full_offset_name in data_df:
                        data["temporal offset"] = data_df[full_offset_name]

                    plot_df = data_df[[col_name_input_part]].assign(**data)

                    plot_df_list.append(plot_df)
            plot_df = pd.concat(plot_df_list, ignore_index=True)
            return plot_df

        only_data = lambda df: df.drop(
            df.columns[mngr.temporal_offset_column_indices], axis=1
        )

        only_temporal_offsets = lambda df: df[
            df.columns[mngr.temporal_offset_column_indices]
        ]

        event_df = mngr.predictions.denormalise_temporal_offsets(event_df)

        x_time_min = (
            only_temporal_offsets(event_df.iloc[-mngr.n_steps_dec :]).min().min(),
            -5,
        )

        x_time_max = 0 + 5

        def do_single_plot(ax, x, y_min, y_max, df, title, legend=False):
            data = get_plot_df(df)

            g = sns.lineplot(
                ax=ax,
                data=data,
                x=x,
                y="feature value",
                style="channel and feature name",
                hue="channel name",
                markers=False,
                legend="brief" if legend else False,
                # dashes="channel and feature name",
                # style="part",
            )

            if x == "input step":
                # add encoder/decoder separator
                ax.axvline(x=mngr.n_steps_enc - 0.5, color="black")
                ax.set_xlim(left=-1)
            else:
                ax.set_xlim(right=1, left=data[x].min() - 1)

            if legend:
                g.legend(loc="center left", bbox_to_anchor=(1, 0.5))
            ax.set_title(title)
            # ax.set_ylim([y_min, y_max])

        figs = []

        fig, axes = plt.subplot_mosaic(
            [
                [
                    "full_time_series_by_step",
                    "full_time_series_by_step",
                    "full_time_series_by_time",
                    "full_time_series_by_time",
                ],
                [
                    "normalised_full_time_series_by_step",
                    "normalised_full_time_series_by_step",
                    "normalised_full_time_series_by_time",
                    "normalised_full_time_series_by_time",
                ],
            ],
            constrained_layout=True,
            # sharey=True,
            figsize=(10 * 2, 6 * 2),
        )

        figs.append(fig)
        fig.suptitle(f"{metadata[mngr.tag_evt_type]} {id}")

        y_min = None
        y_max = None

        assert (
            get_plot_df(denormed_event_df)["temporal offset"].min()
            == get_plot_df(event_df)["temporal offset"].min()
        )

        do_single_plot(
            axes["full_time_series_by_step"],
            "input step",
            y_min,
            y_max,
            denormed_event_df,
            f"original input",
            legend=True,
        )

        if temporal_offset_available:
            do_single_plot(
                axes["full_time_series_by_time"],
                "temporal offset",
                y_min,
                y_max,
                denormed_event_df,
                f"original input",
                # legend=True,
            )

        do_single_plot(
            axes["normalised_full_time_series_by_step"],
            "input step",
            y_min,
            y_max,
            event_df,
            f"normalized input",
            legend=True,
        )

        if temporal_offset_available:
            do_single_plot(
                axes["normalised_full_time_series_by_time"],
                "temporal offset",
                y_min,
                y_max,
                event_df,
                f"normalized input",
                # legend=True,
            )

        # add metadata values as additional plot
        metadata_dict = metadata.to_dict()
        fig = create_text_figure(dict(sorted(metadata_dict.items())))
        figs.append(fig)

        PDFPlotCollection(mngr, figs).save_to_file(output_path)
