import os

import numpy as np
from matplotlib import pyplot as plt

from trans_finder.flares.data_prep_helpers.recreate_aggregated_data import (
    recreate_aggregated_data_from_output_df,
)
from trans_finder.flares.helpers_debugging import get_debug_output_filepath
from trans_finder.flares.plot_list.util.plot_config import gridlines_config, tick_params
from trans_finder.flares.utilities.pdf_plot_collection import PDFPlotCollection


def profile_reconstructed(mngr):
    # first train

    data_train = mngr.data_prep.data_aggregation(
        mngr.data_prep.flares_load_input_data(mngr.dir_input_data_prep_train)
    )
    mngr.data_prep.set_aggregated_data_conf(data_train)

    train_realisation = mngr.data_prep.create_realisation(
        data_train,
        shuffling_strategy="step_based",
        do_randomize_temporal_offsets=False,
        do_shift_input_data_in_time=False,
        add_fluctuations=False,
        add_flares=False,
        add_background_states=False,
        realisation_id="train",
        realisation_seed=0,
        # channel_padding=inference_channel_padding,
    )

    out_dir_train_reconstructed = os.path.join("profiles", "reconstructed_train")
    network_df = mngr.data_prep.write_realisation(
        train_realisation,
        plot_output_dir=out_dir_train_reconstructed,
        fixed_evt_class="custom",
    )

    reconstructed_df = recreate_aggregated_data_from_output_df(mngr, network_df)

    do_profile(mngr, reconstructed_df, out_dir_train_reconstructed)

    # then predict

    data_predict = mngr.data_prep.data_aggregation(
        mngr.data_prep.flares_load_input_data(mngr.dir_input_data_prep_predict)
    )
    mngr.data_prep.set_aggregated_data_conf(data_predict)

    train_realisation = mngr.data_prep.create_realisation(
        data_predict,
        shuffling_strategy=None,
        do_randomize_temporal_offsets=False,
        do_shift_input_data_in_time=False,
        add_fluctuations=False,
        add_flares=False,
        add_background_states=False,
        realisation_id="predict",
        realisation_seed=0,
        # channel_padding=inference_channel_padding,
    )
    out_dir_predict_reconstructed = os.path.join("profiles", "reconstructed_predict")
    network_df = mngr.data_prep.write_realisation(
        train_realisation,
        plot_output_dir=out_dir_predict_reconstructed,
        fixed_evt_class="custom",
    )

    reconstructed_df = recreate_aggregated_data_from_output_df(mngr, network_df)

    do_profile(mngr, reconstructed_df, out_dir_predict_reconstructed)


def do_profile(mngr, profile_data, subdir, autocorr=True):
    if autocorr:
        for channel_tag in mngr.data_prep_channel_tags:
            title = f"{channel_tag}"

            df = profile_data[channel_tag].set_index(mngr.data_prep_day_column_name)

            df.to_hdf(
                get_debug_output_filepath(mngr, subdir, f"{title}.hdf5"), key="data"
            )

            from pandas_profiling import ProfileReport

            ProfileReport(
                df,
                explorative=True,
                # tsmode=True,
                # sortby=mngr.data_prep_day_column_name,
            ).to_file(get_debug_output_filepath(mngr, subdir, f"{title}.html"))

            # Creating Autocorrelation plot
            for col in df.columns:

                def acorr(x, ax=None):
                    if ax is None:
                        ax = plt.gca()

                    x = x - x.mean()

                    autocorr = np.correlate(x, x, mode="full")
                    autocorr = autocorr[x.size :]
                    autocorr /= autocorr.max()

                    return ax.stem(autocorr)

                fig, ax = plt.subplots(figsize=(14, 6))
                ax.grid(**gridlines_config)
                ax.tick_params(**tick_params)
                acorr(df[col].dropna())

                # x = pd.plotting.autocorrelation_plot(df_train[col].dropna())

                ax.set_title(title + " " + col)

                PDFPlotCollection(mngr, [fig]).save_to_file(
                    get_debug_output_filepath(
                        mngr, subdir, f"autocorrelation_{title + '_' + col}.pdf"
                    )
                )

            x = df.plot()
            x.set_title(title)
            x.plot()
            PDFPlotCollection(mngr, [x.figure]).save_to_file(
                get_debug_output_filepath(mngr, subdir, f"data_{title}.pdf")
            )


def profile_input_data(mngr):
    """print distribution of features for input_train, input_predict and background realisations"""

    profile_reconstructed(mngr)

    data_predict = mngr.data_prep.data_aggregation(
        mngr.data_prep.flares_load_input_data(mngr.dir_input_data_prep_predict)
    )
    data_train = mngr.data_prep.data_aggregation(
        mngr.data_prep.flares_load_input_data(mngr.dir_input_data_prep_train)
    )

    do_profile(mngr, data_predict, os.path.join("profiles", "input_predict"))
    do_profile(mngr, data_train, os.path.join("profiles", "input_train"))
