import tensorflow as tf
from tensorflow import keras

from trans_finder.flares.utils import log_table


class LearningRateSchedulingCallback(keras.callbacks.LearningRateScheduler):
    def __init__(
        self,
        log,
        epoch_thresh=None,
        learn_rate_init=None,
        learn_rate_decrease_factor=-0.05,
        learn_rate_new_data=None,
    ):
        super().__init__(self.scheduler)
        self.log = log
        self.epoch_thresh = epoch_thresh
        self.learn_rate_init = learn_rate_init
        self.learn_rate_decrease_factor = learn_rate_decrease_factor
        self.learn_rate_new_data = learn_rate_new_data
        return

    def scheduler(self, epoch, learn_rate):
        # This function keeps the initial learning rate for the first few epochs
        # and decreases it exponentially after that.

        if self.learn_rate_init is not None:
            learn_rate = self.learn_rate_init
            self.learn_rate_init = None

        learn_rate = tf.constant(learn_rate)

        if self.learn_rate_new_data is not None:
            period = self.learn_rate_new_data["period"]
            if self.epoch_thresh is not None and epoch % period >= self.epoch_thresh:
                learn_rate *= 1 + self.learn_rate_decrease_factor

        elif self.epoch_thresh is not None and epoch >= self.epoch_thresh:
            learn_rate *= 1 + self.learn_rate_decrease_factor

        self.log.info(
            [
                ["o", " - learning rate: "],
                ["y", "{:>1.2e}".format(learn_rate.numpy())],
            ]
        )

        return learn_rate


class LoggingCallback(keras.callbacks.Callback):
    """message printouts in various stages of training"""

    def __init__(self, log, epoch_end, n_epoch_msg, train_phase=None):
        self.log = log
        self.epoch_end = epoch_end
        self.n_epoch_msg = n_epoch_msg
        self.train_phase_msg = ""
        if train_phase is not None:
            self.train_phase_msg = f" (up to train_phase({train_phase}))"

        return

    def on_epoch_begin(self, epoch, logs=None):
        self.log.info(
            [
                ["y", f" - epoch{self.train_phase_msg}: "],
                ["r", epoch + 1],
                ["y", " / "],
                ["g", self.epoch_end],
            ]
        )
        return

    def on_epoch_end(self, epoch, logs=None):
        if epoch % self.n_epoch_msg == 0:
            log_table(
                self.log.info,
                logs,
                title=" - current results:",
                add_space=True,
            )
        return


class ModelSavingCallback(keras.callbacks.Callback):
    def __init__(self, saving_function, n_epoch_saving, epoch_end):
        self.save_model = saving_function
        self.n_epoch_saving = n_epoch_saving
        self.epoch_end = epoch_end
        return

    def on_epoch_end(self, epoch, logs=None):
        if epoch % self.n_epoch_saving == 0 or epoch == self.epoch_end - 1:
            self.save_model(self.model)
        return
