from __future__ import annotations

import argparse
import os
import warnings
from multiprocessing import Process

import numpy as np
import tensorflow as tf
import yaml
from pydantic import ValidationError

if os.getenv("KERAS_MODEL_DEBUG_MODE", default=False):
    tf.keras.utils.set_random_seed(1)
    tf.config.experimental.enable_op_determinism()


if os.getenv("TENSORFLOW_FORCE_ONLY_CPU", default=False):
    try:
        # Disable all GPUS
        tf.config.set_visible_devices([], "GPU")
        visible_devices = tf.config.get_visible_devices()
        for device in visible_devices:
            assert device.device_type != "GPU"
    except Exception:
        # Invalid device or cannot modify virtual devices once initialized.
        print("TENSORFLOW_FORCE_ONLY_CPU failed!")

print(f"Tensorflow {tf.config.list_physical_devices('GPU')=}")
print(f"Tensorflow {tf.config.get_visible_devices()=}")

"""
can supress some tf warnings:
    import tensorflow.python.util.deprecation as deprecation
    deprecation._PRINT_DEPRECATION_WARNINGS = False
"""
os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"

warnings.simplefilter(action="ignore", category=FutureWarning)
warnings.filterwarnings("ignore", category=DeprecationWarning)

np.set_printoptions(precision=3, suppress=True)
np.seterr(all="raise")

import trans_finder.flares.manager as manager_mod
from trans_finder.flares.core.config_file import ConfigFile
from trans_finder.flares.core.input_data_spec import InputDataSpec
from trans_finder.flares.log_parser import LogParser


def parse_args():
    def is_valid_file(parser, arg):
        if not os.path.exists(arg):
            parser.error(f"The file {arg} does not exist!")
        elif not arg.endswith(".yaml") and not arg.endswith(".yml"):
            parser.error(f"The file {arg} is not a YAML file!")
        else:
            return arg

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--config_path",
        help="Path to the YAML config file",
        type=lambda x: is_valid_file(parser, x),
        required=True,
    )
    parser.add_argument("--log_file", type=str, default=None)
    parser.add_argument("--log_level", type=str, default="debug")
    parser.add_argument("--use_processes", action="store_true")

    arg_dict = parser.parse_args()

    return vars(arg_dict)


def create_logger(arg_dict):
    log_levels = ["debug", "info", "warning", "warn", "error", "critical"]
    if arg_dict["log_level"] not in log_levels:
        raise argparse.ArgumentTypeError(
            "log_level must be one of: " + ", ".join(log_levels)
        )

    log = LogParser(
        title=__name__,
        log_level=arg_dict["log_level"],
        log_file_name=arg_dict["log_file"],
    )
    return log


def read_input_spec_yaml(conf_dict):
    """parses and validates the dataset specification located under input/dataset_name/"""

    # get path and read yaml
    dataset_name = conf_dict.get("input_dataset_version_name", None)
    if dataset_name is None:
        raise ValueError(
            "input_dataset_version_name needs to de defined as argument or in user conf."
        )
    input_data_spec_path = os.path.join("input", "flares", dataset_name, "conf.yml")
    with open(input_data_spec_path, "r") as dataset_conf_file:
        input_data_spec = yaml.load(dataset_conf_file, Loader=yaml.FullLoader)

    # modify to keep backwards compatibility
    InputDataSpec.backwards_compatibility(input_data_spec)

    # validate contents
    try:
        InputDataSpec.model_validate(input_data_spec)
    except ValidationError as e:
        print(f"Error when validating input_data_spec at {input_data_spec_path}.")
        raise e

    return input_data_spec


def load_config_file(arg_dict):
    """parses and validates config file specified in script argument"""

    with open(arg_dict["config_path"], "r") as dataset_conf_file:
        conf_dict = yaml.load(dataset_conf_file, Loader=yaml.FullLoader)
        ConfigFile.backwards_compatibility(conf_dict)
    try:
        ConfigFile.model_validate(conf_dict)
    except ValidationError as e:
        print(f"Error when validating config file at {arg_dict['config_path']}.")
        raise e

    return conf_dict


def combine_and_validate_params(conf_dict, input_data_spec):
    # Translate pipeline flow config into a complete set list of parameter and validate them
    parameters_per_stage = []
    for run_phase, extra_args in conf_dict["pipeline_flow"]:
        # Combine params from config file, input data spec and extra args
        combined_args = dict(
            run_phase=run_phase,
            **extra_args,
            **conf_dict,
            **input_data_spec,
        )

        # Validate parameters of pipeline flow step
        manager_mod.Manager.model_validate(combined_args)

        # Store for execution in sequence
        parameters_per_stage.append(combined_args)

    return parameters_per_stage


def execute_step(i_step, parameters, arg_dict, conf_dict, parameters_per_stage):
    log = create_logger(arg_dict)

    # log message about upcoming step
    run_phase, extra_args = conf_dict["pipeline_flow"][i_step]
    step_desc = ", ".join(k + "=" + str(v) for k, v in extra_args.items())
    msg = f"Executing pipeline step {i_step + 1}/{len(parameters_per_stage)}: {run_phase} ({step_desc})"
    log.info([["r", "-" * 60]])
    log.info([["r", msg]])
    log.info([["r", "-" * 60]])

    # initialize the manager
    mngr = manager_mod.Manager(**parameters)
    mngr.setup(log=log)

    # run pipeline flow step
    mngr.run()

    # cleanup after pipeline flow step
    mngr.on_termination()


def main():
    """execute the pipeline as a sequence of run_phases as specified in config"""

    arg_dict = parse_args()

    # load config file
    conf_dict = load_config_file(arg_dict)

    # Load dataset conf
    input_data_spec = read_input_spec_yaml(conf_dict)

    # Translate pipeline flow config into a complete set list of parameter and validate them
    parameters_per_stage = combine_and_validate_params(conf_dict, input_data_spec)

    # Execute pipeline flow steps in sequence
    # Use separate processes to ensure there is no shared state and carry-over memory usage
    for i_step, parameters in enumerate(parameters_per_stage):
        if arg_dict["use_processes"]:
            p = Process(
                target=execute_step,
                args=(i_step, parameters, arg_dict, conf_dict, parameters_per_stage),
            )
            p.start()
            p.join()
        else:
            execute_step(i_step, parameters, arg_dict, conf_dict, parameters_per_stage)


if __name__ == "__main__":
    main()
