import itertools
import os
import random
import traceback
from collections import OrderedDict
from datetime import datetime
from pathlib import Path
from typing import Iterable

import numpy as np
import scipy.stats as stats

# for using comet_ml, the module has to be loaded before tensorflow.
# noinspection PyUnresolvedReferences
from trans_finder.flares.optional_comet import initialize_experiment as _

import tensorflow as tf
from tensorflow.estimator import ModeKeys

from trans_finder.flares.calib import Calib
from trans_finder.flares.core.config_file import ConfigFile
from trans_finder.flares.core.constants import Constants
from trans_finder.flares.core.input_data_spec import InputDataSpec
from trans_finder.flares.core.tags import *
from trans_finder.flares.core.types import RunPhase
from trans_finder.flares.create_custom_realisations import CreateCustomRealisations
from trans_finder.flares.data import Data
from trans_finder.flares.data_prep import DataPrep
from trans_finder.flares.debug_keras_model import KerasModelDebugger
from trans_finder.flares.diagnostics import DiagnosticsPlots, get_diagnostics_data
from trans_finder.flares.exceptions import DataScalingMissing
from trans_finder.flares.fit_mixture_model import FitMixtureModelStep
from trans_finder.flares.inspect_inputs import InspectInputs

from trans_finder.flares.plot_lib import PlotLib
from trans_finder.flares.plots import Plots
from trans_finder.flares.predictions import Predictions
from trans_finder.flares.reproduce_data_prep import ReproduceDataPrep
from trans_finder.flares.train import Train
from trans_finder.flares.train_val_split import TrainValSplitter
from trans_finder.flares.utilities.optional_aim import get_aim_run
from trans_finder.flares.utilities.optional_mlflow import initialize_mlflow
from trans_finder.flares.utils import get_columns_wildcard, str_join


class Manager(ConfigFile, InputDataSpec, Constants, extra="allow"):
    """The root object for executing a single step (i.e. run_phase) of a pipeline"""

    # run_phase indicates a task in context of an analysis pipeline
    # (e.g. data augmentation, model training, calibration, inference, ... )
    run_phase: RunPhase

    def setup(self, log):
        self.log = log
        self.data_prep = DataPrep(mngr=self)
        self.data = Data(mngr=self)
        self.train = Train(mngr=self)
        self.diagnostics = DiagnosticsPlots(mngr=self)
        self.predictions = Predictions(mngr=self)
        self.calib = Calib(mngr=self)
        self.plots = Plots(mngr=self)

        try:
            self.main_func_prnt = [["", " "], ["wg", "+"], ["", " "]]
            self.set_rnd_seeds()
            self.set_conf()

        except Exception as e:
            self.log.error([["r", f" - {type(e).__name__}:\n", e]])
            traceback.print_tb(e.__traceback__)
            raise e

    def get_base_name(self):
        return "flares"

    def run_bgmm_fit(self):
        FitMixtureModelStep(self).run()

    def inspect_inputs(self):
        InspectInputs(self).run()

    def create_custom_realisations(self):
        CreateCustomRealisations(self).run()

    def debug_keras_model(self):
        KerasModelDebugger(self).run()

    def reproduce_flares_data_prep(self):
        ReproduceDataPrep(self).run()

    def plot_diagnostics_runphase(self, *args, **kwargs):
        # load diagnostics df
        df_diagnostics = get_diagnostics_data(self, manual_runphase="predict")

        # do plots
        self.diagnostics.plot_diagnostics(df_diagnostics)

    def check_dataset_conf(self):
        """checks for missing entries and raises an error if necessary"""

        required_attributes = [
            "step_features_per_channel",
            "data_prep_fluctuation_cols",
        ]

        any_missing = False
        for attribute in required_attributes:
            if not hasattr(self, attribute):
                any_missing = True
                self.log.error(f"Dataset configuration: missing entry '{attribute}'")

        if any_missing:
            raise ValueError("Dataset configuration incomplete")

    def add_file_logging(self):
        from datetime import datetime

        logPath = os.path.join(self.tag_output_dir, self.tag_model_base, "logs")
        fileName = f"{datetime.now().strftime('%Y-%m-%d--%H-%M-%S')}--{self.run_phase}"

        # ensure logging dir exists
        Path(logPath).mkdir(parents=True, exist_ok=True)

        import logging

        logFormatter = logging.Formatter(
            "%(asctime)s | %(name)s | %(levelname)s | %(message)s"
        )
        rootLogger = logging.getLogger()
        fileHandler = logging.FileHandler("{0}/{1}.log".format(logPath, fileName))
        fileHandler.setFormatter(logFormatter)
        rootLogger.addHandler(fileHandler)

    def set_user_conf(self):
        """set main configuration parameters for a particular analysis"""
        self.add_file_logging()
        self.log.info([["b", " - set-user-conf("], ["p", self.run_phase], ["b", ")"]])

        # function to determine if an inputdataset should be shuffles
        def do_shuffle_dataset(mode):
            do_shuffle = mode in [ModeKeys.TRAIN, ModeKeys.EVAL]

            if self.is_run_wise_sorted_inputs:
                do_shuffle = do_shuffle and self.run_phase == "train"

            msg = f" - for phase({self.run_phase}, {mode}), do_shuffle = {do_shuffle}"
            self.log.debug([["b", msg]])

            return do_shuffle

        self.do_shuffle_dataset = do_shuffle_dataset

        # loss optimizer
        self.optimizer = tf.keras.optimizers.Adam

        """ function for calculating a weighting scheme for ts calculation
            as a default value, all weights are set to unity (uniform 
            inclusive calculation).

            the same weights are applied also to the decoder data, the predictions
            and the event weights which are used for plotting as part of the
            calib stage
        """

        def ts_time_uniform():
            return np.ones(
                (self.n_steps_dec, len(self.df_cols_ts_0)),
            )

        self.ts_weights_enc_dec_anom_det = {
            "tag": "uniform",
            "func": ts_time_uniform,
        }

        # range to perform the kde fit on, given that the classification TS
        # by construction is limited to the range [0, 1]. the kde_scale must
        # be large enough that probability does not leak out (eg the integrated
        # probability density in the kde fit range sums up to 1)
        kde_bins_n = int((self.kde_bins_max - self.kde_bins_min) * self.kde_scale)
        self.kde_sample_bins = np.linspace(
            self.kde_bins_min, self.kde_bins_max, kde_bins_n
        )

        # write prediction output to file at each iteration
        self.write_predict_out_each_iteration = False

        self.tag_custom_realisations = TAG_CUSTOM_REALISATIONS

        # exclude events which have unphysical flux values
        def filter_df_by_flux(df_in):
            df_out = df_in

            # TODO.... check for flux below self.min_flux_fermi and exclude
            # mask = df_out['flux_col'] > self.min_flux_fermi
            # df_out = df_out[mask]

            return df_out

        self.input_data_cut_funcs += [filter_df_by_flux]

        # ensure every channel in step_features_per_channel has as last entry the temporal offset
        if any(
            TAG_TEMPORAL_OFFSET in step_ft_dict.keys()
            for step_ft_dict in self.step_features_per_channel.values()
        ):
            self.log.warn(
                f"""The step_features_per_channel config contains features named {TAG_TEMPORAL_OFFSET}, which is a 
                reserved name. Entries with this name will be ignored. The internal {TAG_TEMPORAL_OFFSET} features are 
                automatically handled and do not need to be configured here. """
            )

        # handle temporal offset features: remove if existing, and insert as last entry in the dict
        for channel_name, channel_conf in self.step_features_per_channel.items():
            channel_conf.pop(TAG_TEMPORAL_OFFSET, None)
            channel_conf[TAG_TEMPORAL_OFFSET] = None

        # default values if it is not part of the configuration
        self.data_prep_static_df_columns = getattr(
            self, "data_prep_static_df_columns", {}
        )

        # whether to add the autoencoder reconstruction error as an additional dimension to the representation
        # a high reconstruction error occurs for input data that the network cannot generalize to
        # with the fitted gaussian mixture model, this ensures high test statistics for these types of input data
        self.bgmm_add_auto_enc_reconstruction_error_as_extra_dim = False

        self.data_step_col_names = []
        self.data_error_up_col_names = []
        self.data_error_down_col_names = []
        for channel_tag in self.step_features_per_channel.keys():
            for feature_name, n_bins in self.step_features_per_channel[
                channel_tag
            ].items():
                if n_bins is None:
                    self.data_step_col_names.append((channel_tag, feature_name))
                    self.data_error_up_col_names.append(
                        (channel_tag, "_".join([feature_name, "error_up"]))
                    )
                    self.data_error_down_col_names.append(
                        (channel_tag, "_".join([feature_name, "error_down"]))
                    )
                else:
                    self.data_step_col_names.extend(
                        (
                            (channel_tag, f"{feature_name}_{n_bin}")
                            for n_bin in range(n_bins)
                        )
                    )
                    self.data_error_up_col_names.extend(
                        (
                            channel_tag,
                            f"{feature_name}_{n_bin}_error_up",
                        )
                        for n_bin in range(n_bins)
                    )
                    self.data_error_down_col_names.extend(
                        (
                            channel_tag,
                            f"{feature_name}_{n_bin}_error_down",
                        )
                        for n_bin in range(n_bins)
                    )

        self.n_main_cols = len(self.data_step_col_names)
        # if len(self.data_step_col_names) != self.n_main_cols:
        #     raise ValueError(
        #         f"{len(self.data_step_col_names)=} must match {self.n_main_cols=}. "
        #         f"Maybe the config for step_features_per_channel is not matching with n_main_cols?"
        #     )

        # function to access light curve columns names from the input DataFrame
        def get_col_name(step_col_names, n_feat, n_step):
            channel_tag, feat = step_col_names[n_feat]
            step = f"{self.data_step_prefix}_{n_step}"
            return f"{channel_tag}_{step}_{feat}"

        def lc_col_func(n_step, n_feat):
            return get_col_name(self.data_step_col_names, n_feat, n_step)

        def lc_error_up_col_func(n_step, n_feat):
            return get_col_name(self.data_error_up_col_names, n_feat, n_step)

        def lc_error_down_col_func(n_step, n_feat):
            return get_col_name(self.data_error_down_col_names, n_feat, n_step)

        self.lc_col_func = lc_col_func
        self.lc_error_up_col_func = lc_error_up_col_func
        self.lc_error_down_col_func = lc_error_down_col_func

        self.step_feature_names = [
            f"{channel_tag}_{feat}" for (channel_tag, feat) in self.data_step_col_names
        ]

        def meta_step_col_func(col_name, n_step):
            return f"{col_name}"
            # return f"{col_name}_{n_step}"

        self.meta_step_col_func = meta_step_col_func

        # first part of the time series of the input data
        self.tag_input_d0 = TAG_INPUT_D0

        # second part of the time series of the input data (potentially contains flare)
        self.tag_input_d1 = TAG_INPUT_D1

        # forecasting of the background over the second part of the input
        self.tag_bck_forecast_d1 = self.tag_pred

        # the input to the autoencoder
        self.tag_auto_enc_input = TAG_AUTO_ENC_INPUT

        # hidden representation learned by the autoencoder given
        # model_input_d1 and model_bck_forecast_d1
        self.tag_cluster_space_sample = TAG_CLUSTER_SPACE_SAMPLE

        self.tag_cluster_space_mean = TAG_CLUSTER_SPACE_MEAN
        self.tag_cluster_space_std = TAG_CLUSTER_SPACE_STD
        self.tag_cluster_space_mean_normed = TAG_CLUSTER_SPACE_MEAN_NORMED

        # # hidden representation of the autoencoder for given the
        # # real onput data (second part of the time series)
        # self.tag_input_d1_hidden = "model_input_d1_hidden"
        # output of the autoencoder, trained to reconstruct the forcasted background
        self.tag_auto_enc_reconstruction = TAG_AUTO_ENC_RECONSTRUCTION

        self.tag_model_forecasting_error = TAG_MODEL_FORECASTING_ERROR

        # the error of the autoencoder reconstruction task
        self.tag_auto_enc_reconstruction_error = TAG_AUTO_ENC_RECONSTRUCTION_ERROR
        self.tag_auto_enc_reconstruction_error_normed = (
            TAG_AUTO_ENC_RECONSTRUCTION_ERROR_NORMED
        )

        # neg-log probability of the real data (second part of the time series)
        # given the hidden representation of the autoencoder
        self.neg_neg_log_prob_input_d1_hidden = "model_neg_log_prob_input_d1_hidden"

        self.tag_bgmm_quality_of_fit = "bgmm_quality_of_fit"

        # number of gaussian components for the mixture model of the
        # hidden representation of the autoencoder
        if self.model_gaus_mix_n_comps < 1:
            raise Exception(
                "model_gaus_mix_n_comps({self.model_gaus_mix_n_comps}) too small"
            )

        # placeholder for a dict which will contain the mcmc state results after fitting
        self.mcmc_states = None

        self.diagnostics_file_name_patt = ["diagnostics", ".hdf5"]

        self.physical_flare_file_name_patt = [".hdf5"]

        self.tag_temporal_offset = TAG_TEMPORAL_OFFSET
        self.tag_fluctuation = TAG_FLUCTUATION
        self.tag_physical_flare = TAG_PHYSICAL_FLARE
        self.tag_background = TAG_BACKGROUND

        if not isinstance(self.lstm_enc_sizes, Iterable):
            self.lstm_enc_sizes = [self.lstm_enc_sizes]
        if not isinstance(self.lstm_dec_sizes, Iterable):
            self.lstm_dec_sizes = [self.lstm_dec_sizes]

        self.temporal_offset_column_indices = []
        for channel_tag in self.step_features_per_channel:
            for idx, (tuple_channel_tag, ft_name) in enumerate(
                self.data_step_col_names
            ):
                if (
                    channel_tag == tuple_channel_tag
                    and ft_name == self.tag_temporal_offset
                ):
                    self.temporal_offset_column_indices.append(idx)
                    break

        if not self.use_reco_err_ts_combo:
            self.calib_stages.remove("calib_step_combo")
            self.meta_bin_calib_stage_cols.pop("calib_step_combo", None)

        self.evt_type_ratios["debug_keras_model"] = self.evt_type_ratios["train"]
        self.evt_type_ratios["inspect_inputs"] = self.evt_type_ratios["train"]

        self.max_evt["debug_keras_model"] = self.max_evt["train"]
        self.max_evt["inspect_inputs"] = (
            self.max_evt["train"]
            if "inspect_inputs" not in self.max_evt
            else self.max_evt["inspect_inputs"]
        )

        self.input_scaling_quant_cut = 0.03

        self.mlflow = initialize_mlflow(self)

        # self.channel_labels = [
        #     self.plot_channel_labels[channel_tag]
        #     + r" $\times 10^{"
        #     + str(self.plot_channel_y_normalisation[channel_tag])
        #     + r"}$"
        #     for channel_tag in self.data_prep_channel_tags
        # ]

        self.time_label = "Time [days]"
        self.plot_flux_label = r"log$_{10}$( Flux $[$ photons cm$^{-2}$ s$^{-1}$ ] )"

        # a file storing the random realisation ids randomly picked for the validation split
        self.tag_val_split_realisation_ids = "tag_val_split_realisation_ids"

        self.train_val_splitter = TrainValSplitter(self)

        self.tag_realisations_metadata_file = "realisations_metadata.csv"

        self.tag_inference_channel_padding = "infer_channel_padding"

        self.data_prep_channel_tags = list(self.step_features_per_channel.keys())

        def get_channel_col_error_names(channel_feature_tuples):
            return {
                tag: [
                    n
                    for (c, n) in channel_feature_tuples
                    if c == tag and not n.startswith(self.tag_temporal_offset)
                ]
                for tag in self.data_prep_channel_tags
            }

        self.data_prep_error_feature_cols = get_channel_col_error_names(
            self.data_step_col_names
        )
        self.data_prep_error_up_cols = get_channel_col_error_names(
            self.data_error_up_col_names
        )
        self.data_prep_error_down_cols = get_channel_col_error_names(
            self.data_error_down_col_names
        )

        self.tag_input_errors_up = TAG_INPUT_ERRORS_UP
        self.tag_input_errors_down = TAG_INPUT_ERRORS_DOWN
        self.tag_model_input_ids = TAG_MODEL_INPUT_IDS

        self.log.info(
            [
                ["", " "],
                ["wb", "    "],
                ["", " model version: "],
                ["", self.train.get_full_model_version_name()],
                ["", " "],
                ["wb", "    "],
            ]
        )

        if hasattr(self, "input_dataset_version_name"):
            self.tag_input_dir = os.path.join(
                self.tag_input_dir, self.input_dataset_version_name
            )

        for channel_tag in self.data_prep_channel_tags:
            if not hasattr(self, "plot_channel_labels"):
                self.plot_channel_labels = dict()

            if channel_tag not in self.plot_channel_labels:
                # self.log.debug(
                #     [
                #         [
                #             "",
                #             f" using default ploting label for channel {channel_tag} ",
                #         ],
                #     ]
                # )
                self.plot_channel_labels[channel_tag] = channel_tag

        for channel_tag in self.data_prep_channel_tags:
            if not hasattr(self, "data_prep_n_days_per_step"):
                self.plot_channel_labels = dict()
                self.data_prep_n_days_per_step = {}

            if channel_tag not in self.data_prep_n_days_per_step:
                self.log.info(
                    [
                        [
                            " - ",
                            f" assuming default time bin width of 1 = 1 day for channel {channel_tag}. "
                            + f"Can be changed in config flag 'data_prep_n_days_per_step'",
                        ],
                    ]
                )
                self.data_prep_n_days_per_step[channel_tag] = 1.0

        for channel_tag in self.data_prep_channel_tags:
            if not hasattr(self, "plot_channel_y_normalisation"):
                self.plot_channel_y_normalisation = dict()

            if channel_tag not in self.plot_channel_y_normalisation:
                self.plot_channel_y_normalisation[channel_tag] = 1.0

        signal_integral_col_names = [
            f"{channel_tag}_signal_integral_dec"
            for channel_tag in self.data_prep_channel_tags
        ]
        self.df_cols_to_keep[0]["cols"].extend(signal_integral_col_names)

        self.tag_event_id = "evt_id"
        self.df_cols_to_keep[0]["cols"].append(self.tag_event_id)

        # adjusting batch sizes as batches grow trough sampling from input errors
        n_row_repetitions = self.model_n_samples_of_input_errors + 1

        def adjust_batch_size(size):
            adjusted = size
            # adjusted = int(np.ceil(size / n_row_repetitions))

            # make sure all repetitions fit in one batch
            adjusted += n_row_repetitions - (adjusted % n_row_repetitions)
            return adjusted

        self.predict_batch_size = adjust_batch_size(self.predict_batch_size)

        self.unique_evt_classes = self.meta_bin_col_params[self.tag_evt_type]["values"][
            0
        ]
        self.unique_evt_classes.sort()  # sorted alphabetically

        # todo cleanup duplicate config setting
        # (setting model_learned_representation_n_dim based on the shape of the data here )
        # n_features is actually defined only after this function. the line below is a duplicate and this
        # should be fixed.
        self.n_features = self.n_main_cols + len(self.meta_step_cols)

        self.autoenc_data_shape = (
            self.n_steps_dec,
            (self.n_features) - len(self.temporal_offset_column_indices),
        )

        self.model_learned_representation_n_dim = int(
            self.model_cluster_space_dimensions_compression_factor
            * np.product(self.autoenc_data_shape)
        )

        self.path_realisation_metadata = os.path.join(
            self.tag_input_dir,
            self.dir_output_prep_train,
            self.tag_realisations_metadata_file,
        )

    def run(self):
        """Run a given stage of the analyis"""

        if self.do_train_scale and self.run_phase == "train":
            try:
                self.data.get_train_scales()
            except DataScalingMissing:
                self.reset_train_scales()
            except Exception as e:
                raise e

            if self.only_do_train_scale:
                return

        try:
            self.data.load_data()

            if self.run_phase == "train":
                if self.train_phases is not None:
                    if self.n_train_epochs > 0:
                        self.train.train_model()
                else:
                    self.train.train_model()

            elif self.run_phase == "bgmm_fit":
                self.run_bgmm_fit()

            elif self.run_phase == "debug_keras_model":
                self.debug_keras_model()

            elif self.run_phase == "inspect_inputs":
                self.inspect_inputs()

            elif self.run_phase == "calib":
                self.calib.derive_calib()

            elif self.run_phase in ["predict", "predict_sim"]:
                self.predictions.get_predictions()

            elif self.run_phase == "predict_plots":
                if self.is_enc_dec_anom_det or self.is_cluster_anom_det:
                    self.plots.plot_predictions()
                elif self.is_enc_dec_clasif:
                    self.plots.plot_predictions(is_sig_plot=False)
                    self.plots.plot_predictions(is_sig_plot=True)
            elif self.run_phase in self.valid_run_phases:
                pass
            else:
                raise Exception("run_phase(", self.run_phase, ") not well-defined ...")

        except Exception as e:
            self.log.error([["r", f" - {type(e).__name__}:", e]])
            raise e

    def on_termination(self):
        """is called before terminating a run phase, any cleanup goes here"""

        # if aim is used, this block releases potential aim-internal locks on files
        aim_run = get_aim_run(self)
        if aim_run is not None:
            aim_run.close()

    def set_conf(self):
        """Set configuration parameters
        Must be implemented to override base.Manager.set_conf()
        """

        try:
            self.set_pre_user_conf()
            self.set_dir_tags()
            self.set_user_conf()
            self.set_post_user_conf()
            self.cross_check_conf()
            self.data.create_out_dirs()
            self.setup_plot_conf()

        except Exception as e:
            raise e

        return

    def set_pre_user_conf(self):
        """default configuration parameters, set before any user
        confs, which may override
        """

        # tag for all time series elements
        self.tag_steps = "steps"

        # tag for encoder time series elements
        self.tag_enc = TAG_ENC

        # tag for decoder time series elements
        self.tag_dec = TAG_DEC

        # tag for weights for time series elements
        self.tag_wgt = TAG_WGT

        # tag for predictions
        self.tag_pred = TAG_PRED

        # tag for deterministic predictions (in case of probabilistic nominal preds)
        self.tag_pred_mean = "static_pred"

        # tag for event classes (eg signal / background)
        self.tag_evt_cats = "evt_cat"

        # tag for one-hot-encoded event classes
        self.tag_evt_hot_cats = "evt_hot_cat"

        # tag for background classes
        self.tag_bck_cats = "bck_cat"

        # tag for signal classes
        self.tag_sig_cats = "sig_cat"

        # tag for flag indicating if an event belongs to one of the background classes
        self.tag_is_bck_cat = "is_bck_cat"

        # tag for the input DataFrame
        self.tag_input_df = "input_df"

        # tag for loss
        self.tag_loss = TAG_LOSS

        # prefix for meta columns
        self.prefix_meta_bins = ""

        # tag for metadaa bin columns
        self.tag_meta_bins = "meta_bins"

        # tag for keeping track of the original metadata group tag, in case of mixed data
        self.flat_meta_key_tag = "flat_meta"

        # tag for keeping track of the original location of entry in input, when
        # splitting / mixing by metadata group
        self.orig_entry_index_tag = "orig_entry_index"

        # tag for flag, indicating is an event passes selection cuts
        self.tag_evt_clip = "pass_evt_clip"

        # tag for test statistics
        self.tag_ts = "ts"

        # keep track of the number of realizations iterations for calicration
        self.n_realization_tag = "n_realization"

        # tag for clustering
        self.tag_cluster = "cluster"

        # tag for p-values
        self.tag_log_p_value = "log_p_value"

        # tag for flag indicating successful p_value calculation
        self.tag_has_p_value = "has_p_value"

        # flag to store general event mask
        self.tag_evt_cut_mask = "evt_cut_mask"

        # flag to store event ts mask (masked events automatically get
        # low ts values corresponding to low-significance p-values)
        self.tag_evt_ts_mask = "evt_ts_mask"

        # tag for naming plot result files
        self.tag_plots = None

        # tag for X2 calculation
        self.tag_chi_2 = "chi_2"

        # tag to use for filtering files
        self.tag_filt_file = "_filt_"

        # general tag for the model
        self.tag_model = "keras_model"

        # model object
        self.model = dict()

        # tag to name model directory
        self.tag_model_dir = "model"

        # tag to name a model version, e.g. for tracking in tensorboard
        self.tag_model_full_version_name = "model_full_version_name"

        # tag to name calibration fit results directory
        self.tag_calib_dir = "calib_fits"

        # tag for variable scaling file name
        self.tag_input_scalings = {
            self.tag_steps: "train_scales_steps",
            self.tag_meta_bins: "train_scales_meta_bins",
        }

        # flag indicating if training scalling is being calculated in this run
        self.prep_train_scales = False

        # cut on value on quantiles for variable scaling (eg, the distribution of
        # features to be normalised is cut between quantile[0+val, 1-val], in order to
        # supress extreme outliers which would make the scaling numerically unfeasible)
        self.input_scaling_quant_cut = 1e-4

        # list of filtering functions on input DataFrame (eg cut out events based on
        # some criteria). this is applied in the first of the pipeline, before
        # any other function handles the input dataset
        self.input_data_cut_funcs = []

        # apply transformations on input data before any other processing
        self.transform_input_feature_funcs = []

        # default values to distinguish missing entries
        self.ts_nan_value = int(-1e6)
        self.p_value_nan_value = int(-1e6)

        # constants for plotting
        self.p_val_5_sigma_two_sided = 5.7e-7
        self.p_val_5_sigma_one_sided = 0.5 * self.p_val_5_sigma_two_sided  # 2.8e-7

        # 5-sigma threshold (as we end up with ndf=1, we get [ts_sqr_thresh=25] )
        self.ts_sqr_thresh = np.power(stats.norm.isf(self.p_val_5_sigma_one_sided), 2)

        # flag to perform sanity check, to make sure that self.meta_bins covers all
        # entries in the data. If False, remove not used meta bins.
        self.do_meta_sanity_check = True

        # dict to store input dataframes
        self.input_data = dict()

    def get_calib_tags(self, calib_stage, tag=None):
        output = {
            "ts": str_join(self.tag_ts, calib_stage),
            "log_p_val": str_join(self.tag_log_p_value, calib_stage),
            "wgt": str_join(self.tag_wgt, calib_stage),
        }
        if tag is not None:
            output = output[tag]

        return output

    def set_meta_bins_from_dict(self, meta_bin_col_params, tag_meta=""):
        """

        Parameters
        ----------
        meta_bin_col_params: dict
        tag_meta: str
            tag that is used a prefix for the meta columns
        with_scope: bool
            if true, expect scope key in the entries of meta_bin_col_params

        Returns
        -------

        """

        meta_bin_cols = [k for k in meta_bin_col_params.keys()]
        meta_n_bins, meta_bins, meta_types = dict(), dict(), dict()

        meta_col_dict = {}
        meta_bin_cols_tag = []
        for meta_bin in meta_bin_cols:
            if tag_meta != "":
                col_name = f"{tag_meta}_{meta_bin}"
            else:
                col_name = meta_bin

            if not isinstance(meta_bin_col_params[meta_bin], dict):
                # is not a meta bin but possibly adds other information...
                continue

            meta_bin_cols_tag.append(col_name)
            meta_col_dict[col_name] = meta_bin

            bin_params = meta_bin_col_params[meta_bin]
            bin_type = bin_params["type"]
            meta_types[col_name] = bin_type

            if bin_type == "linear":
                meta_n_bins[col_name] = bin_params["n"]
                meta_bins[col_name] = np.linspace(
                    bin_params["min"],
                    bin_params["max"],
                    meta_n_bins[col_name] + 1,
                )
            elif bin_type in ["fixed", "list", "range_longitude_deg"]:
                meta_n_bins[col_name] = len(bin_params["values"])
                meta_bins[col_name] = bin_params["values"]
            else:
                raise Exception(f"unsupported bin type for {col_name}, {bin_params}")

        # order the dicts
        meta_bin_cols_tag = sorted(meta_bin_cols_tag)
        meta_n_bins = OrderedDict(sorted(meta_n_bins.items()))
        meta_bins = OrderedDict(sorted(meta_bins.items()))
        meta_types = OrderedDict(sorted(meta_types.items()))

        return meta_bin_cols_tag, meta_n_bins, meta_bins, meta_types, meta_col_dict

    def set_meta_bins(self):
        cols, n_bins, bins, types, col_dict = self.set_meta_bins_from_dict(
            self.meta_bin_col_params, tag_meta=self.prefix_meta_bins
        )

        self.meta_bin_cols = cols
        self.meta_n_bins = n_bins
        self.meta_bins = bins
        self.meta_types = types
        self.meta_col_dict = col_dict

        self.meta_bins_dict = dict(
            meta_bins=self.meta_bins,
            meta_types=self.meta_types,
            meta_bin_cols=self.meta_bin_cols,
            meta_col_dict=self.meta_col_dict,
        )

        # sanity checks
        calib_stages = self.meta_bin_calib_stage_cols.keys()
        if set(calib_stages) != set(self.calib_stages):
            raise Exception(
                " must have etries for each calib_stage as part of meta_bin_calib_stage_cols ..."
            )

        for cols in self.meta_bin_calib_stage_cols.values():
            has_bins = [b in self.meta_bin_col_params.keys() for b in cols]
            if not all(has_bins):
                raise Exception(
                    " all entries in meta_bin_col_params must correspond "
                    + "to entried in meta_bin_calib_stage_cols ..."
                )

    def set_post_user_conf(self):
        """modify configuration parameters after user setting have been applied"""

        # mutually exlusive flags for the main setup of the architecture
        self.is_enc_dec_anom_det = self.tag_model_base == "enc_dec_anom_det"
        self.is_enc_dec_clasif = self.tag_model_base == "enc_dec_clasif"
        self.is_cluster_anom_det = self.tag_model_base == "cluster_anom_det"

        if not (
            self.is_enc_dec_anom_det
            or self.is_enc_dec_clasif
            or self.is_cluster_anom_det
        ):
            raise Exception("must choose one of the allowed general modes")

        # binnings schemes / value ranges for meta-data dimensions
        self.set_meta_bins()

        # extra column names to keep in DataFrame, regardless of the run type
        self.aux_cols_to_keep = [
            self.tag_evt_cut_mask,
            self.tag_evt_type,
            self.tag_evt_cats,
        ]

        # columns related to classification that should not be filtered out of dataframes
        self.clas_cols_to_keep = [self.tag_is_bck_cat]

        # verify that the user has set the following attributes
        check_attrs = [
            self.n_steps_tot,
            self.n_steps_dec,
            self.n_main_cols,
        ]
        if any([x is None for x in check_attrs]):
            raise Exception(f"undefined network arch check_attrs({check_attrs})")

        # how many time-steps are used for the encoder
        self.n_steps_enc = self.n_steps_tot - self.n_steps_dec

        self.has_meta_bins = len(self.meta_bin_cols) > 0

        if not self.has_meta_bins:
            self.add_meta_bin_to_steps = False

        # total number of features for the network (data + metadata columns)
        self.n_features = self.n_main_cols + len(self.meta_step_cols)
        if self.add_meta_bin_to_steps:
            self.n_features += len(self.meta_bin_cols)

        self.setup_calib_conf()

        # placeholder for a dict which will contain the p-value fit results
        self.p_value_fits = dict()

        # the first columns in the time-series array are the main features,
        # then all metadata columns are added on the right
        self.data_col_indices = list(range(self.n_main_cols))

        # titles for metadata bin columns
        self.meta_bin_titles = self.get_meta_bin_titles(
            self.meta_bins,
            self.meta_types,
            [["b", " - get-meta-bin-titles()"]],
        )

        def get_meta_bin_col_func(meta_bin_col, meta_bins_dict=None):
            meta_bins_dict = (
                self.meta_bins_dict if self.meta_bins_dict is None else meta_bins_dict
            )

            return meta_bins_dict["meta_col_dict"][meta_bin_col]

        self.get_meta_bin_col_func = get_meta_bin_col_func

        # classification categories based on event types
        if self.is_enc_dec_clasif:
            self.evt_type_to_cat = dict()

            self.bck_cats = []
            n_cat = 0
            for sb_type in [self.tag_bck_cats, self.tag_sig_cats]:
                for sb_cat in self.evt_type_cats[sb_type]:
                    for evt_type in sb_cat:
                        self.evt_type_to_cat[evt_type] = n_cat

                    if sb_type == self.tag_bck_cats:
                        self.bck_cats += [n_cat]

                    n_cat += 1

            # number of event classes (this is part of the model, and so
            # must be consistent for different self.run_type values)
            self.n_evt_cats = max(self.evt_type_to_cat.values()) + 1

        # for development, set the random seeds to a fixed value
        if self.is_const_random_seed:
            self.set_rnd_seeds(seed=234987324, tags=["np", "tf", "py"])

        # flag to indicate if signal or background samples are used for plotting
        self.is_sig_plot = self.run_phase != "train"

        # reset the calibration directoris
        if self.reset_calib_dir is None:
            self.reset_calib_dir = self.run_phase == "calib"

        # sanity check on the model outputs used for the current calibration scheme
        if self.is_cluster_anom_det and self.tag_bck_forecast_d1 != self.tag_pred:
            raise Exception(" - tag_bck_forecast_d1 incompatible with calib stage ...")

        # test input filter settings
        debug_input_filtering = False
        # debug_input_filtering = True
        if debug_input_filtering:
            self.in_file_filt = 2
            self.in_file_filt_now = 0
            self.in_file_filt_now = 1

            reset_test = False
            # reset_test = True
            if reset_test:
                self.in_file_filt = -1
                self.in_file_filt_now = -1

        # filtering variables, used for parallel processing of a large sample of
        # input files
        for key in ["in_file_filt"]:
            key_now = key + "_now"

            attr_0 = getattr(self, key)
            attr_1 = getattr(self, key_now)

            setattr(self, key, None)
            setattr(self, key_now, None)

            filt_modes = ["calib", "predict"]
            if (attr_0 > 1) and (self.run_phase in filt_modes):
                setattr(self, key, attr_0)
                setattr(self, key_now, attr_1)

                n_key_0, n_key_1 = getattr(self, key_now), getattr(self, key)
                if (n_key_0 < 0) or (n_key_0 >= n_key_1):
                    raise Exception("problem with input variables for filtering()")

                self.log.info(
                    [
                        ["c", " - setting in_file_filt/in_file_filt_now: "],
                        ["y", getattr(self, key)],
                        ["c", " / "],
                        ["y", getattr(self, key_now)],
                    ]
                )

        # size of loaded dataset (filled as part of read_dataset)
        self.n_dataset_samples = dict()

        # set which directories to keep when using reset_tag_dir
        if hasattr(self, "keep_dirs"):
            self.keep_dirs = [getattr(self, dir_now) for dir_now in self.keep_dirs]
        else:
            self.keep_dirs = []

        # set missing dropout values if required
        if self.rnn_recurrent_dropout is None:
            self.rnn_recurrent_dropout = self.rnn_dropout

        if self.input_dec_dropout is None:
            self.input_dec_dropout = self.rnn_dropout

        # for backward compatibility
        self.train_phase = self.train_phases[0]

    def setup_calib_conf(self):
        # name of nominal data columns used for predictions
        data_step_cols = [
            f"{self.data_feat_prefix}_{n_inp}" for n_inp in range(self.n_main_cols)
        ]
        if self.run_phase in ["train"]:
            self.log.info([["y", " - data_step: "], ["o", data_step_cols]])

        # naming scheme for the label columns
        self.df_cols_dec = [
            str_join(self.tag_dec, self.tag_model_base, c) for c in data_step_cols
        ]

        # naming scheme for the feature columns
        self.df_cols_pred = [
            str_join(self.tag_pred, self.tag_model_base, c) for c in data_step_cols
        ]

        # naming scheme for the test statistic column in the output
        calib_tags = self.get_calib_tags(calib_stage=self.calib_stages[0])
        tag_ts_0 = calib_tags["ts"]
        tag_log_p_val_0 = calib_tags["log_p_val"]
        tag_wgt_0 = calib_tags["wgt"]

        if self.is_enc_dec_anom_det:
            # a separate ts value for each of the nominal input data dimensions
            self.df_cols_ts_0 = [str_join(tag_ts_0, c) for c in data_step_cols]

            # plot titles
            self.df_cols_ts_title = dict(
                (c, c.replace("_", " ")) for c in self.df_cols_ts_0
            )

            # naming scheme for the weight columns
            self.df_cols_wgt_0 = [str_join(tag_wgt_0, c) for c in data_step_cols]

        elif self.is_cluster_anom_det:
            # nominal ts value
            if self.use_reco_err_ts_combo:
                self.df_cols_ts_0 = [
                    str_join(tag_ts_0, "mcmc_prob"),
                    str_join(tag_ts_0, "reco_err"),
                ]
            else:
                self.df_cols_ts_0 = [tag_ts_0]

            self.df_cols_ts_title = dict(
                (c, c.replace("_", " ")) for c in self.df_cols_ts_0
            )

            # naming scheme for the weight columns
            self.df_cols_wgt_0 = [
                c.replace(tag_ts_0, tag_wgt_0) for c in self.df_cols_ts_0
            ]

        else:
            # nominal ts value
            self.df_cols_ts_0 = [tag_ts_0]
            self.df_cols_ts_title = {
                tag_ts_0: tag_ts_0.replace("_", " "),
            }

            # naming scheme for the weight columns
            self.df_cols_wgt_0 = [tag_wgt_0]

        # naming scheme for the p-value columns (one for each ts column)
        df_cols_p_val = [
            c.replace(tag_ts_0, tag_log_p_val_0) for c in self.df_cols_ts_0
        ]

        # mapping between ts and p-value / weight names
        self.ts_to_wgt_col = dict(
            (self.df_cols_ts_0[n], self.df_cols_wgt_0[n])
            for n in range(len(self.df_cols_ts_0))
        )

        self.ts_to_p_val_col = dict(
            (self.df_cols_ts_0[n], df_cols_p_val[n])
            for n in range(len(self.df_cols_ts_0))
        )

        self.df_cols_ts = [self.df_cols_ts_0]

        # an additional p-value for the combined ts for anomaly detection
        for n_calib_stage in range(len(self.calib_stages)):
            calib_stage = self.calib_stages[n_calib_stage]
            calib_tags = self.get_calib_tags(calib_stage=calib_stage)
            tag_ts_step_combo = calib_tags["ts"]
            tag_log_p_val_combo = calib_tags["log_p_val"]
            tag_wgt_combo = calib_tags["wgt"]

            self.ts_to_wgt_col[tag_ts_step_combo] = tag_wgt_combo
            self.ts_to_p_val_col[tag_ts_step_combo] = tag_log_p_val_combo
            self.df_cols_ts_title[tag_ts_step_combo] = calib_stage.replace("_", " ")

            if n_calib_stage > 0:
                self.df_cols_ts += [[tag_ts_step_combo]]

        self.ts_to_calib_stage = dict()
        for tag_ts_now in self.df_cols_ts_0:
            self.ts_to_calib_stage[tag_ts_now] = self.calib_stages[0]

        for calib_stage in self.calib_stages:
            tag_ts_now = self.get_calib_tags(calib_stage=calib_stage, tag="ts")
            self.ts_to_calib_stage[tag_ts_now] = calib_stage

        # naming schema for significance columns
        self.p_val_to_significance_col = {
            c: c.replace(self.tag_log_p_value, TAG_SIGNIFICANCE)
            for c in self.ts_to_p_val_col.values()
        }

    def get_meta_bin_titles(self, meta_bins, meta_types, log_title=None):
        meta_bin_titles = dict()
        if log_title:
            self.log.info(log_title)

        for k, v in meta_bins.items():
            if meta_types[k] in ["linear"]:
                meta_bin_titles[k] = [
                    (
                        str(k)
                        + " ("
                        + str(n)
                        + ")"
                        + " ~ {:>2.2f}".format(v[n] + 0.5 * (v[n + 1] - v[n]))
                    )
                    for n in range(len(v) - 1)
                ]
            elif meta_types[k] in ["range_longitude_deg"]:
                meta_bin_titles[k] = [
                    (str(k) + " (" + str(n) + ")" + f" = {v[n]}") for n in range(len(v))
                ]
            elif meta_types[k] in ["fixed", "list"]:
                meta_bin_titles[k] = [
                    (str(k) + " (" + str(n) + ")" + f" = {v[n]}") for n in range(len(v))
                ]

            if self.run_phase in [
                "train",
                "calib",
                "prep_train",
                "prep_calib",
                "prep_predict",
                "prep_sim",
            ]:
                self.log.info([["b", " - meta_bin: "], ["p", meta_bin_titles[k]]])
        return meta_bin_titles

    def cross_check_conf(self):
        """cross-check configuration parameters after user-inputs are defined.
        supplement missing configs where needed
        """

        # cast to int
        self.max_evt = dict((k, int(v)) for k, v in self.max_evt.items())

        if self.is_enc_dec_clasif:
            self.train_batch_size *= 2
            self.max_evt["calib"] *= 2

        if self.run_phase == "calib":
            self.max_evt["train"] = self.max_evt["calib"]

        if self.run_phase != "train" and not self.is_cluster_anom_det:
            self.frac_train_eval = 1

        if self.reset_tag_dir:
            if self.run_phase not in self.reset_tag_dir_phases:
                self.log.warn(
                    [
                        [
                            "r",
                            f" - reset_tag_dir can only be used as part of {self.reset_tag_dir_phases}... ",
                        ],
                        ["r", " changing value to False"],
                    ]
                )
                self.reset_tag_dir = False
            else:
                self.log.warn(
                    [
                        ["r", " - reset_tag_dir is resetting the entire tag!"],
                    ]
                )

        if self.is_enc_dec_clasif:
            if any(t is None for t in self.evt_type_cats.values()):
                raise Exception(
                    "missing self.evt_type_cats setting"
                    + " -> need to define self.evt_type_cats ..."
                )

            cats_bck = list(itertools.chain(*self.evt_type_cats[self.tag_bck_cats]))
            cats_sig = list(itertools.chain(*self.evt_type_cats[self.tag_sig_cats]))

            for run_phase in ["train", "calib"]:
                has_bck = any(
                    [c in self.evt_type_ratios["train"].keys() for c in cats_bck]
                )
                has_sig = any(
                    [c in self.evt_type_ratios["train"].keys() for c in cats_sig]
                )

                if not (has_bck and has_sig):
                    raise Exception(
                        "- missing bck or sig categories in evt_type_ratios for \n"
                        + f"cats_bck: {cats_bck} \n"
                        + f"cats_sig: {cats_sig} \n"
                        + f"evt_type_ratios: {self.evt_type_ratios}"
                    )

        # verify that the attributes following have been set
        check_atts = [
            "in_file_patt",
            "out_file_patt",
            "tag_hdf",
            "lc_col_func",
            "tag_evt_type",
        ]
        for att_now in check_atts:
            try:
                att_val = getattr(self, att_now)
                if att_val is None:
                    raise Exception()
            except Exception:
                raise Exception("missing / None setting:", att_now)

        # valudate that all meta-data dimensions have correspondinf column definitions
        if not all(x in self.meta_bins.keys() for x in self.meta_bin_cols):
            raise Exception("problem with meta_bin_cols setup")

        if self.is_enc_dec_clasif and self.avg_time_steps:
            raise Exception("avg_time_steps not supported for classification")

        return

    def setup_plot_conf(self):
        """setup plotting config"""

        self.plot_lib = PlotLib(log=self.log, tag_base_dir=self.tag_base_output_dir)

        # directory naming scheme for different setups
        dir_type = None
        if self.run_phase == "calib":
            dir_type = "calib_plots"
        elif self.run_phase in ["predict_plots"]:
            dir_type = "predict_plots"

        if dir_type is not None:
            # initialise output pdf file
            plot_file_name = self.data.get_dir_name(dir_type)

            if self.tag_plots is None:
                self.tag_plots = self.dir_output_predict.replace("/", "_")

            file_posfix = (
                str_join(self.tag_model_base.replace("/", "_"), self.tag_plots) + ".pdf"
            )
            plot_file_name = os.path.join(plot_file_name, file_posfix)

            self.plot_lib.set_pdf_pages(plot_file_name)

        # register a class variable which can be useful for jupyter-notebook access
        self.plot_output = None

        return

    def reset_train_scales(self):
        """calculate scaling factors for inputs"""

        self.log.info([*self.main_func_prnt, ["g", " - reset-train-scales()"]])

        # reset the input data with a large number of
        # inputs (by the tag 'train_scales')
        self.prep_train_scales = True
        self.input_data = dict()
        self.data.load_data()
        self.data.set_train_scales()

        # now try again to load the file as a
        # cross-check (avoid infinite loops)
        self.data.get_train_scales(force_check=True)

        # reset the input data with the selected number
        # of inputs (by the tag 'train')
        self.prep_train_scales = False
        self.input_data = dict()

    def get_phase_cols_to_keep(self, df):
        keep_cols = []
        for col in self.clas_cols_to_keep + self.aux_cols_to_keep:
            keep_cols += get_columns_wildcard(df, col)

        for cols_by_pahse in self.df_cols_to_keep:
            if self.run_phase in cols_by_pahse["phases"]:
                for col in cols_by_pahse["cols"]:
                    keep_cols += get_columns_wildcard(df, col)

        return keep_cols

    def set_command_line_overrides(self, cmnd_conf):
        """overwrite config options with none-default arguments of the command line"""
        for k, v in cmnd_conf.items():
            if (k in cmnd_conf["is_def_val"]) or (k == "is_def_val"):
                continue
            elif k == "tag":
                self.tag_dir = v
            else:
                setattr(self, k, v)

    def set_dir_tags(self):
        """set the input/output paths after loading all user configs"""

        # check attr first and make all input args as self
        if self.tag_input_dir is None:
            self.tag_input_dir = os.getcwd()
            self.tag_input_dir = os.path.join(self.tag_input_dir, "input", self.tag_dir)

        if not self.tag_input_dir.endswith("/"):
            self.tag_input_dir += "/"

        if self.tag_base_output_dir is None:
            self.tag_base_output_dir = os.getcwd()

        if not self.tag_base_output_dir.endswith("/"):
            self.tag_base_output_dir += "/"

        self.tag_output_dir = (
            os.path.join(self.tag_base_output_dir, "output", self.tag_dir) + "/"
        )

        return

    def set_rnd_seeds(self, tags=None, seed=None):
        """initialize the random seeds

        by default same setting for numpy every time,
        and a unique seed for tf (for e.g., batch shuffling)

        Parameters
        ----------
        tags : str
            the target of the setting
        seed : int
            the value of the seed to be set
        """
        const_seed = 123456

        if tags is None:
            seed_now = datetime.utcnow() - datetime.utcfromtimestamp(0)
            seed_now = int(seed_now.total_seconds())

            np.random.seed(seed_now)
            tf.random.set_seed(seed_now)
            random.seed(seed_now)
        else:
            try:
                if seed is None:
                    seed = const_seed
                else:
                    try:
                        seed = int(seed)
                    except Exception:
                        raise Exception("seed must be of type int: ", seed)

                has_set = False
                if "np" in tags:
                    has_set = True
                    np.random.seed(seed)
                if "tf" in tags:
                    has_set = True
                    tf.random.set_seed(seed)
                if "py" in tags:
                    has_set = True
                    random.seed(seed)
                if not has_set:
                    raise Exception()
                self.log.info([["b", " - setting random seed: "], ["g", tags, seed]])
            except Exception as e:
                self.log.error(
                    [["r", " problem with parameters for set_rnd_seeds: ", tags, seed]]
                )
                raise e

        return
