import multiprocessing
import time
from datetime import timedelta
from pprint import pprint

import numpy as np
from scipy.stats import norm

from trans_finder.flares.mcmc_fit.evaluate_fit import evaluate_mixture_model_fit
from trans_finder.flares.model_utils.bayesian_gaussian_mixture_model import (
    BayesianGaussianMixtureModel,
)
from trans_finder.flares.state_manager import StateManager


def get_new_params(fit_data, previous_params=None):
    if previous_params is None:
        # the initial configuration
        n_burning_steps = 400

        return dict(
            n_input_dims=fit_data.shape[1],
            n_input_length=fit_data.shape[0],
            gmm_n_components=6,
            gmm_wishard_degrees_of_freedom=fit_data.shape[1],
            hmc_n_parallel_chains=3,
            hmc_n_steps=200,
            hmc_n_burnin_steps=n_burning_steps,
            hmc_n_step_size_adaptation_steps=int(0.9 * n_burning_steps),
            hmc_step_size=0.01,
            hmc_n_leapfrog_integrator_steps=30,
            hmc_max_parallel_iterations=multiprocessing.cpu_count(),
        )

    # how complexity increases every time
    params = previous_params.copy()
    params["hmc_n_steps"] = min(400, int(100 + params["hmc_n_steps"]))
    params["hmc_n_burnin_steps"] = min(1000, int(100 + params["hmc_n_burnin_steps"]))
    params["hmc_n_step_size_adaptation_steps"] = int(0.9 * params["hmc_n_burnin_steps"])
    params["gmm_n_components"] = min(
        2 * params["n_input_dims"],
        int(params["gmm_n_components"] + 1),
    )

    params["n_input_length"] = fit_data.shape[0]

    return params


def run_fit(mngr, params, fit_input, full_input):
    bgmmodel = BayesianGaussianMixtureModel(**params)

    mngr.log.info("MCMC fit starting...")
    pprint(params)
    st = time.time()
    chain_results = bgmmodel.fit(fit_input)
    et = time.time()
    mngr.log.info(f"MCMC fit done. duration: {timedelta(seconds=et - st)}")

    quality_of_fit_fit_input = bgmmodel.calculate_quality_of_fit(fit_input)
    quality_of_fit_full_input = bgmmodel.calculate_quality_of_fit(full_input)
    mngr.log.info(f"MCMC fit - {quality_of_fit_fit_input=}")
    mngr.log.info(f"MCMC fit - {quality_of_fit_full_input=}")

    return (
        bgmmodel,
        chain_results,
        quality_of_fit_full_input,
    )


def do_dynamic_mcmc_fit_with_increasing_complexity(mngr, input_train, input_val):
    """In a loop, do mcmc fits with increasing complexity on growing subsets of the datasets
    until the full dataset's quality of fit stops improving"""

    sample_size = min(5000, len(input_train))
    last_p_value = -1
    do_continue = True
    previous_params = None
    n_fit = 0
    n_min_tries = mngr.mcmc_dynamic_fit_n_min_tries

    while do_continue:
        # get random yet reproducible subsample of desired length
        fit_data = input_train[
            np.random.RandomState(seed=0).permutation(len(input_train))
        ][:sample_size]

        # update mixture model and mcmc parameters
        params = get_new_params(fit_data, previous_params)

        # run mcmc fit
        (
            bgmmodel,
            chain_results,
            quality_of_fit_full_input,
        ) = run_fit(mngr, params, fit_data, full_input=input_train)

        if False:
            # disabled, because it did not improve the quality of fit
            prune_low_contribution_components(
                mngr, bgmmodel, input_train, dist_overlaps
            )

        # the p value comes from a two sample KS test, with null hypothesis being dists are equal and alternative being
        # they are different (two-sided): https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.ks_2samp.html
        # A large difference in the dists corresponds to a small p value.
        # p value: probability of being the same
        p_value_ks_test_avg_across_chains = np.mean(quality_of_fit_full_input)

        prob_different = 1 - p_value_ks_test_avg_across_chains
        last_prob_different = 1 - last_p_value

        # check if it has improved at all
        has_improved = prob_different < last_prob_different

        # check if delta is at least 10% of 2 sigma to see if it is improving significantly
        has_improved_significantly = (
            last_prob_different - prob_different
        ) > 0.1 * 2 * norm.cdf(-2)

        # check for at least 2 sigma significance of being the same distribution
        is_acceptable = prob_different < 2 * norm.cdf(-2)

        do_continue = has_improved_significantly or (
            not is_acceptable and n_fit < n_min_tries
        )

        mngr.log.info(
            f"Quality of fit {is_acceptable=}, {has_improved=}, {has_improved_significantly=}."
        )
        mngr.log.info(
            f"{prob_different=:.2}, p-value={p_value_ks_test_avg_across_chains:.2e}, {do_continue=}."
        )

        # create diagnostics plots for the entire dataset
        _ = evaluate_mixture_model_fit(
            mngr,
            bgmmodel,
            input_train,
            quality_of_fit_full_input,
            chain_results,
            f"dynamic_{n_fit}",
            chain_desc=[
                f"chain #{n_chain} | p_value={quality_of_fit_full_input[n_chain]:.2e} | mixture probs={bgmmodel._state[0][n_chain]}"
                for n_chain in range(params["hmc_n_parallel_chains"])
            ],
        )

        # create diagnostics plots for only the subsample used for the fit
        _ = evaluate_mixture_model_fit(
            mngr,
            bgmmodel,
            fit_data,
            bgmmodel.get_quality_of_fit_ks_2samp(fit_data),
            chain_results,
            f"onlysample_dynamic_{n_fit}",
            chain_desc=[
                f"chain #{n_chain} | p_value={quality_of_fit_full_input[n_chain]:.2e} | mixture probs={bgmmodel._state[0][n_chain]}"
                for n_chain in range(params["hmc_n_parallel_chains"])
            ],
        )

        if has_improved:
            # save fit to disk
            StateManager(mngr=mngr, state_id="bgmm").save_state(bgmmodel._state)
            StateManager(mngr, "bgmm_chain_results").save_state(chain_results)
            StateManager(mngr, mngr.tag_bgmm_quality_of_fit).save_state(
                quality_of_fit_full_input
            )

        if do_continue:
            # prepare next iteration
            previous_params = params
            last_p_value = p_value_ks_test_avg_across_chains
            n_fit += 1
            sample_size = min(int(1.5 * sample_size), len(input_train))

    mngr.log.info(
        "Stopped increasing complexity of the mcmc fit as previously it dit not improve performance."
    )
