from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from matplotlib import pyplot as plt

from trans_finder.flares.checks.generic_checks import calculate_dist_overlap
from trans_finder.flares.helpers_debugging import get_debug_output_filepath
from trans_finder.flares.plot_list.bgmm import (
    plot_per_dimension_input_and_posterior_histograms,
)
from trans_finder.flares.plot_list.util.ecdf import plot_ecdf
from trans_finder.flares.plot_list.util.plot_config import gridlines_config, tick_params
from trans_finder.flares.utilities.pdf_plot_collection import PDFPlotCollection

if TYPE_CHECKING:
    from trans_finder.flares.manager import Manager
    from trans_finder.flares.model_utils.bayesian_gaussian_mixture_model import (
        BayesianGaussianMixtureModel,
    )


def calculate_dim_overlap_scores(bgmm, input_data, posterior_samples=None):
    if posterior_samples is None:
        posterior_samples = bgmm.get_joint_model_dist().sample(len(input_data)).numpy()

    dist_overlaps = []

    for i_chain in range(bgmm.hmc_n_parallel_chains):
        dist_overlap = calculate_dist_overlap(
            input_data,
            posterior_samples[:, i_chain, :],
            bins=1000,
        )
        dist_overlaps.append(dist_overlap)

    return np.array(dist_overlaps)


def plot_comparison_ecdf_datats_posteriorsamplets(
    mngr: Manager,
    bgmmodel: BayesianGaussianMixtureModel,
    data,
    posterior_samples,
    data_desc,
):
    fig, axes = plt.subplots(
        nrows=1,
        figsize=(5 * bgmmodel.hmc_n_parallel_chains, 5),
        ncols=bgmmodel.hmc_n_parallel_chains,
    )
    for i_chain in range(bgmmodel.hmc_n_parallel_chains):
        ts_data = bgmmodel.get_outlier_score(data, i_chain)
        ts_posterior = bgmmodel.get_outlier_score(
            posterior_samples[:, i_chain], i_chain
        )
        ax = axes[i_chain]

        plot_ecdf(ax, ts_data, label=f"data")
        plot_ecdf(ax, ts_posterior, label=f"posterior_samples", linestyle=":")

        ax.grid(**gridlines_config)
        ax.tick_params(**tick_params)
        ax.set_xlabel("ts: negative log prob")
        ax.set_ylabel("CDF")
        # ax.set_yscale("log")
        ax.legend()
    PDFPlotCollection(mngr, figs=[fig]).save_to_file(
        get_debug_output_filepath(mngr, "bgmmodel", f"{data_desc}_ts_cdf.pdf")
    )


def plot_prob_comparison_data_posteriorsample_cdf(
    mngr: Manager,
    bgmmodel: BayesianGaussianMixtureModel,
    data,
    posterior_samples,
    data_desc,
):
    fig, axes = plt.subplots(
        nrows=1,
        figsize=(5 * bgmmodel.hmc_n_parallel_chains, 5),
        ncols=bgmmodel.hmc_n_parallel_chains,
    )
    for i_chain in range(bgmmodel.hmc_n_parallel_chains):
        prob_data = bgmmodel.get_posterior_dist(i_chain).prob(data).numpy()

        print(
            data_desc, "prob mean", i_chain, f"{np.mean(prob_data):.2e}"
        )  # temp remove
        prob_posterior = (
            bgmmodel.get_posterior_dist(i_chain)
            .prob(posterior_samples[:, i_chain])
            .numpy()
        )

        ax = axes[i_chain]

        plot_ecdf(ax, prob_data, label=f"data")
        plot_ecdf(ax, prob_posterior, label=f"posterior_samples", linestyle=":")

        ax.grid(**gridlines_config)
        ax.tick_params(**tick_params)
        ax.set_xlabel("probability")
        ax.set_ylabel("CDF")
        # ax.set_yscale("log")
        ax.legend()
    PDFPlotCollection(mngr, figs=[fig]).save_to_file(
        get_debug_output_filepath(mngr, "bgmmodel", f"{data_desc}_prob_cdf.pdf")
    )


def evaluate_mixture_model_fit(
    mngr: Manager,
    bgmmodel: BayesianGaussianMixtureModel,
    input_data,
    quality_of_fit,
    chain_results,
    data_desc,
    chain_desc=None,
):
    posterior_samples = bgmmodel.get_joint_model_dist().sample(len(input_data)).numpy()
    dist_overlaps = []

    # plot outlier score histograms
    fig, ax = plt.subplots(nrows=1, figsize=(8, 8), ncols=1)
    for i_chain in range(bgmmodel.hmc_n_parallel_chains):
        outlier_scores = bgmmodel.get_outlier_score(input_data, i_chain)
        ax.hist(outlier_scores, bins="auto", histtype="step", label=f"chain {i_chain}")

    ax.hist(
        bgmmodel.get_multi_chain_outlier_score(input_data),
        bins="auto",
        histtype="step",
        label="multi_chain",
        linestyle=":",
    )
    ax.grid(**gridlines_config)
    ax.tick_params(**tick_params)
    ax.set_xlabel("ts: negative log prob")
    ax.set_ylabel("Counts")
    ax.set_yscale("log")
    ax.legend()
    PDFPlotCollection(mngr, figs=[fig]).save_to_file(
        get_debug_output_filepath(mngr, "bgmmodel", f"{data_desc}_ts_dist.pdf")
    )

    # plot cumulative distributions of data ts and model posterior sample ts
    plot_comparison_ecdf_datats_posteriorsamplets(
        mngr, bgmmodel, input_data, posterior_samples, data_desc
    )

    plot_prob_comparison_data_posteriorsample_cdf(
        mngr, bgmmodel, input_data, posterior_samples, data_desc
    )

    return np.array(dist_overlaps)


def create_quality_of_fit_plot(
    mngr,
    bgmmodel,
    input_data,
    data_desc,
    quality_of_fit=None,
    chain_desc=None,
):
    mngr.log.info(
        [
            # self.tag,
            [
                "b",
                f"Plotting per dimension projections of {data_desc} and posterior samples",
            ],
        ]
    )

    if quality_of_fit is None:
        quality_of_fit = bgmmodel.get_quality_of_fit_ks_2samp(input_data)

    n_input_dims = input_data.shape[1]
    n_chains = len(quality_of_fit)

    fig, axes = plt.subplots(
        nrows=n_chains,
        ncols=n_input_dims,
        # sharey=True,
        figsize=(5 * n_input_dims, 6 * n_chains),
    )
    fig.suptitle(
        f"max p-value: {np.max(quality_of_fit):.2e} (chain #{1 + np.argmax(quality_of_fit)})",
        fontsize=24,
    )

    plt.tight_layout(pad=3)

    if n_chains == 1:
        axes = [axes]
    for axes_row, n_chain in zip(axes, range(bgmmodel.hmc_n_parallel_chains)):
        posterior_samples = bgmmodel.get_posterior_dist(n_chain).sample(
            len(input_data)
            # min(50000, len(input_data))
        )

        plot_per_dimension_input_and_posterior_histograms(
            input_samples=input_data,
            posterior_samples=posterior_samples,
            axes=axes_row,
        )

        if chain_desc:
            axes_row[int(0.5 * len(axes_row))].set_title(chain_desc[n_chain])
        axes_row[0].legend()

    filename = get_debug_output_filepath(
        mngr, "bgmmodel", f"{data_desc}_vs_posterior_projections.pdf"
    )

    PDFPlotCollection(mngr, [fig]).save_to_file(filename)
