import numpy as np

from trans_finder.flares.mcmc_fit.evaluate_fit import calculate_dim_overlap_scores
from trans_finder.flares.model_utils.bayesian_gaussian_mixture_model import (
    BayesianGaussianMixtureModel,
)


def remove_component(mix_probs, indices):
    """set component prob at every selected index to 0.0 and renormalize"""
    mask = np.ones_like(mix_probs)
    mask[indices] = 0
    new_mix_probs = mix_probs * mask
    new_mix_probs = new_mix_probs / np.sum(new_mix_probs)
    return new_mix_probs


def prune_low_contribution_components(
    mngr, bgmm: BayesianGaussianMixtureModel, dataset, nominative_performance=None
):
    """gradually set mixture components with lowest mixture probability to zero and check if it increases
    the quality of fit. Sets mixture model state with best quality of fit, which may be the original.
    """
    mngr.log.info(
        "Dynamic pruning: Testing if removal of low-probability components improves fit..."
    )

    mix_probs, loc, precision = bgmm._state
    n_chains = bgmm.hmc_n_parallel_chains
    is_improving = [True] * n_chains
    ith_component = -1
    component_indexes = [np.argsort(mix_probs[i_chain]) for i_chain in range(n_chains)]

    last_chain_performance = np.mean(
        (
            calculate_dim_overlap_scores(bgmm, dataset)
            if nominative_performance is None
            else nominative_performance
        ),
        axis=1,
    )

    best_mix_probs = mix_probs.copy()

    while any(is_improving):
        ith_component += 1
        altered_probs = mix_probs.copy()

        # prune at max all but one component
        if ith_component == bgmm.gmm_n_components - 1:
            break

        # set n smallest component of each chain to zero
        for i_chain in range(n_chains):
            if not is_improving[i_chain]:
                continue

            component_indices = component_indexes[i_chain][: ith_component + 1]

            altered_probs[i_chain] = remove_component(
                mix_probs[i_chain], component_indices
            )

        # evaluate fit for all parallel chains
        trial_state = (altered_probs, loc, precision)
        bgmm.update_state(trial_state)

        candidates_performance = np.mean(
            calculate_dim_overlap_scores(bgmm, dataset), axis=1
        )

        # for each chain, check if the pruning improved performance
        for i_chain in range(n_chains):
            if not is_improving[i_chain]:
                continue

            last_chain_performance[i_chain] = candidates_performance[i_chain]

            is_improving[i_chain] = (
                candidates_performance[i_chain] > last_chain_performance[i_chain]
            )

            if is_improving[i_chain]:
                mngr.log.info(
                    f"Chain #{i_chain} improved by pruning the n={ith_component} smallest components (from {last_chain_performance[i_chain]:.6} to {candidates_performance[i_chain]:.6})."
                )

                best_mix_probs[i_chain] = trial_state[0][i_chain]

    bgmm.update_state((best_mix_probs, loc, precision))
    mngr.log.info(
        f"Dynamic pruning finished. Removed up to {ith_component} components."
    )
