import numpy as np
from tensorflow.keras import backend as K
from tensorflow.keras.callbacks import Callback

from trans_finder.flares.core.tags import TAG_AUTO_ENC_INPUT_BEFORENORM


class AutoencInputNormCallback(Callback):
    def __init__(self, input_data, max_rows=20000):
        super().__init__()
        self.input_data = input_data
        self.max_rows = max_rows

    def on_epoch_end(self, epoch, logs=None):
        # Prepare the data for prediction
        # Take only the maximum of 15k rows or less if less are available
        sample_input_data = self.input_data.take(
            min(self.max_rows, len(self.input_data))
        )

        # Predict the model output
        y_pred = self.model.predict(sample_input_data)

        # Get the desired output
        desired_output = y_pred[TAG_AUTO_ENC_INPUT_BEFORENORM]

        # Calculate the quantiles, one for each feature (combining all time steps)
        q_low, q_high = np.percentile(desired_output, [5, 95], axis=(0, 1))

        # Calculate the difference between the 95th and 5th quantile.
        # This will give us a measure of the dispersion of the output values.
        diff = q_high - q_low

        # Now, we want to replicate the difference across all time steps.
        # The np.tile function can be used to repeat the diff array along a new axis.
        # The resulting array will have a shape of (time_steps, features), with identical values across all time steps.
        diff_replicated = np.tile(diff, (desired_output.shape[1], 1))

        # Update the non-trainable weight in the model
        for weight in self.model.weights:
            if weight.name == "autoenc_input_norm:0":
                # print(f"Before update, autoenc_input_norm: {K.get_value(weight)}")
                K.set_value(weight, diff_replicated)
                # print(f"After update, autoenc_input_norm: {K.get_value(weight)}")
