from __future__ import annotations

import functools
from typing import List

import numpy as np
import scipy

# os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
import tensorflow as tf
import tensorflow_probability as tfp
from sklearn.cluster import KMeans
from tensorflow_probability import bijectors as tfb
from tensorflow_probability import distributions as tfd

from trans_finder.flares.checks.generic_checks import calculate_dist_overlap

# physical_devices = tf.config.experimental.list_physical_devices("GPU")
# if len(physical_devices) > 0:
#     tf.config.experimental.set_memory_growth(physical_devices[0], True)
# tf.get_logger().setLevel("INFO")


def get_mvn_chol_precision_tril_dist(loc, chol_precision_tril):
    return tfp.distributions.TransformedDistribution(
        distribution=tfp.distributions.Independent(
            tfp.distributions.Normal(tf.zeros_like(loc), scale=tf.ones_like(loc)),
            reinterpreted_batch_ndims=1,
        ),
        bijector=tfp.bijectors.Chain(
            [
                tfp.bijectors.Shift(shift=loc),
                tfp.bijectors.Invert(
                    tfp.bijectors.ScaleMatvecTriL(
                        scale_tril=chol_precision_tril, adjoint=True
                    )
                ),
            ]
        ),
    )


class BayesianGaussianMixtureModel:
    dtype = np.float32

    unconstraining_bijectors = [
        tfb.SoftmaxCentered(),
        # tfb.Identity(),
        tfb.Identity(),
        tfb.Chain(
            [
                tfb.TransformDiagonal(tfb.Softplus()),
                tfb.FillTriangular(),
            ]
        ),
    ]

    def __init__(
        self,
        n_input_dims: int,
        n_input_length: int,
        hmc_n_steps: int,
        hmc_n_parallel_chains: int,
        hmc_n_leapfrog_integrator_steps: int,
        hmc_step_size: float,
        hmc_max_parallel_iterations: int,
        hmc_n_burnin_steps: int,
        hmc_n_step_size_adaptation_steps: int,
        gmm_n_components: int,
        gmm_wishard_degrees_of_freedom: int,
        gmm_prior_active_clusters_ratio: float = 1.0,
        gmm_prior_component_loc_mean: float = 0.0,
        gmm_prior_component_loc_std: float = 0.5,
    ):
        if gmm_wishard_degrees_of_freedom < n_input_dims:
            raise ValueError("gmm_wishard_degrees_of_freedom must be >= n_input_dims")

        self.n_input_dims = n_input_dims
        self.gmm_n_components = gmm_n_components
        self.gmm_wishard_degrees_of_freedom = gmm_wishard_degrees_of_freedom
        self.gmm_prior_component_loc_mean = gmm_prior_component_loc_mean
        self.gmm_prior_component_loc_std = gmm_prior_component_loc_std
        self.gmm_prior_active_clusters_ratio = gmm_prior_active_clusters_ratio

        self.hmc_n_parallel_chains = hmc_n_parallel_chains
        self.hmc_n_steps = hmc_n_steps
        self.hmc_n_burnin_steps = hmc_n_burnin_steps
        self.hmc_n_leapfrog_integrator_steps = hmc_n_leapfrog_integrator_steps
        self.hmc_n_step_size_adaptation_steps = hmc_n_step_size_adaptation_steps
        self.hmc_step_size = hmc_step_size
        self.hmc_max_parallel_iterations = hmc_max_parallel_iterations

        self._state = None

        self.internal_dist = tfd.JointDistributionNamed(
            dict(
                mix_probs=tfd.Dirichlet(
                    concentration=np.ones(self.gmm_n_components, self.dtype),
                ),
                loc=tfd.Independent(
                    tfd.Normal(
                        loc=np.ones(
                            shape=(self.gmm_n_components, n_input_dims),
                            dtype=self.dtype,
                        ),
                        scale=np.ones(
                            shape=(self.gmm_n_components, n_input_dims),
                            dtype=self.dtype,
                        ),
                    ),
                    reinterpreted_batch_ndims=2,
                ),
                precision=tfd.Independent(
                    tfd.WishartTriL(
                        df=self.gmm_wishard_degrees_of_freedom,
                        scale_tril=np.stack(
                            [np.eye(n_input_dims, dtype=self.dtype)]
                            * self.gmm_n_components
                        ),
                        input_output_cholesky=True,
                    ),
                    reinterpreted_batch_ndims=1,
                ),
                s=lambda mix_probs, loc, precision: tfd.Sample(
                    tfd.MixtureSameFamily(
                        mixture_distribution=tfd.Categorical(probs=mix_probs),
                        components_distribution=get_mvn_chol_precision_tril_dist(
                            loc=loc, chol_precision_tril=precision
                        ),
                    ),
                    sample_shape=n_input_length,
                ),
            )
        )

        self._posterior_prob_normalisation_factors = None

    def compare_input_and_posterior_sample_dist_overlap(self, mngr, input_data):
        posterior_samples = self.get_joint_model_dist().sample(len(input_data)).numpy()

        mean_dim_dist_overlaps = [
            np.mean(
                calculate_dist_overlap(input_data, posterior_samples[:, i_chain, :])
            )
            for i_chain in range(self.hmc_n_parallel_chains)
        ]

        fit_error_string = np.array2string(
            np.array(mean_dim_dist_overlaps),
            formatter={"float_kind": lambda x: "%.2f" % x},
        )
        mngr.log.info(
            [
                mngr.tag,
                [
                    "b",
                    f"avg input/output dist overlap = {np.mean(mean_dim_dist_overlaps):.2%} | per chain: {fit_error_string}",
                ],
            ]
        )

        return mean_dim_dist_overlaps

    def create_initial_state(
        self,
        initial_cluster_centroids=None,
        initial_cluster_scales=None,
    ):
        initial_locs = []
        initial_mix_probs = []
        initial_scales = []
        for i in range(self.hmc_n_parallel_chains):
            # we allow clusters to be active or inactive initially. The initial probability of inactive clusters is
            # set to a small nonzero number, to allow the model to sample from them also initially
            active_clusters_prob_sum = 0.95
            inactive_clusters_prob_sum = 1 - active_clusters_prob_sum

            n_active_clusters = int(
                np.round(self.gmm_prior_active_clusters_ratio * self.gmm_n_components)
            )
            n_inactive_clusters = int(self.gmm_n_components - n_active_clusters)

            initial_probs = []

            if n_active_clusters > 0:
                initial_probs.append(
                    np.repeat(
                        active_clusters_prob_sum / n_active_clusters,
                        repeats=n_active_clusters,
                    )
                )
            if n_inactive_clusters > 0:
                initial_probs.append(
                    np.repeat(
                        inactive_clusters_prob_sum / n_inactive_clusters,
                        repeats=n_inactive_clusters,
                    )
                )

            # normalise the mix probs distribution, which is required by the model
            initial_probs = np.concatenate(initial_probs)
            initial_probs /= np.sum(initial_probs)

            # print(self.gmm_prior_active_clusters_ratio, initial_probs)

            initial_mix_probs.append(
                # tf.fill(
                #     [self.gmm_n_components],
                #     value=np.array(1.0 / self.gmm_n_components, self.dtype),
                #     name="mix_probs",
                # )
                tf.constant(
                    np.array(
                        initial_probs,
                        dtype=self.dtype,
                    ),
                    name="mix_probs",
                )
            )

            initial_locs.append(
                tf.constant(
                    (
                        initial_cluster_centroids.astype(self.dtype)
                        if initial_cluster_centroids is not None
                        else np.array(
                            # np.zeros((self.gmm_n_components, self.n_input_dims)),
                            np.random.normal(
                                self.gmm_prior_component_loc_mean,
                                self.gmm_prior_component_loc_std,
                                size=[
                                    self.gmm_n_components,
                                    self.n_input_dims,
                                ],
                            ),
                            dtype=self.dtype,
                        )
                    ),
                    name="loc",
                )
            )

            initial_scales.append(
                tf.cast(initial_cluster_scales, self.dtype, name="chol_precision")
                if initial_cluster_scales is not None
                else tf.linalg.eye(
                    self.n_input_dims,
                    batch_shape=[self.gmm_n_components],
                    dtype=self.dtype,
                    name="chol_precision",
                )
            )

        return [
            tf.stack(value_list)
            for value_list in [initial_mix_probs, initial_locs, initial_scales]
        ]

    def get_joint_model_dist(self):
        mix_probs, loc, precision = self._state
        return self.get_joint_model_dist_from_state(mix_probs, loc, precision)

    @staticmethod
    def get_joint_model_dist_from_state(mix_probs, loc, precision):
        return tfd.MixtureSameFamily(
            mixture_distribution=tfd.Categorical(probs=mix_probs),
            components_distribution=get_mvn_chol_precision_tril_dist(
                loc=loc, chol_precision_tril=precision
            ),
        )

    def _hmc_joint_log_prob(self, observations, mix_probs, loc, chol_precision):
        return self.internal_dist.log_prob(
            mix_probs=mix_probs, loc=loc, precision=chol_precision, s=observations
        )

    def update_state_from_chain_results(self, chain_results):
        [mix_probs, loc, chol_precision], is_accepted = chain_results

        mean_mix_probs = tf.reduce_mean(mix_probs, axis=0).numpy()
        mean_loc = tf.reduce_mean(loc, axis=0).numpy()
        mean_chol_precision = tf.reduce_mean(chol_precision, axis=0).numpy()

        self.update_state([mean_mix_probs, mean_loc, mean_chol_precision])

    def update_state(self, state):
        self._state = state

    def fit(self, data):
        self.adapt(data)

        unnormalized_posterior_log_prob = functools.partial(
            self._hmc_joint_log_prob, data
        )

        @tf.function(autograph=False)
        def wrapped_sample_chain():
            hmc_kernel = tfp.mcmc.SimpleStepSizeAdaptation(
                tfp.mcmc.TransformedTransitionKernel(
                    inner_kernel=tfp.mcmc.HamiltonianMonteCarlo(
                        target_log_prob_fn=unnormalized_posterior_log_prob,
                        step_size=self.hmc_step_size,
                        num_leapfrog_steps=self.hmc_n_leapfrog_integrator_steps,
                    ),
                    bijector=self.unconstraining_bijectors,
                ),
                num_adaptation_steps=self.hmc_n_step_size_adaptation_steps,
            )

            return tfp.mcmc.sample_chain(
                num_results=self.hmc_n_steps,
                num_burnin_steps=self.hmc_n_burnin_steps,
                current_state=self._state,
                kernel=hmc_kernel,
                trace_fn=lambda _, pkr: pkr.inner_results.inner_results.is_accepted,
                parallel_iterations=self.hmc_max_parallel_iterations,
            )

        chain_results = wrapped_sample_chain()

        [mix_probs, loc, chol_precision], is_accepted = chain_results
        acceptance_rate = tf.reduce_mean(tf.cast(is_accepted, dtype=tf.float32)).numpy()
        print(
            "acceptance_rate:", acceptance_rate, " with step size ", self.hmc_step_size
        )
        # mean_mix_probs = tf.reduce_mean(mix_probs, axis=0).numpy()
        # print("avg mix probs:", mean_mix_probs)

        self.update_state_from_chain_results(chain_results)

        return chain_results

    def compute_sample_stats(self, x):
        sample_mean = tf.reduce_mean(x, axis=0, keepdims=True)
        s = x - sample_mean
        sample_cov = tf.linalg.matmul(s, s, adjoint_a=True) / tf.cast(
            x.shape[0], s.dtype
        )
        sample_scale = tf.linalg.cholesky(sample_cov)
        sample_mean = sample_mean[0]
        return [
            sample_mean,
            sample_cov,
            sample_scale,
        ]

    def adapt(self, input_data):
        """setting informed priors as model state based on input data"""
        if self._state is None:
            kmeans = KMeans(n_clusters=self.gmm_n_components)
            kmeans.fit(input_data)
            initial_cluster_centroids = kmeans.cluster_centers_

            initial_cluster_scales = tf.repeat(
                tf.expand_dims(self.compute_sample_stats(input_data)[2], 0),
                self.gmm_n_components,
                axis=0,
            )
            if np.isnan(initial_cluster_scales).any():
                print(
                    "nan in sample cluster scales, will use uniform prior for scales "
                )
                initial_cluster_scales = None

            self._state = self.create_initial_state(
                initial_cluster_centroids=initial_cluster_centroids,
                initial_cluster_scales=initial_cluster_scales,
            )

    def get_single_chain_state(self, n_chain):
        return [d[n_chain] for d in self._state]

    def get_grid_points(
        self,
        grid_range,
        grid_resolution,
    ):
        coordinate_arrays = [
            np.arange(grid_range[0], grid_range[1], grid_resolution)
        ] * self.n_input_dims

        grid_coords = np.meshgrid(*coordinate_arrays)
        coordinates = np.array([cords.ravel() for cords in grid_coords]).T

        return coordinates

    def get_posterior_prob_normalisation_factors(self, grid_range, grid_resolution):
        if self._posterior_prob_normalisation_factors is not None:
            return self._posterior_prob_normalisation_factors
        else:
            coordinates = self.get_grid_points(
                grid_range=grid_range,
                grid_resolution=grid_resolution,
            )
            dx = grid_resolution**self.n_input_dims

            posterior_prob_normalisation_factors = []
            for n_chain in range(self.hmc_n_parallel_chains):
                chain_state = self.get_single_chain_state(n_chain)
                unnormalized_posterior = self.get_joint_model_dist_from_state(
                    *chain_state
                )

                integral = np.sum(unnormalized_posterior.prob(coordinates) * dx)

                posterior_prob_normalisation_factors.append(1 / integral)

            self._posterior_prob_normalisation_factors = np.stack(
                posterior_prob_normalisation_factors
            )
            return self._posterior_prob_normalisation_factors

    def get_posterior_dist(self, n_chain):
        chain_state = self.get_single_chain_state(n_chain)
        return self.get_joint_model_dist_from_state(*chain_state)

    def get_normalized_probs(self, points, grid_range, grid_resolution):
        posterior_prob_normalisation_factors = (
            self.get_posterior_prob_normalisation_factors(
                grid_range=grid_range,
                grid_resolution=grid_resolution,
            )
        )

        normalized_probs = []
        for n_chain in range(self.hmc_n_parallel_chains):
            unnormalized_posterior = self.get_posterior_dist(n_chain)
            normalized_prob = (
                unnormalized_posterior.prob(points)
                * posterior_prob_normalisation_factors[n_chain]
            )
            normalized_probs.append(normalized_prob)

        return np.stack(normalized_probs)

    def get_outlier_score(self, points: np.ndarray, n_chain=None):
        posterior = self.get_posterior_dist(n_chain)
        return -1 * posterior.log_prob(points).numpy()

    def get_multi_chain_outlier_score(self, points: np.ndarray):
        scores = [
            self.get_outlier_score(points, n_chain)
            for n_chain in range(self.hmc_n_parallel_chains)
        ]

        return np.mean(scores, axis=0)

    def get_quality_of_fit_ks_2samp(self, input_data):
        posterior_samples = self.get_joint_model_dist().sample(len(input_data)).numpy()

        # bins = np.linspace(-1, 1, num=200)

        # observation_hists = np.apply_along_axis(
        #     lambda a: np.histogram(a, bins=bins)[0], 0, input_data
        # )
        # observation_hists = observation_hists / observation_hists.sum(axis=0)
        n_chains = posterior_samples.shape[1]
        per_chain_goodness_of_fit = np.zeros(n_chains)
        for n_chain in range(n_chains):
            # chain_hists = np.apply_along_axis(
            #     lambda a: np.histogram(a, bins=bins)[0],
            #     0,
            #     posterior_samples[:, n_chain, :],
            # )
            # chain_hists = chain_hists / chain_hists.sum(axis=0)

            per_dim_goodness_of_fit = []

            for n_dim in range(posterior_samples.shape[-1]):
                result = scipy.stats.ks_2samp(
                    input_data[..., n_dim],
                    posterior_samples[..., n_chain, n_dim],
                    alternative="two-sided",
                )
                per_dim_goodness_of_fit.append(result.pvalue)

            per_chain_goodness_of_fit[n_chain] = np.max(
                np.array(per_dim_goodness_of_fit)
            )

        return per_chain_goodness_of_fit

    def calculate_quality_of_fit(self, data, posterior_samples=None) -> List[float]:
        """by comparing the empirical distributions of the input data probability vs
        posterior sample probability"""

        # get posterior samples
        if posterior_samples is None:
            posterior_samples = self.get_joint_model_dist().sample(len(data)).numpy()
        assert self.hmc_n_parallel_chains == posterior_samples.shape[1]

        # to ks test between ts of data and ts of posterior samples
        quality_of_fit = []
        for i_chain in range(self.hmc_n_parallel_chains):
            prob_data = self.get_posterior_dist(i_chain).prob(data).numpy()
            prob_posterior = (
                self.get_posterior_dist(i_chain)
                .prob(posterior_samples[:, i_chain])
                .numpy()
            )

            test = scipy.stats.ks_2samp(
                prob_data,
                prob_posterior,
                alternative="two-sided",
            )
            quality_of_fit.append(test.pvalue)

        return quality_of_fit

    def get_quality_of_fit_metric(self, input_data, grid_range, grid_resolution):
        # posterior_prob_normalisation_factors = (
        #     self.get_posterior_prob_normalisation_factors(
        #         grid_range=grid_range,
        #         grid_resolution=grid_resolution,
        #     )
        # )

        normalized_data_probs = []
        normalized_grid_probs = []

        grid_coordinates = self.get_grid_points(
            grid_range=grid_range,
            grid_resolution=grid_resolution,
        )

        quality_of_fit_by_chain = []

        for n_chain in range(self.hmc_n_parallel_chains):
            unnormalized_posterior = self.get_posterior_dist(n_chain)
            normalized_prob = (
                unnormalized_posterior.prob(input_data)
                # * posterior_prob_normalisation_factors[n_chain]
            )
            # normalized_data_probs.append(normalized_prob)

            normalized_grid_prob = (
                unnormalized_posterior.prob(grid_coordinates)
                # * posterior_prob_normalisation_factors[n_chain]
            )

            mean_data_prob = np.mean(normalized_prob)
            mean_grid_prob = np.mean(normalized_grid_prob)

            quality_of_fit = mean_data_prob / mean_grid_prob

            quality_of_fit_by_chain.append(quality_of_fit)

            # normalized_grid_probs.append(normalized_grid_prob)

        return np.array(quality_of_fit_by_chain)

    def get_effective_sample_size(self, chain_results):
        ess = tfp.mcmc.effective_sample_size(chain_results[0], cross_chain_dims=[1] * 3)

        result = {
            model_param_name: ess_tensor.numpy()
            for model_param_name, ess_tensor in zip(
                ["mix_probs", "loc", "chol_precision"], ess
            )
        }

        return result

    def get_rhat_score(self, chain_results):
        rhat = tfp.mcmc.diagnostic.potential_scale_reduction(
            chain_results[0], independent_chain_ndims=1
        )

        result = {
            model_param_name: rhat_tensor.numpy()
            for model_param_name, rhat_tensor in zip(
                ["mix_probs", "loc", "chol_precision"], rhat
            )
        }

        return result


# chain_results = bgmmodel.run_mcmc(observations)
#
# [mix_probs, loc, chol_precision], is_accepted = chain_results

# trace_plot(loc, parameter_name="loc", num_burnin_steps=hmc_num_burnin_steps)
# trace_plot(
#     chol_precision,
#     parameter_name="chol_precision",
#     num_burnin_steps=hmc_num_burnin_steps,
# )

# %%

# acceptance_rate = tf.reduce_mean(tf.cast(is_accepted, dtype=tf.float32)).numpy()
# mean_mix_probs = tf.reduce_mean(mix_probs, axis=0).numpy()
# mean_loc = tf.reduce_mean(loc, axis=0).numpy()
# mean_chol_precision = tf.reduce_mean(chol_precision, axis=0).numpy()
# precision = tf.linalg.matmul(chol_precision, chol_precision, transpose_b=True)
#
# print("acceptance_rate:", acceptance_rate)
# print("avg mix probs:", mean_mix_probs)
# print("avg loc:\n", mean_loc)
# print("avg chol(precision):\n", mean_chol_precision)
