from __future__ import annotations

from typing import TYPE_CHECKING

import tensorflow as tf
from keras import backend as K
from tensorflow import keras

from trans_finder.flares.core.tags import *
from trans_finder.flares.model_utils.forecaster import Forecaster
from trans_finder.flares.model_utils.lstm_vae_model import (
    NegativeLogProbability,
    build_decoder,
    build_encoder,
    fn_kl_loss,
)
from trans_finder.flares.model_utils.temporal_offsets import (
    AddInputOffsetsBack,
    RemoveTemporalOffsetsLayer,
)
from trans_finder.flares.model_utils.utils import fn_rmse_batch, fn_rmse_samples

if TYPE_CHECKING:
    from trans_finder.flares.manager import Manager


def build_model_forecast_combined_vae(mngr: Manager):
    vae = ForecastCombinedVAE(
        mngr.autoenc_data_shape,
        beta=mngr.model_cluster_space_regularisation_factor,
        n_features=mngr.n_features,
        temporal_offset_column_indices=mngr.temporal_offset_column_indices,
        latent_space_dim=mngr.model_learned_representation_n_dim,
        forecast_enc_sizes=mngr.lstm_enc_sizes,
        forecast_dec_sizes=mngr.lstm_dec_sizes,
        autoenc_enc_sizes=mngr.autoenc_enc_sizes,
        autoenc_dec_sizes=mngr.autoenc_dec_sizes,
        n_steps_dec=mngr.n_steps_dec,
        use_tf_prob=mngr.use_tf_prob,
        train_phase=mngr.train_phase,
        regularizer_l2_scale=mngr.regularizer_l2_scale,
        dropout=mngr.rnn_dropout,
    )

    losses = None
    return vae, losses


def fn_forecasting_loss(forecast, d1):
    return tf.reduce_mean(keras.losses.mean_squared_error(forecast, d1))


def fn_coeff_determination(y_true, y_pred):
    SS_res = K.sum(K.square(y_true - y_pred))
    SS_tot = K.sum(K.square(y_true - K.mean(y_true)))
    return 1 - SS_res / (SS_tot + K.epsilon())


class ForecastCombinedVAE(keras.Model):
    def __init__(
        self,
        data_shape,
        beta,
        n_features,
        temporal_offset_column_indices,
        autoenc_enc_sizes,
        autoenc_dec_sizes,
        forecast_enc_sizes,
        forecast_dec_sizes,
        latent_space_dim,
        n_steps_dec,
        use_tf_prob,
        train_phase,
        regularizer_l2_scale,
        dropout,
    ):
        super(ForecastCombinedVAE, self).__init__()

        self.data_shape = data_shape
        self.beta = beta
        self.n_features = n_features
        self.temporal_offset_column_indices = temporal_offset_column_indices
        self.autoenc_enc_sizes = autoenc_enc_sizes
        self.autoenc_dec_sizes = autoenc_dec_sizes
        self.forecast_enc_sizes = forecast_enc_sizes
        self.forecast_dec_sizes = forecast_dec_sizes
        self.latent_space_dim = latent_space_dim
        self.n_steps_dec = n_steps_dec
        self.use_tf_prob = use_tf_prob
        self.train_phase = train_phase
        self.regularizer_l2_scale = regularizer_l2_scale
        self.dropout = dropout

        self.remove_temporal_offsets_layer = RemoveTemporalOffsetsLayer(
            n_features, temporal_offset_column_indices
        )
        self.add_temporal_offsets_back_layer = AddInputOffsetsBack(
            n_features, temporal_offset_column_indices
        )

        # the normalization to divide the autoencoder input by
        self.autoenc_input_norm = self.add_weight(
            name="autoenc_input_norm",
            shape=data_shape,
            initializer="ones",
            trainable=False,
        )

        self.encoder = build_encoder(
            input_shape=data_shape,
            sizes=autoenc_enc_sizes,
            latent_dim=latent_space_dim,
            beta=beta,
            regularizer_l2_scale=regularizer_l2_scale,
            dropout=dropout,
        )
        self.decoder = build_decoder(
            latent_dim=latent_space_dim,
            output_shape=data_shape,
            sizes=autoenc_dec_sizes,
            regularizer_l2_scale=regularizer_l2_scale,
            dropout=dropout,
        )

        # forecaster = build_background_prediction_model_baseline_lastencoderstep(mngr)
        self.forecaster = Forecaster(
            n_features,
            temporal_offset_column_indices,
            forecast_enc_sizes,
            forecast_dec_sizes,
            n_steps_dec=n_steps_dec,
            use_tf_prob=use_tf_prob,
            regularizer_l2_scale=regularizer_l2_scale,
            dropout=dropout,
        )  # build_background_prediction_model(mngr)

        # self.forecaster = MultiStepLastBaseline(
        #     n_features, n_steps_dec, temporal_offset_column_indices
        # )

        self.norm_ae_input = keras.layers.Normalization(axis=None)

    def get_config(self):
        return dict(
            data_shape=self.data_shape,
            beta=self.beta,
            n_features=self.n_features,
            temporal_offset_column_indices=self.temporal_offset_column_indices,
            autoenc_enc_sizes=self.autoenc_enc_sizes,
            autoenc_dec_sizes=self.autoenc_dec_sizes,
            forecast_enc_sizes=self.forecast_enc_sizes,
            forecast_dec_sizes=self.forecast_dec_sizes,
            latent_space_dim=self.latent_space_dim,
            n_steps_dec=self.n_steps_dec,
            use_tf_prob=self.use_tf_prob,
            train_phase=self.train_phase,
            regularizer_l2_scale=self.regularizer_l2_scale,
            dropout=self.dropout,
        )

    @classmethod
    def from_config(cls, config):
        return cls(**config)

    def set_ae_input_norm(self, input_dataset):
        if self.norm_ae_input.is_adapted:
            print("Loading stored normalization params for autoencoder inputs...")
            return

        inputs = input_dataset.map(lambda x, y: x)

        def get_ae_input(x):
            d1 = x[TAG_DEC]
            forecast = self.forecaster(x)[TAG_PRED]
            return tf.subtract(d1, forecast)

        ae_input = inputs.map(get_ae_input)
        print("Adapting normalization params of autoencoder inputs...")
        self.norm_ae_input.adapt(ae_input)

    def get_autoencoder_input(self, inputs):
        outputs_forecasting = self.forecaster(inputs)
        forecast = self.remove_temporal_offsets_layer(outputs_forecasting[TAG_PRED])
        forecasting_loss = outputs_forecasting[TAG_LOSS]

        d1 = self.remove_temporal_offsets_layer(inputs[TAG_DEC])
        wgts = self.remove_temporal_offsets_layer(inputs[TAG_WGT])

        x = tf.subtract(d1, forecast)

        autoenc_input_beforenorm = x
        # x = self.remove_temporal_offsets_layer(x)

        # x = tf.divide(x, 1 + tf.minimum(tf.abs(forecast), tf.abs(d1)))
        x = tf.divide(x, self.autoenc_input_norm)

        # offset_weights = GetOffsetWeightsLayer(self.mngr)(auto_enc_encoder_input)

        # Normalization based on forecast:
        # # normalize autoencoder input based on forecast
        # auto_enc_encoder_input = tf.divide(
        #     auto_enc_encoder_input, tf.add(tf.abs(forecast), 1)
        # )

        # Normalization using keras.layers.Normalization
        x = self.norm_ae_input(x)

        # apply weights based on temporal offsets
        x = tf.multiply(x, wgts)

        autoenc_input = x

        # normalize autoencoder inputs
        # fixed_autoenc_input = self.autoencnorm(fixed_autoenc_input)

        # forecast = self.remove_temporal_offsets_layer(forecast)

        return autoenc_input, autoenc_input_beforenorm, forecast, forecasting_loss

    def call(self, inputs, training=False, **kwargs):
        d0 = inputs[TAG_ENC]
        d1 = inputs[TAG_DEC]
        wgts = inputs[TAG_WGT]
        input_ids = inputs[TAG_MODEL_INPUT_IDS]

        (
            autoenc_input,
            autoenc_input_beforenorm,
            forecast,
            forecasting_loss,
        ) = self.get_autoencoder_input(inputs)

        z_mean, z_log_var, z_sample = self.encoder(autoenc_input)

        z_std = tf.exp(0.5 * z_log_var)

        if training:
            z = z_sample
        else:
            z = z_mean

        x_mean, x_log_var = self.decoder(z)

        reconstruction_loss = NegativeLogProbability()(
            [x_mean, x_log_var, autoenc_input]
        )
        reco_error = tf.reduce_sum(reconstruction_loss, axis=1)

        # mean across batch
        reconstruction_loss = tf.reduce_mean(reconstruction_loss)

        reconstruction = x_mean

        forecasting_error = fn_rmse_samples(
            self.remove_temporal_offsets_layer(d1), forecast
        )

        # forecasting_loss = fn_forecasting_loss(forecast, d1)
        self.add_metric(forecasting_loss, "forecasting_loss")
        # self.add_metric(fn_rmse_batch(d1, forecast), "forecasting_rmse")
        self.add_metric(
            fn_coeff_determination(self.remove_temporal_offsets_layer(d1), forecast),
            "forecasting_r2",
        )

        self.add_metric(
            fn_coeff_determination(autoenc_input, reconstruction),
            "reconstruction_r2",
        )
        self.add_metric(
            reconstruction_loss,
            "reconstruction_loss",
        )
        kl_loss = fn_kl_loss(z_mean, z_log_var, self.beta)
        self.add_metric(kl_loss, "kl_loss")
        self.add_metric(tf.reduce_mean(tf.exp(0.5 * z_log_var)), "cluster_space_std")

        if self.train_phase == 0:
            self.add_loss(tf.reduce_mean(forecasting_loss))
        elif self.train_phase == 1:
            self.add_loss(reconstruction_loss)
            self.add_loss(kl_loss)

        # add temporal offsets back to the reconstruction, for plotting convenience
        # fixed_reconstruction = AddInputOffsetsBack(self.mngr)([reconstruction, d1])

        model_output = {
            # first part of the time series of the input data
            TAG_INPUT_D0: d0,
            # second part of the time series of the input data (potentially contains flare)
            TAG_INPUT_D1: d1,
            # the forecast of the background in the second part of the input
            TAG_PRED: self.add_temporal_offsets_back_layer([forecast, d1]),
            TAG_BCK_FORECAST_D1_NOOFFSETS: forecast,
            # # the input to the autoencoder
            TAG_AUTO_ENC_INPUT: autoenc_input,
            TAG_AUTO_ENC_INPUT_BEFORENORM: autoenc_input_beforenorm,
            # # hidden representation learned by the autoencoder given model_input_d1 and model_bck_forecast_d1
            TAG_CLUSTER_SPACE_SAMPLE: z_sample,
            TAG_CLUSTER_SPACE_MEAN: z_mean,
            TAG_CLUSTER_SPACE_STD: z_std,
            # # output of the autoencoder, trained to reconstruct auto_enc_input
            TAG_AUTO_ENC_RECONSTRUCTION: reconstruction,
            TAG_AUTO_ENC_RECONSTRUCTION_ERROR: reco_error,
            # mean absolute percentage error of forecast
            TAG_MODEL_FORECASTING_ERROR: forecasting_error,
            # # log probability of the real data (second part of the time series)
            # # given the hidden representation of the autoencoder
            TAG_MODEL_INPUT_IDS: input_ids,
        }
        return model_output


class MultiStepLastBaseline(tf.keras.Model):
    def __init__(
        self, n_features, n_steps_dec, temporal_offset_column_indices, **kwargs
    ):
        super().__init__(**kwargs)
        self.n_steps_dec = n_steps_dec
        self.temporal_offset_column_indices = temporal_offset_column_indices
        self.n_features = n_features

    def call(self, inputs, **kwargs):
        d0 = inputs[TAG_ENC]
        d1 = inputs[TAG_DEC]

        X = RemoveTemporalOffsetsLayer(
            self.n_features, self.temporal_offset_column_indices
        )(d0)

        X = tf.tile(tf.reduce_mean(d0, axis=1, keepdims=True), [1, self.n_steps_dec, 1])
        X = AddInputOffsetsBack(self.n_features, self.temporal_offset_column_indices)(
            [X, d1]
        )

        loss = fn_forecasting_loss(X, d1)

        return {
            TAG_LOSS: loss,
            TAG_PRED: X,
            "forecasting_rmse": fn_rmse_batch(d1, X),
        }
