from pprint import pprint

import tensorflow as tf
import tensorflow_probability as tfp
from tensorflow import keras

from trans_finder.flares.core.tags import *
from trans_finder.flares.model_utils.temporal_offsets import (
    AddInputOffsetsBack,
    RemoveTemporalOffsetsLayer,
)


class Forecaster(keras.layers.Layer):
    def __init__(
        self,
        n_features,
        temporal_offset_column_indices,
        enc_sizes,
        dec_sizes,
        n_steps_dec,
        use_tf_prob,
        regularizer_l2_scale,
        dropout,
    ):
        super(Forecaster, self).__init__()
        self.n_features = n_features
        self.temporal_offset_column_indices = temporal_offset_column_indices
        self.enc_sizes = enc_sizes
        self.dec_sizes = dec_sizes
        self.n_steps_dec = n_steps_dec
        self.use_tf_prob = use_tf_prob

        self.remove_temporal_offsets_layer = RemoveTemporalOffsetsLayer(
            n_features, temporal_offset_column_indices
        )
        self.add_temporal_offsets_back_layer = AddInputOffsetsBack(
            n_features, temporal_offset_column_indices
        )

        self.encoder_layers = [
            keras.layers.LSTM(
                enc_sizes[n_size],
                return_state=True,
                return_sequences=n_size < len(enc_sizes) - 1,
                dropout=dropout,
                recurrent_dropout=dropout,
                kernel_regularizer=keras.regularizers.L2(regularizer_l2_scale),
                name=f"forecaster_lstm_enc_{n_size}",
            )
            for n_size in range(len(enc_sizes))
        ]

        self.repeat_vector = keras.layers.RepeatVector(n_steps_dec)

        self.decoder_layers = [
            keras.layers.LSTM(
                dec_sizes[n_size],
                return_state=False,
                return_sequences=True,
                dropout=dropout,
                recurrent_dropout=dropout,
                kernel_regularizer=keras.regularizers.L2(regularizer_l2_scale),
                name=f"forecaster_lstm_dec_{n_size}",
            )
            for n_size in range(len(dec_sizes))
        ]

        out_dim = n_features - len(temporal_offset_column_indices)

        if use_tf_prob:
            # number of parameters for a normal distribution serves to multiply
            # the inputs, so that we get a layer with trainable weights for each param
            n_prob_param = 2

            self.output_layers = [
                keras.layers.Dense(
                    out_dim * n_prob_param,
                    name="forecaster_prob_layer_dense",
                )
            ]

            # probability layer, with an output which samples the distribution
            def dist_func(t):
                return tfp.distributions.Normal(
                    loc=t[..., :out_dim],
                    scale=tf.exp(t[..., out_dim:]),
                )

            self.prob_layer_sample = tfp.layers.DistributionLambda(
                make_distribution_fn=dist_func,
                convert_to_tensor_fn=tfp.distributions.Distribution.sample,
                name="forecaster_prob_layer_sample",
            )

            self.prob_layer_mean = tfp.layers.DistributionLambda(
                make_distribution_fn=dist_func,
                convert_to_tensor_fn=tfp.distributions.Distribution.mean,
                name="forecaster_prob_layer_mean",
            )

        else:
            self.output_layers = [keras.layers.Dense(out_dim, name="forecaster_dense")]

    def get_config(self):
        return dict(
            n_features=self.n_features,
            temporal_offset_column_indices=self.temporal_offset_column_indices,
            enc_sizes=self.enc_sizes,
            dec_sizes=self.dec_sizes,
            n_steps_dec=self.n_steps_dec,
            use_tf_prob=self.use_tf_prob,
        )

    @classmethod
    def from_config(cls, config):
        return cls(**config)

    def __call__(self, inputs):
        # debugging, todo: remove
        debug_mask = None
        if tf.executing_eagerly():
            np_input_ids = inputs[TAG_MODEL_INPUT_IDS].numpy()
            debug_input_id = 19306.0
            if debug_input_id in np_input_ids:
                debug_mask = np_input_ids == debug_input_id
                print("input seen in model call() method: ")
                pprint({k: v[debug_mask].numpy() for k, v in inputs.items()})
                # print("=" * 20, " input_enc: ")
                # print("input_enc hash", get_hash_string([repr(input_encs[debug_mask])]))
                # print(
                #     TAG_ENC,
                #     repr(input_encs[debug_mask].numpy()),  # [:80],
                #     #                    " [...]",
                # )

        # for residual forecasting:
        # repeat last encoder step values for n_steps_dec
        # decoder_base = tf.tile(input_encs_now[:, -1:, :], [1, mngr.n_steps_dec, 1])

        # x = inputs[TAG_ENC]
        x = self.remove_temporal_offsets_layer(inputs[TAG_ENC])

        for encoder_layer in self.encoder_layers:
            outputs = encoder_layer(
                x,
            )
            # print("enc lstm layer -----")
            # print(outputs)
            whole_seq_output, hidden_state, cell_state = outputs
            x = whole_seq_output

        x = self.repeat_vector(x)

        for decoder_layer in self.decoder_layers:
            x = decoder_layer(
                x,
                # initial_state=initial_state,
            )

        for output_layer in self.output_layers:
            x = output_layer(x)

        # output_with_offsets = x

        # for residual forecasting:
        # output_pred = decoder_base + output_pred

        if debug_mask is not None:
            # print(
            #     "fc_model weights:",
            #     str(self.encoder_layers[0].weights[0].numpy())  # [:80],
            #     # " [...]",
            # )
            # print("fc_model weights checksum:", get_hash_string(repr(self.weights)))
            # print(
            #     "fc_decoder_layers[0] weights:",
            #     get_hash_string(repr(self.decoder_layers[0].weights)),
            # )
            print("=" * 20, " forecast: ")
            # print("forecast hash ", get_hash_string([repr(x[debug_mask])]))
            print(str(x[debug_mask].numpy()))  # [:80], " [...]")

        # probability output
        if self.use_tf_prob:
            sample = self.prob_layer_sample(x)

            # loss tensor, calculated explicitly as part of the model
            truth_nooffsets = self.remove_temporal_offsets_layer(inputs[TAG_DEC])

            loss = -1 * sample.log_prob(truth_nooffsets)
            loss = tf.reduce_sum(loss, axis=[-1, -2])

            mean = self.prob_layer_mean(x)
            mean_with_offsets = self.add_temporal_offsets_back_layer(
                [mean, inputs[TAG_DEC]]
            )

            return {
                TAG_PRED: mean_with_offsets,
                TAG_LOSS: loss,
            }

        else:
            # deterministic output
            output_with_offsets = self.add_temporal_offsets_back_layer(
                [x, inputs[TAG_DEC]]
            )

            # loss tensor, calculated explicitly as part of the model
            loss = tf.square(inputs[TAG_DEC] - output_with_offsets, name="loss_0")
            loss = tf.reduce_mean(keras.layers.Flatten()(loss), axis=1)

            # output_wgts = input_wgts
            # loss = tf.math.multiply(loss, output_wgts, name="loss_1")
            # loss = tf.reduce_sum(loss, axis=-1, name="loss_2")
            #
            # if not mngr.avg_time_steps:
            #     loss = tf.reduce_sum(loss, axis=-1, name="loss_4")

            return {
                TAG_LOSS: loss,
                TAG_PRED: output_with_offsets,
            }
