import numpy as np
import tensorflow as tf
from keras import Input
from keras.layers import LSTM, Dense, Layer, RepeatVector
from keras.models import Model
from tensorflow import keras


class Sampling(Layer):
    def call(self, inputs, **kwargs):
        z_mean, z_log_var = inputs
        batch = tf.shape(z_mean)[0]
        dim = tf.shape(z_mean)[1]
        epsilon = tf.keras.backend.random_normal(shape=(batch, dim))
        return z_mean + tf.exp(0.5 * z_log_var) * epsilon


class NegativeLogProbability(Layer):
    def call(self, inputs, **kwargs):
        z_mean, z_log_var, z_sample = inputs
        var = tf.exp(z_log_var)
        log_prob = 0.5 * (
            tf.math.log(2 * tf.constant(np.pi))
            + z_log_var
            + tf.square(z_sample - z_mean) / var
        )
        return tf.reduce_sum(log_prob, axis=-1)


def evt_weighted_mean(data, evt_wgts):
    wgt_loss = tf.math.multiply(data, evt_wgts)
    wgt_loss = tf.reduce_sum(input_tensor=wgt_loss)
    wgt_loss = tf.math.divide(wgt_loss, tf.reduce_sum(evt_wgts))

    return wgt_loss


def output_reconstruction_error(auto_enc_input, reconstruction):
    reco_error = tf.abs(reconstruction - auto_enc_input) / (1 + tf.abs(auto_enc_input))

    # mean over timesteps and features
    reco_error = tf.reduce_mean(reco_error, axis=[1, 2])

    return reco_error


def fn_reconstruction_loss(reconstruction, autoencoder_input):
    return tf.reduce_mean(
        keras.losses.mean_squared_error(reconstruction, autoencoder_input), axis=1
    )


def fn_kl_loss(z_mean, z_log_var, beta):
    kl_loss = 1 + z_log_var - tf.square(z_mean) - tf.exp(z_log_var)
    kl_loss = -0.5 * beta * tf.reduce_mean(tf.reduce_sum(kl_loss, axis=1))
    return kl_loss


class VAE(keras.Model):
    """this class takes encoder and decoder models and define the complete variational autoencoder architecture"""

    def __init__(self, mngr, encoder, decoder, **kwargs):
        super(VAE, self).__init__(**kwargs)
        self.mngr = mngr
        self.encoder = encoder
        self.decoder = decoder

        self.beta = mngr.model_cluster_space_regularisation_factor

    def call(self, inputs, **kwargs):
        d0 = inputs[self.mngr.tag_enc]
        d1 = inputs[self.mngr.tag_dec]
        wgts = inputs[self.mngr.tag_wgt]

        z_mean, z_log_var, z = self.encoder(d1)
        reconstruction = self.decoder(z_mean)

        reco_error = output_reconstruction_error(d1, reconstruction)

        model_output = {
            # first part of the time series of the input data
            self.mngr.tag_input_d0: d0,
            # second part of the time series of the input data (potentially contains flare)
            self.mngr.tag_input_d1: d1,
            # the forecast of the background in the second part of the input
            self.mngr.tag_bck_forecast_d1: d1,
            # # the input to the autoencoder
            self.mngr.tag_auto_enc_input: d1,
            # # hidden representation learned by the autoencoder given model_input_d1 and model_bck_forecast_d1
            self.mngr.tag_cluster_space_sample: z,
            # # output of the autoencoder, trained to reconstruct auto_enc_input
            self.mngr.tag_auto_enc_reconstruction: reconstruction,
            self.mngr.tag_auto_enc_reconstruction_error: reco_error,
            # # log probability of the real data (second part of the time series)
            # # given the hidden representation of the autoencoder
            # self.mngr.neg_neg_log_prob_input_d1_hidden: neg_log_prob_input_d1_hidden,
        }
        return model_output


def build_encoder(
    input_shape,
    sizes,
    latent_dim,
    beta,
    regularizer_l2_scale,
    dropout,
):
    encoder_input = Input(
        shape=input_shape,
        name="input_encs",
    )

    x = encoder_input

    for i_size in range(len(sizes)):
        lstm_size = sizes[i_size]

        # do sequence to vector only on the final encoder layer
        return_sequences = i_size < len(sizes) - 1

        x = LSTM(
            lstm_size,
            return_sequences=return_sequences,
            dropout=dropout,
            recurrent_dropout=dropout,
            kernel_regularizer=keras.regularizers.L2(regularizer_l2_scale),
            name=f"vae_encoder_lstm_{i_size}",
        )(x)

    z_mu = Dense(latent_dim)(x)
    z_log_var = Dense(latent_dim)(x)

    if beta == 0:
        z = z_mu
    else:
        z = Sampling()([z_mu, z_log_var])

    encoder = Model(encoder_input, [z_mu, z_log_var, z], name="vae_encoder")

    return encoder


def build_decoder(
    output_shape,
    latent_dim,
    sizes,
    regularizer_l2_scale,
    dropout,
):
    decoder_input = Input(
        shape=(latent_dim,),
        name="latent_z",
    )

    x = decoder_input
    x = RepeatVector(output_shape[0])(x)

    for n_size in range(len(sizes)):
        lstm_size = sizes[n_size]

        x = keras.layers.LSTM(
            lstm_size,
            return_sequences=True,
            dropout=dropout,
            recurrent_dropout=dropout,
            kernel_regularizer=keras.regularizers.L2(regularizer_l2_scale),
            name=f"vae_lambda_dec_lstm_{n_size}",
        )(x)

    x_mu = Dense(output_shape[1])(x)
    x_log_var = Dense(output_shape[1])(x)

    decoder = Model(decoder_input, [x_mu, x_log_var], name="vae_decoder")

    return decoder
