from tensorflow.keras import layers
from tensorflow.python.keras import backend as keras_back

from trans_finder.flares.model_utils.utils import add_scaled_loss


class SamplingLayer(layers.Layer):
    """Uses (z_mean, z_log_var) to sample z, the vector encoding a digit."""

    def __init__(self, latent_dim, latent_loss_weight, beta=1, **kwargs):
        super().__init__(**kwargs)
        self.latent_dim = latent_dim

        self.beta = beta

        self.mean_layer = layers.Dense(self.latent_dim, name="z_mean")
        self.log_var_layer = layers.Dense(self.latent_dim, name="z_log_var")
        self.latent_loss_weight = latent_loss_weight

    def call(self, x, training=None, **kwargs):
        mean = self.mean_layer(x)
        log_var = self.log_var_layer(x)

        if mean.shape[0] == None or log_var.shape[0] == None:
            # for when the input shape is only partially known
            return mean + 0 * log_var  # Keras needs the *0 so the gradient is not None

        # kl divergence:
        latent_loss = -0.5 * (
            1 + log_var - keras_back.square(mean) - keras_back.exp(log_var)
        )
        latent_loss = keras_back.sum(latent_loss, axis=-1)  # sum over latent dimension
        # latent_loss = keras_back.mean(latent_loss, axis=0)  # avg over batch

        # use beta to force less usage of vector space:
        latent_loss = self.beta * latent_loss

        add_scaled_loss(self, latent_loss, "1_latent_kl", self.latent_loss_weight)
        # self.add_loss(latent_loss)

        # if not training:
        #     return mean

        epsilon = keras_back.random_normal(shape=log_var.shape, mean=0.0, stddev=1.0)
        stddev = keras_back.exp(log_var * 0.5)
        return mean + stddev * epsilon
