import ast

import pylab
from matplotlib import gridspec as gridspec
from matplotlib import pyplot as plt


class PlotLib(object):
    def __init__(self, log, tag_base_dir, *args, **kwargs):
        self.log = log
        self.tag_base_dir = tag_base_dir

        self.styles_colors = {
            "main0": [
                "#F756B6",
                "#5B90DC",
                "#F9AE51",
                "#0C9FB3",
                "#E39699",
                "#3B507D",
                "#FD6989",
                "#794D88",
                "#8A978E",
                "#BA2C54",
            ],
            "main1": [
                "#D6088F",
                "#424D9C",
                "#178084",
                "#AF99DA",
                "#F58D46",
                "#634B5B",
                "#0C9FB3",
                "#7C438A",
                "#328CD6",
                "#8D0F25",
            ],
            "purples": [
                "#A57BB7",
                "#343D80",
                "#EA60BF",
                "#B7308E",
                "#E099C3",
                "#7C438A",
                "#AF99DA",
                "#4D428E",
                "#56276D",
            ],
            "greens": [
                "#268F92",
                "#ABC14D",
                "#8A978E",
                "#0C9FB3",
                "#BDA962",
                "#B0CB9E",
                "#769168",
                "#5E93A5",
                "#178084",
                "#B7BBAD",
            ],
            "autumn": [
                "#3C8DAB",
                "#A4657A",
                "#4E615D",
                "#A9434D",
                "#424D9C",
                "#DC575A",
                "#1D2D38",
                "#634B5B",
                "#56276D",
                "#577580",
            ],
            "pastels": [
                "#E39699",
                "#658BBA",
                "#8ABDB8",
                "#E3BD96",
                "#AFA9C9",
                "#3C8DAB",
                "#D982A0",
                "#C4A1B9",
                "#5B98A8",
                "#96B09D",
            ],
        }

        self.styles_colors["default"] = self.styles_colors["main0"]

        self.styles_lines = dict()
        for k, v in self.styles_colors.items():
            self.styles_lines[k] = ["-", "--", "-.", ":"]

        self.styles_markers = dict()
        for k, v in self.styles_colors.items():
            self.styles_markers[k] = [
                "o",
                "v",
                "s",
                "X",
                "d",
                "*",
                "p",
                "^",
                "<",
                "D",
                "x",
                ">",
            ]

        # make sure we repeat combinations are unique
        for k, v in self.styles_colors.items():
            if len(v) % len(self.styles_lines[k]) == 0:
                v.append("#000000")

        self.single_save_all = False
        self.single_save_count = 0

        self.fig_params = {
            "font_size": 12,
            "ax_lbl_font_size": 12,
            "leg_font_size": 10,
            "tick_spacing": 1,
            "bck_color": "white",
            "line_width": 0.5,
            "marker_size": 9,
        }

        pylab.rcParams.update(
            {
                "font.size": self.fig_params["font_size"],
                "axes.facecolor": self.fig_params["bck_color"],
                "axes.titlesize": self.fig_params["font_size"],
                "axes.labelsize": self.fig_params["ax_lbl_font_size"],
                "legend.fontsize": self.fig_params["leg_font_size"],
                "figure.facecolor": self.fig_params["bck_color"],
                "lines.markersize": self.fig_params["marker_size"],
                "legend.handlelength": 2,
            }
        )

        return

    def get_single_save_count(self):
        self.single_save_count += 1
        return "_" + str(self.single_save_count - 1)

    def get_colors(self, tag=None):
        return self.styles_colors if tag is None else self.styles_colors[tag]

    def set_line_style(self, base_set, index, plot_style=None):
        n_lines = len(self.styles_lines[base_set])
        line = self.styles_lines[base_set][index % n_lines]
        if plot_style is not None:
            plot_style["linestyle"] = line
        return line

    def set_marker_style(self, base_set, index, plot_style=None):
        n_markers = len(self.styles_markers[base_set])
        marker = self.styles_markers[base_set][index % n_markers]
        if plot_style is not None:
            plot_style["marker"] = marker
        return marker

    def get_styles(self, tag=None):
        out = {
            "styles_lines": (
                self.styles_lines if tag is None else self.styles_lines[tag]
            ),
            "styles_markers": (
                self.styles_markers if tag is None else self.styles_markers[tag]
            ),
        }
        return out

    def set_pdf_pages(self, pdf_pages):
        self.pdf_pages = pdf_pages

    def save_pdf_pages(self, pdf_pages, out_file_name="", figs=[]):
        if len(figs) > 0:
            for fig in figs:
                pdf_pages.savefig(fig)
                plt.close(fig)
        else:
            pdf_pages.savefig()

        self.log.info(
            [
                ["g", " - saved plot to "],
                ["y", out_file_name.replace(self.tag_base_dir, "")],
            ]
        )

        return

    def add_fig(self, n_rows=1, n_cols=1, opts=None):
        fig_size = (10, 8)

        gridspec_opts = {
            "hspace": 0.45,
            "wspace": 0.45,
            "left": 0.15,
            "right": 0.90,
            "bottom": 0.10,
            "top": 0.9,
        }
        if n_cols == n_rows == 1:
            gridspec_opts["left"] += 0.05
            gridspec_opts["bottom"] += 0.05
        elif n_rows == 1:
            gridspec_opts["bottom"] += 0.1

        if opts is not None:
            if "fig_size" in opts:
                fig_size = opts["fig_size"]

            if "fig_size_cor" in opts:
                assert 0
            if "aspect_ratio_cor" in opts:
                assert 0
            if "gridspec_opts" in opts:
                gridspec_opts.update(opts["gridspec_opts"])

            if "height_ratios" in opts:
                gridspec_opts.update(opts["height_ratios"])
            if "width_ratios" in opts:
                gridspec_opts.update(opts["width_ratios"])

        spec = gridspec.GridSpec(ncols=n_cols, nrows=n_rows, **gridspec_opts)
        fig = plt.figure(figsize=fig_size)

        return fig, spec

    def add_subplot(self, fig, spec):
        ax = fig.add_subplot(spec)

        # ax.tick_params(axis='x', labelsize=self.fig_params['font_size'])
        # ax.tick_params(axis='y', labelsize=self.fig_params['font_size'])

        ax.grid(
            **{
                "color": "#4f8c9d",
                "alpha": 0.8,
                "linestyle": ":",
                "linewidth": self.fig_params["line_width"],
            }
        )
        # ax.grid(False)

        return ax

    def get_pdf_pages(self, tag=None):
        out_file_name = self.pdf_pages

        if tag is not None:
            file_split = out_file_name.split(".")
            assert len(file_split) >= 2

            file_base = ".".join(file_split[0:-1])
            file_ext = file_split[-1]
            out_file_name = file_base + "_" + tag + "." + file_ext

        return out_file_name

    def get_plot_style(
        self, index=0, tag="main", style_set=None, has_line=True, has_marker=False
    ):
        base_set, color, line, marker = None, None, None, None
        style_opt = []
        if style_set is None:
            base_set = "default"
        else:
            style_opt = style_set.replace(" ", "").split(";")
            for opt in style_opt:
                if opt in self.styles_colors.keys():
                    base_set = opt
            if base_set is None:
                base_set = "default"
        assert base_set is not None and base_set in self.styles_colors.keys()

        col_set = self.styles_colors[base_set]
        lin_set = self.styles_lines[base_set]
        mrk_set = self.styles_markers[base_set]

        shift = {"color": 0, "line": 0, "marker": 0}
        for opt in style_opt:
            for col_lin in shift.keys():
                if f"shift_{col_lin}" not in opt:
                    continue
                assert opt.count("=") == 1
                shift[col_lin] = opt.split("=")[1]
                assert shift[col_lin].isdigit()
                shift[col_lin] = int(shift[col_lin])

        for opt in style_opt:
            if "sub_set" in opt:
                assert opt.count("=") == 1
                sub_set = opt.split("=")[1]
                assert sub_set.isdigit()
                sub_set = int(sub_set)

                index_col = int(index // sub_set)
                index_lin = index % sub_set

                color = col_set[shift["color"] + (index_col % len(col_set))]
                line = lin_set[shift["line"] + (index_lin % len(lin_set))]
                marker = mrk_set[shift["marker"] + (index_lin % len(mrk_set))]

            index_lists = ["index_list_color", "index_list_line", "index_list_marker"]
            if any([x in opt for x in index_lists]):
                assert opt.count("=") == 1
                index_list = opt.split("=")[1]
                index_list = ast.literal_eval(index_list)
                index_list = [int(i) for i in index_list]
                index_now = index_list[index % len(index_list)]
                # print('--', index_list,index, index_now)

                if "index_list_color" in opt:
                    color = col_set[shift["color"] + (index_now % len(col_set))]
                elif "index_list_line" in opt:
                    line = lin_set[shift["line"] + (index_now % len(lin_set))]
                elif "index_list_marker" in opt:
                    marker = mrk_set[shift["marker"] + (index_now % len(mrk_set))]

        if color is None:
            color = col_set[shift["color"] + (index % len(col_set))]

        if line is None:
            line = lin_set[shift["line"] + (index % len(lin_set))]

        if marker is None:
            marker = mrk_set[shift["marker"] + (index % len(mrk_set))]

        if not has_line:
            line = "None"

        if not has_marker:
            marker = "None"

        if tag == "main_data":
            style = {
                "color": color,
                "alpha": 0.8,
                "linestyle": line,
                "marker": marker,
                "linewidth": 2,
            }
        elif tag == "err_data":
            style = {
                "color": color,
                "alpha": 0.1,
            }
        else:
            assert 0

        return style

    def get_n_rows_cols(self, n_grps):
        n_rows, n_cols = 0, 0
        if n_grps <= 3:
            n_rows, n_cols = 1, n_grps
        elif n_grps in [4]:
            n_rows, n_cols = 2, 2
        elif n_grps in [5, 6]:
            n_rows, n_cols = 3, 2
        elif n_grps in [7, 8, 9]:
            n_rows, n_cols = 3, 3
        elif n_grps in [10, 11, 12]:
            n_rows, n_cols = 4, 3
        else:
            for n_rows_cols in range(4, 8):
                if n_grps <= n_rows_cols**2:
                    n_rows, n_cols = n_rows_cols, n_rows_cols
                    break

        assert (n_rows * n_cols) > 0
        return n_rows, n_cols
