import itertools
from typing import List

import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt

from trans_finder.flares.plot_list.util.histogram_groupedby_label import (
    do_detailed_histograms_grouped_by_label,
)
from trans_finder.flares.plot_list.util.plot_config import gridlines_config, tick_params
from trans_finder.flares.utilities.pdf_plot_collection import PDFPlotCollection


def plot_comparison_input_and_posterior_sample_logprob(
    mngr, bgmmodel, input_data, out_file_name
):
    fig, axes = plt.subplots(
        1,
        bgmmodel.hmc_n_parallel_chains,
        figsize=(8 * bgmmodel.hmc_n_parallel_chains, 8),
    )

    posterior_samples = bgmmodel.get_joint_model_dist().sample(len(input_data)).numpy()
    for i_chain in range(bgmmodel.hmc_n_parallel_chains):
        input_logprob = (
            bgmmodel.get_posterior_dist(i_chain).log_prob(input_data).numpy()
        )
        posterior_sample_logprob = bgmmodel.get_posterior_dist(i_chain).log_prob(
            posterior_samples[:, i_chain, :]
        )

        ax = axes[i_chain]

        ax.hist(
            input_logprob,
            bins="auto",
            histtype="step",
            label="input",
            # linestyle=":",
        )
        ax.hist(
            posterior_sample_logprob,
            bins="auto",
            histtype="step",
            label="posterior samples",
            # linestyle=":",
        )
        ax.grid(**gridlines_config)
        ax.tick_params(**tick_params)
        ax.set_xlabel("log_prob")
        ax.set_ylabel("Counts")
        ax.set_yscale("log")
        ax.legend()

    PDFPlotCollection(mngr, [fig]).save_to_file(out_file_name)


def plot_per_dimension_input_and_posterior_histograms(
    input_samples,
    posterior_samples,
    plot_axis_range=None,
    axes=None,
    hist_kwargs=None,
    subplots_kwargs=None,
):
    if hist_kwargs is None:
        hist_kwargs = {}
    if subplots_kwargs is None:
        subplots_kwargs = {}

    n_input_dims = input_samples.shape[1]
    assert n_input_dims == posterior_samples.shape[1]

    if axes is None:
        fig, axes = plt.subplots(
            nrows=1,
            ncols=n_input_dims,
            # sharey=True,
            figsize=(5 * n_input_dims, 7),
            **subplots_kwargs,
        )

    else:
        fig = None

    if n_input_dims == 1:
        axes = [axes]

    for dim, ax in zip(range(n_input_dims), axes):
        # g = sns.histplot(
        #     ax=ax,
        #     data={
        #         "posterior samples": posterior_samples[:, dim],
        #         "input data": input_samples[:, dim],
        #     },
        #     legend=dim == 0,
        #     # stat="density",
        #     # common_norm=False,
        #     element="step",
        #     binrange=plot_axis_range,
        #     binwidth=0.01,
        # )

        ax.hist(
            [posterior_samples[:, dim], input_samples[:, dim]],
            bins="auto",
            histtype="step",
            label=["posterior data", "input data"],
            # density=True,
            **hist_kwargs,
        )
        ax.grid(**gridlines_config)
        ax.tick_params(**tick_params)
        # ax.set_xlabel("ts: negative log prob")
        ax.set_ylabel(
            "Density"
            if "density" in hist_kwargs and hist_kwargs["density"] == True
            else "Counts"
        )

        ax.set_yscale("log")
        ax.grid(**gridlines_config)
        ax.tick_params(**tick_params)
        ax.set_xlabel(f"dimension {dim}")
        # plt.xlabel(f"dimension {dim}")

    return fig


def plot_quality_of_fit_histogram(quality_of_fit: np.array):
    fig, ax = plt.subplots()

    sns.histplot(
        ax=ax,
        data=quality_of_fit,
        stat="density",
        bins=10,
    )
    plt.xlabel(f"quality of fit (mean data prob / mean grid prob)")

    return fig


def bgmm_plot_outlier_score_histograms(
    mngr,
    data: np.array,
    out_file_name: str,
    x_label,
    labels: pd.Series = None,
    ts_significance_levels=None,
    vertical_lines=None,
):
    do_detailed_histograms_grouped_by_label(
        mngr,
        histogram_data=data,
        data_desc=x_label,
        out_file_name=out_file_name,
        labels=labels,
        vertical_lines=vertical_lines,
    )


def bgmm_plot_outlier_score_ecdfplot(
    mngr,
    data: np.array,
    x_label,
    out_file_name: str,
    group_label_col_name: str,
    labels: pd.Series = None,
    vertical_lines=None,
):
    fig, ax = plt.subplots(
        nrows=1,
        ncols=1,
        figsize=(10, 5),
    )

    plot_df = pd.DataFrame({x_label: data, group_label_col_name: labels})
    sns.ecdfplot(
        ax=ax,
        data=plot_df,
        x=x_label,
        hue=group_label_col_name,
        legend=True,
    )

    ax.grid(**gridlines_config)
    if vertical_lines is not None:
        for label, x_pos in vertical_lines.items():
            if x_pos is None:
                msg = f"Encountered None in vertical line '{label}'. Skipping it."
                mngr.log.info([["b", msg]])
                continue
            ax.axvline(x=x_pos, c="black", linewidth=0.5)

    PDFPlotCollection(mngr, [fig]).save_to_file(out_file_name)


def trace_plot(
    trace: List,
    parameter_name: str,
    num_burnin_steps: int = None,
):
    figures = []

    parameter_array = trace.numpy()

    shape = parameter_array.shape[1:]
    for idx in itertools.product(*[range(s) for s in shape]):
        selector = ..., *idx
        s = parameter_array[selector]
        fig, ax = plt.subplots(figsize=(10, 4))
        ax.plot(s)
        ax.set_title(f"trace plot of {parameter_name}{str(idx)}")
        # ax.set_xlabel("steps")
        # ax.set_ylabel("parameter value")
        if num_burnin_steps:
            plt.axvspan(0, num_burnin_steps, color="grey", alpha=0.5)

        # aesthetics
        try:
            ax.set_facecolor("white")
            # ax.grid()
        except Exception as e:
            pass

        figures.append(fig)
        plt.show()

    return figures

    # # Trace plots
    # colors = ["b", "g", "r"]
    # for i in range(n_components):
    #     plt.plot(parameter_array[:, i], c=colors[i], alpha=0.3)
    #     # plt.hlines(true_w[i], 0, 1000, zorder=4, color=colors[i], label="$w_{}$".format(i))
    # plt.legend(loc="upper right")
    # plt.title(parameter_name)
    # plt.show()
    #
    # # Histogram of samples
    # for i in range(n_components):
    #     sns.distplot(parameter_array[:, i], color=colors[i])
    # ymax = plt.ylim()[1]
    # # for i in range(3):
    # #     plt.vlines(true_w[i], 0, ymax, color=colors[i])
    # plt.ylim(0, ymax)
    # plt.title(parameter_name)
    # plt.show()
