import matplotlib.patheffects as pe
from matplotlib import colors

from trans_finder.flares.core.tags import (
    TAG_LOWERQUANT,
    TAG_MEDIAN,
    TAG_SIGNIFICANCE_COMBINED,
    TAG_UPPERQUANT,
)
from trans_finder.flares.plot_list.util.plot_config import (
    COLORS_QUANTITATIVE_BRIGHT,
    DEFAULT_SCALE,
    MEDIUM_SIZE,
    SMALL_SIZE,
    gridlines_config,
    tick_params,
)


def plot_channel_group_metrics(
    mngr,
    ax,
    df_cut,
    ts_col_name,
    channel_groups_to_skip,
    plot_kwargs=None,
    plot_conf=None,
):
    if plot_kwargs is None:
        plot_kwargs = {}

    if plot_conf is None:
        plot_conf = {}

    fill_kwargs = plot_kwargs.copy()
    fill_kwargs["linewidth"] = 0
    if "kw" in fill_kwargs:
        del fill_kwargs["lw"]

    col_name_fn = lambda mask_name: f"{ts_col_name}_{mask_name}"

    full_input_ts = ts_col_name

    masked_input_ts = [
        col_name_fn(mask_name) for mask_name in mngr.channel_groups.keys()
    ]

    cols = masked_input_ts + [full_input_ts]

    plot_markers = ("show_markers" in plot_conf) and plot_conf["show_markers"]

    styles = {
        group_name: mngr.plot_lib.get_plot_style(
            index=idx,  # offset to not have the pink
            tag="main_data",
            style_set="main_1",
            has_line=True,
            has_marker=plot_markers,
        )
        for idx, group_name in enumerate(mngr.channel_groups.keys())
    }

    # metric_tag = TAG_SIGNIFICANCE_CLUSTER_SPACE
    metric_tag = TAG_SIGNIFICANCE_COMBINED

    lower_cutoff = 0  # in sigma
    # lower_cutoff = 1  # in sigma

    # group_colors = {k: v["color"] for k, v in styles.items()}
    group_colors = {
        k: COLORS_QUANTITATIVE_BRIGHT[i % len(COLORS_QUANTITATIVE_BRIGHT)]
        for i, k in enumerate(styles.keys())
    }

    print(repr(group_colors))

    # group_colors[TAG_INPUT_MASK_NONE] = "#000000"
    # group_colors[list(group_colors.keys())[0]] = "#000000"

    for mask_name in mngr.channel_groups.keys():
        if mask_name in channel_groups_to_skip:
            to_skip = plot_conf.get("skip_channels_in_results", [])
            if mask_name in to_skip:
                continue

        df_masked = df_cut[df_cut["input_mask"] == mask_name]

        cutoff_mask = df_masked["_".join([metric_tag, TAG_MEDIAN])] < lower_cutoff

        if lower_cutoff > 0:
            df_masked = df_masked.assign(
                **{
                    c: df_masked[c].mask(cutoff_mask, 0)
                    for c in [
                        "_".join([metric_tag, TAG_LOWERQUANT]),
                        "_".join([metric_tag, TAG_UPPERQUANT]),
                        "_".join([metric_tag, TAG_MEDIAN]),
                    ]
                }
            )

        ax.plot(
            df_masked["time_right_edge"].to_numpy(),
            df_masked["_".join([metric_tag, TAG_MEDIAN])].to_numpy(),
            # styles[mask_name]["linestyle"],
            label=mngr.channel_labels.get(mask_name, mask_name),
            color=colors.to_rgba(group_colors[mask_name], 0.99),
            **(
                dict(
                    marker="x",
                    # marker=styles[mask_name]["marker"],
                    markersize=4,
                    markerfacecolor="none",
                )
                if plot_markers
                else {}
            ),
            **plot_kwargs,
        )

        ax.fill_between(
            df_masked["time_right_edge"].to_numpy(),
            df_masked["_".join([metric_tag, TAG_LOWERQUANT])].to_numpy(),
            df_masked["_".join([metric_tag, TAG_UPPERQUANT])].to_numpy(),
            alpha=0.4,  # Adjust transparency of the shaded region
            color=group_colors[mask_name],
            # color='gray',  # Adjust the color of the shaded region
            **fill_kwargs,
        )

    leg = ax.legend(
        loc=plot_conf.get("legend_position", "upper center"),
        ncol=plot_conf.get("legend_n_cols", len(cols) / 3),
        alignment="right",
        bbox_to_anchor=plot_conf.get("legend_bbox_to_anchor", None),
        prop={"size": plot_conf.get("legend_fontsize", DEFAULT_SCALE * SMALL_SIZE)},
    )  # bbox_to_anchor=(0.5, -0.05), )
    # plt.yscale("log")

    # set the linewidth of each legend object
    for legobj in leg.legendHandles:
        legobj.set_linewidth(2.0)

    ax.grid(**gridlines_config)
    ax.tick_params(**tick_params)

    # ax.set_ylabel(mngr.label_significance_combo)

    # add channel name label
    text = ax.text(
        *(0.025, 0.9),
        f"{mngr.label_significance_combo}",
        transform=ax.transAxes,
        fontsize=MEDIUM_SIZE * DEFAULT_SCALE,
        verticalalignment="center",
        horizontalalignment="left",
        alpha=0.9,
        # bbox=dict(
        #     boxstyle="round",
        #     facecolor="white",
        #     alpha=0.5,
        #     edgecolor="black",
        #     # color="black",
        # ),
        path_effects=[pe.withStroke(linewidth=0.75, foreground="white", alpha=0.75)],
    )
