from pathlib import Path

import matplotlib.patheffects as pe
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import umap
from sklearn.decomposition import PCA
from umap import UMAP

from trans_finder.flares.core.tags import (
    TAG_INPUT_MASK,
    TAG_INPUT_MASK_NONE,
    TAG_SIGNIFICANCE_CLUSTER_SPACE,
    TAG_SIGNIFICANCE_COMBINED,
    TAG_SIGNIFICANCE_RECO_ERROR,
)
from trans_finder.flares.helpers_debugging import get_debug_output_filepath
from trans_finder.flares.plot_list.util.plot_config import (
    COLORS_HIGH_CONTRAST,
    DEFAULT_SCALE,
    MEDIUM_SIZE,
    SMALL_SIZE,
    set_fontsizes,
)
from trans_finder.flares.utilities.pdf_plot_collection import PDFPlotCollection


def do_cluster_space_n_signal_steps_plots(
    mngr,
    plot_df,
    df_diagnostics: pd.DataFrame,
):
    tag = ["b", f" - flares.plot_list.{Path(__file__).stem} - "]
    mngr.log.info([tag, ["", f"do_cluster_space_n_signal_steps_plots()"]])

    signal_step_plot_df = plot_df.assign(
        signal_step_sum=df_diagnostics["signal_step_sum"]
    )

    evt_classes = ("bck", "signal_dec")

    combination_df = signal_step_plot_df[
        signal_step_plot_df[mngr.tag_evt_type].isin(evt_classes)
    ]

    g = do_single_plot(mngr, combination_df, hue="signal_step_sum")

    file_name = "cluster_space_n_signal_steps_plot.pdf"
    pdf_file_path = get_debug_output_filepath(mngr, "cluster_space", file_name)
    PDFPlotCollection(mngr, [g.fig]).save_to_file(pdf_file_path)


def do_cluster_space_signal_integral_dec_plots(
    mngr,
    plot_df,
    df_diagnostics: pd.DataFrame,
):
    tag = ["b", f" - flares.plot_list.{Path(__file__).stem} - "]
    mngr.log.info([tag, ["", f"do_cluster_space_signal_integral_dec_plots()"]])

    signal_step_plot_df = plot_df.assign(
        signal_strength_categorical=df_diagnostics["signal_strength_categorical"]
    )

    combinations = [("bck", "signal_dec")]
    for evt_classes in combinations:
        combination_df = signal_step_plot_df[
            signal_step_plot_df[mngr.tag_evt_type].isin(evt_classes)
        ]

        if len(combination_df) == 0:
            continue

        figs = []

        # y_vars = ["latent dim 2"]
        # x_vars = [f"latent dim {n+1}" for n in [0, 1, 3, 4, 5, 6, 7, 8, 9]]
        g = sns.PairGrid(
            combination_df,
            diag_sharey=False,
            height=5,
            hue="signal_strength_categorical",
            # y_vars=y_vars,
        )
        g.map_upper(sns.scatterplot, marker="x")
        try:
            g.map_lower(sns.kdeplot, common_norm=False)
        except Exception as e:
            pass
        # g.map_diag(sns.histplot, stat="percent")
        g.map_diag(sns.kdeplot, common_norm=False)
        g.add_legend()

        figs.append(g.fig)

        file_name = "cluster_space_signal_integral_dec_plot.pdf"
        pdf_file_path = get_debug_output_filepath(mngr, "cluster_space", file_name)
        PDFPlotCollection(mngr, figs).save_to_file(pdf_file_path)


def do_single_plot(mngr, df, hue):
    g = sns.PairGrid(
        df,
        diag_sharey=False,
        height=5,
        hue=hue,
    )
    g.map_upper(sns.scatterplot, marker="x")
    try:
        g.map_lower(sns.kdeplot, common_norm=False)
    except Exception as e:
        pass
    # g.map_diag(sns.histplot, stat="percent")
    g.map_diag(sns.kdeplot, common_norm=False)
    g.add_legend()

    return g


def do_projection_plot(mngr, df, hue_col):
    auto_enc_representations = df[
        [
            f"{mngr.tag_cluster_space_mean}_{i}"
            for i in range(mngr.model_learned_representation_n_dim)
        ]
    ].to_numpy()

    reducer = umap.UMAP(n_components=2)
    embedding = reducer.fit_transform(auto_enc_representations)

    plot_df = df.assign(
        umap_dim_0=embedding[:, 0],
        umap_dim_1=embedding[:, 1],
    )

    g = sns.JointGrid(
        data=plot_df,
        x="umap_dim_0",
        y="umap_dim_1",
        hue=hue_col,
        height=10,
    )
    g.plot_joint(sns.scatterplot, marker="x")
    g.plot_marginals(sns.kdeplot, common_norm=False)

    file_name = f"cluster_space_umap.pdf"
    pdf_file_path = get_debug_output_filepath(mngr, "cluster_space_umap", file_name)
    PDFPlotCollection(mngr, [g.fig]).save_to_file(pdf_file_path)


def variants(mngr, df_diagnostics):
    # only use full input mask
    df = df_diagnostics[df_diagnostics[TAG_INPUT_MASK] == TAG_INPUT_MASK_NONE]

    print(df["realisation_id"].unique())

    print(len(df))

    if mngr.dir_input_predict == mngr.tag_custom_realisations:
        realisations = [
            "bck_no_plateaus",
            "systematic_all_flare_templates_cta_and_fermi_1",
            "systematic_all_flare_templates_only_cta_1",
            "systematic_all_flare_templates_only_fermi_1",
            "systematic_all_flare_templates_cta_and_fermi_2",
            "systematic_all_flare_templates_only_cta_2",
            "systematic_all_flare_templates_only_fermi_2",
            "systematic_all_flare_templates_cta_and_fermi_3",
            "systematic_all_flare_templates_only_cta_3",
            "systematic_all_flare_templates_only_fermi_3",
            "systematic_all_flare_templates_cta_and_fermi_4",
            "systematic_all_flare_templates_only_cta_4",
            "systematic_all_flare_templates_only_fermi_4",
            "systematic_all_flare_templates_cta_and_fermi_5",
            "systematic_all_flare_templates_only_cta_5",
            "systematic_all_flare_templates_only_fermi_5",
        ]
        present_realisations = df.realisation_id.unique()
        if all(r in present_realisations for r in realisations):
            # filter out parts with a high reconstruction error
            df = df[df[TAG_SIGNIFICANCE_RECO_ERROR] < 1.5]

            yield (
                "systematic_all_flare_templates",
                df[df["realisation_id"].isin(realisations)],
                TAG_SIGNIFICANCE_CLUSTER_SPACE,
            )

    elif mngr.dir_input_predict == mngr.dir_output_data_prep_predict:
        yield "infer_combined", df, TAG_SIGNIFICANCE_COMBINED

        # filter out parts with a high reconstruction error
        df = df[df[TAG_SIGNIFICANCE_RECO_ERROR] < 2]
        yield "infer_clusterspace", df, TAG_SIGNIFICANCE_CLUSTER_SPACE
    else:
        yield "predict_combined", df, TAG_SIGNIFICANCE_COMBINED

        # filter out parts with a high reconstruction error
        df = df[df[TAG_SIGNIFICANCE_RECO_ERROR] < 2]
        yield "predict_clusterspace", df, TAG_SIGNIFICANCE_CLUSTER_SPACE


def plot_perdim_grouped_distributions(mngr, df_diagnostics, latent_cols):
    # only use full input mask
    df = df_diagnostics[df_diagnostics[TAG_INPUT_MASK] == TAG_INPUT_MASK_NONE]

    realisations_background = [
        "bck_no_plateaus",
    ]

    realisations_flaring = [
        # "systematic_all_flare_templates_cta_and_fermi_1",
        # "systematic_all_flare_templates_cta_and_fermi_2",
        # "systematic_all_flare_templates_cta_and_fermi_3",
        # "systematic_all_flare_templates_cta_and_fermi_4",
        "systematic_all_flare_templates_cta_and_fermi_5",
    ]

    if mngr.skip_flare_templates:
        return

    # select_original_flux_events(mngr, df_diagnostics, latent_cols)

    # n_samples = 100  # todo
    n_samples = 10000

    df_flaring = df[df["realisation_id"].isin(realisations_flaring)]

    if len(df_flaring) == 0:
        print("Skipping plot_perdim_grouped_distributions() as no data is available.")
        return

    df_flaring = df_flaring[df_flaring[TAG_SIGNIFICANCE_CLUSTER_SPACE].gt(2)].sample(
        n=n_samples
    )

    df_background = df[df["realisation_id"].isin(realisations_background)]
    df_background = df_background[
        df_background[TAG_SIGNIFICANCE_CLUSTER_SPACE].lt(2)
    ].sample(n=n_samples)

    max_n_latent_cols = 7

    # assign column for the group, then concatenate the dataframes and set the column name for the hue arg
    col_group = "Data subset"
    df_flaring[col_group] = "Flaring"
    df_background[col_group] = "Background"
    cols = latent_cols[: min(max_n_latent_cols, len(latent_cols))] + [col_group]
    df = pd.concat([df_flaring[cols], df_background[cols]])
    df.columns = df.columns.str.replace("cluster_space_mean_", "Dimension ")
    # sns.set_theme(font_scale=2.0)
    print("Plotting cluster space PairGrid (this can take a moment...)")

    with sns.plotting_context(
        rc={
            "axes.labelsize": MEDIUM_SIZE * DEFAULT_SCALE,
            "xtick.labelsize": SMALL_SIZE * DEFAULT_SCALE,
            "ytick.labelsize": SMALL_SIZE * DEFAULT_SCALE,
        }
    ):
        g = sns.PairGrid(
            df,
            diag_sharey=False,
            height=2.5,
            hue=col_group,
            palette=sns.color_palette(COLORS_HIGH_CONTRAST[:2:][::-1]),
        )
        g.fig.dpi = 300

        def hide_current_axis(*args, **kwds):
            plt.gca().set_visible(False)

        g.map_upper(hide_current_axis)

        try:
            g.map_lower(sns.histplot, rasterized=True)
            g.map_lower(sns.kdeplot, common_norm=False)
        except Exception as e:
            pass
        g.map_diag(sns.kdeplot, common_norm=False)

        g.add_legend(
            loc="upper right",
            bbox_to_anchor=(0.966, 0.15),
            title="",
            fontsize=MEDIUM_SIZE * DEFAULT_SCALE,
        )  # bbox_to_anchor=(0, 0),

    set_fontsizes(scale=DEFAULT_SCALE)

    file_name = f"latent_dims_grid.pdf"
    pdf_file_path = get_debug_output_filepath(mngr, "cluster_space", file_name)
    PDFPlotCollection(mngr, [g.fig]).save_to_file(pdf_file_path)


def do_cluster_space_plots(
    mngr,
    df_diagnostics: pd.DataFrame,
):
    latent_cols = [
        f"{mngr.tag_cluster_space_mean}_{i}"
        for i in range(mngr.model_learned_representation_n_dim)
    ]

    plot_perdim_grouped_distributions(mngr, df_diagnostics, latent_cols)

    if mngr.dir_output_predict == "predict_out_infer":
        return

    for name, df, hue_col in variants(mngr, df_diagnostics):
        print(len(df))

        if len(df) == 0:
            print("No data left in do_cluster_space_plots..")
            return

        max_fit_events = 250000
        if len(df) > max_fit_events:
            df = df.sample(n=max_fit_events)

        print(len(df))

        # Compute PCA
        print("Fitting PCA")
        pca = PCA(n_components=2)
        pca.fit(df[latent_cols])

        print("Fitting UMAP")
        # Compute UMAP
        umap = UMAP(n_components=2)  # , random_state=42)
        umap.fit(df[latent_cols])

        # max_plot_events = 100000
        # if len(df) > max_plot_events:
        #     df = df.sample(n=max_plot_events)

        umap_result = umap.transform(df[latent_cols])
        pca_result = pca.transform(df[latent_cols])

        plot_pca_umap_independently(
            mngr,
            pca_result,
            umap_result,
            df[hue_col],
            name + ".pdf",
            hue_label=hue_col,
            pca_explained_variance=pca.explained_variance_ratio_,
            umap_params={
                "n_neighbors": umap.n_neighbors,
                "min_dist": umap.min_dist,
                "metric": umap.metric,
            },
        )


def inside_label(ax, text):
    # add channel name label
    text = ax.text(
        *(0.019, 0.933),
        text,
        transform=ax.transAxes,
        fontsize=MEDIUM_SIZE * DEFAULT_SCALE,
        verticalalignment="center",
        horizontalalignment="left",
        alpha=0.9,
        # bbox=dict(
        #     boxstyle="round",
        #     facecolor="white",
        #     alpha=0.5,
        #     edgecolor="black",
        #     # color="black",
        # ),
        path_effects=[pe.withStroke(linewidth=0.75, foreground="white", alpha=0.75)],
    )


def plot_pca_umap_independently(
    mngr,
    pca_result,
    umap_result,
    hue,
    file_name,
    hue_label: str,
    pca_explained_variance=None,
    umap_params=None,
):
    fig_kwargs = dict(
        figsize=(7, 4),
        dpi=300,
        gridspec_kw={
            "hspace": 0.00,
            "wspace": 0.00,
            "left": 0.02,
            "right": 0.85,
            "bottom": 0.02,
            "top": 0.94,
        },
    )

    # Create a figure with subplots
    # Set up the figure and gridspec
    fig1, ax1 = plt.subplots(1, 1, **fig_kwargs)
    fig2, ax2 = plt.subplots(1, 1, **fig_kwargs)

    axes = [ax1, ax2]
    figs = [fig1, fig2]

    # Unified colormap for both PCA and UMAP
    min_size = 1
    max_size = 15
    # Assuming 'series' is your Pandas Series of positive floats
    rescaled_hue = min_size + (hue - hue.min()) * (max_size - min_size) / (
        hue.max() - hue.min()
    )

    scatter_kwargs = dict(
        cmap="viridis",
        # vmin=hue.min(),
        # vmax=hue.max(),
        c=np.minimum(hue, 5.001),
        # cmap="inferno",
        # cmap="magma",
        s=rescaled_hue,
        alpha=0.9,
        # linewidths=0,
        marker=".",
        edgecolor="none",
        rasterized=True,
    )

    # PCA scatter plot
    scatter_pca = axes[0].scatter(pca_result[:, 0], pca_result[:, 1], **scatter_kwargs)

    inside_label(axes[0], "PCA")

    # UMAP scatter plot
    scatter_umap = axes[1].scatter(
        umap_result[:, 0],
        umap_result[:, 1],
        **scatter_kwargs,
    )
    inside_label(axes[1], "UMAP")

    for fig in figs:
        # Add a colorbar
        cbar_ax = fig.add_axes(
            [0.875, 0.05, 0.03, 0.85]
        )  # [left, bottom, width, height]
        cbar = fig.colorbar(scatter_pca, cax=cbar_ax, orientation="vertical")
        cbar.set_label(mngr.label_significance_mm, fontsize=MEDIUM_SIZE * DEFAULT_SCALE)

    for ax in axes:
        ax.set_xticklabels([])
        ax.set_yticklabels([])
        ax.tick_params(length=0)

    plt.subplots_adjust(
        **{
            "hspace": 0.2,
            "wspace": 0.05,
            "left": 0.05,
            "right": 0.95,
            "bottom": 0.25,
            "top": 0.9,
        }
    )

    PDFPlotCollection(mngr, [fig1]).save_to_file(
        get_debug_output_filepath(mngr, "cluster_space", "pca.pdf")
    )
    PDFPlotCollection(mngr, [fig2]).save_to_file(
        get_debug_output_filepath(mngr, "cluster_space", "umap.pdf")
    )


def plot_pca_umap_sidebyside_sharedcolormap(
    mngr, pca_result, umap_result, hue, file_name, hue_label: str
):
    # Create a figure with subplots
    # Set up the figure and gridspec
    fig, axes = plt.subplots(1, 2, figsize=(15, 6))

    # Unified colormap for both PCA and UMAP

    min_size = 1
    max_size = 15
    # Assuming 'series' is your Pandas Series of positive floats
    rescaled_hue = min_size + (hue - hue.min()) * (max_size - min_size) / (
        hue.max() - hue.min()
    )

    scatter_kwargs = dict(
        cmap="viridis",
        # vmin=hue.min(),
        # vmax=hue.max(),
        c=np.minimum(hue, 5),
        # cmap="inferno",
        # cmap="magma",
        s=rescaled_hue,
        alpha=0.9,
        # linewidths=0,
        marker=".",
        edgecolor="none",
        rasterized=True,
    )

    # PCA scatter plot
    scatter_pca = axes[0].scatter(pca_result[:, 0], pca_result[:, 1], **scatter_kwargs)
    axes[0].set_title("PCA")
    # axes[0].axis('off')  # Turn off axis

    # UMAP scatter plot
    scatter_umap = axes[1].scatter(
        umap_result[:, 0],
        umap_result[:, 1],
        **scatter_kwargs,
    )
    axes[1].set_title("UMAP")

    # Add a colorbar at the bottom
    cbar_ax = fig.add_axes([0.1, 0.15, 0.8, 0.03])  # [left, bottom, width, height]
    cbar = fig.colorbar(scatter_pca, cax=cbar_ax, orientation="horizontal")
    cbar.set_label(hue_label)

    for ax in axes.flat:
        ax.set_xticklabels([])
        ax.set_yticklabels([])
        ax.tick_params(length=0)

    plt.subplots_adjust(
        **{
            "hspace": 0.2,
            "wspace": 0.05,
            "left": 0.05,
            "right": 0.95,
            "bottom": 0.25,
            "top": 0.9,
        }
    )

    pdf_file_path = get_debug_output_filepath(mngr, "cluster_space", file_name)
    PDFPlotCollection(mngr, [fig]).save_to_file(pdf_file_path)


def by_event_class_combination(mngr, plot_df):
    for combination in [
        # ("bck_pure", "signal_dec"),
        # ("bck_pure", "signal_old_dec", "signal_old_enc", "signal_old_enc_dec"),
        # ("bck_pure", "signal_old_dec"),
        # ("bck_pure", "signal_old_enc"),
        # ("signal_dec", "signal_old_dec"),
        #
        # (mngr.tag_background, "signal_dec"),
        # (mngr.tag_background, "signal_enc"),
        # (mngr.tag_background, "signal_enc_dec"),
        # ("signal_enc", "signal_dec"),
    ]:
        combination_df = plot_df[plot_df[mngr.tag_evt_type].isin(combination)]
        if len(combination_df) == 0:
            continue

        mngr.log.info(
            [
                tag,
                [
                    "",
                    f"plotting event_class comparison of " + " and ".join(combination),
                ],
            ]
        )

        g = do_single_plot(mngr, combination_df, hue=mngr.tag_evt_type)

        file_name = "cluster_space_" + "_".join(combination) + ".pdf"
        pdf_file_path = get_debug_output_filepath(mngr, "cluster_space", file_name)
        PDFPlotCollection(mngr, [g.fig]).save_to_file(pdf_file_path)
