import os

import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt

from trans_finder.flares.helpers_debugging import get_debug_output_filepath
from trans_finder.flares.plot_list.util.text_figure import create_text_figure
from trans_finder.flares.utilities.pdf_plot_collection import PDFPlotCollection


def do_debug_plot_of_data_prep_events(
    mngr,
    event_data_values,
    event_metadata_values,
    output_data_col_names,
    output_metadata_col_names,
    realisation,
    i_event,
    channel_column_names,
    old_signal_steps,
    new_signal_steps,
    signal_std_and_mean_per_channel,
):
    n_plots = mngr.debug_data_prep_plots_per_realisation
    if (
        mngr.run_phase != "reproduce_data_prep"
        and (
            not mngr.debug_create_data_prep_plots
            or realisation.id
            not in range(*mngr.debug_plot_of_data_prep_events_realisation_range)
            or (n_plots > 0 and i_event > n_plots)
        )
        or (
            mngr.run_phase == "reproduce_data_prep"
            and i_event not in mngr.data_prep_reproduce_indices
        )
    ):
        return

    if realisation.id == 0 and i_event == 0:
        mngr.log.info(
            [
                [
                    "b",
                    f" - flares.plot_list.data_prep_event"
                    f" - Plotting the {n_plots} first events of the realisations in range {mngr.debug_plot_of_data_prep_events_realisation_range}...",
                ]
            ]
        )

    # only plot the first x and the last x events of a realisation

    cols_to_drop = [
        mngr.data_prep_day_column_name,
        # mngr.tag_temporal_offset,
        "ts",
        "spectral_index",
    ]
    plot_channel_col_names = [
        name for name in channel_column_names if name not in cols_to_drop
    ]

    col_name_input_part = "input part"

    df_col_names = []
    plot_col_names = []
    for channel_tag in mngr.data_prep_channel_tags:
        df_col_names.extend(
            [f"{channel_tag}_{ft_name}" for ft_name in channel_column_names]
        )
        plot_col_names.extend(
            [f"{channel_tag}_{ft_name}" for ft_name in plot_channel_col_names]
        )

    df_data = np.hstack(
        np.concatenate(event_data_values).reshape(
            len(mngr.data_prep_channel_tags), mngr.n_steps_tot, -1
        )
    )

    df = pd.DataFrame(df_data)
    df.columns = df_col_names
    df = df.loc[:, plot_col_names]

    df[col_name_input_part] = ["d0"] * mngr.n_steps_enc + ["d1"] * mngr.n_steps_dec

    feature_names_by_channel_tag = {
        tag: plot_channel_col_names for tag in mngr.data_prep_channel_tags
    }

    def get_plot_df(data_df):
        plot_df_list = []
        for channel_tag, ft_names in feature_names_by_channel_tag.items():
            for ft_name in ft_names:
                if ft_name == mngr.tag_temporal_offset:
                    continue

                full_ft_name = f"{channel_tag}_{ft_name}"
                full_offset_name = f"{channel_tag}_{mngr.tag_temporal_offset}"
                plot_df = data_df[[col_name_input_part]].assign(
                    **{
                        "input step": data_df.index,
                        "channel name": channel_tag,
                        "feature name": ft_name,
                        "channel and feature name": full_ft_name,
                        "feature value": data_df[full_ft_name],
                        # "temporal offset" : data_df[full_offset_name],
                        # "temporal offset": -np.cumsum(
                        #     data_df[full_offset_name][::-1]
                        # )[::-1],
                        "temporal offset": data_df[full_offset_name],
                        # "real time": data_df[mngr.data_prep_day_column_name],
                    }
                )

                plot_df_list.append(plot_df)
        plot_df = pd.concat(plot_df_list, ignore_index=True)
        return plot_df

    evt_class = event_metadata_values[
        output_metadata_col_names.index(mngr.tag_evt_type)
    ]

    figs = []

    fig, axes = plt.subplots(
        nrows=3,
        ncols=1,
        figsize=(12, 3 * 3),
        constrained_layout=True,
        # sharey=True,
        # sharex=True,
    )
    figs.append(fig)

    g = sns.lineplot(
        ax=axes[0],
        data=get_plot_df(df),
        x="temporal offset",
        y="feature value",
        style="channel and feature name",
        hue="channel name",
        markers=True,
        # dashes="channel and feature name",
        # style="part",
    )
    g.legend(loc="center left", bbox_to_anchor=(1, 0.5))

    # fig, ax = plt.subplots(
    #     nrows=1,
    #     ncols=1,
    #     figsize=(10, 6),
    #     constrained_layout=True,
    #     sharey=True,
    #     sharex=True,
    # )

    plt.title(evt_class)

    g = sns.lineplot(
        ax=axes[1],
        data=get_plot_df(df),
        x="input step",
        y="feature value",
        style="channel and feature name",
        hue="channel name",
        markers=True,
        # dashes="channel and feature name",
        # style="part",
    )
    g.legend(loc="center left", bbox_to_anchor=(1, 0.5))
    # vertical line to separate encoder / decoder
    axes[1].axvline(x=mngr.n_steps_enc - 0.5, color="black")
    # figs.append(fig)

    # add metadata values as additional plot
    metadata_dict = dict(zip(output_metadata_col_names, event_metadata_values))
    for channel_tag in mngr.data_prep_channel_tags:
        (
            df_bck_median,
            df_bck_std,
        ) = signal_std_and_mean_per_channel[channel_tag]

        for column in df_bck_std.keys():
            metadata_dict[f"{channel_tag}_{column}_mean"] = np.round(
                df_bck_median[column], 4
            )
            metadata_dict[f"{channel_tag}_{column}_std"] = np.round(
                df_bck_std[column], 4
            )

    # fig, ax = plt.subplots(
    #     nrows=1,
    #     ncols=1,
    #     figsize=(10, 3),
    #     constrained_layout=True,
    # )

    old_signal_steps.rename(
        columns={tag: f"old_{tag}" for tag in realisation.channel_tags}
    ).astype(int).plot(ax=axes[2])
    new_signal_steps.rename(
        columns={tag: f"new_{tag}" for tag in realisation.channel_tags}
    ).astype(int).plot(ax=axes[2], style=["-."] * len(realisation.channel_tags))
    axes[2].axvline(x=mngr.n_steps_enc - 0.5, color="black")
    axes[2].set_ylim(-0.2, 1.2)
    plt.title("signal steps (0 = background, 1=signal)")
    axes[2].legend(loc="center left", bbox_to_anchor=(1, 0.5))
    # figs.append(fig)

    fig = create_text_figure(dict(sorted(metadata_dict.items())))
    figs.append(fig)

    dir = os.path.join("data_prep", f"realisation_{realisation.id}")

    out_file_name = get_debug_output_filepath(mngr, dir, f"{i_event}_{evt_class}.pdf")
    PDFPlotCollection(mngr, figs).save_to_file(out_file_name)
