import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from trans_finder.flares.data_prep_helpers.fluctuation import Fluctuation
from trans_finder.flares.data_prep_helpers.physical_flare import PhysicalFlare
from trans_finder.flares.helpers_debugging import get_debug_output_filepath
from trans_finder.flares.utilities.pdf_plot_collection import PDFPlotCollection


def plot_single_realisation_in_one_figure(
    mngr,
    realisation,
    reconstructed_dfs,
    ax,
    mark_plateaus=True,
    mark_injection_channels_separately=True,
    mark_injection_channels_combined=False,
    legend="brief",
    time_range=None,
    add_all_markers=False,
    fixed_y_range=None,
):
    realisation_data = reconstructed_dfs

    time_col_name = "Time [days]"
    y_label = r"log$_{10}$( Flux $[$ photons cm$^{-2}$ s$^{-1}$ ] )"

    norm = mngr.plot_channel_y_normalisation

    selected_channel_dfs = []

    feature_names_by_channel_tag = {}

    for channel_tag, channel_df in realisation_data.items():
        relevant_cols = mngr.data_prep_fluctuation_cols[channel_tag]

        channel_df = channel_df[[mngr.data_prep_day_column_name]].assign(
            **{
                f"{channel_tag}_{col}": channel_df[col] + norm[channel_tag]
                for col in relevant_cols
            }
        )

        selected_channel_dfs.append(channel_df)

        for ft_name in relevant_cols + [mngr.data_prep_day_column_name]:
            if channel_tag not in feature_names_by_channel_tag:
                feature_names_by_channel_tag[channel_tag] = []
            feature_names_by_channel_tag[channel_tag].append(ft_name)

    def get_plot_df(data_df):
        plot_df_list = []
        for channel_tag, ft_names in feature_names_by_channel_tag.items():
            for ft_name in ft_names:
                if ft_name == mngr.data_prep_day_column_name:
                    continue

                full_ft_name = f"{channel_tag}_{ft_name}"
                plot_channel_and_ft_name = (
                    f"log({channel_tag}_{ft_name}) + {norm[channel_tag]}"
                )
                plot_ft_name = f"log({ft_name}) + {norm[channel_tag]}"

                if full_ft_name not in data_df:
                    continue

                plot_df = data_df[[mngr.data_prep_day_column_name]].assign(
                    **{
                        time_col_name: data_df[mngr.data_prep_day_column_name],
                        "channel name": channel_tag,
                        "feature name": plot_ft_name,
                        "ft_name": ft_name,
                        "channel and feature name": plot_channel_and_ft_name,
                        y_label: data_df[full_ft_name],
                    }
                )

                plot_df_list.append(plot_df)
        plot_df = pd.concat(plot_df_list, ignore_index=True).reset_index(drop=True)
        return plot_df

    if time_range is not None:
        selected_channel_dfs = [
            df[df["time"].between(time_range[0], time_range[1])]
            for df in selected_channel_dfs
        ]

    plot_df = pd.concat(
        [get_plot_df(df) for df in selected_channel_dfs], ignore_index=True
    )

    g = sns.lineplot(
        ax=ax,
        data=plot_df,
        x=time_col_name,
        y=y_label,
        style="ft_name",
        hue="channel name",
        markers=add_all_markers,
        legend=legend,
        # dashes="channel and feature name",
        # style="part",
        markersize=5,
    )

    max_time_val = plot_df[time_col_name].max()
    min_time_val = plot_df[time_col_name].min()
    ax.set_xlim(left=min_time_val - 10, right=max_time_val + 10)

    if fixed_y_range:
        ax.set_ylim(*fixed_y_range)

    mark_injections(
        ax,
        realisation,
        mark_plateaus,
        mark_injection_channels_separately,
        mark_injection_channels_combined,
    )

    if legend is not None:
        g.legend(loc="upper left")  # , bbox_to_anchor=(1, 0.5))

    if not add_all_markers:
        # actual number of data points divided by possible number of data points
        sparsity = {
            channel_tag: len(df)
            / (
                (
                    df[mngr.data_prep_day_column_name].max()
                    - df[mngr.data_prep_day_column_name].min()
                )
                / mngr.data_prep_n_days_per_step[channel_tag]
            )
            for channel_tag, df in reconstructed_dfs.items()
        }

        for channel_tag in reconstructed_dfs:
            if sparsity[channel_tag] > 0.5:
                # only add markers if there are less than 50% possible data points
                continue
            # marker = (4, 0, 45)
            marker = "P"

            for idx in get_channel_line_indices(mngr)[channel_tag]:
                ax.lines[idx].set_marker(marker)

    return ax


def get_channel_line_indices(mngr):
    result = {tag: [] for tag in mngr.data_prep_channel_tags}
    line_idx = 0

    for channel_tag, ft_name in mngr.data_step_col_names:
        if ft_name != mngr.tag_temporal_offset:
            result[channel_tag].append(line_idx)
            line_idx += 1

    return result


def plot_single_realisation(mngr, realisation, reconstructed_df: pd.DataFrame):
    realisation_data = reconstructed_df

    fig, axes = plt.subplots(
        nrows=4,
        ncols=1,
        figsize=(20, 3 * 4),
        constrained_layout=True,
        sharex=True,
    )

    ax_idx = 0
    time_col_name = mngr.data_prep_day_column_name
    for channel_tag, channel_df in realisation_data.items():
        channel_df_with_injections = channel_df

        relevant_cols = mngr.data_prep_fluctuation_cols[channel_tag]
        for col in relevant_cols:
            plot_col_name = f"{channel_tag} {col}"
            df = channel_df_with_injections[[time_col_name]].assign(
                **{plot_col_name: channel_df_with_injections[col]}
            )
            sns.scatterplot(
                ax=axes[ax_idx], data=df, x=time_col_name, y=plot_col_name, marker="x"
            )
            # mark_injections(axes[ax_idx], channel_tag, realisation)

            ax_idx += 1

    out_file_name = get_debug_output_filepath(
        mngr, "data_prep", f"realisation_{realisation.id}.pdf"
    )

    PDFPlotCollection(mngr, [fig]).save_to_file(out_file_name)


def plot_realisation_list_overview(mngr, realisations):
    figs = []

    n_realisations = len(realisations)
    fig, axes = plt.subplots(
        nrows=n_realisations,
        ncols=1,
        figsize=(30, 4 * n_realisations),
        constrained_layout=True,
    )

    if n_realisations == 1:
        axes = [axes]

    for i, (ax, realisation) in enumerate(zip(axes, realisations)):
        channel_names = mngr.data_prep_channel_tags

        ax.set_ylim([-0.1, len(channel_names) * 0.1])
        ax.get_yaxis().set_visible(False)

        for channel_tag in realisation.real_times:
            series = realisation.real_times[channel_tag]

            y = np.repeat(channel_names.index(channel_tag) / 10, len(series))
            s = np.repeat(20, len(series))

            ax.scatter(series, y=y, label=channel_tag, marker="|", s=s)

            n_fluctuations = len(
                [t for t in realisation.signal_injections if isinstance(t, Fluctuation)]
            )
            n_flares = len(
                [
                    t
                    for t in realisation.signal_injections
                    if isinstance(t, PhysicalFlare)
                ]
            )

            desc_title = " ".join(
                [
                    f"Realisation {i}",
                    f"{'with randomized offsets' if realisation.has_randomized_real_times else ''}",
                    f"{'with shifted real times' if realisation.has_shifted_real_times else ''}",
                    f"+ {n_fluctuations} fluctuations" if n_fluctuations > 0 else "",
                    f"+ {n_flares} physical flares" if n_flares > 0 else "",
                    f"-> yielding {realisation.get_n_events()} events",
                ]
            )

            ax.set_title(desc_title)
            ax.set_xlabel("real time")

            ax.legend()

        ax.axvline(
            x=realisation.get_first_temporal_injection_start_time(
                mngr=mngr, randomize=False
            ),
            ymin=0,
            ymax=1,
            linestyle="--",
            color="black",
            label=f"{mngr.data_prep.n_steps_shuffling_window} steps in every channel",
        )

        for temporal_injection in realisation.signal_injections:
            if isinstance(temporal_injection, Fluctuation):
                color = "brown"
            elif isinstance(temporal_injection, PhysicalFlare):
                color = "green"
            else:
                color = "grey"

            ax.axvspan(
                temporal_injection.start_time,
                temporal_injection.end_time,
                alpha=0.2,
                color=color,
            )

        for long_term_state in realisation.long_term_states:
            color = "blue"

            ax.axvspan(
                long_term_state.start_time,
                long_term_state.end_time,
                alpha=0.2,
                color=color,
                ymax=0.1,
            )

    fig.suptitle(
        "Randomized data realisations with time intervals of fluctuations marked brown and flares marked green."
    )
    figs.append(fig)

    out_file_name = get_debug_output_filepath(mngr, "data_prep", f"realisations.pdf")

    PDFPlotCollection(mngr, figs).save_to_file(out_file_name)


def mark_injections(
    ax,
    realisation,
    mark_plateaus,
    mark_injection_channels_separately,
    mark_injection_channels_combined,
):
    for channel_tag in realisation.channel_tags:
        if mark_injection_channels_separately:
            for temporal_injection in realisation.signal_injections:
                if channel_tag not in temporal_injection.channel_start_times:
                    continue
                color = "green"
                alpha = 0.1

                ax.axvspan(
                    temporal_injection.channel_start_times[channel_tag],
                    temporal_injection.channel_end_times[channel_tag],
                    alpha=alpha,
                    color=color,
                )

        if mark_plateaus:
            for temporal_injection in realisation.long_term_states:
                color = "blue"
                alpha = 0.025

                ax.axvspan(
                    temporal_injection.channel_start_times[channel_tag],
                    temporal_injection.channel_end_times[channel_tag],
                    alpha=alpha,
                    color=color,
                )

    if mark_injection_channels_combined:
        for temporal_injection in realisation.signal_injections:
            color = "green"
            alpha = 0.1

            ax.axvspan(
                temporal_injection.start_time,
                temporal_injection.end_time,
                alpha=alpha,
                color=color,
            )
