import numpy as np
from matplotlib import cycler
from matplotlib import pyplot as plt
from matplotlib.lines import Line2D

from trans_finder.flares.plot_list.util.plot_config import (
    gridlines_config,
    set_fontsizes,
    tick_params,
)

desc = """
Here, the top panel shows an exemplary input light curve 
comprising two channels, and a time pivot $\tref$. The 
first channel has a dense coverage and thus a short 
window size. The second channel is more sparsely sampled,
 resulting in the window beginning at a previous point in 
 time, where the latest data is available. It also covers 
 a wider time interval, that is required to capture $n+m$ 
 data points. 
The lower panel shows the resulting time-series structure 
separated into the context window $\tau_{enc}$ and search 
window $\tau_{dec}$.
"""


def mark_tref(ax, t_ref: float, hide_box=False, offset=3, y=1.55):
    ax.axvline(
        x=t_ref + 0.5,
        color="black",
        linewidth=1.5,
        # linestyle="dashed",
        ymax=1.0,
    )

    if not hide_box:
        ax.text(
            # x=t_ref,
            # y=-0.1,
            x=t_ref + (offset),
            y=y,
            s=r"$t_{ref}$",
            fontsize=11,
            fontweight="bold",
            verticalalignment="center",
            # horizontalalignment="center",
            transform=ax.get_xaxis_transform(),
            bbox=dict(
                facecolor="white",
                edgecolor="black",
                boxstyle="round,pad=0.4 ",
                # linewidth=2,
            ),
        )


def mark_x_interval(ax, range, facecolor="grey", alpha=0.25, top=True):
    height = 1.0
    ax.axvspan(
        xmin=range[0],
        xmax=range[1],
        alpha=alpha,
        facecolor=facecolor,
        ymin=1 - height if top else 0,
        ymax=1 if top else height,
        lw=0,
    )


def plot_data(ax, x, y, error_up, error_down, data_mask=None):
    point_plot_style = {
        "color": "black",
        "alpha": 0.9,
        "linestyle": "None",
        "marker": "o",
        "linewidth": 0,
        "ecolor": (0.1843137254901961, 0.30980392156862746, 0.30980392156862746, 0.9),
        "elinewidth": 1,
        "capthick": 0,
        "markerfacecolor": "black",
        "markersize": 3,
        "markeredgewidth": 0,
        "fillstyle": "full",
    }

    handle = ax.errorbar(
        x=x if data_mask is None else x[data_mask],
        y=y if data_mask is None else y[data_mask],
        yerr=np.stack(
            (
                error_down if data_mask is None else error_down[data_mask],
                error_up if data_mask is None else error_up[data_mask],
            )
        ),
        # label=mngr.channel_labels[label],
        label="Flux [arbitrary units]",
        **point_plot_style,
    )
    return handle


def mark_encoder_decoder_shading_by_time(ax, t, start=None, end=None, **kwargs):
    mark_x_interval(
        ax,
        [
            (start if start is not None else t[-n_steps_tot] - 0.5),
            t[-n_steps_dec] - 0.5,
        ],
        facecolor="teal",
        alpha=0.1,
        **kwargs,
    )
    mark_x_interval(
        ax,
        [t[-n_steps_dec] - 0.5, (t_ref + 0.5 if end is None else end)],
        facecolor="teal",
        alpha=0.25,
        **kwargs,
    )


def get_offset_weights(t, power_param):
    recent_period = 1 * n_steps_dec

    offsets = np.abs(t - t_ref)
    offsets = offsets - recent_period
    offsets = np.maximum(0, offsets)

    wgt = np.divide(1, 1 + offsets)
    wgt = np.power(wgt, power_param)
    return wgt


if __name__ == "__main__":
    # generate random data for plot
    size_sparse = 40
    size_dense = 120
    y_min = 0
    y_max = 100

    n_steps_tot = 15
    n_steps_enc = 10
    n_steps_dec = 5

    t_ref = size_dense

    np.random.seed(8)
    y_dense = np.random.uniform(y_min, y_max, size_dense)
    y_sparse = np.random.uniform(y_min, y_max, size_dense)

    def create_errors(size):
        return np.random.uniform(20, 60, size)

    e_up_sparse = create_errors(size_sparse)
    e_down_sparse = create_errors(size_sparse)
    e_up_dense = create_errors(size_dense)
    e_down_dense = create_errors(size_dense)

    for array, rho in [
        (y_dense, 0.6),
        (y_sparse, 0.6),
    ]:
        for t in range(1, len(array)):
            array[t] = rho * array[t - 1] + np.random.uniform(1, 100)

    for array, rho in [
        (e_up_sparse, 0.2),
        (e_down_sparse, 0.2),
        (e_up_dense, 0.3),
        (e_down_dense, 0.2),
    ]:
        for t in range(1, len(array)):
            array[t] = rho * array[t - 1] + np.random.uniform(20, 60)

    t_dense = np.arange(size_dense)
    t_sparse = np.random.choice(np.arange(size_dense), size_sparse, replace=False)
    t_sparse.sort()
    t_sparse = t_sparse + (0.5 * (size_dense - np.max(t_sparse)))
    y_sparse = np.random.choice(y_sparse, size_sparse)
    # x_sparse -= 1.5 * (x_max - x_min)

    axes_list = [
        "Light Curve 1",
        "Light Curve 2",
        ".",
        "Window 1",
        "Window 2",
        ".",
        "Input 1",
        "Input 2",
        ".",
    ]
    set_fontsizes()
    fig, axes = plt.subplot_mosaic(
        [[r] for r in axes_list],
        height_ratios=[2, 2, 1.6, 2, 2, 1.6, 2, 2, 0.01],
        # nrows=3,
        # ncols=1,
        # sharex=True,
        sharey=True,
        figsize=(6.5, 8),
        # constrained_layout=True,
        gridspec_kw={
            "hspace": 0.0,
            "wspace": 0.00,
            "left": 0.07,
            "right": 0.97,
            "bottom": 0.18,
            "top": 0.95,
        },
    )

    # configure shared x axes
    axes["Light Curve 1"].sharex(axes["Light Curve 2"])
    axes["Window 1"].sharex(axes["Window 2"])
    axes["Input 1"].sharex(axes["Input 2"])
    for ax_id in ["Light Curve 1", "Window 1", "Input 1"]:
        for label in axes[ax_id].get_xticklabels():
            label.set_visible(False)

    first_plot = "Light Curve 1"
    axes[first_plot].set_xlim(
        [0 - (0.03 * size_dense), size_dense + (0.12 * size_dense)]
    )

    plot_kwargs = dict(marker="o", markersize=3)

    data_handle = plot_data(
        axes["Light Curve 1"], t_dense, y_dense, e_up_dense, e_down_dense
    )
    plot_data(axes["Light Curve 2"], t_sparse, y_sparse, e_up_sparse, e_down_sparse)
    # axes["Light Curve 1"].plot(t_dense, y_dense, **plot_kwargs)
    # axes["Light Curve 2"].plot(t_sparse, y_sparse, **plot_kwargs)

    mark_tref(axes["Light Curve 1"], t_ref, hide_box=True)
    mark_tref(axes["Light Curve 2"], t_ref)
    mark_encoder_decoder_shading_by_time(axes["Light Curve 1"], t_dense)
    mark_encoder_decoder_shading_by_time(axes["Light Curve 2"], t_sparse, top=False)

    # ===========
    # Second panel, isolated window
    plot_data(
        axes["Window 1"],
        t_dense,
        y_dense,
        e_up_dense,
        e_down_dense,
        data_mask=slice(-n_steps_tot, None),
    )
    plot_data(
        axes["Window 2"],
        t_sparse,
        y_sparse,
        e_up_sparse,
        e_down_sparse,
        data_mask=slice(-n_steps_tot, None),
    )

    mark_encoder_decoder_shading_by_time(axes["Window 1"], t_dense[-n_steps_tot:])
    mark_encoder_decoder_shading_by_time(
        axes["Window 2"], t_sparse[-n_steps_tot:], top=False, start=0.1
    )
    axes["Window 2"].set_xlim(
        [t_sparse[-n_steps_tot] - 1, t_ref]
        # [t_sparse[-n_steps_tot] - 1, t_ref + 0.035 * t_ref]
    )  # - 1 for aesthetic reasons, some margin on the left of panel 2

    mark_tref(axes["Window 1"], t_ref, hide_box=True)
    mark_tref(axes["Window 2"], t_ref, hide_box=True)

    # ===========
    # third panel, network inputs
    indices_dense = np.arange(size_dense)
    indices_dense[-n_steps_tot:] = np.arange(n_steps_tot)
    indices_sparse = np.arange(size_sparse)
    indices_sparse[-n_steps_tot:] = np.arange(n_steps_tot)

    plot_data(
        axes["Input 1"],
        indices_dense,
        y_dense,
        e_up_dense,
        e_down_dense,
        data_mask=slice(-n_steps_tot, None),
    )
    plot_data(
        axes["Input 2"],
        indices_sparse,
        y_sparse,
        e_up_sparse,
        e_down_sparse,
        data_mask=slice(-n_steps_tot, None),
    )
    # axes["Input 1"].plot(t_input, y_dense[-n_steps_tot:], **plot_kwargs)
    # axes["Input 2"].plot(t_input, y_sparse[-n_steps_tot:], **plot_kwargs)

    mark_encoder_decoder_shading_by_time(
        axes["Input 1"], indices_dense[-n_steps_tot:], start=-1, end=15
    )
    mark_encoder_decoder_shading_by_time(
        axes["Input 2"],
        indices_sparse[-n_steps_tot:],
        top=False,
        start=-1,
        end=15,
    )
    axes["Input 1"].set_xlim([-0.5, 14.5])

    for name in ["Window 1", "Window 2", "Light Curve 1", "Light Curve 2"]:
        axes[name].grid(**gridlines_config)

    for name in axes:
        axes[name].set_yticks([])
        axes[name].tick_params(**tick_params)

    axes["Light Curve 2"].set_xlabel("Time [days]")
    axes["Window 2"].set_xlabel("Time [days]")
    axes["Input 2"].set_xlabel("Time step in window")

    # fig.supylabel("Normalized Flux + Weights")
    for ax, label in zip(axes.values(), ["C1", "C2", "C1", "C2", "C1", "C2"]):
        if label is not None:
            ax.set_ylabel(label + "      ", rotation=0, ha="center", va="center")

    weight_style_cycler = cycler(
        "color",
        [
            "#5B90DC",
            "#F9AE51",
            # "#0C9FB3",
            # "#E39699",
            # "#3B507D",
            "#FD6989",
            "#794D88",
            # "#8A978E",
            # "#BA2C54",
        ],
    ) + cycler("ls", ["solid", "dashed", "dashdot", "dotted"])

    def plot_weights(ax, x, y, powers):
        ax.set_prop_cycle(weight_style_cycler)

        print(int(np.round(x.min())), int(np.round(x.max())))
        ref_handle = ax.axhline(
            0,
            color="black",
            # xmax=int(np.round(x.max())),
            # xmin=int(np.round(x.min())),
            linestyle=":",
            alpha=0.85,
            label=rf"$\omega = 0$",
            # label=rf"$\omega(t, \eta_{{\mathrm{{decay}}}}) = 0$",
        )
        # ax.axhline(1, color="black", linestyle=":", alpha=0.9)

        for power in powers:
            handle = ax.plot(
                x,
                get_offset_weights(y, power_param=power),
            )
        ax.set_ylim([-0.1, 1.1])
        # ax.set_ylim([-0.0, 1.0])
        ax.set_yticks([])
        ax.tick_params(**tick_params)
        return handle, ref_handle

    # Lambda function for linear interpolation
    interpolate_dense = lambda x: np.interp(
        np.arange(x[0], x[-1] + 0.1, 0.1),  # new x-coordinates
        np.linspace(x[0], x[-1], len(x)),  # original x-coordinates
        x,  # original y-values
    )

    weight_ax = axes["Window 2"].twinx()
    weight_handle, ref_handle = plot_weights(
        weight_ax,
        interpolate_dense(t_sparse[-n_steps_tot:]),
        interpolate_dense(t_sparse[-n_steps_tot:]),
        powers=[1, 0.5, 2],
    )

    plot_weights(
        axes["Window 1"].twinx(),
        interpolate_dense(t_dense[-n_steps_tot:]),
        interpolate_dense(t_dense[-n_steps_tot:]),
        powers=[1, 0.5, 2],
    )

    plot_weights(
        axes["Input 2"].twinx(),
        indices_sparse[-n_steps_tot:],
        t_sparse[-n_steps_tot:],
        powers=[1, 0.5, 2],
    )

    plot_weights(
        axes["Input 1"].twinx(),
        indices_dense[-n_steps_tot:],
        t_dense[-n_steps_tot:],
        powers=[1, 0.5, 2],
    )

    handles, labels = weight_ax.get_legend_handles_labels()
    fig.legend(
        handles=[
            data_handle,
            *(
                Line2D(
                    [],
                    [],
                    **c,
                    label=rf"$\omega(t, \gamma_{{\mathrm{{decay}}}}={power})$",
                )
                for c, power in zip(weight_style_cycler, [1, 0.5, 2])
            ),
        ],
        loc="lower center",
        ncol=2,
        bbox_to_anchor=(0.5, 0.02),
    )

    fig.savefig("figure2.pdf", dpi=300)
    # plt.show()
