import matplotlib.pyplot as plt
import numpy as np

from trans_finder.flares.model_utils.remove_temporal_offsets import (
    remove_temporal_offsets,
)

# def plot_2d_histogram(ax, truth, prediction, bins=50, clip_outliers=0.00):
#     # Define lower and upper quantiles
#     lower_q = np.percentile(np.concatenate([truth, prediction]), 100 * clip_outliers)
#     upper_q = np.percentile(
#         np.concatenate([truth, prediction]), 100 * (1 - clip_outliers)
#     )
#
#     # Clip values
#     truth = np.clip(truth, lower_q, upper_q)
#     prediction = np.clip(prediction, lower_q, upper_q)
#
#     # Define bin edges
#     min_val = min(truth.min(), prediction.min())
#     max_val = max(truth.max(), prediction.max())
#     edges = np.linspace(min_val, max_val, bins)
#
#     # Calculate histogram
#     h, _, _ = np.histogram2d(truth, prediction, bins=[edges, edges])
#
#     ax.imshow(
#         h,
#         interpolation="nearest",
#         origin="lower",
#         extent=[edges[0], edges[-1], edges[0], edges[-1]],
#     )
#
#     ax.set_xlabel("Truth")
#     ax.set_ylabel("Prediction")
#
#     # ax.set_xlim(edges[0], edges[-1])
#     # ax.set_ylim(edges[0], edges[-1])
#
#
# if __name__ == "__main__":
#
#     fig, ax = plt.subplots()
#     truth = np.random.rand(1000)
#     prediction = 1 + np.random.rand(1000)
#     plot_2d_histogram(ax, truth, prediction)
#     plt.show()


def calculate_feature_bins_and_clipping(truth, prediction, outlier_clip, bins):
    n_features = truth.shape[2]
    feature_bin_edges = []
    feature_clipping_values = []

    for dim in range(n_features):
        feature_data = np.concatenate([truth[:, :, dim], prediction[:, :, dim]])

        lower_q, upper_q = np.percentile(
            feature_data, [100 * outlier_clip, 100 * (1 - outlier_clip)]
        )
        min_val, max_val = (
            max(feature_data.min(), lower_q),
            min(feature_data.max(), upper_q),
        )

        feature_bin_edges.append(np.linspace(min_val, max_val, bins))
        feature_clipping_values.append((lower_q, upper_q))

    return feature_bin_edges, feature_clipping_values


def plot_2d_histogram(ax, truth, prediction, bins, clip_values):
    if len(truth) != len(prediction):
        raise ValueError("Truth and prediction arrays must have the same length.")

    truth, prediction = np.clip(truth, *clip_values), np.clip(prediction, *clip_values)
    h, xedges, yedges = np.histogram2d(truth, prediction, bins=[bins, bins])

    extent = [xedges[0], xedges[-1], yedges[0], yedges[-1]]
    ax.imshow(
        h.T,
        cmap="viridis",
        extent=extent,
        origin="lower",
        aspect="equal",
        interpolation="nearest",
    )

    ax.set_xlabel("Truth")
    ax.set_ylabel("Prediction")

    ax.grid(True)


def plot_2dhistogram_of_ndarray_per_step_and_feature(
    mngr,
    truth,
    prediction,
    ft_names,
    outlier_clip,
    title=None,
    hist_kwargs=None,
    skip_offset_indices=False,
    clip_around_zero=None,
):
    if hist_kwargs is None:
        hist_kwargs = {}

    if skip_offset_indices:
        truth = remove_temporal_offsets(mngr, truth)
        prediction = remove_temporal_offsets(mngr, prediction)

        ft_names = [
            n for n in mngr.step_feature_names if mngr.tag_temporal_offset not in n
        ]

    n_time_steps, n_features = truth.shape[1], truth.shape[2]
    feature_bin_edges, feature_clipping_values = calculate_feature_bins_and_clipping(
        truth, prediction, outlier_clip, hist_kwargs.get("bins", 50)
    )

    fig, axes = plt.subplots(
        nrows=n_time_steps,
        ncols=n_features,
        figsize=(4.5 * n_features, 3.5 * n_time_steps),
    )

    if title:
        fig.suptitle(title)
    plt.tight_layout()

    for time_step, axes_row in enumerate(axes):
        for idx, ax in enumerate(axes_row):
            # if skip_offset_indices and idx in mngr.temporal_offset_column_indices:
            #     continue

            # # clip values from large temporal offsets
            # mask = truth[:, time_step, idx] > 0.1

            if clip_around_zero is not None:
                mask = np.abs(truth[:, time_step, idx]) > clip_around_zero
            else:
                mask = np.ones_like(truth[:, time_step, idx], dtype=bool)
            plot_2d_histogram(
                ax,
                truth[:, time_step, idx][mask],
                prediction[:, time_step, idx][mask],
                bins=feature_bin_edges[idx],
                clip_values=feature_clipping_values[idx],
            )
            ax.set_xlabel(ft_names[idx])

    return fig, axes
