import os.path

import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt
from matplotlib.lines import Line2D

from trans_finder.flares.helpers_debugging import get_debug_output_filepath
from trans_finder.flares.plot_list.data_prep_realisation import (
    get_channel_line_indices,
    plot_single_realisation_in_one_figure,
)
from trans_finder.flares.utilities.pdf_plot_collection import PDFPlotCollection


def plot_input_from_network_df(
    mngr,
    network_data_row,
    time_right_edge,
    input_ax_bytime,
    input_ax_bystep,
    legend="brief",
):
    plot_df = pd.DataFrame()

    y_label = r"log$_{10}$( Flux $[$ photons cm$^{-2}$ s$^{-1}$ ] )"

    offset_label = r"Time $-\; t_{ref}$ [days]"

    step_label = "Input step"

    for channel_tag, feature_col in mngr.data_step_col_names:
        if feature_col == mngr.tag_temporal_offset:
            continue

        for step in range(mngr.n_steps_tot):
            plot_df = plot_df.append(
                {
                    offset_label: network_data_row[
                        f"{channel_tag}_t_{step}_{mngr.tag_temporal_offset}"
                    ],
                    y_label: network_data_row[f"{channel_tag}_t_{step}_{feature_col}"]
                    + mngr.plot_channel_y_normalisation[channel_tag],
                    "feature name": feature_col,
                    step_label: step + 1,
                    "channel name": channel_tag,
                    "channel and feature name": f"{channel_tag}_{feature_col}",
                    "input_part": "enc" if step < mngr.n_steps_enc else "dec",
                },
                ignore_index=True,
            )

    g = sns.lineplot(
        ax=input_ax_bytime,
        data=plot_df,
        x=offset_label,
        y=y_label,
        style="channel and feature name",
        hue="channel name",
        markers=True,
        legend=legend,
        # dashes="channel and feature name",
        # style="part",
    )
    # g.legend(loc="center left", bbox_to_anchor=(1, 0.5))

    g = sns.lineplot(
        ax=input_ax_bystep,
        data=plot_df[plot_df["input_part"] == "enc"],
        x=step_label,
        y=y_label,
        style="channel and feature name",
        hue="channel name",
        markers=True,
        legend=legend,
        # dashes="channel and feature name",
        # style="part",
    )

    g = sns.lineplot(
        ax=input_ax_bystep,
        data=plot_df[plot_df["input_part"] == "dec"],
        x=step_label,
        y=y_label,
        style="channel and feature name",
        hue="channel name",
        markers=True,
        legend=legend,
        # dashes="channel and feature name",
        # style="part",
    )

    input_ax_bystep.axvline(x=mngr.n_steps_enc + 1 - 0.5, color="black")
    input_ax_bystep.set_xticks(range(1, mngr.n_steps_tot + 1))
    # plt.xlim(left=-1)

    # g.legend(loc="center left", bbox_to_anchor=(1, 0.5))
    # ax.set_title(title)
    # ax.set_ylim([y_min, y_max])


def add_channel_legend(ax, mngr):
    channel_indices = get_channel_line_indices(mngr)

    all_channels_indices = [
        channel_indices[channel_tag][0] for channel_tag in mngr.data_prep_channel_tags
    ]

    custom_lines = [
        Line2D(
            [0],
            [0],
            color=ax.lines[any_channel_index].get_color(),
            lw=plt.getp(ax.lines[any_channel_index], "linewidth"),
        )
        for any_channel_index in all_channels_indices
    ]

    ax.legend(
        handles=custom_lines,
        # title="Channels",
        loc="upper left",
        labels=mngr.channel_labels,
    )


def plot_inputs_with_sliding_window_context(
    mngr,
    network_data,
    reconstructed_df,
    realisation,
    show_shuffling_window,
    temporal_offset_fixed_min,
    show_signal_injection,
):
    if realisation.id not in mngr.plot_inputs_with_sliding_window_context_conf:
        return

    plot_conf = mngr.plot_inputs_with_sliding_window_context_conf[realisation.id]

    plot_reconstructed_dfs = {
        tag: reconstructed_df[tag][
            (reconstructed_df[tag]["time"] >= plot_conf["context"]["start_time"])
            & (reconstructed_df[tag]["time"] <= plot_conf["context"]["end_time"])
        ]
        for tag in realisation.channel_tags
    }

    for time_right_edge in network_data["time_right_edge"][
        (network_data["time_right_edge"] >= plot_conf["inputs"]["start_time"])
        & (network_data["time_right_edge"] <= plot_conf["inputs"]["end_time"])
    ]:
        input_series = network_data[
            network_data["time_right_edge"] == time_right_edge
        ].iloc[0]

        fermi_start = input_series["fermi_t_0_time"]
        fermi_end = input_series[f"fermi_t_{mngr.n_steps_tot - 1}_time"]
        cta_start = input_series["cta_t_0_time"]
        cta_end = input_series[f"cta_t_{mngr.n_steps_tot - 1}_time"]

        if show_shuffling_window:

            def get_shuffling_window_start(channel_tag, end_time):
                reconstructed_df[channel_tag]["data_point_index"] = range(
                    len(reconstructed_df[channel_tag])
                )
                shuffling_start_idx = int(
                    reconstructed_df[channel_tag].at[end_time, "data_point_index"]
                    - mngr.data_prep.n_steps_shuffling_window
                )

                if shuffling_start_idx < 0:
                    return 0  # reconstructed_df[channel_tag]["time"].iloc[0]

                return reconstructed_df[channel_tag][
                    reconstructed_df[channel_tag]["data_point_index"]
                    == shuffling_start_idx
                ]["time"].iloc[0]

            fermi_shuffling_window_start = get_shuffling_window_start(
                "fermi", fermi_end
            )
            cta_shuffling_window_start = get_shuffling_window_start("cta", cta_end)

        fig, axes = plt.subplots(
            nrows=3,
            ncols=1,
            figsize=(15, 9),
            constrained_layout=True,
            sharey=True,
        )

        context_ax = axes[0]
        input_ax_bytime = axes[1]
        input_ax_bystep = axes[2]

        plot_single_realisation_in_one_figure(
            mngr,
            realisation,
            plot_reconstructed_dfs,
            context_ax,
            mark_plateaus=False,
            mark_injection_channels_separately=False,
            mark_injection_channels_combined=show_signal_injection,
            legend=None,
        )

        add_channel_legend(context_ax, mngr)

        def highlight_interval_line_whiskers(ax, y, xstart, xstop, color):
            """Plot timelines at y from xstart to xstop with given color."""
            delta = 0.1

            ax.axhline(
                y=ax.transAx.transform(y),
                xmin=xstart,
                xmax=xstop,
                color=color,
                lw=1,
                transform=ax.get_xaxis_transform(),
            )
            ax.axvline(
                x=xstart,
                ymin=y + delta,
                ymax=y - delta,
                color=color,
                lw=1,
                transform=ax.get_xaxis_transform(),
            )
            ax.axvline(
                x=xstop,
                ymin=y + delta,
                ymax=y - delta,
                color=color,
                lw=1,
                transform=ax.get_xaxis_transform(),
            )

        # make encoder decoder lines not connect

        t_ref = max(fermi_end, cta_end)
        context_ax.axvline(
            x=t_ref,
            color="black",
            # linewidth=2,
            ymax=1.0,
        )

        context_ax.text(
            # x=t_ref,
            # y=-0.1,
            x=t_ref + 4,
            y=0.9,
            s=r"$t_{ref}$",
            fontsize=11,
            fontweight="bold",
            verticalalignment="center",
            # horizontalalignment="center",
            transform=context_ax.get_xaxis_transform(),
            bbox=dict(facecolor="white", edgecolor="black", boxstyle="round,pad=0.4 "),
        )

        # eps = 0.5
        # context_ax.axvline(
        #     x=max(fermi_end, cta_end) - 0.5 + eps + eps,
        #     color="black",
        #     linewidth=2 * eps,
        # )
        # context_ax.axvline(
        #     x=max(fermi_end, cta_end) - 0.5 - eps + eps,
        #     color="black",
        #     linewidth=2 * eps,
        # )

        context_ax.axvspan(
            xmin=fermi_start,
            xmax=fermi_end,
            alpha=0.25,
            facecolor="blue",
            ymin=0.50,
            ymax=1,
            lw=0,
        )

        context_ax.axvspan(
            xmin=cta_start,
            xmax=cta_end,
            alpha=0.25,
            facecolor="orange",
            ymin=0.0,
            ymax=0.5,
            lw=0,
        )

        if show_shuffling_window:

            def show_with_whiskers(
                ax,
                window_start,
                window_end,
                color,
                y,
                height=0.25,
                alpha=0.5,
                x_width=1.5,
                y_width=0.015,
            ):
                ax.axvspan(
                    xmin=window_start + x_width,
                    xmax=window_end - x_width,
                    alpha=alpha,
                    facecolor=color,
                    ymin=y,
                    ymax=y + y_width,
                    fill=True,
                    lw=0,
                )

                ax.axvspan(
                    xmin=window_start,
                    xmax=window_start + x_width,
                    alpha=alpha,
                    facecolor=color,
                    ymin=y,
                    ymax=y + height,
                    # ymin=y - (height / 2),
                    # ymax=y + (height / 2),
                    fill=True,
                    lw=0,
                )

                ax.axvspan(
                    xmin=window_end - x_width,
                    xmax=window_end,
                    alpha=alpha,
                    facecolor=color,
                    ymin=y,
                    ymax=y + height,
                    fill=True,
                    lw=0,
                )

            show_with_whiskers(
                context_ax,
                fermi_shuffling_window_start,
                fermi_end,
                "blue",
                1,
                height=-0.075,
                y_width=-0.02,
            )
            show_with_whiskers(
                context_ax,
                cta_shuffling_window_start,
                cta_end,
                "orange",
                0.00,
                height=0.075,
                y_width=0.02,
                alpha=0.6,
            )

        out_file_name = get_debug_output_filepath(
            mngr,
            os.path.join(
                "inputs_with_window_context",
                realisation.id,
            ),
            f"{time_right_edge}_input_with_window_context.pdf",
        )

        plot_input_from_network_df(
            mngr,
            input_series,
            time_right_edge,
            input_ax_bytime,
            input_ax_bystep,
            legend=None,
        )

        context_ax.set(ylabel=None)
        input_ax_bystep.set(ylabel=None)

        if temporal_offset_fixed_min is not None:
            limits = input_ax_bytime.get_xlim()
            input_ax_bytime.set_xlim(
                min(limits[0], temporal_offset_fixed_min), limits[1]
            )

        context_ax.lines[2].set_marker(input_ax_bystep.lines[2].get_marker())
        context_ax.lines[3].set_marker(input_ax_bystep.lines[3].get_marker())

        y_center = np.mean(input_ax_bystep.get_ylim())
        input_ax_bystep.text(
            mngr.n_steps_enc + 1 - 0.6667,
            y_center,
            # "\n".join("encoder"),
            "encoder",
            rotation=-90,
            # fontsize=14,
            # fontweight="bold",
            va="center",
            ha="center",
        )

        input_ax_bystep.text(
            mngr.n_steps_enc + 1 - 0.3334,
            y_center,
            # "\n".join("decoder"),
            "decoder",
            rotation=90,
            # fontsize=14,
            # fontweight="bold",
            va="center",
            ha="center",
        )

        PDFPlotCollection(mngr, [fig]).save_to_file(out_file_name)

    pass
