from typing import Dict

import matplotlib.patches as mpatches
import pandas as pd
from matplotlib import pyplot as plt

from trans_finder.flares.helpers_debugging import get_debug_output_filepath
from trans_finder.flares.plot_list.lightcurve_combined_normalized import (
    plot_lightcurve_combined_normalized,
)
from trans_finder.flares.plot_list.lightcurve_in_panels import plot_lightcurve_in_panels
from trans_finder.flares.plot_list.scenario_highlights import get_gridspec_kw
from trans_finder.flares.utilities.pdf_plot_collection import PDFPlotCollection


def plot_lightcurve(
    mngr,
    data: Dict[str, pd.DataFrame],
    name: str,
    output_dir: str,
    event_classes=None,
    plot_combined=False,
    plot_uncertainty=True,
):
    if plot_combined:
        fig, ax = plt.subplots(
            nrows=1,
            ncols=1,
            figsize=(24, max(4.0, 1.33 * len(mngr.data_prep_channel_tags))),
            constrained_layout=True,
            sharex="all",
        )

        plot_lightcurve_combined_normalized(mngr, data, ax)
        extra_annotations(mngr, [ax], event_classes)

        out_file_name = get_debug_output_filepath(
            mngr,
            output_dir,
            f"{name}_lc_combined.pdf",
        )

        PDFPlotCollection(mngr, [fig]).save_to_file(out_file_name)

    # -----
    channels = list(data.keys())
    size_unit = 3 * len(channels)
    fig, axes = plt.subplots(
        nrows=len(channels),
        ncols=1,
        figsize=((size_unit / 9) * 16, size_unit),
        # constrained_layout=True,
        sharex="all",
    )

    plot_lightcurve_in_panels(mngr, data, axes, plot_uncertainty=plot_uncertainty)
    plt.subplots_adjust(**get_gridspec_kw(bottom=0))

    extra_annotations(mngr, axes, event_classes)

    out_file_name = get_debug_output_filepath(
        mngr,
        output_dir,
        f"{name}_lc_panels.pdf",
    )

    PDFPlotCollection(mngr, [fig]).save_to_file(out_file_name)


def get_evt_class_colormap(mngr, present_event_class_list, white_background=True):
    """get a colormap per event class"""
    colormap = {
        evt_class: mngr.plot_lib.get_plot_style(
            index=mngr.unique_evt_classes.index(evt_class),
            tag="main_data",
            style_set="main1",
            has_line=False,
            has_marker=True,
        )["color"]
        for evt_class in mngr.unique_evt_classes
        if evt_class in present_event_class_list
    }

    if white_background:
        if "bck_pure" in colormap:
            colormap["bck_pure"] = "#FFFFFF"
        if "bck_fluct" in colormap:
            colormap["bck_fluct"] = "#FFFFFF"

    return colormap


def extra_annotations(mngr, *args):
    if mngr.run_phase != "prep_predict":
        mark_evt_classes(mngr, *args)


def mark_evt_classes(mngr, axes, evt_classes):
    # mark evt type that are not background (pure or plateau) with a color , if prep_train
    first = True

    # get a colormap per event class

    for ax in axes:
        if evt_classes is not None:
            colormap = get_evt_class_colormap(mngr, evt_classes.unique().tolist())

            start = None
            previous = None
            final_t = evt_classes.index[-1]
            for t, evt_class in evt_classes.items():
                if start is None:
                    start = t
                    previous = evt_class
                    continue

                if evt_class == previous and t != final_t:
                    continue

                # event class has changed or no more data, plot interval up until now

                ax.axvline(
                    start,
                    alpha=0.66,
                    color=colormap[previous],
                    lw=1,
                    linestyle=":",
                )

                ax.axvspan(
                    start,
                    t,
                    alpha=0.95,
                    color=colormap[previous],
                    lw=0.0,
                    ymin=0,
                    ymax=0.066,
                )
                start = t
                previous = evt_class

            if first:
                # Create a Patch object for each label and color and add them to the legend
                handles = []
                for label, color in colormap.items():
                    patch = mpatches.Patch(color=color, label=label)
                    handles.append(patch)
                ax.legend(handles=handles)

        first = False
