from typing import Dict

import matplotlib.patheffects as pe
import numpy as np
import pandas as pd
from matplotlib import colors
from matplotlib import pyplot as plt

from trans_finder.flares.plot_list.util.plot_config import (
    DEFAULT_SCALE,
    MEDIUM_SIZE,
    SMALL_SIZE,
    gridlines_config,
)

DEFAULT_YLABELCOORDS = (-0.085, 0.5)


def transform(data_space):
    return np.squeeze(data_space.to_numpy())


def plot_lightcurve_in_panels(
    mngr,
    reconstructed_dfs: Dict[str, pd.DataFrame],
    axes,
    plot_conf=None,
    plot_uncertainty=True,
    cut_masks=None,
    markersize=4,
    markeredgewidth=0.8,
    elinewidth=0.5,
):
    """cut_masks is a boolean mask per channel that marks data points that have been cut out in the
    manual background cut.
    """
    if plot_conf is None:
        plot_conf = {}

    marker_groups = list(reconstructed_dfs.keys())

    for ax, (channel_tag, channel_df) in zip(axes, reconstructed_dfs.items()):
        plot_style = mngr.plot_lib.get_plot_style(
            index=marker_groups.index(channel_tag),
            tag="main_data",
            style_set="main_0",
            has_line=not plot_uncertainty,
            has_marker=True,
        )

        common_style = {
            "color": "black",
            "alpha": 0.9,
            "linewidth": 0 if plot_uncertainty else 1,
            "markersize": markersize,
            "markeredgewidth": 0,
            "fillstyle": "full",
        }

        if plot_uncertainty:
            plot_style.update(
                dict(
                    ecolor=colors.to_rgba("darkslategray", 0.9),
                    elinewidth=elinewidth,
                    capthick=0,
                    marker="o",
                    markerfacecolor="black",
                )
            )
        else:
            common_style.update(
                dict(
                    linestyle="-",
                    marker="o",
                    markersize=2 * markersize,
                    markerfacecolor="black",
                    color=colors.to_rgba("darkslategray", 0.9),
                )
            )

        plot_style.update(common_style)

        y = transform(channel_df[mngr.data_prep_error_feature_cols[channel_tag]])
        x = transform(channel_df[mngr.data_prep_day_column_name])

        # Adjust y scale in exponential format
        offset_exp = plot_conf.get("y_offsets_exponential", {}).get(channel_tag, None)
        if offset_exp is not None:
            y *= offset_exp

        label = channel_tag

        if cut_masks is None:
            groups = [(None, {})]
        else:
            groups = [
                (~cut_masks[channel_tag], {}),
                (
                    cut_masks[channel_tag],
                    dict(
                        marker="x",
                        markerfacecolor="red",
                        markeredgecolor="red",
                        markersize=markersize,
                        # linewidth=1,
                        markeredgewidth=markeredgewidth,
                        ecolor=colors.to_rgba("firebrick", 0.9),
                    ),
                ),
            ]

        for mask, group_plot_style in groups:
            style = plot_style.copy()
            style.update(group_plot_style)

            if plot_uncertainty:
                error_up = transform(
                    channel_df[mngr.data_prep_error_up_cols[channel_tag]]
                )
                error_down = transform(
                    channel_df[mngr.data_prep_error_down_cols[channel_tag]]
                )
                yerr = np.stack((error_down, error_up))
                ax.errorbar(
                    x=x[mask] if mask is not None else x,
                    y=y[mask] if mask is not None else y,
                    yerr=yerr[:, mask] if mask is not None else yerr,
                    label=mngr.channel_labels[label],
                    **style,
                )
            else:
                ax.plot(
                    x[mask] if mask is not None else x,
                    y[mask] if mask is not None else y,
                    label=label,
                    **style,
                )

        ax.grid(**gridlines_config)
        # ax.tick_params(**tick_params)

        unit: str = mngr.channel_units[channel_tag].strip()

        # add channel name label
        text = ax.text(
            *(0.019, 0.8),
            (
                f"{mngr.channel_labels[label]}"
                + (f" $[${unit}$]$" if len(unit) > 0 else "")
            ),
            transform=ax.transAxes,
            fontsize=SMALL_SIZE * DEFAULT_SCALE,
            verticalalignment="center",
            horizontalalignment="left",
            alpha=0.9,
            path_effects=[
                pe.withStroke(linewidth=0.75, foreground="white", alpha=0.75)
            ],
        )

        if not plot_conf.get("hide_data_ylabel", {}).get(channel_tag, False):
            y_label = mngr.channel_y_labels[channel_tag]

            ax.set_ylabel(
                y_label,
                fontsize=DEFAULT_SCALE * MEDIUM_SIZE,
            )
            ax.yaxis.set_label_coords(
                *plot_conf.get("ylabel_coords", DEFAULT_YLABELCOORDS)
            )
        # ax.tick_params(axis="y", which="major", pad=DEFAULT_TICKPAD)
        # ax.yaxis.labelpad = plot_conf.get("label_pad", DEFAULT_LABELPAD)

        if plot_conf.get("hide_data_yticklabel", {}).get(channel_tag, False):
            # hide y tick label
            ax.set_yticklabels([])

        fixed_ylim_range = plot_conf.get("data_ylims", dict()).get(channel_tag, None)
        if fixed_ylim_range is not None:
            ax.set_ylim(*fixed_ylim_range)

            if (
                plot_conf.get("annotate_outlier", False)
                and np.max(y) > fixed_ylim_range[1]
            ):
                # annotation in case a data point is above the upper edge of the plot
                ydata = np.max(y)
                xdata = x[np.argmax(y)]

                ax.annotate(
                    f"Outlier with {mngr.channel_y_labels[channel_tag].lower().strip('s')}={ydata:.1f} at\nMJD {55170 + xdata:,.0f} is not shown.",
                    (xdata, fixed_ylim_range[1]),
                    xytext=(
                        xdata + plot_conf.get("annotate_outlier_xoffset", 0),
                        fixed_ylim_range[1]
                        - 0.40 * (fixed_ylim_range[1] - fixed_ylim_range[0]),
                    ),
                    textcoords="data",
                    bbox=dict(boxstyle="round", fc=(1, 1, 1, 0.9)),
                    arrowprops=dict(
                        arrowstyle="wedge,tail_width=0.3",
                        connectionstyle="angle3,angleA=0,angleB=-90",
                        fc=(0.1, 0.1, 0.1),
                        ec="none",
                    ),
                    fontsize=DEFAULT_SCALE * SMALL_SIZE,
                )

        else:
            ylim = ax.get_ylim()
            extra_y = plot_conf.get("extend_y_range_upwards_by", 0.0)
            ax.set_ylim(
                (
                    ylim[0],
                    ylim[1] + (extra_y * (ylim[1] - ylim[0])),
                )
            )

        fixed_yticks = plot_conf.get("data_yticks", dict()).get(channel_tag, None)
        if fixed_yticks is not None:
            ax.set_yticks(fixed_yticks)

    plt.xlabel(
        plot_conf.get("x_label", "Time [days]"), fontsize=DEFAULT_SCALE * MEDIUM_SIZE
    )
