import contextlib

import matplotlib.pyplot as plt
import numpy as np

from trans_finder.flares.plot_list.util.plot_config import gridlines_config, tick_params


def violin_plot_grid(
    data,
    labels,
    data_desc,
    sharey=True,
):
    # Create a figure and a set of subplots, one for each feature
    fig, axs = plt.subplots(
        data.shape[2],
        1,
        figsize=(1.5 * data.shape[1], 3.5 * data.shape[2]),
        sharey=sharey,
    )

    # Calculate the global y-axis limits if sharing y-axis
    if sharey:
        ymin = np.min(data) - np.ptp(data) * 0.1
        ymax = np.max(data) + np.ptp(data) * 0.1

    # Iterate over each feature
    for feature_index in range(data.shape[2]):
        with contextlib.suppress(FloatingPointError):
            # Create violin plot for each time step
            parts = axs[feature_index].violinplot(
                [
                    data[:, time_step, feature_index]
                    for time_step in range(data.shape[1])
                ],
                showmedians=True,
            )

            # Color the body of the violin plot
            for pc in parts["bodies"]:
                pc.set_facecolor("blue")
                pc.set_edgecolor("black")
                pc.set_alpha(0.7)

            # Color the median, 2 sigma quantiles and max
            parts["cmedians"].set_edgecolor("red")
            parts["cmins"].set_edgecolor("green")
            parts["cmaxes"].set_edgecolor("orange")

            # Extend the y-axis limits of the plot by 10% up and down
            if not sharey:
                ymin, ymax = axs[feature_index].get_ylim()
                yrange = ymax - ymin
                extension = yrange * 0.1
                axs[feature_index].set_ylim(ymin - extension, ymax + extension)

            # Add a label just above each maximum and below each minimum
            for time_step in range(data.shape[1]):
                max_value = np.max(data[:, time_step, feature_index])
                axs[feature_index].text(
                    time_step + 1,
                    max_value + (ymax - ymin) * 0.01,
                    f"{max_value:.2e}",
                    ha="center",
                    va="bottom",
                )

                min_value = np.min(data[:, time_step, feature_index])
                axs[feature_index].text(
                    time_step + 1,
                    min_value - (ymax - ymin) * 0.01,
                    f"{min_value:.2e}",
                    ha="center",
                    va="top",
                )
        # Set x-ticks and labels
        axs[feature_index].set_xticks(np.arange(1, data.shape[1] + 1))
        axs[feature_index].set_xticklabels(
            [f"step {i}" for i in range(1, data.shape[1] + 1)]
        )

        # Adding other labels
        axs[feature_index].set_ylabel(data_desc)
        axs[feature_index].set_title(labels[feature_index])

        # Call the provided functions
        axs[feature_index].grid(**gridlines_config)
        axs[feature_index].tick_params(**tick_params)
        axs[feature_index].set_axisbelow(True)

    return fig, axs


if __name__ == "__main__":
    # Example usage
    actual = np.random.rand(100, 3, 5)
    predicted = np.random.rand(100, 3, 5)
    labels = ["feature1", "feature2", "feature3", "feature4", "feature5"]
    # Calculate residuals
    residuals = actual - predicted

    violin_plot_grid(residuals, labels, "Residuals")

    # Adjust the layout
    plt.tight_layout()
    # Show the plot
    plt.show()
