import itertools

import matplotlib.patheffects as pe
import numpy as np
from matplotlib import pyplot as plt

from trans_finder.flares.core.tags import (
    TAG_INPUT_MASK,
    TAG_INPUT_MASK_NONE,
    TAG_LOWERQUANT,
    TAG_MEDIAN,
    TAG_SIGNIFICANCE_CLUSTER_SPACE,
    TAG_SIGNIFICANCE_COMBINED,
    TAG_TS_CLUSTER_SPACE,
    TAG_UPPERQUANT,
)
from trans_finder.flares.helpers_debugging import get_debug_output_filepath
from trans_finder.flares.plot_list.channel_group_results import (
    plot_channel_group_metrics,
)
from trans_finder.flares.plot_list.lightcurve_in_panels import (
    DEFAULT_YLABELCOORDS,
    plot_lightcurve_in_panels,
)
from trans_finder.flares.plot_list.util.plot_config import (
    COLORS_QUANTITATIVE_DARK,
    DEFAULT_SCALE,
    MEDIUM_SIZE,
    SMALL_SIZE,
    gridlines_config,
    tick_params,
)
from trans_finder.flares.utilities.pdf_plot_collection import PDFPlotCollection

highlight_default_figwidth = 6


def get_gridspec_kw(bottom=None, left=None, right=None):
    if left is None:
        left = 0.12

    if bottom is None:
        bottom = 0.1

    if right is None:
        right = 0.985

    return {
        "hspace": 0.00,
        "wspace": 0.00,
        "left": left,
        "right": right,
        "bottom": bottom,
        "top": 0.98,
    }


def do_plot_scenario_highlights(mngr, realisation, reconstructed_df, output_dir):
    if mngr.run_phase == "prep_predict":
        # inference data
        for highlight_name, plot_conf in mngr.infer_plot_highlights.items():
            plot_scenario_highlights_lightcurve(
                mngr,
                plot_conf,
                highlight_name,
                realisation,
                reconstructed_df,
                output_dir,
            )

    elif (
        mngr.run_phase == "create_custom_realisations"
        and realisation.id in mngr.plot_scenario_highlights
    ):
        plot_conf = mngr.plot_scenario_highlights[realisation.id]

        plot_scenario_highlights_lightcurve(
            mngr,
            plot_conf,
            realisation.id,
            realisation,
            reconstructed_df,
            output_dir,
        )


def get_highlight_default_figwidth(mngr):
    return 0.8 * len(mngr.data_prep_channel_tags)


def show_cuts_and_get_mask(mngr, ax, plot_conf, channel_tag, channel_df):
    time_col = mngr.data_prep_day_column_name
    times = channel_df[time_col].values
    mask_is_background_manual = np.zeros_like(times).astype(bool)

    # horizontal cuts (in time)
    try:
        cut_data = mngr.dataset_metadata["channel_excluded_times"][channel_tag]
        cut_data = [v for v in cut_data.values()]
        cut_data = list(zip(*cut_data))
    except:
        print("No channel_excluded_times to mark in channel " + channel_tag)
        cut_data = []

    for start, end in cut_data:
        mask = (times >= start - mngr.data_prep.data_prep_analysis_min_time) & (
            times <= end - mngr.data_prep.data_prep_analysis_min_time
        )
        mask_is_background_manual = mask_is_background_manual | mask
        # ax.axvspan(
        #     start - mngr.data_prep.data_prep_analysis_min_time,
        #     end - mngr.data_prep.data_prep_analysis_min_time,
        #     alpha=0.25,
        #     color="grey",
        #     lw=0,
        #     ymin=0,
        #     ymax=1,
        # )

    # vertical cuts (by value threshold)
    try:
        vertical_cut = mngr.dataset_metadata["channel_training_horizontal_cutoffs"][
            channel_tag
        ]
        is_upper_threshold = (
            channel_tag in mngr.dataset_metadata["channel_scale_is_inverted"]
        )
    except Exception as e:
        print(
            "No valid channel_training_horizontal_cutoffs found to mark in channel "
            + channel_tag,
            e,
        )

        return mask_is_background_manual

        # ax.axhspan(
        #     ax.get_ylim()[0] if is_upper_threshold else vertical_cut,
        #     vertical_cut if is_upper_threshold else ax.get_ylim()[1],
        #     alpha=0.25,
        #     color="grey",
        #     lw=0,
        #     xmin=0,
        #     xmax=1,
        # )

    ax.axhline(vertical_cut, color="grey", ls="--", xmin=0, xmax=1, lw=0.8)
    values = channel_df[mngr.data_prep_error_feature_cols[channel_tag]]
    mask = (values >= vertical_cut) if is_upper_threshold else (values <= vertical_cut)

    # any feature outside the range is enough to mark the point as background
    mask_is_background_manual = mask_is_background_manual | ~mask.any(axis=1).values

    return mask_is_background_manual.astype(bool)


def show_zoom_in_indicators(mngr, ax, plot_conf):
    time_ranges = []
    zoom_in_names = plot_conf.get("indicate_zoom_scenarios", [])
    for name in zoom_in_names:
        zoom_conf = mngr.infer_plot_highlights[name]
        time_range = (
            zoom_conf["time_to_highlight"] - zoom_conf["time_context_width"],
            zoom_conf["time_to_highlight"] + zoom_conf["time_context_width"],
        )
        time_ranges.append(time_range)

    time_ranges.extend(plot_conf.get("highlight_time_ranges", []))

    for time_range in time_ranges:
        ax.axvspan(
            time_range[0],
            time_range[1],
            alpha=0.25,
            color="grey",
            lw=0,
            ymin=0,
            ymax=1,
        )

    pass


def plot_scenario_highlights_lightcurve(
    mngr, plot_conf, highlight_name, realisation, reconstructed_df, output_dir
):
    channel_groups_to_skip = plot_conf.get("channel_groups_to_skip", [])

    channels_to_skip = set(
        itertools.chain.from_iterable(
            mngr.channel_groups[g]
            for g in channel_groups_to_skip
            if g in mngr.channel_groups.keys()
        )
    )
    # n_existing_and_skipped = len([c in reconstructed_df for c in channels_to_skip])

    reconstructed_df = {
        k: v for k, v in reconstructed_df.items() if k not in channels_to_skip
    }

    figsize = (
        tuple(plot_conf["figsize"])
        if "figsize" in plot_conf
        else (highlight_default_figwidth, get_highlight_default_figwidth(mngr))
    )

    fig, axes = plt.subplots(
        nrows=len(reconstructed_df),  # - n_existing_and_skipped,
        ncols=1,
        figsize=figsize,
        sharex="all",
        constrained_layout=False,
        gridspec_kw=get_gridspec_kw(
            bottom=0,
            left=plot_conf.get("gridspec_left", None),
            right=plot_conf.get("gridspec_right", None),
        ),
        # gridspec_kw=gridspec_kw,
    )

    for ax in axes:
        ax.grid(**gridlines_config)
        ax.tick_params(**tick_params)
        ax.yaxis.set_label_coords(-0.09, 0.5)

    time_range = (
        plot_conf["time_to_highlight"] - plot_conf["time_context_width"],
        plot_conf["time_to_highlight"] + plot_conf["time_context_width"],
    )

    fixed_y_range = (
        tuple(plot_conf["fixed_y_range"]) if "fixed_y_range" in plot_conf else None
    )
    if fixed_y_range:
        for ax in axes:
            ax.set_ylim(*fixed_y_range)

    cut_masks = {}

    for channel_tag, ax in zip(reconstructed_df.keys(), axes):
        show_zoom_in_indicators(mngr, ax, plot_conf)

        cut_masks[channel_tag] = show_cuts_and_get_mask(
            mngr, ax, plot_conf, channel_tag, reconstructed_df[channel_tag]
        )

    plot_lightcurve_in_panels(
        mngr,
        reconstructed_df,
        axes,
        plot_conf=plot_conf,
        cut_masks=cut_masks,
        markersize=plot_conf.get("markersize", 3),
        markeredgewidth=plot_conf.get("markeredgewidth", 1.2),
        elinewidth=plot_conf.get("elinewidth", 1.0),
    )

    ax.set_xlim(time_range[0], time_range[1])

    for channel_tag, ax in zip(reconstructed_df.keys(), axes):
        try:
            if channel_tag in mngr.dataset_metadata.get(
                "channel_scale_is_inverted", []
            ):
                ax.invert_yaxis()
        except Exception as e:
            print("Could not read dataset_metadata.channel_scale_is_inverted ", e)
            pass

    filename = plot_conf.get("filename", highlight_name)

    out_file_name = get_debug_output_filepath(
        mngr,
        output_dir,
        f"{filename}_lc.pdf",
    )

    PDFPlotCollection(mngr, [fig]).save_to_file(out_file_name)


def inside_label(mngr, ax, text):
    # add channel name label
    text = ax.text(
        *(0.019, 0.9),
        text,
        transform=ax.transAxes,
        fontsize=SMALL_SIZE * DEFAULT_SCALE,
        verticalalignment="center",
        horizontalalignment="left",
        alpha=0.9,
        # bbox=dict(
        #     boxstyle="round",
        #     facecolor="white",
        #     alpha=0.5,
        #     edgecolor="black",
        #     # color="black",
        # ),
        path_effects=[pe.withStroke(linewidth=0.75, foreground="white", alpha=0.75)],
    )


def plot_highlight_results(mngr, df_out, output_dir, highlight_name, plot_conf=None):
    if plot_conf is None:
        plot_conf = {}

    plot_kwargs = plot_conf.get("plot_kwargs", {})

    metric_height = (
        plot_conf["metric_height"] if ("metric_height" in plot_conf) else 2.5 * 2
    )
    if "figsize" in plot_conf:
        figsize = (
            plot_conf["figsize"][0],
            metric_height,
        )

    else:
        figsize = (highlight_default_figwidth, metric_height)

    # fig, ax = plt.subplots(
    #     nrows=1,
    #     ncols=1,
    #     figsize=figsize,
    #     constrained_layout=True,
    # )

    if "time_to_highlight" in plot_conf and "time_context_width" in plot_conf:
        time_range = (
            plot_conf["time_to_highlight"] - plot_conf["time_context_width"],
            plot_conf["time_to_highlight"] + plot_conf["time_context_width"],
        )

    else:
        time_range = (df_out.time_left_edge.min(), df_out.time_right_edge.max())

    # df_cut = df_cut[
    #     (df_cut["time_right_edge"] > time_range[0])
    #     & (df_cut["time_right_edge"] < time_range[1])
    # ]

    df_cut = df_out[df_out[TAG_INPUT_MASK] == TAG_INPUT_MASK_NONE]

    fig, axes = plt.subplots(
        nrows=2,
        ncols=1,
        figsize=figsize,
        constrained_layout=False,
        sharex=True,
        gridspec_kw=get_gridspec_kw(
            left=plot_conf.get("gridspec_left", None),
            right=plot_conf.get("gridspec_right", None),
            bottom=plot_conf.get("gridspec_bottom", None),
        ),
    )

    for ax in axes:
        ax.grid(**gridlines_config)
        ax.tick_params(**tick_params)
        # print(time_range)
        # diff = time_range[1] - time_range[0]
        # ax.set_xlim(time_range[0], time_range[1] + diff * 0.05)

        show_zoom_in_indicators(mngr, ax, plot_conf)

    axes[0].set_xlim(time_range[0], time_range[1])
    axes[0].tick_params(labelbottom=False, bottom=True)

    significance_plot_kwargs = plot_kwargs.copy()
    if "show_markers" in plot_conf and plot_conf["show_markers"]:
        significance_plot_kwargs.update(
            dict(
                linestyle="-",
                marker="x",
                markersize=4,
                markerfacecolor="none",
                # color=colors.to_rgba("darkslategray", 0.9),
            )
        )

    fill_kwargs = plot_kwargs.copy()
    fill_kwargs["linewidth"] = 0
    if "kw" in fill_kwargs:
        del fill_kwargs["lw"]

    axes[0].plot(
        df_cut["time_right_edge"].to_numpy(),
        df_cut["_".join([TAG_SIGNIFICANCE_COMBINED, TAG_MEDIAN])].to_numpy(),
        **significance_plot_kwargs,
        label=mngr.label_significance_combo,
        alpha=0.9,
        color=COLORS_QUANTITATIVE_DARK[0],
    )
    axes[0].fill_between(
        df_cut["time_right_edge"].to_numpy(),
        df_cut["_".join([TAG_SIGNIFICANCE_COMBINED, TAG_LOWERQUANT])].to_numpy(),
        df_cut["_".join([TAG_SIGNIFICANCE_COMBINED, TAG_UPPERQUANT])].to_numpy(),
        alpha=0.25,  # Adjust transparency of the shaded region
        color=COLORS_QUANTITATIVE_DARK[0],
        # color='gray',  # Adjust the color of the shaded region
        **fill_kwargs,
    )

    # significance over time
    axes[0].plot(
        df_cut["time_right_edge"].to_numpy(),
        df_cut["_".join([TAG_SIGNIFICANCE_CLUSTER_SPACE, TAG_MEDIAN])].to_numpy(),
        "--",
        alpha=0.9,
        label=mngr.label_significance_mm,
        color=COLORS_QUANTITATIVE_DARK[2],
        **significance_plot_kwargs,
    )

    axes[0].fill_between(
        df_cut["time_right_edge"].to_numpy(),
        df_cut["_".join([TAG_SIGNIFICANCE_CLUSTER_SPACE, TAG_LOWERQUANT])].to_numpy(),
        df_cut["_".join([TAG_SIGNIFICANCE_CLUSTER_SPACE, TAG_UPPERQUANT])].to_numpy(),
        color=COLORS_QUANTITATIVE_DARK[2],
        alpha=0.25,  # Adjust transparency of the shaded region
        # color='gray',  # Adjust the color of the shaded region
        **fill_kwargs,
    )
    # axes[0].set_ylabel("Significance")
    # inside_label(mngr, axes[0], "Significance")
    if not plot_conf.get("hide_result_ylabel", False):
        axes[0].set_ylabel(
            "Significance",
            fontsize=DEFAULT_SCALE * MEDIUM_SIZE,
        )
        axes[1].set_ylabel(
            "Significance",
            fontsize=DEFAULT_SCALE * MEDIUM_SIZE,
        )

    leg = axes[0].legend(
        loc=plot_conf.get("legend_position", "upper center"),
        ncol=plot_conf.get("legend_n_cols", 3),
        alignment="right",
        bbox_to_anchor=plot_conf.get("legend_bbox_to_anchor", None),
        prop={"size": plot_conf.get("legend_fontsize", DEFAULT_SCALE * SMALL_SIZE)},
    )

    # set the linewidth of each legend object
    for legobj in leg.legendHandles:
        legobj.set_linewidth(2.0)

    # channel masking metrics
    channel_groups_to_skip = plot_conf.get("channel_groups_to_skip", [])
    plot_channel_group_metrics(
        mngr,
        axes[1],
        df_out,
        ts_col_name=TAG_TS_CLUSTER_SPACE,
        channel_groups_to_skip=channel_groups_to_skip,
        plot_kwargs=plot_kwargs,
        plot_conf=plot_conf,
    )

    plt.xlabel(
        plot_conf.get("x_label", mngr.time_label), fontsize=DEFAULT_SCALE * MEDIUM_SIZE
    )

    for ax in axes.flat:
        ax.set_ylim(bottom=0)
        ax.tick_params(axis="y", which="major")
        # ax.yaxis.labelpad = plot_conf.get("labelpad_results", DEFAULT_LABELPAD)
        ax.yaxis.set_label_coords(*plot_conf.get("ylabel_coords", DEFAULT_YLABELCOORDS))

        # ax.yaxis.labelpad = plot_conf.get("label_pad", DEFAULT_LABELPAD)

    fixed_ylim_top = plot_conf.get("fixed_ylim_top", None)

    if fixed_ylim_top is not None:
        axes[0].set_ylim(top=fixed_ylim_top)
        axes[1].set_ylim(top=fixed_ylim_top)
    else:
        axes[1].set_ylim(top=max(axes[1].get_ylim()[1], axes[0].get_ylim()[1]))

    # for ax in axes:
    #     ax.yaxis.set_label_coords(-0.025, 0.5)

    # fig.align_ylabels()

    out_file_name = get_debug_output_filepath(
        mngr,
        output_dir,
        f"{highlight_name}_results.pdf",
    )

    PDFPlotCollection(mngr, [fig]).save_to_file(out_file_name)


def do_plot_scenario_highlights_results(
    mngr, df_diagnostics, realisation_id, output_dir
):
    if mngr.dir_input_predict == mngr.dir_output_data_prep_predict:
        # inference data
        for highlight_name, plot_conf in mngr.infer_plot_highlights.items():
            plot_highlight_results(
                mngr, df_diagnostics, output_dir, highlight_name, plot_conf
            )

    elif realisation_id in mngr.plot_scenario_highlights:
        # any other case
        plot_conf = mngr.plot_scenario_highlights[realisation_id]
        df_cut = df_diagnostics[df_diagnostics["realisation_id"] == realisation_id]
        plot_highlight_results(mngr, df_cut, output_dir, realisation_id, plot_conf)
