import numpy as np
from matplotlib import pyplot as plt
from scipy.special import erf

from trans_finder.flares.core.tags import TAG_MEDIAN, TAG_SIGNIFICANCE_CLUSTER_SPACE
from trans_finder.flares.helpers_debugging import get_debug_output_filepath
from trans_finder.flares.plot_list.util.plot_config import gridlines_config, tick_params
from trans_finder.flares.utilities.pdf_plot_collection import PDFPlotCollection


def plot_significance_distribution_per_realisation(
    mngr, df_diagnostics, realisation_desc
):
    for realisation_id in df_diagnostics.realisation_id.unique():
        realisation_df = df_diagnostics[df_diagnostics.realisation_id == realisation_id]

        plot_significance_distribution(
            mngr, realisation_df, f"{realisation_desc}_{realisation_id}"
        )


def plot_significance_distribution(mngr, df_diagnostics, data_desc):
    significance_col = "_".join([TAG_SIGNIFICANCE_CLUSTER_SPACE, TAG_MEDIAN])

    significance = df_diagnostics[significance_col]

    fig_pdf, ax = plt.subplots(
        nrows=1,
        ncols=1,
    )

    expectation_plot_style = mngr.plot_lib.get_plot_style(
        index=1,
        tag="main_data",
        style_set="autumn",
    )

    measurement_plot_style = mngr.plot_lib.get_plot_style(
        index=0,
        tag="main_data",
        style_set="autumn",
    )

    n_bins = min(200, len(significance))

    x = np.linspace(min(significance), max(significance), n_bins + 1)

    significance_hist, _ = np.histogram(
        significance,
        bins=x,
    )

    measures = np.array(
        [len(significance[significance >= x]) / len(significance) for x in x]
    )
    ax.plot(
        x,
        measures,
        label=r"ratio of $\sigma \geq x$",
        **measurement_plot_style,
    )

    # theoretical_max = 1.0 / len(significance)
    # plt.axhline(
    #     y=theoretical_max,
    #     c=expectation_plot_style["color"],
    #     linewidth=2,
    #     linestyle=":",
    # )
    # plt.text(
    #     0.4,
    #     theoretical_max,
    #     f"theoretical: 1 out of n={len(significance)} events",
    #     fontsize=11,
    #     va="center",
    #     ha="center",
    #     backgroundcolor="w",
    #     transform=ax.get_yaxis_transform(),
    # )

    # Expected fraction of population outside range
    # (https://en.wikipedia.org/wiki/68–95–99.7_rule)
    y_ref = 1 - erf(x / np.sqrt(2))
    # y_ref = 1 - norm.cdf(x)
    ax.plot(
        x,
        y_ref,
        label=r"$1 - \operatorname {erf} \left({\frac {x}{\sqrt {2}}}\right)$",
        **expectation_plot_style,
    )

    ax.set_xlabel(r"$\sigma$")
    ax.set_title(f"{data_desc} (n={len(significance)} events)")
    ax.legend()
    ax.grid(**gridlines_config)
    ax.tick_params(**tick_params)
    ax.set_yscale("log")

    # ----
    #
    # fig_pdf_division, ax = plt.subplots(
    #     nrows=1,
    #     ncols=1,
    # )
    #
    # ax.plot(
    #     x,
    #     measures / y_ref,
    #     label="PDF measured",
    #     **measurement_plot_style,
    # )
    #
    # ax.plot(
    #     x,
    #     np.ones_like(x),
    #     label=r"Expected fraction",
    #     **expectation_plot_style,
    # )
    #
    # ax.set_xlabel(r"$\sigma$")
    # ax.set_title(data_desc)
    # ax.legend()
    # ax.grid(**gridlines_config)
    # ax.tick_params(**tick_params)
    # # ax.set_yscale("log")

    PDFPlotCollection(mngr, [fig_pdf]).save_to_file(
        get_debug_output_filepath(mngr, significance_col + "_dist", f"{data_desc}.pdf")
    )
