import os

import matplotlib.patheffects as pe
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt

from trans_finder.flares.core.tags import (
    TAG_INPUT_MASK,
    TAG_INPUT_MASK_NONE,
    TAG_SIGNIFICANCE_CLUSTER_SPACE,
    TAG_SIGNIFICANCE_COMBINED,
    TAG_SIGNIFICANCE_RECO_ERROR,
)
from trans_finder.flares.helpers_debugging import get_debug_output_filepath
from trans_finder.flares.plot_list.label_cleanup import clean_label
from trans_finder.flares.plot_list.util.plot_config import (
    COLORS_HIGH_CONTRAST,
    DEFAULT_SCALE,
    MEDIUM_SIZE,
    SMALL_SIZE,
    gridlines_config,
    tick_params,
)
from trans_finder.flares.state_manager import StateManager
from trans_finder.flares.utilities.pdf_plot_collection import PDFPlotCollection
from trans_finder.flares.utilities.tensorboard_utills import tensorboard_log_figure

template_x_label = r"Flux [ TeV$^{-1}$ cm$^{-2}$ s$^{-1}$ ]"
FIXED_Y_RANGE = [-1.5, 21]
GROUP_COLORS = [COLORS_HIGH_CONTRAST[2], COLORS_HIGH_CONTRAST[0], "#228833"]


def get_scale_1_flux(mngr):
    scale_1_template_path = os.path.join(
        mngr.tag_input_dir,
        mngr.dir_input_flare_templates,
        mngr.flare_template_conf["real_scaling_template_name"],
    )
    flux_real_scale = pd.read_hdf(scale_1_template_path)["sim_flux_input"].iat[0]
    return flux_real_scale


def plot_systematic_all_flare_templates(mngr, df_diagnostics: pd.DataFrame):
    if mngr.skip_flare_templates:
        return

    all_metadata = StateManager(
        mngr,
        "systematic_flare_evaluation_metadata",
        dir_tag="custom_realisations_input",
    ).load_state()

    template_conf = mngr.flare_template_conf

    # Define the layout
    layout = """
    CCAA
    CCBB
    """

    mapping = {
        TAG_SIGNIFICANCE_CLUSTER_SPACE: "A",
        TAG_SIGNIFICANCE_RECO_ERROR: "B",
        TAG_SIGNIFICANCE_COMBINED: "C",
    }
    # (10, 6)
    figsize = (7.5, 4.5)

    # significance by flux at specific flare duration
    for flare_duration in template_conf["systematic_evaluation_durations"]:
        fig, axes = plt.subplot_mosaic(
            layout,
            figsize=figsize,
            sharex=True,
            sharey=False,
            gridspec_kw={
                "hspace": 0.1,
                "wspace": 0.4,
                "left": 0.08,
                "right": 0.97,
                "bottom": 0.2,
                "top": 0.91,
            },
        )
        for significance_tag, axis_name in mapping.items():
            ax = axes[axis_name]

            plot_systematic_all_flare_templates_for_significance(
                mngr,
                ax,
                all_metadata,
                df_diagnostics,
                significance_tag,
                flare_duration,
                show_percentages=axis_name != "B",
            )

        axes["A"].set_xlabel(None)
        axes["A"].set_xticklabels([])

        axes["C"].set_ylabel(
            "Significance",
            fontsize=DEFAULT_SCALE * MEDIUM_SIZE,
        )

        axes["C"].legend(
            loc="lower center",
            ncol=3,
            bbox_to_anchor=(1.11, -0.26),
            fontsize=SMALL_SIZE * DEFAULT_SCALE,
        )
        fig.subplots_adjust(wspace=0.25, hspace=0.1, bottom=0.15)

        plot_name = f"flux_scalings_for_duration_{flare_duration}.pdf"
        tensorboard_log_figure(mngr, fig, plot_name)
        PDFPlotCollection(mngr, [fig]).save_to_file(
            get_debug_output_filepath(
                mngr, os.path.join("systematic_all_flare_templates"), plot_name
            )
        )

    # significance by duration at reference flux
    fig, axes = plt.subplot_mosaic(
        layout,
        sharex=True,
        sharey=False,
        figsize=figsize,
        gridspec_kw={
            "hspace": 0.1,
            "wspace": 0.4,
            "left": 0.08,
            "right": 0.97,
            "bottom": 0.2,
            "top": 0.91,
        },
    )
    for significance_tag, axis_name in mapping.items():
        ax = axes[axis_name]
        plot_summary_systematic_flare_templates_reference_flux(
            mngr, ax, all_metadata, df_diagnostics, significance_tag
        )

    axes["A"].set_xlabel(None)
    axes["A"].set_xticklabels([])

    axes["C"].set_ylabel(
        "Significance",
        fontsize=DEFAULT_SCALE * MEDIUM_SIZE,
    )

    def make_deduplicated_legend(legend_opts=None, label_dict=None):
        if legend_opts is None:
            legend_opts = {}

        handles, axis_labels = axes["C"].get_legend_handles_labels()
        by_label = dict(zip(axis_labels, handles))
        axes["C"].legend(
            by_label.values(),
            [clean_label(mngr, label) for label in by_label.keys()],
            fontsize=SMALL_SIZE * DEFAULT_SCALE,
            **legend_opts,
        )

    #
    make_deduplicated_legend(
        legend_opts=dict(loc="lower center", ncol=3, bbox_to_anchor=(1.1, -0.26))
    )

    plot_name = f"durations_at_reference_flux.pdf"
    tensorboard_log_figure(mngr, fig, plot_name)
    PDFPlotCollection(mngr, [fig]).save_to_file(
        get_debug_output_filepath(
            mngr, os.path.join("systematic_all_flare_templates"), plot_name
        )
    )


def plot_systematic_all_flare_templates_for_significance(
    mngr,
    ax,
    all_metadata,
    df_diagnostics: pd.DataFrame,
    significance_tag,
    flare_duration,
    show_percentages=True,
):
    template_conf = mngr.flare_template_conf

    # fig, ax = plt.subplots(
    #     figsize=(4, 4 if significance_tag == TAG_SIGNIFICANCE_COMBINED else 2)
    # )
    ax.grid(**gridlines_config)
    ax.tick_params(**tick_params)

    base_name = "systematic_all_flare_templates"

    labels = list(template_conf["systematic_evaluation_combinations"].keys())

    if show_percentages:
        # Create a secondary x-axis
        ax2 = ax.twiny()

    for combo_name, metadata in all_metadata.items():
        combo_label = combo_name[len(base_name) + 1 :]
        if not combo_name.endswith(str(flare_duration)):
            continue

        df = df_diagnostics[df_diagnostics["custom_realisation_name"] == combo_name]
        if len(df) == 0:
            mngr.log.warn(
                [
                    [
                        "r",
                        f" - plot_systematic_all_flare_templates - No data for '{combo_label}'",
                    ]
                ]
            )
        meta_df = pd.DataFrame(metadata, columns=["start", "end", "strength"])

        final_df = gather_significances_for_flare(df, meta_df, significance_tag)

        grouped_by_strength = final_df.drop(columns=["start", "end"]).groupby(
            "strength"
        )

        # median = grouped_by_strength.median()
        mean = grouped_by_strength.mean()

        uncertainty_up = np.abs(grouped_by_strength.quantile(0.136) - mean)
        uncertainty_down = np.abs(grouped_by_strength.quantile(0.864) - mean)

        combo_label_without_duration = "_".join(combo_label.split("_")[:-1])

        plot_style = mngr.plot_lib.get_plot_style(
            index=labels.index(combo_label_without_duration),
            tag="main_data",
            style_set="main_0",
            has_line=False,
            has_marker=True,
        )

        plot_style.update(
            dict(
                # alpha=0.9,
                # lw=2,
                # capsize=5,
                capthick=0.2,
                # fmt="o",
                markersize=5,
                # ls="none",
                # markerfacecolor="none",
            )
        )

        plot_style["color"] = GROUP_COLORS[labels.index(combo_label_without_duration)]

        ax.errorbar(
            y=mean[significance_tag],
            x=mean.index,
            yerr=np.hstack((uncertainty_down, uncertainty_up)).T,
            label=clean_label(mngr, combo_label_without_duration),
            **plot_style,
        )

        ax.set_ylim(*FIXED_Y_RANGE)

        if show_percentages:
            # Set the limits and ticks for the secondary axis
            ax2.set_xlim(ax.get_xlim())
            ax2.set_xticks(mean.index)  # Align ticks with the primary axis
            # Calculate the scaled index as a percentage and format it
            scaled_index_percentage = (
                mean.index.to_numpy() / get_scale_1_flux(mngr) * 100
            )
            percentage_labels = [
                (
                    ""
                    if int(round(value)) in [20, 30, 40, 60, 70, 80, 90]
                    else f"{int(round(value))}%"
                )
                for value in scaled_index_percentage
            ]

            mapping = {"1%": "1%  ", "10%": "    10%"}

            # Set the tick labels, replacing specified values with empty strings
            ax2.set_xticklabels([mapping.get(l, l) for l in percentage_labels])

    ax.ticklabel_format(scilimits=(-2, 2), style="sci")

    # ax.legend(loc="upper left")
    ax.set_xlabel(
        clean_label(mngr, template_x_label), fontsize=SMALL_SIZE * DEFAULT_SCALE
    )
    inside_label(
        ax,
        clean_label(
            mngr,
            significance_tag,
        ),
        y=0.92 if significance_tag == TAG_SIGNIFICANCE_COMBINED else 0.85,
    )

    ax.axvline(x=get_scale_1_flux(mngr), c="grey", linestyle="--", linewidth=1)


def inside_label(ax, text, y=None):
    if y is None:
        y = 0.9

    # add channel name label
    text = ax.text(
        *(0.05, y),
        text,
        transform=ax.transAxes,
        fontsize=MEDIUM_SIZE * DEFAULT_SCALE,
        verticalalignment="center",
        horizontalalignment="left",
        alpha=0.9,
        # bbox=dict(
        #     boxstyle="round",
        #     facecolor="white",
        #     alpha=0.5,
        #     edgecolor="black",
        #     # color="black",
        # ),
        path_effects=[pe.withStroke(linewidth=0.75, foreground="white", alpha=0.75)],
    )


def gather_significances_for_flare(df, meta_df, significance_tag):
    # For each start and end time in meta_df, we get all corresponding model evaluations
    # we pick the time with the max median significance, and then store all evaluations at that time
    # in the final_output df

    # Initialize an empty list to store the intermediate DataFrames
    dfs_to_merge = []

    df = df[df[TAG_INPUT_MASK] == TAG_INPUT_MASK_NONE]

    # Iterate over meta_df
    for idx, row in meta_df.iterrows():
        # Filter df based on time_right_edge and calculate medians
        filtered_df = df[
            df["time_right_edge"].between(row["start"], row["end"], inclusive="both")
        ]
        # get the median significance for every event
        grouped_medians = filtered_df.groupby("time_right_edge")[
            significance_tag
        ].median()

        # Get the time_right_edge with the maximum median significance
        assert len(grouped_medians) > 0
        max_median_time = grouped_medians.idxmax()

        # Filter rows with this time_right_edge and add a column with the current index from meta_df
        max_median_df = filtered_df[
            filtered_df["time_right_edge"] == max_median_time
        ].copy()
        max_median_df["meta_index"] = idx

        # Add the DataFrame to the list
        dfs_to_merge.append(max_median_df)

    # Concatenate all the DataFrames
    concatenated_df = pd.concat(dfs_to_merge)

    # Merge with meta_df based on the index
    final_df = meta_df.merge(
        concatenated_df[[significance_tag, "meta_index"]],
        left_index=True,
        right_on="meta_index",
        how="inner",
    ).drop("meta_index", axis=1)

    return final_df


def plot_summary_systematic_flare_templates_reference_flux(
    mngr, ax, all_metadata, df_diagnostics: pd.DataFrame, significance_tag
):
    reference_strength = get_scale_1_flux(mngr)

    template_conf = mngr.flare_template_conf

    # fig, ax = plt.subplots()
    ax.grid(**gridlines_config)
    ax.tick_params(**tick_params)

    base_name = "systematic_all_flare_templates"
    out_dir = os.path.join(base_name, significance_tag)

    labels = list(template_conf["systematic_evaluation_combinations"].keys())

    for flare_duration in template_conf["systematic_evaluation_durations"]:
        for combo_name, metadata in all_metadata.items():
            combo_label = combo_name[len(base_name) + 1 :]
            if not combo_name.endswith(str(flare_duration)):
                continue

            df = df_diagnostics[df_diagnostics["custom_realisation_name"] == combo_name]
            if len(df) == 0:
                mngr.log.warn(
                    [
                        [
                            "r",
                            f" - plot_systematic_all_flare_templates - No data for '{combo_label}'",
                        ]
                    ]
                )
            meta_df = pd.DataFrame(metadata, columns=["start", "end", "strength"])

            final_df = gather_significances_for_flare(df, meta_df, significance_tag)

            grouped_by_strength = final_df[
                final_df["strength"] == reference_strength
            ].drop(columns=["start", "end"])[significance_tag]

            # median = grouped_by_strength.median()
            mean = grouped_by_strength.mean()

            uncertainty_up = np.abs(grouped_by_strength.quantile(0.16) - mean)
            uncertainty_down = np.abs(grouped_by_strength.quantile(0.84) - mean)

            combo_label_without_duration = "_".join(combo_label.split("_")[:-1])

            plot_style = mngr.plot_lib.get_plot_style(
                index=labels.index(combo_label_without_duration),
                tag="main_data",
                style_set="main_0",
                has_line=False,
                has_marker=True,
            )

            plot_style.update(
                dict(
                    # alpha=0.9,
                    # lw=2,
                    # capsize=5,
                    capthick=0.2,
                    # fmt="o",
                    markersize=5,
                    # ls="none",
                )
            )
            plot_style["color"] = GROUP_COLORS[
                labels.index(combo_label_without_duration)
            ]

            ax.errorbar(
                y=mean,
                x=flare_duration,
                yerr=np.expand_dims(
                    np.array((uncertainty_down, uncertainty_up)), axis=1
                ),
                label=combo_label_without_duration,
                **plot_style,
            )
            ax.ticklabel_format(scilimits=(-2, 2), style="sci")

            ax.set_ylim(*FIXED_Y_RANGE)

    # def make_deduplicated_legend(legend_opts=None, label_dict=None):
    #     if legend_opts is None:
    #         legend_opts = {}
    #
    #     handles, axis_labels = plt.gca().get_legend_handles_labels()
    #     by_label = dict(zip(axis_labels, handles))
    #     plt.legend(by_label.values(), by_label.keys(), **legend_opts)
    #
    # make_deduplicated_legend(legend_opts=dict(loc="lower right"))

    ax.set_xlabel(
        clean_label(mngr, "Days observed [count]"), fontsize=SMALL_SIZE * DEFAULT_SCALE
    )

    inside_label(
        ax,
        clean_label(mngr, significance_tag),
        y=0.92 if significance_tag == TAG_SIGNIFICANCE_COMBINED else 0.85,
    )

    # fig_name = f"at_reference_flux.pdf"
    # tensorboard_log_figure(mngr, fig, fig_name)
    # PDFPlotCollection(mngr, [fig]).save_to_file(
    #     get_debug_output_filepath(mngr, out_dir, fig_name)
    # )
