from __future__ import annotations

import copy
from datetime import datetime
from math import log10
from typing import TYPE_CHECKING, Iterable, List

import astropy.units as asp_unit
import matplotlib.pyplot as plt
import mpmath
import numpy as np
import pandas as pd
import scipy.stats as stats
from astropy.cosmology import Planck13
from matplotlib.backends.backend_pdf import PdfPages

from trans_finder.flares.callbacks.callback_write_model_outputs_to_csv import (
    write_model_outputs__and_errors_to_csv_for_debugging,
)
from trans_finder.flares.helpers_debugging import get_debug_output_filepath
from trans_finder.flares.plot_list.autoencoder import autoenc_create_all_plots
from trans_finder.flares.utils import str_join

if TYPE_CHECKING:
    from trans_finder.flares.log_parser import LogParser
    from trans_finder.flares.manager import Manager

mpmath.mp.dps = 100  # decimal digits of precision


class Plots:
    def __init__(self, mngr):
        self.mngr: Manager = mngr
        self.log: LogParser = mngr.log

        self.tag_to_title = {}

    def plot_predictions(self, is_sig_plot=None):
        """plot validation results

        for sginificance calculation, we use stats.chi2.isf()
        since this is for two sided intervals, we take:
        stats.chi2.isf(5.7e-7, df=1) --> 25.05
        stats.chi2.isf(5.7e-7, df=2) --> 28.76

        Parameters
        ----------
        is_sig_plot : bool
            is working in classification mode, this flag indicates
            if we are plotting the signal sample
        """

        self.log.info([*self.mngr.main_func_prnt, ["g", "plot-predictions()"]])

        if is_sig_plot is None:
            is_sig_plot = self.mngr.is_sig_plot

        # some constants and tags
        chi_2_ndf_rnn = 1
        chi_2_ndf_ctools = 1
        numerical_p_val_lim = 1e-20

        signif_sqr_ctools = "sig_sqr_ctools"
        signif_sqr_rnn = (
            "sig_sqr_clasif" if self.mngr.is_enc_dec_clasif else "sig_sqr_anom_det"
        )

        if self.do_ts_combine:
            tag_log_p_val_rnn = self.tag_log_p_val_combo
        else:
            tag_log_p_val_rnn = (
                f"{self.mngr.tag_log_p_value}_{self.mngr.tag_model_base}"
            )

        # event weight tag
        if self.do_ts_combine:
            wgt_col = self.tag_wgt_combo
        else:
            if len(self.mngr.df_cols_wgt_0) == 0:
                wgt_col = None
            elif len(self.mngr.df_cols_wgt_0) == 1:
                wgt_col = self.mngr.df_cols_wgt_0[0]
            else:
                msg = """
                 - unexpected number of weights for classification ...
                 is this consistent with get-predictions() ?
                """
                raise Exception(msg)

        self.set_title(signif_sqr_ctools, "ctools")
        self.set_title(
            self.mngr.tag_chi_2, r"$\frac{1}{2}\chi^{2}_{(" + str(chi_2_ndf_rnn) + ")}$"
        )
        if self.mngr.is_enc_dec_anom_det:
            self.set_title(signif_sqr_rnn, "Anom-det.")
        elif self.mngr.is_enc_dec_clasif:
            self.set_title(signif_sqr_rnn, "Clas.")
        elif self.mngr.is_enc_dec_clasif:
            self.set_title(signif_sqr_rnn, "Clust.")

        self.set_title("src_sim_flux", "flux")
        self.set_title("src_sim_spec_index", "spectral index")
        self.set_title("src_sim_time_index", "temporal index")

        # get the input data and filter
        data = self.mngr.input_data[self.mngr.tag_input_df][self.mngr.run_phase]
        if self.mngr.is_enc_dec_clasif:
            if is_sig_plot:
                data = data[~data[self.mngr.tag_is_bck_cat]]
            else:
                data = data[data[self.mngr.tag_is_bck_cat]]

        # print(data[['ts_clas', 'log_p_value_clas']])
        # print(data.iloc[0])

        if len(data) == 0:
            raise Exception("empty input dataset...")

        # filter data where TS/p-values have not been calculated (due e.g.,
        # to missing calibration groups), or data masked
        mask = data[self.mngr.tag_has_p_value].astype(bool)

        # filter by ts cut
        if self.mngr.tag_evt_ts_mask in data.columns:
            mask = mask & data[self.mngr.tag_evt_ts_mask]

        # filter by general event cut for signals
        if is_sig_plot:
            mask = mask & data[self.mngr.tag_evt_cut_mask]

        data = data[mask].reset_index(drop=True).copy()

        if len(data) == 0:
            raise Exception("empty input dataset...")

        is_debug = False
        if is_debug:
            self.log.info([["wr", " - reducing stats for debugging..."]])
            data = data.iloc[0:1000].copy()

        self.log.info(
            [
                ["b", "data (", len(data), "entries), with columns:\n"],
                ["c", data.columns.tolist()],
                ["b", "\n"],
            ]
        )

        # some safety checks on p-values and significance values
        log_p_values = data[tag_log_p_val_rnn].astype(float)
        p_values = np.power(10, log_p_values)

        p_values[p_values > 1] = 1
        # print(data[p_values < numerical_p_val_lim])
        p_values[p_values < numerical_p_val_lim] = numerical_p_val_lim

        signif = np.asarray(stats.norm.isf(p_values))
        signif[(signif < 0) | (np.isneginf(signif))] = 0
        data.loc[:, signif_sqr_rnn] = np.power(signif, 2)

        # func to get the ctools TS-values from the data
        def get_ts_ctools(ts):
            ts = ts[-1] if isinstance(ts, Iterable) else ts
            return ts

        ts_values = data[self.mngr.ts_compare].apply(get_ts_ctools)

        # derive the significance corresponding to the TS-values
        # if [chi_2_ndf_ctools>1], we convert the p-values to the simple
        # Gaussian case, where
        # [stats.norm.sf(np.sqrt(x))] and [0.5 * stats.chi2.sf(x, df=1)]
        # give the same answer...
        if chi_2_ndf_ctools == 1:
            # if this is for 1 degree of freedom, simply take the
            # TS as the square of the significance (quicker than using the
            # survival function as done for ndf>1 under the else condition)
            signif_sqr = ts_values
            signif_sqr[(signif_sqr < 0) | (np.isinf(signif_sqr))] = 0
            data.loc[:, signif_sqr_ctools] = signif_sqr
        else:
            # if ndf>1, explicitly calculate the one-sided p-value for the TS
            # from the survival function with the corresponding ndf
            p_values = 0.5 * stats.chi2.sf(ts_values, df=chi_2_ndf_ctools)
            p_values[p_values > 1] = 1
            p_values[p_values < numerical_p_val_lim] = numerical_p_val_lim

            signif = np.asarray(stats.norm.isf(p_values))
            signif[(signif < 0) | (np.isneginf(signif))] = 0
            data.loc[:, signif_sqr_ctools] = np.power(signif, 2)

        # some additional data transforms
        sig_x_names_0 = ["src_sim_flux", "src_sim_spec_index", "src_sim_time_index"]

        sig_x_names = []
        for x_name in sig_x_names_0:
            if x_name not in data:
                continue

            x_name_now = x_name
            if x_name in ["src_sim_flux"]:
                x_name_now = str_join(x_name, "log10")

                x_label = r"$\mathrm{log_{10}}$" + "(" + self.get_title(x_name) + ")"
                self.set_title(x_name_now, x_label)

                if x_name_now not in data:
                    data.loc[:, x_name_now] = data[x_name].apply(lambda x: log10(x))

            if x_name in ["src_sim_spec_index", "src_sim_time_index"]:
                x_name_now = str_join(x_name, "abs")

                self.set_title(x_name_now, self.get_title(x_name))

                data.loc[:, x_name_now] = np.abs(data[x_name])

            sig_x_names += [x_name_now]

        figs = []
        # for signal
        if is_sig_plot:
            data_tags = [signif_sqr_ctools, signif_sqr_rnn]
            wgt_tags = [wgt_col for _ in range(len(data_tags))]

            for x_name in sig_x_names:
                opts = {
                    "x_name": x_name,
                    "x_label": self.get_title(x_name),
                    "legend_opts": {
                        "title": "Signal",
                    },
                }

                # # +++++++++++++++++++++++++++++++++++++++++++++++++
                # data = data.query('src_sim_flux > 1e-5').copy()
                # # data = data.query('src_sim_flux > 1e-2').copy()
                # # # data = data.query('p_value_clas < ').copy()
                # # # print(data.iloc[0])
                # print('WWWWWWWWW - cuts-0....')
                # # +++++++++++++++++++++++++++++++++++++++++++++++++

                output = self.plot_is_det_comp(
                    data=data,
                    data_tags=data_tags,
                    wgt_tags=wgt_tags,
                    opts=opts,
                )

                figs += [output["fig"]]

                # meta_cols = []
                meta_cols = ["point_zen_deg", "point_shift_deg"]
                meta_cols = [c for c in meta_cols if c in data]

                plot_meta = True
                # plot_meta = False
                plot_meta = plot_meta and len(meta_cols) > 0
                if plot_meta:
                    if "src_sim_flux" in x_name:
                        self.plot_metadata_binned(
                            data=data,
                            data_tags=data_tags,
                            wgt_tags=wgt_tags,
                            meta_cols=meta_cols,
                            figs=figs,
                            func=self.plot_is_det_comp,
                            opts=opts,
                        )

        # for background
        if not is_sig_plot:
            # the column names to plot and their respective event-weights
            data_tags = [signif_sqr_ctools, signif_sqr_rnn]
            # data_tags = [signif_sqr_rnn]
            wgt_tags = [wgt_col for _ in range(len(data_tags))]

            opts = {
                "legend_opts": {
                    "title": "Background",
                },
            }

            # inclusive plot
            output = self.plot_signif_sqr_comp(
                data=data,
                data_tags=data_tags,
                wgt_tags=wgt_tags,
                opts=opts,
            )
            figs += [output["fig"]]

            # meta_cols = []
            meta_cols = ["point_zen_deg", "point_shift_deg"]
            # meta_cols = [ 'point_zen_deg']
            # meta_cols = [ 'point_shift_deg']
            meta_cols = [c for c in meta_cols if c in data]

            # plots binned in metadata cut dimensions
            plot_meta = True
            # plot_meta = False
            plot_meta = plot_meta and len(meta_cols) > 0
            if plot_meta:
                self.plot_metadata_binned(
                    data=data,
                    data_tags=data_tags,
                    wgt_tags=wgt_tags,
                    meta_cols=meta_cols,
                    figs=figs,
                    func=self.plot_signif_sqr_comp,
                    opts=opts,
                )

        # store the results to file
        is_sig_tag = "sig" if is_sig_plot else "bck"
        out_file_name = self.mngr.plot_lib.get_pdf_pages(
            tag=str_join("predict_plots", is_sig_tag)
        )

        with PdfPages(out_file_name) as pdf_pages:
            self.mngr.plot_lib.save_pdf_pages(pdf_pages, out_file_name, figs=figs)

        return

    def plot_metadata_binned(
        self, data, data_tags, wgt_tags, meta_cols, func, figs, opts={}
    ):
        """plot by func for binned data (allows for nested binning)

        Parameters
        ----------
        data : pd.DataFrame
            the input data
        data_tags : list
            the df column names to be plotted
        wgt_tags : list
            optional weight column names corresponding to the data tags
        meta_cols : list of str
            the names of columns in the df on which (nested) cuts are mare
        func : func
            plotting function to call for the binned data
        figs : list
            list of figure objects
        opts : dict
            optional settings for plotting / styling
        """

        self.log.info([["y", " - plot-binned(): "], ["p", data_tags]])

        n_meta_cols = len(meta_cols)

        if n_meta_cols == 1:
            output = self.plot_data_cuts(
                data=data,
                meta_col=meta_cols[0],
                data_tags=data_tags,
                wgt_tags=wgt_tags,
                func=func,
                figs=figs,
                opts=opts,
            )

            figs += [output["fig"]]

        elif n_meta_cols == 2:
            meta_col_0 = meta_cols[0]
            meta_col_1 = meta_cols[1]

            meta_cuts_0 = [
                pd.cut(
                    data[meta_col_0],
                    self.mngr.meta_bins[meta_col_0],
                    include_lowest=True,
                    retbins=False,
                    labels=False,
                )
            ]

            grps_0 = data.groupby(meta_cuts_0)
            keys_0 = list(grps_0.groups.keys())
            for n_grp_0 in range(len(keys_0)):
                key_0 = keys_0[n_grp_0]
                indx = grps_0.get_group(name=key_0).index
                data_cut_0 = data.loc[indx].reset_index(drop=True)

                title_cut_0 = self.mngr.meta_bin_titles[meta_col_0][n_grp_0]
                self.log.info([["b", " - cut:  "], ["p", title_cut_0]])

                output = self.plot_data_cuts(
                    data=data_cut_0,
                    meta_col=meta_col_1,
                    data_tags=data_tags,
                    wgt_tags=wgt_tags,
                    func=func,
                    figs=figs,
                    opts=opts,
                )

                output["fig"].suptitle(title_cut_0)

                figs += [output["fig"]]

        else:
            raise Exception(
                "problem with meta_cols:",
                meta_cols,
                "(up to 2 cut dims are allwed ...)",
            )

        return

    def plot_data_cuts(self, data, data_tags, wgt_tags, meta_col, func, figs, opts={}):
        """plot by func for binned data

        Parameters
        ----------
        data : pd.DataFrame
            the input data
        data_tags : list
            the df column names to be plotted
        wgt_tags : list
            optional weight column names corresponding to the data tags
        meta_col : str
            the name of the column in the DataFrame on which cuts are made
        func : func
            plotting function to call for the binned data
        figs : list
            list of figure objects
        opts : dict
            optional settings for plotting / styling
        """

        meta_cuts = [
            pd.cut(
                data[meta_col],
                self.mngr.meta_bins[meta_col],
                include_lowest=True,
                retbins=False,
                labels=False,
            )
        ]

        grps = data.groupby(meta_cuts)
        keys = list(grps.groups.keys())

        n_grps = len(keys)
        n_rows, n_cols = self.mngr.plot_lib.get_n_rows_cols(n_grps + 1)
        opts_now = {
            "fig_spec": None,
            "fig": {
                "n_rows": n_rows,
                "n_cols": n_cols,
                "opts": {
                    "gridspec_opts": {
                        "hspace": 0.4,
                        "wspace": 0.3,
                    },
                },
            },
        }
        opts_now.update(opts)

        for n_grp in range(-1, n_grps):
            if n_grp < 0:
                data_cut = data
                title_cut = "Inclusive"
            else:
                key = keys[n_grp]
                indx = grps.get_group(name=key).index
                data_cut = data.loc[indx].reset_index(drop=True)
                title_cut = self.mngr.meta_bin_titles[meta_col][n_grp]

            spec_n_row = int((n_grp + 1) // n_cols)
            spec_n_col = (n_grp + 1) - spec_n_row * n_cols

            legend_opts = None
            if n_grp == -1:
                if opts is not None and "legend_opts" in opts:
                    legend_opts = opts["legend_opts"]

            opts_now.update(
                {
                    "spec": {
                        "spec_n_row": spec_n_row,
                        "spec_n_col": spec_n_col,
                    },
                    "legend_opts": legend_opts,
                }
            )

            output = func(
                data=data_cut,
                data_tags=data_tags,
                wgt_tags=wgt_tags,
                opts=opts_now,
            )
            opts_now["fig_spec"] = output

            output["ax"].set_title(title_cut)

        return output

    def plot_signif_sqr_comp(self, data, data_tags, wgt_tags, opts={}):
        """plot a significance distribution

        Parameters
        ----------
        data : pd.DataFrame
            the input data
        data_tags : list
            the df column names to be plotted
        wgt_tags : list
            optional weight column names corresponding to the data tags
        opts : dict
            optional settings for plotting / styling

        Returns
        -------
        dict
            the figure and relevant objects
        """

        self.log.info([["g", " - plot-signif-comp(): "], ["c", data_tags]])

        # add/override options based on the opt input
        def update_opts(opt, grp):
            if grp in opts:
                opt.update(opts[grp])

        # the general style group
        style_set = "autumn"
        # style_set = 'purples'
        # style_set = None

        # labels for the axes
        x_label = r"$\sigma^{2}$"
        y_label = r"1/N dN/d$\sigma^{2}$"

        # flag to overly a chi2 distribution
        add_chi_2 = True
        # flag for adding histograms in addition to the curves
        draw_bar_hist = False
        # add annotation for this range of significance values
        signif_lines = list(range(2, 7))

        # minimal number of entries to consider for a distribution
        min_n_vals = 10

        # number of bins for the horizontal axis, axis limits, etc.
        n_bins = 50
        x_lims, y_lims = None, None
        x_lims = [-0.5, None]
        y_lims = [None, 2.5]

        opts_fig = {
            "n_rows": 1,
            "n_cols": 1,
        }
        update_opts(opts_fig, "fig")

        if "fig_spec" in opts and (opts["fig_spec"] is not None):
            fig = opts["fig_spec"]["fig"]
            spec = opts["fig_spec"]["spec"]
        else:
            fig, spec = self.mngr.plot_lib.add_fig(**opts_fig)

        spec_n_row = self.get_opt_or_def(opts, "spec", "spec_n_row", 0)
        spec_n_col = self.get_opt_or_def(opts, "spec", "spec_n_col", 0)

        legend_opts = opts["legend_opts"] if "legend_opts" in opts else {}

        ax = self.mngr.plot_lib.add_subplot(fig, spec[spec_n_row, spec_n_col])

        # add all data tags
        has_drawn = False
        for n_tag in range(len(data_tags)):
            tag_now = data_tags[n_tag]
            wgt_tag = wgt_tags[n_tag]

            out_vals = np.asarray(data[tag_now])

            if len(out_vals) < min_n_vals:
                continue

            if wgt_tag is None:
                out_wgts = np.asarray([1 for _ in range(len(out_vals))])
            else:
                out_wgts = np.asarray(data[wgt_tag])
            sum_wgts = np.sum(out_wgts)

            if sum_wgts < 1e-10:
                self.log.error(
                    [
                        ["wr", " - sum_wgts = ", sum_wgts, " ?!?"],
                        ["p", "\n", tag_now, "\n"],
                        ["r", data, "\n"],
                    ]
                )
                raise Exception("sum_wgts too small ?!?")

            has_drawn = True

            # axis properties
            x_width = 0.25
            x_min = 0
            x_n_bins = int(((out_vals.max() - x_min) // x_width))
            x_n_bins = max(x_n_bins, 10)
            x_max = x_min + x_n_bins * x_width
            x_max = max(x_max, 1)
            # print('out_vals.max()',tag_now,out_vals.min(),out_vals.max())

            # normalise, so that the pdf will also be the
            # probabiity mass function [--> integral==sum==1]
            out_wgts = out_wgts / (sum_wgts * x_width)

            x = np.linspace(x_min, x_max, x_n_bins + 1)
            y, _ = np.histogram(
                out_vals,
                weights=out_wgts,
                bins=x,
            )

            x_mid = [x[n] + 0.5 * (x[n + 1] - x[n]) for n in range(len(x) - 1)]

            plot_style = self.mngr.plot_lib.get_plot_style(
                index=n_tag,
                tag="main_data",
                style_set=style_set,
            )
            label = self.get_title(tag_now)

            _ = ax.plot(
                x_mid,
                y,
                **plot_style,
                label=label,
            )

            # plot an additional histogram
            if draw_bar_hist:
                his_style = copy.deepcopy(plot_style)
                his_style["alpha"] = 0.2
                _ = ax.hist(
                    out_vals,
                    x,
                    **his_style,
                    label=label,
                    density=True,
                )

        # overly a chi2 distribution for background comparisons
        if has_drawn and add_chi_2:
            n_tag_chi_2 = n_tag + 1
            x_max_chi_2 = ax.get_xlim()[1]

            x = np.linspace(0, x_max_chi_2, n_bins + 1)

            # at this point we only compare to ndf=1, since even if
            # [chi_2_ndf_ctools>1] above, we convert the p-values.
            y = stats.norm.sf(np.sqrt(x))
            # y = 0.5 * stats.chi2.sf(x, df=1)

            plot_style = self.mngr.plot_lib.get_plot_style(
                index=n_tag_chi_2,
                tag="main_data",
                style_set=style_set,
            )
            label = self.get_title(self.mngr.tag_chi_2)

            _ = ax.plot(
                x,
                y,
                **plot_style,
                label=label,
            )

        # styling
        if has_drawn:
            _ = ax.set_xlabel(x_label)
            _ = ax.set_ylabel(y_label)
            _ = ax.set_yscale("log")

            if legend_opts is not None:
                handles, labels = ax.get_legend_handles_labels()
                ax.legend(handles, labels, **legend_opts)
            # _ = plt.legend()

            def set_axis_lims(lims, new_lims, set_lims):
                if new_lims is None:
                    return
                if new_lims[0] is not None and lims[0] < new_lims[0]:
                    new_lims[0] = lims[0]
                if new_lims[1] is not None and lims[1] > new_lims[1]:
                    new_lims[1] = lims[1]
                _ = set_lims(*new_lims)
                return

            set_axis_lims(lims=ax.get_xlim(), new_lims=x_lims, set_lims=ax.set_xlim)
            set_axis_lims(lims=ax.get_ylim(), new_lims=y_lims, set_lims=ax.set_ylim)

        # add annotation for a range of significance values
        if has_drawn:
            x_lims = ax.get_xlim()
            y_lims = ax.get_ylim()
            for n_signif_line in range(len(signif_lines)):
                signif_now = signif_lines[n_signif_line]
                signif_sqr_now = signif_now**2

                # pv_now = 0.5 * stats.chi2.sf(signif_sqr_now, df=1)
                pv_now = stats.norm.sf(np.sqrt(signif_sqr_now))
                if pv_now < y_lims[0]:
                    continue

                line_style_set = "greens"
                plot_style = self.mngr.plot_lib.get_plot_style(
                    index=n_signif_line,
                    tag="main_data",
                    style_set=line_style_set,
                )
                linewidth = 2.5
                linestyle = "--"
                # linestyle = plot_style['linestyle']
                color = "black"
                # color = plot_style['color']
                line_alpha = 0.15

                _ = ax.axhline(
                    y=pv_now,
                    linewidth=linewidth,
                    alpha=line_alpha,
                    linestyle=linestyle,
                    color=color,
                )
                # _ = ax.axvline(
                #     x=signif_sqr_now, linewidth=linewidth,
                #     alpha=line_alpha, linestyle=linestyle, color=color,
                # )

                text_x = x_lims[1] - 0.01 * (x_lims[1] - x_lims[0])
                text_y = 1.1 * pv_now
                text_s = str(signif_now) + r"$\sigma$"
                text_alpha = 0.45
                _ = ax.text(
                    x=text_x,
                    y=text_y,
                    s=text_s,
                    horizontalalignment="right",
                    verticalalignment="bottom",
                    alpha=text_alpha,
                    color=color,
                )

        output = {
            "fig": fig,
            "spec": spec,
            "ax": ax,
        }

        return output

    def plot_grb_flux_refs(self, opts={}):
        self.log.info(
            [
                ["g", " - plot-grb-flux-refs(): "],
            ]
        )

        sim_lumi = 1e-2
        sim_z = np.linspace(0, 0.2, 16)
        sim_z = np.asarray(
            [
                (sim_z[n] + 0.5 * (sim_z[n + 1] - sim_z[n]))
                for n in range(len(sim_z) - 1)
            ]
        )

        data_tags = []
        data_tags += [
            {
                "name": "GRB_080916C",
                "z": 4.35,
                "lumi": 6.5e4,
                "flux": 500e-5,
            }
        ]
        data_tags += [
            {
                "name": "GRB_090323",
                "z": 3.57,
                "lumi": 4.1e4,
                "flux": 6e-5,
            }
        ]
        data_tags += [
            {
                "name": "GRB_110731A",
                "z": 2.83,
                "lumi": 7.2e3,
                "flux": 220e-5,
            }
        ]
        data_tags += [
            {
                "name": "GRB_090902B",
                "z": 1.822,
                "lumi": 3.4e4,
                "flux": 600e-5,
            }
        ]
        data_tags += [
            {
                "name": "GRB_090510",
                "z": 0.9,
                "lumi": 7.3e2,
                "flux": 3900e-5,
            }
        ]

        def get_sim_flux(n_tag):
            ref_lumi = data_tags[n_tag]["lumi"]
            ref_flux = data_tags[n_tag]["flux"]
            ref_z = np.repeat(data_tags[n_tag]["z"], len(sim_z))

            ref_dist = Planck13.luminosity_distance(ref_z)
            sim_dist = Planck13.luminosity_distance(sim_z)
            ref_dist = (ref_dist / asp_unit.Mpc).value
            sim_dist = (sim_dist / asp_unit.Mpc).value

            lumi_ratio = sim_lumi / ref_lumi
            flux_ratio = lumi_ratio / pow((sim_dist / ref_dist), 2)

            src_sim_flux = ref_flux * flux_ratio

            return src_sim_flux

        style_set = "purples"

        x_label = r"$z$"
        y_label = r"Flux"

        def update_opts(opt, grp):
            if grp in opts:
                opt.update(opts[grp])

        opts_fig = {
            "n_rows": 1,
            "n_cols": 1,
        }
        update_opts(opts_fig, "fig")

        legend_opts = opts["legend_opts"] if "legend_opts" in opts else {}

        if "fig_spec" in opts and (opts["fig_spec"] is not None):
            fig = opts["fig_spec"]["fig"]
            spec = opts["fig_spec"]["spec"]
        else:
            fig, spec = self.mngr.plot_lib.add_fig(**opts_fig)

        spec_n_row = self.get_opt_or_def(opts, grp="spec", tag="spec_n_row", default=0)
        spec_n_col = self.get_opt_or_def(opts, grp="spec", tag="spec_n_col", default=0)

        ax = self.mngr.plot_lib.add_subplot(fig, spec[spec_n_row, spec_n_col])

        for n_tag in range(len(data_tags)):
            tag_now = data_tags[n_tag]["name"]

            x = sim_z
            y = get_sim_flux(n_tag)

            # label = self.get_title(tag_now)
            label = tag_now

            plot_style = self.mngr.plot_lib.get_plot_style(
                index=n_tag,
                tag="main_data",
                style_set=style_set,
            )
            _ = ax.plot(
                x,
                y,
                **plot_style,
                label=label,
            )

            plot_style = self.mngr.plot_lib.get_plot_style(
                index=n_tag,
                tag="err_data",
                style_set=style_set,
            )

        _ = ax.set_xlabel(x_label)
        _ = ax.set_ylabel(y_label)
        _ = ax.set_yscale("log")

        if legend_opts is not None:
            handles, labels = ax.get_legend_handles_labels()
            ax.legend(handles, labels, **legend_opts)

        output = {
            "fig": fig,
            "spec": spec,
            "ax": ax,
        }

        return output

    def plot_is_det_comp(self, data, data_tags, wgt_tags, opts={}):
        self.log.info([["g", " - plot-is-det-comp(): "], ["c", data_tags]])

        def get_det_col(tag):
            return str_join("is_det", tag)

        # detection fraction defined by ts threshold (e.g., anything above 5sigma)
        def det_ts_func(x):
            return 1 if (x >= self.mngr.ts_sqr_thresh) else 0

        # the function used to evaluate the detection fraction from a sample
        def np_avg(data, weights):
            return np.average(data, weights=weights)

        def update_opts(opt, grp):
            if grp in opts:
                opt.update(opts[grp])

        # uncertainty using bootstrapping
        def get_p_det_err(data, weights, det_func):
            # how many realisations to use
            n_smpls = 20
            # what if the fraction of entries per realisation
            frac_smpl = 0.7

            # number of events to sample
            n_evts = len(data)
            n_smpl_evt = int(n_evts * frac_smpl)

            # sanity check
            if n_evts < 10 or n_smpl_evt < 3:
                return [0, 0]

            # reset the index so that we get the correct filter for weight indices
            data = data.reset_index(drop=True)

            # derive list of realisations of the average
            p_err = []
            for _ in range(n_smpls):
                data_now = data.sample(n=n_smpl_evt)
                wgts_now = weights[data_now.index]
                p_err += [det_func(data=data_now, weights=wgts_now)]

            quants = np.quantile(p_err, [0.16, 0.5, 0.84])

            return [quants[2] - quants[1], quants[1] - quants[0]]

        # the general style group
        style_set = "purples"

        n_quant_bins = 10
        x_name = opts["x_name"]

        x_label = self.get_opt_or_def(opts, tag="x_label", default=x_name)
        y_label = r"$p_{\mathrm{det}}$"

        opts_fig = {
            "n_rows": 1,
            "n_cols": 1,
        }
        update_opts(opts_fig, "fig")

        legend_opts = opts["legend_opts"] if "legend_opts" in opts else {}

        if "fig_spec" in opts and (opts["fig_spec"] is not None):
            fig = opts["fig_spec"]["fig"]
            spec = opts["fig_spec"]["spec"]
        else:
            fig, spec = self.mngr.plot_lib.add_fig(**opts_fig)

        spec_n_row = self.get_opt_or_def(opts, grp="spec", tag="spec_n_row", default=0)
        spec_n_col = self.get_opt_or_def(opts, grp="spec", tag="spec_n_col", default=0)

        ax = self.mngr.plot_lib.add_subplot(fig, spec[spec_n_row, spec_n_col])

        output = {
            "fig": fig,
            "spec": spec,
            "ax": ax,
        }

        # split up the data into bins
        quants = np.linspace(0, 1, n_quant_bins + 1)
        x_bins = data[x_name].quantile(quants).values
        x_bins = np.unique(x_bins)

        # if there are too few events, cant continue
        if np.var(x_bins) < 1e-10:
            return output

        x_cuts = [
            pd.cut(
                data[x_name],
                x_bins,
                include_lowest=True,
                retbins=False,
                labels=False,
            )
        ]

        # go over all data samples
        for n_tag in range(len(data_tags)):
            tag_now = data_tags[n_tag]
            wgt_tag = wgt_tags[n_tag]

            det_col = get_det_col(tag_now)

            if wgt_tag is None:
                out_wgts = np.asarray([1 for _ in range(len(data[tag_now]))])
            else:
                out_wgts = np.asarray(data[wgt_tag])
            sum_wgts = np.sum(out_wgts)

            if sum_wgts < 1e-10:
                self.log.error(
                    [
                        ["wr", " - sum_wgts = ", sum_wgts, " ?!?"],
                        ["p", "\n", tag_now, "\n"],
                        ["r", data, "\n"],
                    ]
                )
                raise Exception("sum_wgts too small ?!?")

            det_0 = data[tag_now].apply(det_ts_func)
            # if self.clip_evt_lc_func is not None:
            #     det_1 = data['evt_lc'].apply(self.clip_evt_lc_func)
            #     data.loc[:, det_col] = det_0 * det_1
            # else:
            #     data.loc[:, det_col] = det_0
            data.loc[:, det_col] = det_0

            pd_grps = data.groupby(x_cuts)
            grp_keys = list(pd_grps.groups.keys())
            # print(pd_grps.agg('count').fillna(value=0)) # ; return
            # print(grp_keys)

            # derive the detection fraction for each bin
            x, y, e_n, e_p = [], [], [], []
            for n_grp in range(len(grp_keys)):
                grp_key = grp_keys[n_grp]
                grp = pd_grps.get_group(name=grp_key)

                wgts_now = out_wgts[grp.index]
                if np.sum(wgts_now) < 1e-10:
                    continue

                # bin centers
                x_bin_now = x_bins[n_grp] + 0.5 * (x_bins[n_grp + 1] - x_bins[n_grp])
                x += [x_bin_now]

                # detection fraction
                is_det_avg = np_avg(data=grp[det_col], weights=wgts_now)
                y += [is_det_avg]

                # uncertainty on detection fraction
                e = get_p_det_err(data=grp[det_col], weights=wgts_now, det_func=np_avg)
                e_n += [y[-1] - e[0]]
                e_p += [y[-1] + e[1]]

            label = self.get_title(tag_now)

            plot_style = self.mngr.plot_lib.get_plot_style(
                index=n_tag,
                tag="main_data",
                style_set=style_set,
            )
            _ = ax.plot(
                x,
                y,
                **plot_style,
                label=label,
            )

            plot_style = self.mngr.plot_lib.get_plot_style(
                index=n_tag,
                tag="err_data",
                style_set=style_set,
            )
            _ = ax.fill_between(
                x,
                e_n,
                e_p,
                **plot_style,
            )

        _ = ax.set_xlabel(x_label)
        _ = ax.set_ylabel(y_label)
        # _ = ax.set_yscale('log')

        if legend_opts is not None:
            handles, labels = ax.get_legend_handles_labels()
            ax.legend(handles, labels, **legend_opts)

        return output

    def get_opt_or_def(self, opts, grp=None, tag=None, default=None):
        if grp is None:
            return opts[tag] if tag in opts else default
        if grp not in opts:
            return default
        return opts[grp][tag] if tag in opts[grp] else default

    def get_title(self, tag):
        return self.tag_to_title[tag] if tag in self.tag_to_title.keys() else tag

    def set_title(self, tag, title):
        self.tag_to_title[tag] = title
        return

    def plot_model_output_representations(
        self, model_outputs: dict, evt_classes: List[str]
    ):
        # raise NotImplementedError(
        #     "How to access the model outputs of the predict phase? "
        # )
        # perdict_data = self.mngr.data.get_perdict_data(
        #     n_predict=-1,
        #     mode=ModeKeys.PREDICT,
        # )
        # model_inputs = perdict_data["inputs"]
        # model_outputs = self.mngr.model["keras_model"].predict(model_inputs)

        # labels = perdict_data["df"]["evt_class"].tolist()

        labels = evt_classes

        for i in range(len(labels)):
            if (
                labels[i] in self.mngr.evt_type_ratios["train"]
                and self.mngr.evt_type_ratios["train"][labels[i]] > 0
            ):
                labels[i] = labels[i] + " (training)"

        if self.mngr.debug_create_model_output_plots:
            auto_enc_hidden = model_outputs[self.mngr.tag_cluster_space_sample]
            auto_enc_input = model_outputs[self.mngr.tag_auto_enc_input]
            auto_enc_reconstruction = model_outputs[
                self.mngr.tag_auto_enc_reconstruction
            ]

            out_file_name = get_debug_output_filepath(
                self.mngr, "model_plots", f"predict_hidden_repr.pdf"
            )

            autoenc_create_all_plots(
                self.mngr,
                auto_enc_input,
                auto_enc_hidden,
                auto_enc_reconstruction,
                out_file_name,
                labels=labels,
            )

        if self.mngr.debug_create_model_output_csv_files:
            write_model_outputs__and_errors_to_csv_for_debugging(
                self.mngr, model_outputs
            )

    def plot_fit_p_values(self, data_ts, data_preds, is_verb=True):
        """plot the distributions of inputs, TS- and p-values for the calibration sample

        Parameters
        ----------
        data_ts : pd.DataFrame
            input data including columns for TS-values and weights
        data_preds : pd.DataFrame
            input data including columns for inputs and weights
        is_verb : bool
            flag indicating is extra logging info should be done
        """

        self.plot_meta_bin_fit_p_values(
            data_ts=data_ts,
            is_verb=is_verb,
        )

        calib_stage = self.mngr.calib_stages[0]
        grps_0, keys_0 = self.mngr.calib.get_flat_meta_groups(
            df=data_ts, calib_stage=calib_stage
        )

        for n_grp_0 in range(len(keys_0)):
            grp_key = keys_0[n_grp_0]
            indx = grps_0.get_group(name=grp_key).index
            df_grp = data_ts.loc[indx]

            self.plot_fit_p_values_inclusive_meta_groups(
                data_ts=df_grp,
                data_preds=data_preds[grp_key],
                grp_key=grp_key,
                is_verb=is_verb,
            )

    def plot_meta_bin_fit_p_values(self, data_ts, is_verb=True):
        """plot the distributions of inputs, TS- and p-values for the calibration sample

        Parameters
        ----------
        data_ts : pd.DataFrame
            input data including columns for TS-values and weights
        is_verb : bool
            flag indicating is extra logging info should be done
        """

        save_postfix = self.mngr.calib.get_meta_grp_postfix(
            calib_stage="meta_bin_p_values", grp_key=""
        )
        out_file_name = self.mngr.plot_lib.get_pdf_pages(
            tag=str_join("plot", save_postfix)
        )

        ts_tags, calib_stages = [], []
        for n_calib_stage in reversed(range(len(self.mngr.calib_stages))):
            if n_calib_stage == 0:
                if self.mngr.is_enc_dec_anom_det:
                    continue
                elif self.mngr.is_cluster_anom_det and self.mngr.use_reco_err_ts_combo:
                    continue

            calib_stage = self.mngr.calib_stages[n_calib_stage]
            calib_stages += [calib_stage]

            tag_ts_step_combo = self.mngr.get_calib_tags(
                calib_stage=calib_stage, tag="ts"
            )
            ts_tags += [tag_ts_step_combo]

        figs = []
        fig_opts = {
            "gridspec_opts": {
                "hspace": 0.35,
                "wspace": 0.35,
            },
        }
        style_set = "pastels"

        # consistent way to define histogram bins for the plots
        def get_x_bins(data, x_width=None, n_bins=None):
            try:
                data_max = data.max()
                data_min = data.min()
                if (n_bins is None) and (x_width is None):
                    raise
                if (n_bins is not None) and (x_width is not None):
                    raise
                if x_width is None:
                    if n_bins < 1:
                        raise
                    x_width = (data_max - data_min) / n_bins
                    x_width = max(x_width, 1e-6)

                x_min = min(data_min, 0)
                x_n_bins = int((data_max - x_min) // x_width)
                x_n_bins = max(x_n_bins, 10)
                x_max = x_min + x_n_bins * x_width
                x_max = max(x_max, 1)
                x_bins = np.linspace(x_min, x_max, x_n_bins + 1)
                # print('x_min, x_max, x_n_bins',x_min, x_max, x_n_bins)
            except Exception:
                raise Exception("problem with get_x_bins inputs:", x_width, n_bins)

            return x_bins

        for n_tag in range(len(ts_tags)):
            tag_ts = ts_tags[n_tag]
            calib_stage = calib_stages[n_tag]

            grps_0, keys_0 = self.mngr.calib.get_flat_meta_groups(
                df=data_ts, calib_stage=calib_stage
            )

            # plot p-value as a function of TS-value for each TS
            n_rows, n_cols = self.mngr.plot_lib.get_n_rows_cols(len(keys_0))

            fig, spec = self.mngr.plot_lib.add_fig(
                n_rows=n_rows,
                n_cols=n_cols,
                opts=fig_opts,
            )
            figs += [fig]

            wgt_tag = self.mngr.ts_to_wgt_col[tag_ts]

            ts_wgts, ts_wgt_sum = dict(), dict()

            for n_grp_0 in range(len(keys_0)):
                grp_key = keys_0[n_grp_0]
                indx = grps_0.get_group(name=grp_key).index
                df_grp = data_ts.loc[indx]

                mask = ~df_grp[tag_ts].isnull()
                df_grp = df_grp[mask]

                out_vals = np.asarray(df_grp[tag_ts])

                if wgt_tag in df_grp.columns:
                    ts_wgts[tag_ts] = np.asarray(df_grp[wgt_tag])
                else:
                    ts_wgts[tag_ts] = np.ones(df_grp[tag_ts].shape)
                ts_wgt_sum[tag_ts] = np.sum(ts_wgts[tag_ts])

                n_cdf_bins = 400
                quant_range = [0, 1]
                cdf_x, cdf_y = [], []

                cdf_range = self.mngr.calib.get_quants(
                    data=out_vals,
                    quants=quant_range,
                    weights=ts_wgts[tag_ts],
                )
                # print('cdf_range',quant_range,cdf_range)

                cdf_bins_x = np.linspace(cdf_range[0], cdf_range[1], n_cdf_bins)
                cdf_x += list(cdf_bins_x)

                out_norm = 1
                if ts_wgt_sum[tag_ts] > 0:
                    out_norm /= ts_wgt_sum[tag_ts]

                for ts_val in cdf_bins_x:
                    wgts_in = ts_wgts[tag_ts][out_vals >= ts_val]
                    cdf_y_now = np.sum(wgts_in) * out_norm
                    cdf_y += [cdf_y_now]

                n_row_col_shift = 0
                n_row_now = (n_grp_0 + n_row_col_shift) // n_cols
                n_col_now = (n_grp_0 + n_row_col_shift) - n_row_now * n_cols

                label = f"""bin {grp_key}"""

                time_str = datetime.utcnow().strftime("(UTC) %Y%m%d, %H:%M:%S")
                time_str = str("(" + time_str + ")")
                fig.suptitle(
                    time_str
                    + " "
                    + f"""{self.mngr.df_cols_ts_title[ts_tags[n_tag]]}""",
                    fontsize=16,
                    color="#F756B6",
                )

                plot_style = self.mngr.plot_lib.get_plot_style(
                    index=n_grp_0,
                    tag="main_data",
                    style_set=style_set,
                )

                ax = self.mngr.plot_lib.add_subplot(fig, spec[n_row_now, n_col_now])

                _ = ax.plot(
                    cdf_x,
                    cdf_y,
                    **plot_style,
                    label=label,
                )

                min_range, max_range = np.min(cdf_x), np.max(cdf_x)
                d_range = max_range - min_range
                min_range -= 0.2 * d_range
                max_range += 0.2 * d_range
                ts_vals = np.linspace(min_range, max_range, n_cdf_bins + 1)

                eval_output = self.mngr.calib.eval_p_values(
                    df_calib=ts_vals,
                    ts_tags=tag_ts,
                    calib_fit_tag=grp_key,
                )
                fit_vals = eval_output["log_p_vals"]

                # to avoid numerical errors on taking the yscale
                mask = fit_vals > -30
                fit_vals = fit_vals[mask]
                ts_vals = ts_vals[mask]

                fit_vals = np.power(10, fit_vals).flatten()

                plot_style["color"] = "black"
                plot_style["linestyle"] = ":"
                plot_style["linewidth"] = float(plot_style["linewidth"]) * 1.1

                label = "Fit"

                _ = ax.plot(
                    ts_vals,
                    fit_vals,
                    **plot_style,
                    label=label,
                )

                _ = plt.legend()
                _ = ax.set_xlabel("TS")
                _ = ax.set_ylabel("p-value")
                _ = ax.set_yscale("log")

        with PdfPages(out_file_name) as pdf_pages:
            self.mngr.plot_lib.save_pdf_pages(pdf_pages, out_file_name, figs=figs)

    def plot_fit_p_values_inclusive_meta_groups(
        self, data_ts, data_preds, grp_key, is_verb=True
    ):
        """plot the distributions of inputs, TS- and p-values for the calibration sample

        Parameters
        ----------
        data_ts : pd.DataFrame
            input data including columns for TS-values and weights
        data_preds : pd.DataFrame
            input data including columns for inputs and weights
        grp_key : tuple
            unique key (cut combination) for the input data
        is_verb : bool
            flag indicating is extra logging info should be done
        """

        save_postfix = self.mngr.calib.get_meta_grp_postfix(
            calib_stage="inclusive_meta_groups", grp_key=grp_key
        )
        out_file_name = self.mngr.plot_lib.get_pdf_pages(
            tag=str_join("plot", save_postfix)
        )

        # all calib stages which share exactly the same meta-bins as the first calib stage
        base_calib_stages = []
        for calib_stage, cols in self.mngr.meta_bin_calib_stage_cols.items():
            has_bins = [b in cols for b in self.mngr.meta_bin_col_params.keys()]
            if all(has_bins) and calib_stage != self.mngr.calib_stages[0]:
                base_calib_stages += [calib_stage]

        # make a local copy of the TS tags of the first calibration stage, and add to it
        # for all following stages which share the same base meta-bins
        ts_tags = [] + self.mngr.df_cols_ts_0
        for calib_stage in base_calib_stages:
            tag_ts_step_combo = self.mngr.get_calib_tags(
                calib_stage=calib_stage, tag="ts"
            )
            ts_tags += [tag_ts_step_combo]

        tags_prd = self.mngr.df_cols_pred
        tags_lbl = self.mngr.df_cols_dec

        figs = []
        fig_opts = {
            "gridspec_opts": {
                "hspace": 0.35,
                "wspace": 0.35,
            },
        }
        style_set = "pastels"

        # consistent way to define histogram bins for the plots
        def get_x_bins(data, x_width=None, n_bins=None):
            try:
                data_max = data.max()
                data_min = data.min()
                if (n_bins is None) and (x_width is None):
                    raise
                if (n_bins is not None) and (x_width is not None):
                    raise
                if x_width is None:
                    if n_bins < 1:
                        raise
                    x_width = (data_max - data_min) / n_bins
                    x_width = max(x_width, 1e-6)

                x_min = min(data_min, 0)
                x_n_bins = int((data_max - x_min) // x_width)
                x_n_bins = max(x_n_bins, 10)
                x_max = x_min + x_n_bins * x_width
                x_max = max(x_max, 1)
                x_bins = np.linspace(x_min, x_max, x_n_bins + 1)
                # print('x_min, x_max, x_n_bins',x_min, x_max, x_n_bins)
            except Exception:
                raise Exception("problem with get_x_bins inputs:", x_width, n_bins)

            return x_bins

        # plot p-value as a function of TS-value for each TS
        n_rows, n_cols = self.mngr.plot_lib.get_n_rows_cols(len(ts_tags))

        fig, spec = self.mngr.plot_lib.add_fig(
            n_rows=n_rows,
            n_cols=n_cols,
            opts=fig_opts,
        )
        figs += [fig]

        ts_wgts, ts_wgt_sum = dict(), dict()

        for n_tag in range(len(ts_tags)):
            tag_ts = ts_tags[n_tag]
            wgt_tag = self.mngr.ts_to_wgt_col[tag_ts]

            data_now = data_ts
            mask = ~data_now[tag_ts].isnull()
            data_now = data_now[mask]

            out_vals = np.asarray(data_now[tag_ts])

            if wgt_tag in data_now.columns:
                ts_wgts[tag_ts] = np.asarray(data_now[wgt_tag])
            else:
                ts_wgts[tag_ts] = np.ones(data_now[tag_ts].shape)
            ts_wgt_sum[tag_ts] = np.sum(ts_wgts[tag_ts])

            n_cdf_bins = 400
            quant_range = [0, 1]
            cdf_x, cdf_y = [], []

            cdf_range = self.mngr.calib.get_quants(
                data=out_vals,
                quants=quant_range,
                weights=ts_wgts[tag_ts],
            )
            # print('cdf_range',quant_range,cdf_range)

            cdf_bins_x = np.linspace(cdf_range[0], cdf_range[1], n_cdf_bins)
            cdf_x += list(cdf_bins_x)

            out_norm = 1
            if ts_wgt_sum[tag_ts] > 0:
                out_norm /= ts_wgt_sum[tag_ts]

            for ts_val in cdf_bins_x:
                wgts_in = ts_wgts[tag_ts][out_vals >= ts_val]
                cdf_y_now = np.sum(wgts_in) * out_norm
                cdf_y += [cdf_y_now]

            n_row_col_shift = 0
            n_row_now = (n_tag + n_row_col_shift) // n_cols
            n_col_now = (n_tag + n_row_col_shift) - n_row_now * n_cols

            label = (
                f"""{self.mngr.df_cols_ts_title[ts_tags[n_tag]]} (#{len(out_vals)})"""
            )
            label = self.mngr.df_cols_ts_title[ts_tags[n_tag]]

            time_str = datetime.utcnow().strftime("(UTC) %Y%m%d, %H:%M:%S")
            time_str = str("(" + time_str + ")")
            fig.suptitle(
                time_str + " " + ("cuts: " + str(grp_key)),
                fontsize=16,
                color="#F756B6",
            )

            plot_style = self.mngr.plot_lib.get_plot_style(
                index=n_tag,
                tag="main_data",
                style_set=style_set,
            )

            ax = self.mngr.plot_lib.add_subplot(fig, spec[n_row_now, n_col_now])

            _ = ax.plot(
                cdf_x,
                cdf_y,
                **plot_style,
                label=label,
            )

            min_range, max_range = np.min(cdf_x), np.max(cdf_x)
            d_range = max_range - min_range
            min_range -= 0.2 * d_range
            max_range += 0.2 * d_range
            ts_vals = np.linspace(min_range, max_range, n_cdf_bins + 1)

            eval_output = self.mngr.calib.eval_p_values(
                df_calib=ts_vals,
                ts_tags=tag_ts,
                calib_fit_tag=grp_key,
            )
            fit_vals = eval_output["log_p_vals"]

            # avoid numerical underflow errors with large negative exponents by filtering too large exponents
            exponent_limit = np.log10(np.finfo(fit_vals.dtype).tiny)
            mask = fit_vals > exponent_limit
            n_discarded = np.sum(mask)
            ts_vals = ts_vals[mask]
            fit_vals = fit_vals[mask]

            if n_discarded > 0:
                self.log.warn(
                    [
                        [
                            "r",
                            f"Discarding {n_discarded} rows in plot_fit_p_values_inclusive_meta_groups "
                            f"to circumvent avoid numerical underflow.",
                        ]
                    ]
                )

            fit_vals = np.power(10, fit_vals).flatten()

            # set lower treshold on yscale for plotting
            mask = fit_vals > 1e-30
            n_discarded = np.sum(mask)
            ts_vals = ts_vals[mask]
            fit_vals = fit_vals[mask]

            if n_discarded > 0:
                self.log.warn(
                    [
                        [
                            "r",
                            f"Discarding {n_discarded} rows in plot_fit_p_values_inclusive_meta_groups "
                            f"due to lower-yscale-threshold.",
                        ]
                    ]
                )

            plot_style["color"] = "black"
            plot_style["linestyle"] = ":"
            plot_style["linewidth"] = float(plot_style["linewidth"]) * 1.1

            label = "Fit" if n_tag == 0 else None

            _ = ax.plot(
                ts_vals,
                fit_vals,
                **plot_style,
                label=label,
            )

            _ = plt.legend()
            _ = ax.set_xlabel("TS")
            _ = ax.set_ylabel("p-value")
            _ = ax.set_yscale("log")

        # plot significance as a function of TS-value and compare to a chi2 dist.
        n_rows, n_cols = self.mngr.plot_lib.get_n_rows_cols(len(ts_tags))

        fig, spec = self.mngr.plot_lib.add_fig(
            n_rows=n_rows,
            n_cols=n_cols,
            opts=fig_opts,
        )
        figs += [fig]

        for n_tag in range(len(ts_tags)):
            tag_ts = ts_tags[n_tag]

            data_now = data_ts
            mask = ~data_now[tag_ts].isnull()
            data_now = data_now[mask]

            eval_output = self.mngr.calib.eval_p_values(
                df_calib=data_now,
                ts_tags=tag_ts,
                calib_fit_tag=grp_key,
            )
            df_calib = eval_output["df_calib"]
            log_p_val_tags = eval_output["log_p_val_tags"]

            # sanity check
            if len(log_p_val_tags) != 1:
                raise Exception(
                    f" - unexpected output for eval_p_values({ts_tags})", eval_output
                )

            log_p_vals = df_calib[log_p_val_tags[0]]

            # log_p_vals = log_p_vals.astype(float)
            p_vals = np.power(10, log_p_vals)

            # signif_sqr = stats.chi2.isf(p_vals, df=1).flatten()
            signif_sqr = stats.chi2.isf(0.5 * p_vals, df=1).flatten()

            # def ts_isf_to_sig_sqr(ts):
            #     # return np.sqrt(ts) if ts > 0 else 0
            #     return ts

            # ts_isf_to_sig_sqr = np.vectorize(ts_isf_to_sig_sqr, otypes=[float])
            # out_vals = ts_isf_to_sig_sqr(signif_sqr)
            out_vals = signif_sqr

            out_wgts = np.asarray(ts_wgts[tag_ts])

            sum_wgts = np.sum(out_wgts)

            x_width = 1
            x_bins = get_x_bins(out_vals, x_width=x_width)
            # x_min = min(out_vals.min(), 0)
            # x_n_bins = int((out_vals.max() - x_min) // x_width)
            # x_n_bins = max(x_n_bins, 10)
            # x_max = x_min + x_n_bins * x_width
            # x_max = max(x_max, 1)
            # # print('x_min, x_max, x_n_bins',x_min, x_max, x_n_bins)
            # x = np.linspace(x_min, x_max, x_n_bins + 1)

            # normalise, so that the pdf will also be the
            # probabiity mass function [--> integral==sum==1]
            out_wgts = out_wgts / x_width
            if sum_wgts > 0:
                out_wgts /= sum_wgts

            y, _ = np.histogram(
                out_vals,
                weights=out_wgts,
                bins=x_bins,
            )
            x = [
                x_bins[n] + 0.5 * (x_bins[n + 1] - x_bins[n])
                for n in range(len(x_bins) - 1)
            ]
            # for n in range(len(y)):
            #     print(n,x[n],y[n])

            plot_style = self.mngr.plot_lib.get_plot_style(
                index=n_tag,
                tag="main_data",
                style_set=style_set,
            )

            n_row_col_shift = 0
            # n_row_col_shift = len(ts_tags)
            n_row_now = int((n_tag + n_row_col_shift) // n_cols)
            n_col_now = (n_tag + n_row_col_shift) - n_row_now * n_cols

            # label = str(ts_tags[n_tag]) + ' (#' + str(len(out_vals)) + ')'
            label = (
                f"""{self.mngr.df_cols_ts_title[ts_tags[n_tag]]} (#{len(out_vals)})"""
            )

            ax = self.mngr.plot_lib.add_subplot(fig, spec[n_row_now, n_col_now])

            _ = ax.plot(
                x,
                y,
                **plot_style,
                label=label,
            )

            # x_max_chi_2 = min(ax.get_xlim()[1], 30)
            # try:
            #     x_max_chi_2 = x_max_chi_2**2
            # except:
            #     pass

            x_max_chi_2 = ax.get_xlim()[1]

            n_bins = 100
            x = np.linspace(0, x_max_chi_2, n_bins + 1)
            y = stats.chi2.sf(x, df=1)
            # y = 0.5 * stats.chi2.sf(x, df=1)
            # for n in range(len(y)):
            #     print(n,x[n],y[n])

            label = r"$\chi^{2}_{(1)}$" if n_tag == 0 else None
            # (?# label = r'$\frac{1}{2}\chi^{2}_{(1)}$' if n_tag==0 else None)

            _ = ax.plot(
                x,
                y,
                alpha=0.7,
                linewidth=1,
                color="black",
                linestyle="-",
                label=label,
            )

            _ = plt.legend()

            _ = ax.set_xlabel(r"$\sigma^{2}$")
            _ = ax.set_ylabel(r"1/N dN/d$\sigma^{2}$")
            _ = ax.set_yscale("log")

        # plot distributions of the decoder values and their corresponding predictions
        if self.mngr.is_enc_dec_anom_det or self.mngr.is_cluster_anom_det:
            fig_all, spec_all = self.mngr.plot_lib.add_fig(
                n_rows=2,
                n_cols=2,
                opts=fig_opts,
            )
            figs += [fig_all]
            ax_all = [
                self.mngr.plot_lib.add_subplot(fig_all, spec_all[i, j])
                for (i, j) in [[0, 0], [0, 1], [1, 0], [1, 1]]
            ]

            n_rows, n_cols = self.mngr.plot_lib.get_n_rows_cols(len(tags_prd))

            fig_pairs, spec_pairs = self.mngr.plot_lib.add_fig(
                n_rows=n_rows,
                n_cols=n_cols,
                opts=fig_opts,
            )
            figs += [fig_pairs]

            for n_tag in range(len(tags_prd)):
                n_row_now = int((n_tag) // n_cols)
                n_col_now = (n_tag) - n_row_now * n_cols

                ax = self.mngr.plot_lib.add_subplot(
                    fig_pairs,
                    spec_pairs[n_row_now, n_col_now],
                )

                plot_tags = [tags_lbl[n_tag], tags_prd[n_tag]]
                for n_plot_tag in range(len(plot_tags)):
                    tag_now = plot_tags[n_plot_tag]
                    style_n_shift = 0
                    # style_n_shift = n_plot_tag * len(tags_prd)

                    plot_style = self.mngr.plot_lib.get_plot_style(
                        index=n_tag + style_n_shift,
                        tag="main_data",
                        style_set=style_set,
                    )
                    self.mngr.plot_lib.set_line_style(style_set, n_plot_tag, plot_style)

                    out_vals = np.asarray(data_preds[tag_now])
                    out_wgts = np.asarray(ts_wgts[tag_ts])

                    sum_wgts = np.sum(out_wgts)
                    if sum_wgts < 1e-10:
                        self.log.error(
                            [
                                ["wr", " - sum_wgts = ", sum_wgts, " ?!?"],
                                ["p", "\n", tag_ts, " / ", tag_now, "\n"],
                                ["r", data_preds[tag_now], "\n"],
                                ["y", ts_wgts[tag_ts], "\n"],
                            ]
                        )
                        raise Exception("sum_wgts too small ?!?")

                    n_bins = 10
                    x_bins = get_x_bins(out_vals, n_bins=n_bins)

                    # normalise, so that the pdf will also be the
                    # probabiity mass function [--> integral==sum==1]
                    out_wgts = out_wgts / (sum_wgts * x_width)

                    # y, x = np.histogram(out_vals, weights=out_wgts, bins=10)
                    y, _ = np.histogram(
                        out_vals,
                        weights=out_wgts,
                        bins=x_bins,
                    )
                    x = [
                        x_bins[n] + 0.5 * (x_bins[n + 1] - x_bins[n])
                        for n in range(len(x_bins) - 1)
                    ]

                    plot_style["linewidth"] = float(plot_style["linewidth"]) * (
                        1 + 0.3 * n_plot_tag
                    )

                    _ = ax.plot(
                        x,
                        y,
                        **plot_style,
                        label=tag_now,
                    )

                    # single type
                    _ = ax_all[n_plot_tag].plot(
                        x,
                        y,
                        **plot_style,
                        label=tag_now,
                    )
                    _ = ax_all[n_plot_tag].set_xlabel(
                        "decoder" if (n_plot_tag == 0) else "predictions"
                    )

                    # all together
                    _ = ax_all[2].plot(
                        x,
                        y,
                        **plot_style,
                        label=tag_now,
                    )
                    _ = ax_all[2].set_xlabel("decoder, predictions")

                    # just for the legend...
                    _ = ax_all[3].plot(
                        [0],
                        [0],
                        **plot_style,
                        label=tag_now,
                    )
                    _ = ax_all[3].set_xlabel("")

                for ax_now in ax_all:
                    _ = ax_now.set_ylabel("PDF")
                _ = ax_all[-1].legend()

                _ = ax.legend()
                _ = ax.set_xlabel("decoder, predictions")
                _ = ax.set_ylabel("PDF")
                # _ = ax.set_yscale('log')

        # plot the distributions of TS-values
        if self.mngr.is_enc_dec_anom_det or self.mngr.is_cluster_anom_det:
            n_rows, n_cols = self.mngr.plot_lib.get_n_rows_cols(len(ts_tags))

            fig_pairs, spec_pairs = self.mngr.plot_lib.add_fig(
                n_rows=n_rows,
                n_cols=n_cols,
                opts=fig_opts,
            )
            figs += [fig_pairs]

            for n_tag in range(len(ts_tags)):
                tag_ts = ts_tags[n_tag]

                data_now = data_ts
                mask = ~data_now[tag_ts].isnull()
                data_now = data_now[mask]

                out_vals = np.asarray(data_now[tag_ts])
                out_wgts = np.asarray(ts_wgts[tag_ts])

                sum_wgts = np.sum(out_wgts)
                if sum_wgts < 1e-10:
                    self.log.error(
                        [
                            ["wr", " - sum_wgts = ", sum_wgts, " ?!?"],
                            ["p", "\n", tag_ts, " / ", "\n"],
                            ["r", data_now[tag_ts], "\n"],
                            ["y", ts_wgts[tag_ts], "\n"],
                        ]
                    )
                    raise Exception("sum_wgts too small ?!?")

                n_bins = 10
                x_bins = get_x_bins(out_vals, n_bins=n_bins)

                # normalise, so that the pdf will also be the
                # probabiity mass function [--> integral==sum==1]
                out_wgts = out_wgts / (sum_wgts * x_width)

                y, _ = np.histogram(
                    out_vals,
                    weights=out_wgts,
                    bins=x_bins,
                )
                x = [
                    x_bins[n] + 0.5 * (x_bins[n + 1] - x_bins[n])
                    for n in range(len(x_bins) - 1)
                ]

                # y, x = np.histogram(out_vals, weights=out_wgts, bins=10)
                # x = [x[n] + 0.5 * (x[n + 1] - x[n]) for n in range(len(x) - 1)]
                # # y = y / sum_wgts

                n_row_col_shift = 0
                # n_row_col_shift = len(tags_prd)
                n_row_now = int((n_tag + n_row_col_shift) // n_cols)
                n_col_now = (n_tag + n_row_col_shift) - n_row_now * n_cols

                ax = self.mngr.plot_lib.add_subplot(
                    fig_pairs,
                    spec_pairs[n_row_now, n_col_now],
                )

                plot_style = self.mngr.plot_lib.get_plot_style(
                    index=n_tag,
                    tag="main_data",
                    style_set=style_set,
                )

                label = self.mngr.df_cols_ts_title[ts_tags[n_tag]]

                _ = ax.plot(
                    x,
                    y,
                    **plot_style,
                    label=label,
                )

                _ = plt.legend()
                _ = ax.set_xlabel("TS")
                _ = ax.set_ylabel("PDF")
                _ = ax.set_yscale("log")

        with PdfPages(out_file_name) as pdf_pages:
            self.mngr.plot_lib.save_pdf_pages(pdf_pages, out_file_name, figs=figs)
