from __future__ import annotations

import os
import os.path
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
from tensorflow.estimator import ModeKeys

from trans_finder.flares.channel_group_masking import masked_channel_groups_inputs
from trans_finder.flares.checks.generic_checks import (
    plot_dim_hists_of_ndarrays,
    temp_load,
)
from trans_finder.flares.core.tags import TAG_CUSTOM_REALISATIONS
from trans_finder.flares.exceptions import InputOutOfRange
from trans_finder.flares.helpers_debugging import (
    clear_directory,
    get_debug_output_filepath,
)
from trans_finder.flares.plot_list.performance_by_time_sliding_window import (
    do_custom_realisation_time_series_plots,
)
from trans_finder.flares.plot_list.util.text_figure import create_text_figure
from trans_finder.flares.utilities.pdf_plot_collection import PDFPlotCollection
from trans_finder.flares.utils import get_n_write_file, str_join

if TYPE_CHECKING:
    from trans_finder.flares.log_parser import LogParser
    from trans_finder.flares.manager import Manager


class Predictions:
    def __init__(self, mngr):
        self.mngr: Manager = mngr
        self.log: LogParser = mngr.log

        self.n_batch_in_error_analysis = -1

    def get_predictions_output_filter(self, df_out):
        return df_out  # .groupby(self.mngr.tag_event_id).first().reset_index()

    def on_predict_model_outputs(self, df_out, perdict_data, model_outputs):
        if (
            self.mngr.run_phase == "predict"
            and self.mngr.dir_input_predict == TAG_CUSTOM_REALISATIONS
            and self.mngr.debug_plot_predict_model_outputs
        ):
            # for cross-checking, plot model output distributions of custom realisations

            out_dir = os.path.join(TAG_CUSTOM_REALISATIONS, "model_outputs")

            bgmm_fit_outputs_train = temp_load("bgmm_fit_outputs_train")
            bgmm_fit_outputs_val = temp_load("bgmm_fit_outputs_val")

            for key in bgmm_fit_outputs_train.keys():
                plot_dim_hists_of_ndarrays(
                    self.mngr,
                    [
                        bgmm_fit_outputs_train[key],
                        bgmm_fit_outputs_val[key],
                        model_outputs[key],
                    ],
                    ["bgmm_fit_train", "bgmm_fit_val", "custom_realisations"],
                    get_debug_output_filepath(self.mngr, out_dir, f"temp_{key}.pdf"),
                    # subplots_kwargs=dict(sharex=True),
                    hist_kwargs=dict(density=True),
                )

            # plot_dim_hists_of_ndarrays(
            #     self.mngr,
            #     [
            #         bgmm_fit_outputs_train["model_auto_enc_input"],
            #         bgmm_fit_outputs_train["model_auto_enc_reconstruction"],
            #     ],
            #     ["model_auto_enc_input", "model_auto_enc_reconstruction"],
            #     get_debug_output_filepath(
            #         self.mngr, "bgmmodel", "temp_bgmm_train_ae.pdf"
            #     ),
            #     hist_kwargs=dict(density=True),
            # )

            plot_dim_hists_of_ndarrays(
                self.mngr,
                [
                    model_outputs["model_auto_enc_input"],
                    model_outputs["model_auto_enc_reconstruction"],
                ],
                ["model_auto_enc_input", "model_auto_enc_reconstruction"],
                get_debug_output_filepath(
                    self.mngr, out_dir, "temp_custom_realisations_ae.pdf"
                ),
                hist_kwargs=dict(density=True),
            )

    def get_predictions(self):
        self.log.info([*self.mngr.main_func_prnt, ["g", " - get-predictions() "]])

        ts_cols = list(self.mngr.ts_to_p_val_col.keys())
        if self.mngr.is_enc_dec_anom_det or (
            self.mngr.is_cluster_anom_det and self.mngr.use_reco_err_ts_combo
        ):
            tag_ts_step_0 = self.mngr.get_calib_tags(
                calib_stage=self.mngr.calib_stages[0], tag="ts"
            )
            ts_cols.remove(tag_ts_step_0)

        p_val_cols = [self.mngr.ts_to_p_val_col[t] for t in ts_cols]

        max_write_evt = self.mngr.max_write_evt * (
            2 if self.mngr.is_enc_dec_clasif else 1
        )

        # make a local copy and add to it
        wgt_cols = [] + self.mngr.df_cols_wgt_0
        for n_calib_stage in range(1, len(self.mngr.calib_stages)):
            tag_wgt_combo = self.mngr.get_calib_tags(
                calib_stage=self.mngr.calib_stages[n_calib_stage], tag="wgt"
            )
            wgt_cols += [tag_wgt_combo]

        tag_p_val = self.mngr.get_calib_tags(
            calib_stage=self.mngr.calib_stages[-1], tag="log_p_val"
        )

        pd_df_write = pd.DataFrame()

        n_predict = -1
        n_accepted_evts, n_skipped_evts = 0, 0

        # loop over all events
        while True:
            try:
                perdict_data = self.mngr.data.get_perdict_data(
                    n_predict=n_predict,
                    mode=ModeKeys.PREDICT,
                )
            except InputOutOfRange:
                pd_df_write = self.write_output_df(pd_df_write)
                break

            # evaluate the data without any masking
            out_inputs = [perdict_data["inputs"]]
            out_dfs = [perdict_data["df"].assign(input_mask="full")]

            # add extra evaluations with input masks
            for mask_name, modified_inputs in masked_channel_groups_inputs(
                self.mngr, predict_in=perdict_data["inputs"]
            ):
                out_inputs.append(modified_inputs)
                out_dfs.append(perdict_data["df"].assign(input_mask=mask_name))

            df_out = pd.concat(out_dfs).reset_index(drop=True)
            inputs = {}
            for k in perdict_data["inputs"].keys():
                inputs[k] = np.concatenate([v[k] for v in out_inputs], axis=0)

            n_predict += len(perdict_data["df"])

            # create new columns for the new output (initial value
            # determines dtype, which is important for hdf optimisation)
            for col in p_val_cols:
                df_out.loc[:, col] = self.mngr.p_value_nan_value
            for col in wgt_cols:
                df_out.loc[:, col] = 0
            # df_out.loc[:, self.tag_metadata_grp] = ""
            df_out.loc[:, self.mngr.tag_has_p_value] = True
            df_out.loc[:, self.mngr.tag_evt_ts_mask] = True

            # add the flat representation of the metadata groups, needed for p-value derivation
            # (must happen before get_ts_from_preds, since this colums might be needed for diagnostics)
            self.mngr.calib.add_flat_meta_keys(df_calib=df_out)

            # construct the input for the estimator
            model_inputs = inputs
            # get the response of the network
            model_outputs = self.mngr.model[self.mngr.tag_model].predict(
                model_inputs, self.mngr.predict_batch_size
            )

            ts_output = self.mngr.calib.get_ts_from_preds(
                model_inputs=model_inputs,
                model_outputs=model_outputs,
                df_full=df_out,
            )
            tss = ts_output["ts"]
            wgts = ts_output["wgt"]

            df_ts_now = pd.DataFrame(tss, columns=self.mngr.df_cols_ts_0)
            # print(df_ts_now.agg('describe')) # ; return

            clip_mask = None
            if self.mngr.tag_evt_ts_mask in ts_output.keys():
                clip_mask = ts_output[self.mngr.tag_evt_ts_mask].astype(bool)
                df_ts_now.loc[:, self.mngr.tag_evt_ts_mask] = clip_mask.astype(bool)

            # merge the weights and perform the combination
            # and p-value estimation
            df_ts_now = df_ts_now.merge(
                pd.DataFrame(wgts, columns=self.mngr.df_cols_wgt_0),
                left_index=True,
                right_index=True,
            )

            # now the merge operation itself
            df_out.loc[:, df_ts_now.columns] = df_ts_now

            # derive the p-values for all calibration stages (including derivation of TS values
            # for all calibration stages, exclusind the first one, which is already present)
            eval_output = self.mngr.calib.get_p_values_by_stage(
                df_calib=df_out,
                calib_stage=self.mngr.calib_stages[-1],
            )
            # print(df_out.columns)
            # print(df_out)
            # print(tag_p_val)
            # print(df_out[tag_p_val].values[0])
            # print(df_out[tag_p_val].values[0])
            # print(df_out[ ~df_out[tag_p_val].isnull()  ])

            if len(eval_output["exceptions"]) > 0:
                exp_msg = " -- ".join(str(x) for x in eval_output["exceptions"])
                self.log.warn(
                    [["wr", " - missing calibrations: "], ["b", "  ", exp_msg]]
                )

            missing = np.isclose(df_out[tag_p_val], self.mngr.p_value_nan_value)

            n_missing = np.sum(missing)
            n_skipped_evts += n_missing
            n_accepted_evts += len(missing) - n_missing

            df_out.loc[:, self.mngr.tag_has_p_value] = ~missing

            # n_accepted_evts += len(indx)
            # print(log_p_vals_now.columns)
            # print(df_out.columns)

            self.on_predict_model_outputs(df_out, perdict_data, model_outputs)

            df_out = self.get_predictions_output_filter(df_out)

            # write out the modified df
            pd_df_write = pd.concat([pd_df_write, df_out], sort=False)
            pd_df_write = pd_df_write.sort_index().reset_index(drop=True)

            # print(pd_df_write[pd_df_write["evt_class"] == "bck_0"])
            # print(
            #     pd_df_write[pd_df_write["evt_class"] == "bck_0"][
            #         "log_p_value_calib_cluster_single"
            #     ].describe()
            # )
            # print("-" * 120)
            # print(pd_df_write[pd_df_write["evt_class"] == "bck_fluct"])
            # print(
            #     pd_df_write[pd_df_write["evt_class"] == "bck_fluct"][
            #         "log_p_value_calib_cluster_single"
            #     ].describe()
            # )
            # print("-" * 120)
            # assert 0

            self.do_error_analysis(model_outputs, df_out)

            if (
                len(pd_df_write) >= max_write_evt
                or self.mngr.write_predict_out_each_iteration
            ):
                pd_df_write = self.write_output_df(pd_df_write)

        self.log.info([["g", " --> number of accepted events: ", n_accepted_evts]])
        if n_skipped_evts > 0:
            self.log.warn(
                [["wr", " --> number of skipped events: "], ["r", " ", n_skipped_evts]]
            )

        # optional diagnostics plotting after predictions (and possibly after calibration)
        self.mngr.diagnostics.plot_diagnostics()

    def write_output_df(self, df_in, output_dir=None):
        """write outout hdf output files

        Parameters
        ----------
        df_in : pd.DataFrame
            input data to write to file

        Returns
        -------
        dict
            updated DataFrame and counter for further writing
        """

        # nothing to be done if there are no data
        if len(df_in) == 0:
            return df_in

        # write out a single file
        def write_one(df_now, file_name):
            df_now.to_hdf(
                file_name,
                key=self.mngr.tag_hdf,
                mode="w",
            )
            self.log.info(
                [
                    ["y", " - # events: " + str(len(df_now)) + " written to: "],
                    ["b", file_name.replace(self.mngr.tag_base_output_dir, "")],
                ]
            )
            return

        n_file = get_n_write_file(
            dir_path=output_dir, out_file_patt=self.mngr.out_file_patt
        )

        file_name = self.get_out_file_name(n_file, df_in=df_in)

        if output_dir is None:
            if self.mngr.run_phase in ["predict"]:
                file_path = output_dir = self.mngr.data.get_dir_name("predict_out")
            elif self.mngr.run_phase in ["predict_sim"]:
                file_path = output_dir = self.mngr.data.get_dir_name("predict_sim_out")

        file_path = os.path.join(output_dir, file_name)

        write_one(df_in, file_path)

        return pd.DataFrame()

    def get_out_file_name(self, n_file, **kwargs):
        in_file_filt_str = ""
        if self.mngr.in_file_filt_now is not None:
            in_file_filt_str = self.mngr.data.get_n_file_filt()

        file_name = str_join(
            self.mngr.out_file_patt[0],
            in_file_filt_str,
            str("%08d" % n_file),
            self.mngr.out_file_patt[1],
        ).replace("__", "_")

        return file_name

    def get_desired_ts_by_evt_class(self, evt_class):
        ts_targets = {
            "bck_pure": -np.inf,
            # "fluct_enc": None,
            # "fluct_dec": np.inf,
            # "fluct_enc_dec": None,
            # "flare_enc": None,
            # "flare_dec": np.inf,
            # "flare_enc_dec": None,
            "signal_enc": None,
            "signal_dec": np.inf,
            "signal_enc_dec": None,
        }

        if evt_class not in ts_targets:
            raise ValueError(
                f"get_desired_ts_by_evt_class() not specified for '{evt_class}'"
            )

        return ts_targets[evt_class]

    def denormalise_temporal_offsets(self, df) -> pd.DataFrame:
        if not self.mngr.do_train_scale or df.empty:
            return df

        column_indices = self.mngr.temporal_offset_column_indices

        norms = self.mngr.data.train_feat_norms[self.mngr.tag_steps]
        shift = norms["shift"][0, column_indices]
        scale = norms["scale"][0, column_indices]

        df.iloc[:, column_indices] /= scale
        df.iloc[:, column_indices] -= shift

        return df

    def do_error_analysis(self, model_outputs, df_out):
        """called during prediction"""

        do_custom_realisation_time_series_plots(model_outputs, df_out)

        # tag = ["b", f" - flares.{Path(__file__).stem} - "]

        # self.log.info([tag, ["y", "do_error_analysis()"]])
        n_examples = self.mngr.error_analysis_n_examples
        ts_col_name = self.mngr.df_cols_ts_0[0]
        self.n_batch_in_error_analysis += 1

        # df_out = df_out.assign(signal_step_sum=self.mngr.diagnostics.get_signal_step_sum(df_out))

        # n_batch_limit = self.error_analysis_n_batch_limit

        # if n_batch_limit > 1:
        #     # custom_plateau_fluct
        #     rows = df_out[df_out.evt_class == "custom_plateau_fluct"]
        #     for index in rows.index:
        #         # evt_class = group.at[index, self.mngr.tag_evt_type]
        #         output_dir = os.path.join("custom_plateau_fluct")
        #         id = f"batch_{self.n_batch_in_error_analysis}_index_{index}"
        #         filename = f"{rows.loc[index, 'time_right_edge']}_{rows.loc[index, self.mngr.tag_evt_type]}_ts_{rows.loc[index, ts_col_name]:.2f}.pdf"
        #         model_output = {k: model_outputs[k][index] for k in model_outputs}
        #         self.do_error_analysis_plot_of_event(
        #             model_output,
        #             rows.loc[index],
        #             id,
        #             get_debug_output_filepath(self.mngr, output_dir, filename),
        #         )

        # if 0 < n_batch_limit <= self.n_batch_in_error_analysis:
        #     self.log.info(
        #         [
        #             self.tag,
        #             [
        #                 "y",
        #                 f"Skipping error analysis batch {self.n_batch_in_error_analysis} to save runtime.",
        #             ],
        #         ]
        #     )
        #     return

        if len(self.mngr.error_analysis_custom_realisations) > 0:
            base_output_dir = "error_analysis_custom_realisation"
            if self.n_batch_in_error_analysis == 0:
                clear_directory(self.mngr, base_output_dir)

            for (
                custom_name,
                range_dict,
            ) in self.mngr.error_analysis_custom_realisations.items():
                custom_realisation_df = df_out[
                    (df_out["custom_realisation_name"] == custom_name)
                ]

                if range_dict["start_time"] != 0 or range_dict["end_time"] != 0:
                    custom_realisation_df = custom_realisation_df[
                        (
                            custom_realisation_df["time_right_edge"]
                            >= range_dict["start_time"]
                        )
                        & (
                            custom_realisation_df["time_right_edge"]
                            <= range_dict["end_time"]
                        )
                    ]

                for index in custom_realisation_df.index:
                    # evt_class = group.at[index, self.mngr.tag_evt_type]
                    output_dir = os.path.join(
                        base_output_dir,
                        custom_name,
                    )
                    id = f"{custom_realisation_df.loc[index, 'time_right_edge']}"
                    filename = f"{id}_ts_{custom_realisation_df.loc[index, ts_col_name]:.2f}.pdf"
                    model_output = {k: model_outputs[k][index] for k in model_outputs}
                    self.do_error_analysis_plot_of_event(
                        model_output,
                        custom_realisation_df.loc[index],
                        id,
                        get_debug_output_filepath(self.mngr, output_dir, filename),
                    )

                pass

        if self.mngr.do_error_analysis_test_statistic:
            base_output_dir = "error_analysis_ts"
            if self.n_batch_in_error_analysis == 0:
                clear_directory(self.mngr, base_output_dir)

            # high and low test statistic grouped by event classes
            label_high = "high_ts"
            label_low = "low_ts"
            for category in [label_high, label_low]:
                # for signal_step_sum, group in df_out.groupby("signal_step_sum"):
                for evt_class, group in df_out.groupby(self.mngr.tag_evt_type):
                    # ts_target = self.get_desired_ts_by_evt_class(evt_class)
                    # if ts_target is None:
                    #     continue

                    if category == label_high:
                        rows = group[ts_col_name].nlargest(n_examples).index
                    elif category == label_low:
                        rows = group[ts_col_name].nsmallest(n_examples).index

                    for index in rows:
                        # evt_class = group.at[index, self.mngr.tag_evt_type]
                        output_dir = os.path.join(
                            base_output_dir,
                            category,
                            evt_class,  # , str(signal_step_sum)
                        )
                        id = f"seed_{group.loc[index, 'realisation_seed']}_row_{group.loc[index, 'realisation_row_index']}"
                        filename = f"ts_{group.loc[index, ts_col_name]:.2f}_{id}.pdf"
                        model_output = {
                            k: model_outputs[k][index] for k in model_outputs
                        }
                        self.do_error_analysis_plot_of_event(
                            model_output,
                            group.loc[index],
                            id,
                            get_debug_output_filepath(self.mngr, output_dir, filename),
                        )

        if self.mngr.do_error_analysis_reconstruction_error:
            base_output_dir = "error_analysis_reconstruction"
            if self.n_batch_in_error_analysis == 0:
                clear_directory(self.mngr, base_output_dir)

            # highest and lowest reconstruction error
            label_high = "high_reconstruction_error"
            label_low = "low_reconstruction_error"
            for category in [label_high, label_low]:
                reco_error = pd.Series(
                    model_outputs[self.mngr.tag_auto_enc_reconstruction_error]
                )

                if category == label_high:
                    rows = reco_error.nlargest(n_examples).index
                elif category == label_low:
                    rows = reco_error.nsmallest(n_examples).index

                for index in rows:
                    output_dir = os.path.join(
                        base_output_dir,
                        category,
                        df_out.loc[index][self.mngr.tag_evt_type],
                    )
                    id = f"seed_{df_out.loc[index, 'realisation_seed']}_row_{df_out.loc[index, 'realisation_row_index']}"
                    filename = f"re_{reco_error[index]:.2f}_{id}.pdf"
                    model_output = {k: model_outputs[k][index] for k in model_outputs}
                    self.do_error_analysis_plot_of_event(
                        model_output,
                        df_out.loc[index],
                        id,
                        get_debug_output_filepath(self.mngr, output_dir, filename),
                    )

    def do_error_analysis_plot_of_event(self, model_output, metadata, id, output_path):
        ts_col_name = self.mngr.df_cols_ts_0[0]

        col_name_input_part = "input part"

        # fn_get_feature_name_from_step_and_column = self.mngr.lc_col_func

        column_names = ["_".join(t) for t in self.mngr.data_step_col_names]
        column_names_without_offsets = [
            "_".join(t)
            for t in self.mngr.data_step_col_names
            if t[1] != self.mngr.tag_temporal_offset
        ]

        temporal_offset_available = any(
            ftname == self.mngr.tag_temporal_offset
            for (channel, ftname) in self.mngr.data_step_col_names
        )

        event_df = pd.DataFrame(
            np.vstack(
                [
                    model_output[self.mngr.tag_input_d0],
                    model_output[self.mngr.tag_input_d1],
                ]
            ).T,
            column_names,
        ).T
        event_df[col_name_input_part] = ["d0"] * self.mngr.n_steps_enc + [
            "d1"
        ] * self.mngr.n_steps_dec

        # auto_enc_cols = [f"bck_forecast_{cn}" for cn in column_names] + [
        #     f"d1_{cn}" for cn in column_names
        # ]
        # autoenc_input_df = pd.DataFrame(
        #     model_output[self.mngr.tag_auto_enc_input].T,
        #     auto_enc_cols,
        # ).T

        bck_forecast_df = pd.DataFrame(
            model_output[self.mngr.tag_bck_forecast_d1].T,
            column_names,
        ).T
        bck_forecast_df[col_name_input_part] = "d1"
        bck_forecast_df.index += self.mngr.n_steps_enc

        has_bck_forecast_in_autoencoder = (
            model_output[self.mngr.tag_auto_enc_reconstruction].shape[1]
            == 2 * self.mngr.n_features
        )

        if has_bck_forecast_in_autoencoder:
            reconstruction_observation = pd.DataFrame(
                model_output[self.mngr.tag_auto_enc_reconstruction][
                    :, self.mngr.n_features :
                ].T,
                column_names,
            ).T
        else:
            reconstruction_observation = pd.DataFrame(
                model_output[self.mngr.tag_auto_enc_reconstruction].T,
                column_names_without_offsets,
            ).T
        reconstruction_observation[col_name_input_part] = "d1"
        reconstruction_observation.index += self.mngr.n_steps_enc

        if has_bck_forecast_in_autoencoder:
            reconstruction_bck_forecast = pd.DataFrame(
                model_output[self.mngr.tag_auto_enc_reconstruction][
                    :, : self.mngr.n_features
                ].T,
                column_names,
            ).T
            reconstruction_bck_forecast[col_name_input_part] = "d1"
            reconstruction_bck_forecast.index += self.mngr.n_steps_enc
        else:
            reconstruction_bck_forecast = pd.DataFrame()

        feature_names_by_channel_tag = {}
        for channel_tag, ft_name in self.mngr.data_step_col_names:
            if channel_tag not in feature_names_by_channel_tag:
                feature_names_by_channel_tag[channel_tag] = []
            feature_names_by_channel_tag[channel_tag].append(ft_name)

        def get_plot_df(data_df):
            plot_df_list = []
            for channel_tag, ft_names in feature_names_by_channel_tag.items():
                for ft_name in ft_names:
                    if ft_name == self.mngr.tag_temporal_offset:
                        continue

                    full_ft_name = f"{channel_tag}_{ft_name}"
                    full_offset_name = f"{channel_tag}_{self.mngr.tag_temporal_offset}"

                    data = {
                        "input step": data_df.index,
                        "channel name": channel_tag,
                        "feature name": ft_name,
                        "channel and feature name": full_ft_name,
                        "feature value": data_df[full_ft_name],
                    }
                    if full_offset_name in data_df:
                        data["temporal offset"] = data_df[full_offset_name]

                        # data["temporal offset"] = -np.cumsum(
                        #     data_df[full_offset_name][::-1]
                        # )[::-1]

                    plot_df = data_df[[col_name_input_part]].assign(**data)

                    plot_df_list.append(plot_df)
            plot_df = pd.concat(plot_df_list, ignore_index=True)
            return plot_df

        only_data = lambda df: df.drop(
            df.columns[self.mngr.temporal_offset_column_indices], axis=1
        )

        only_temporal_offsets = lambda df: df[
            df.columns[self.mngr.temporal_offset_column_indices]
        ]

        y_min = (
            min(
                only_data(reconstruction_observation).min(numeric_only=True).min(),
                (
                    np.inf
                    if reconstruction_bck_forecast.empty
                    else only_data(reconstruction_bck_forecast)
                    .min(numeric_only=True)
                    .min()
                ),
                only_data(bck_forecast_df).min(numeric_only=True).min(),
                only_data(event_df).min(numeric_only=True).min(),
                -2,
            )
            - 0.2
        )
        y_max = (
            max(
                only_data(reconstruction_observation).max(numeric_only=True).max(),
                (
                    -np.inf
                    if reconstruction_bck_forecast.empty
                    else only_data(reconstruction_bck_forecast)
                    .max(numeric_only=True)
                    .max()
                ),
                only_data(bck_forecast_df).max(numeric_only=True).max(),
                only_data(event_df).max(numeric_only=True).max(),
                2,
            )
            + 0.2
        )

        if temporal_offset_available:
            event_df = self.denormalise_temporal_offsets(event_df)
            bck_forecast_df = self.denormalise_temporal_offsets(bck_forecast_df)
            reconstruction_bck_forecast = self.denormalise_temporal_offsets(
                reconstruction_bck_forecast
            )
            reconstruction_observation = self.denormalise_temporal_offsets(
                reconstruction_observation
            )

        x_time_min = (
            min(
                only_temporal_offsets(reconstruction_observation).min().min(),
                (
                    np.inf
                    if reconstruction_bck_forecast.empty
                    else only_temporal_offsets(reconstruction_bck_forecast).min().min()
                ),
                only_temporal_offsets(bck_forecast_df).min().min(),
                only_temporal_offsets(event_df.iloc[-self.mngr.n_steps_dec :])
                .min()
                .min(),
            )
            - 5
        )

        x_time_max = 0 + 5

        def do_single_plot(ax, x, y_min, y_max, df, title, legend=False):
            g = sns.lineplot(
                ax=ax,
                data=get_plot_df(df),
                x=x,
                y="feature value",
                style="channel and feature name",
                hue="channel name",
                markers=True,
                legend="brief" if legend else False,
                # dashes="channel and feature name",
                # style="part",
            )

            if x == "input step":
                # add encoder/decoder separator
                ax.axvline(x=self.mngr.n_steps_enc - 0.5, color="black")
                plt.xlim(left=-1)

            if legend:
                g.legend(loc="center left", bbox_to_anchor=(1, 0.5))
            ax.set_title(title)
            ax.set_ylim([y_min, y_max])

        figs = []

        # input step as x-axis

        # fig, axes = plt.subplots(
        #     nrows=2,
        #     ncols=1,
        #     figsize=(10, 8),
        #     constrained_layout=True,
        #     sharey=True,
        #     sharex=True,
        # )

        fig, axes = plt.subplot_mosaic(
            [
                [
                    "full_time_series_by_step",
                    "full_time_series_by_step",
                    "full_time_series_by_time",
                    "full_time_series_by_time",
                ],
                [
                    "input_d1_step",
                    "input_forecast_step",
                    "input_d1_by_time",
                    "input_forecast_by_time",
                ],
                [
                    "reco_d1_step",
                    "reco_forecast_step",
                    "reco_d1_time",
                    "reco_forecast_time",
                ],
                # ["input_step", "input_step", "input_time", "input_time"],
                # ["reco_step", "reco_step", "reco_time", "reco_time"],
                # [
                #     "input_forecast_step",
                #     "reco_forecast_step",
                #     "input_forecast_time",
                #     "reco_forecast_time",
                # ],
            ],
            constrained_layout=True,
            sharey=True,
            figsize=(10 * 2, 8 * 2),
        )

        figs.append(fig)
        reconstruction_error = model_output[self.mngr.tag_auto_enc_reconstruction_error]
        fig.suptitle(
            f"{metadata[self.mngr.tag_evt_type]} event with {reconstruction_error=}"
        )

        do_single_plot(
            axes["full_time_series_by_step"],
            "input step",
            y_min,
            y_max,
            event_df,
            f"full time series",
            legend=True,
        )

        do_single_plot(
            axes["reco_d1_step"],
            "input step",
            y_min,
            y_max,
            reconstruction_observation,
            f"reconstruction of d1",
        )

        do_single_plot(
            axes["input_d1_step"],
            "input step",
            y_min,
            y_max,
            event_df.iloc[-self.mngr.n_steps_dec :],
            f"d1",
        )

        if temporal_offset_available:
            do_single_plot(
                axes["full_time_series_by_time"],
                "temporal offset",
                y_min,
                y_max,
                event_df,
                f"full time series",
                # legend=True,
            )

            do_single_plot(
                axes["reco_d1_time"],
                "temporal offset",
                y_min,
                y_max,
                reconstruction_observation,
                f"reconstruction of d1",
            )
            axes["reco_d1_time"].set_xlim([x_time_min, x_time_max])

            do_single_plot(
                axes["input_d1_by_time"],
                "temporal offset",
                y_min,
                y_max,
                event_df.iloc[-self.mngr.n_steps_dec :],
                f"d1",
            )
            axes["input_d1_by_time"].set_xlim([x_time_min, x_time_max])

        if has_bck_forecast_in_autoencoder:
            # background forecast

            do_single_plot(
                axes["input_forecast_step"],
                "input step",
                y_min,
                y_max,
                bck_forecast_df,
                f"background forecast",
                legend=True,
            )

            if temporal_offset_available:
                do_single_plot(
                    axes["input_forecast_by_time"],
                    "temporal offset",
                    y_min,
                    y_max,
                    bck_forecast_df,
                    f"background forecast",
                )
                axes["input_forecast_by_time"].set_xlim([x_time_min, x_time_max])

            # reconstruction of background forecast

            do_single_plot(
                axes["reco_forecast_step"],
                "input step",
                y_min,
                y_max,
                reconstruction_bck_forecast,
                f"reconstruction of background forecast",
                legend=True,
            )

            if temporal_offset_available:
                do_single_plot(
                    axes["reco_forecast_time"],
                    "temporal offset",
                    y_min,
                    y_max,
                    reconstruction_bck_forecast,
                    f"reconstruction of background forecast",
                )
                axes["reco_forecast_time"].set_xlim([x_time_min, x_time_max])

        # if temporal_offset_available:
        #     # share x axis for selected plots
        #     axes_labels_to_share = [
        #         "input_d1_by_time",
        #         "input_forecast_by_time",
        #         "reco_d1_time",
        #         "reco_forecast_time",
        #     ]
        #     for label in axes_labels_to_share:
        #         if label == axes_labels_to_share[0]:
        #             continue
        #
        #         axes[label].sharex(axes[axes_labels_to_share[0]])
        #         # axes[label].set_xticklabels([])

        # add metadata values as additional plot
        metadata_dict = metadata.to_dict()
        fig = create_text_figure(dict(sorted(metadata_dict.items())))
        figs.append(fig)

        PDFPlotCollection(self.mngr, figs).save_to_file(output_path)
