from __future__ import annotations

import csv
import os
from datetime import datetime
from pathlib import Path
from typing import TYPE_CHECKING, Iterable

import numpy as np
import tensorflow as tf
from keras.callbacks import ModelCheckpoint
from tensorflow import keras
from tensorflow.estimator import ModeKeys
from tensorflow.keras.callbacks import CSVLogger, TensorBoard
from tensorflow.python.framework.errors_impl import NotFoundError

from trans_finder.flares.callbacks.callback_create_model_output_plots import (
    CreateModelOutputPlotsCallback,
)
from trans_finder.flares.callbacks.callback_debug_inspect_data_normalisation import (
    InspectDataNormalisationCallback,
)
from trans_finder.flares.callbacks.callback_write_model_outputs_to_csv import (
    WriteModelOutputsToCSVCallback,
)
from trans_finder.flares.debug_keras_model import KerasModelDebugger
from trans_finder.flares.model_utils.autoenc_input_norm_callback import (
    AutoencInputNormCallback,
)
from trans_finder.flares.model_utils.forecast_combined_VAE import (
    ForecastCombinedVAE,
    build_model_forecast_combined_vae,
)
from trans_finder.flares.model_utils.utils import get_batch_size, plot_tf_model_overview
from trans_finder.flares.optional_comet import initialize_experiment
from trans_finder.flares.state_manager import StateManager
from trans_finder.flares.utilities.optional_aim import log_params
from trans_finder.flares.utils import log_table

if TYPE_CHECKING:
    from trans_finder.flares.log_parser import LogParser
    from trans_finder.flares.manager import Manager


class Train:
    def __init__(self, mngr):
        self.mngr: Manager = mngr
        self.log: LogParser = mngr.log
        self.tag = ["b", f" - flares.{Path(__file__).stem} - "]

    def load_model(self, dataset):
        """load the model with possible stored weights, in case it has already been trained
        use keras format for subclassed models. It should work without, but like this
        there will be better error messages when parts are not properly saved&loaded.
        """

        try:
            model_dir = self.mngr.data.get_dir_name("model_export")
            self.mngr.model[self.mngr.tag_model] = keras.models.load_model(
                model_dir, custom_objects={"ForecastCombinedVAE": ForecastCombinedVAE}
            )

            if self.mngr.run_phase != "train" and self.mngr.tensorflow_run_eagerly:
                self.mngr.model[self.mngr.tag_model].compile(run_eagerly=True)

            # validate that all variable values have been restored from the checkpoint
            # load_status.expect_partial()
            # load_status.assert_consumed()

            self.log.info(
                [
                    ["c", " - importing model "],
                    ["g", model_dir.replace(self.mngr.tag_base_output_dir, "")],
                ]
            )

            # model = self.mngr.model[self.mngr.tag_model]
            # pprint(model.predict(temp_model_input))
            # print("---------" * 10)
            # exit(0)

        except (ImportError, IOError, NotFoundError) as e:
            if self.mngr.run_phase != "train" or (
                self.mngr.run_phase == "train" and self.mngr.train_phase == 1
            ):
                self.log.error(
                    [
                        ["r", " - no trained model found in "],
                        ["r", model_dir.replace(self.mngr.tag_base_output_dir, "")],
                    ]
                )
                raise e

            self.log.info([["c", " - creating new model ... "]])
            self.build_model()

            # if not self.mngr.model[self.mngr.tag_model].built:
            #     # If the model uses keras subclassing, it may not have been built yet.
            #     # Building the model by calling it once. Otherwise loading pre-trained weights will not work.
            #     single_batch_input = next(iter(dataset))[0]
            #     self.mngr.model[self.mngr.tag_model](single_batch_input)
            #
            #     assert self.mngr.model[self.mngr.tag_model].built

        return

    def train_model(self):
        self.log.info([*self.mngr.main_func_prnt, ["g", " - train-model()"]])

        physical_devices = tf.config.list_physical_devices("GPU")
        self.log.info(str(physical_devices))

        dataset_train = self.mngr.data.read_dataset(mode=ModeKeys.TRAIN)
        dataset_eval = self.mngr.data.read_dataset(mode=ModeKeys.EVAL)

        self.load_model(dataset_train)

        # update train_phase varaible of model in case it was loaded from disk
        self.mngr.model[self.mngr.tag_model].train_phase = self.mngr.train_phase

        if self.mngr.train_phase == 0:
            self.mngr.model[self.mngr.tag_model].forecaster.trainable = True
            self.mngr.model[self.mngr.tag_model].encoder.trainable = False
            self.mngr.model[self.mngr.tag_model].decoder.trainable = False

        elif self.mngr.train_phase == 1:
            self.mngr.model[self.mngr.tag_model].forecaster.trainable = False
            self.mngr.model[self.mngr.tag_model].encoder.trainable = True
            self.mngr.model[self.mngr.tag_model].decoder.trainable = True

        self.mngr.model[self.mngr.tag_model].compile(
            optimizer=self.mngr.optimizer(learning_rate=self.mngr.learning_rate),
            run_eagerly=self.mngr.tensorflow_run_eagerly,
            loss=None,
        )

        phase_epochs = []
        for n_train_phase in range(self.mngr.train_phase + 1):
            csv_logger_name = self.mngr.data.get_file_name(
                dir_type="model_history",
                file_name=f"train_history_{n_train_phase}.csv",
            )
            try:
                with open(csv_logger_name, newline="") as csvfile:
                    data = list(csv.DictReader(csvfile))
                    phase_epochs += [len(data)]
                    if len(data) == 0:
                        raise FileNotFoundError
                    if "lr" not in data[-1].keys():
                        raise FileNotFoundError
                    if self.mngr.learn_rate_new_data is None:
                        learn_rate_init = float(data[-1]["lr"])
                    else:
                        modulo = len(data) % self.mngr.learn_rate_new_data["period"]
                        entry = (
                            modulo
                            if modulo != 0
                            else self.mngr.learn_rate_new_data["period"]
                        )
                        learn_rate_init = float(data[-entry]["lr"]) * (
                            1 + self.mngr.learn_rate_new_data["decrease_factor"]
                        )
            except (csv.Error, FileNotFoundError) as e:
                phase_epochs += [0]
                learn_rate_init = self.mngr.learning_rate

        if self.mngr.learn_rate_init is not None:
            learn_rate_init = self.mngr.learn_rate_init

        # the current initial epoch is derived from all training phases
        epoch_init = sum(phase_epochs)
        epoch_end = self.mngr.n_train_epochs + epoch_init

        # the threshold is defined as the number of epochs for the current
        # training phase
        if self.mngr.learn_rate_epoch_thresh_phases is None:
            epoch_thresh = self.mngr.learn_rate_epoch_thresh
        else:
            epoch_thresh = self.mngr.learn_rate_epoch_thresh_phases[
                self.mngr.train_phase
            ]

        epoch_thresh += sum(phase_epochs[:-1])

        callbacks = [
            # self.get_logging_callback(epoch_end),
            # LearningRateSchedulingCallback(
            #     log=self.log,
            #     learn_rate_init=learn_rate_init,
            #     epoch_thresh=epoch_thresh,
            #     learn_rate_decrease_factor=self.mngr.learn_rate_decrease_factor,
            #     learn_rate_new_data=self.mngr.learn_rate_new_data,
            # ),
            # ReduceLROnPlateau(monitor="val_loss", factor=0.2, patience=5, min_lr=0.001),
            TensorBoard(
                **self.get_tensorboard_params(),
            ),
            CSVLogger(
                filename=csv_logger_name,
                append=True,
            ),
            ModelCheckpoint(
                filepath=self.mngr.data.get_dir_name("model_export"),
                save_weights_only=False,
                save_best_only=self.mngr.model_save_best_weights_only,
                monitor=self.get_stopping_criteria_loss(),
                mode="min",
                verbose=1,
            ),
        ]

        if self.mngr.train_phase == 0:
            callbacks.insert(0, AutoencInputNormCallback(dataset_train))

        callbacks += self.get_user_keras_callbacks(dataset_train, dataset_eval)

        log_params(self.mngr)

        if self.mngr.train_phase == 1:
            self.mngr.model[self.mngr.tag_model].set_ae_input_norm(dataset_train)

        # print("---------" * 10, ".predict()")
        #
        # self.mngr.model[self.mngr.tag_model].predict(
        #     dataset_train, batch_size=get_batch_size(self.mngr, force_predict_size=True)
        # )

        self.log.info(
            f"{self.mngr.train_phase=} model.fit(batch_size={self.mngr.train_batch_size},initial_epoch={epoch_init},epochs={epoch_end})"
        )

        # calculate how many batches correspond to one realisation
        # and then set the size of an epoch based on n_realisations_per_train_epoch
        n_batches_per_epoch = (
            int(
                np.ceil(
                    len(dataset_train)
                    / max(
                        1, self.mngr.model_n_samples_of_input_errors
                    )  # if zero, this is ignored
                    / self.mngr.data_prep_train_n_realisations["background"]
                )
            )
            * self.mngr.n_realisations_per_train_epoch
        )

        self.mngr.model[self.mngr.tag_model].fit(
            dataset_train,
            steps_per_epoch=n_batches_per_epoch,
            epochs=epoch_end,
            initial_epoch=epoch_init,
            validation_data=dataset_eval,
            callbacks=callbacks,
            verbose=2,
        )

        self.log.info("Evaluating on validation dataset...")
        result = self.mngr.model[self.mngr.tag_model].evaluate(
            dataset_eval,
            verbose=2,
            batch_size=get_batch_size(self.mngr, force_predict_size=True),
        )
        if not isinstance(result, Iterable):
            result = [result]

        result = dict(zip(self.mngr.model[self.mngr.tag_model].metrics_names, result))
        log_table(
            self.log.info,
            result,
            title=" - final results (validation dataset):",
            add_space=True,
        )

        self.log.info(
            "Predicting on validation dataset for detailed performance metrics..."
        )
        eval_outputs = self.mngr.model[self.mngr.tag_model].predict(
            dataset_eval, get_batch_size(self.mngr, force_predict_size=True)
        )
        if self.mngr.train_phase == 0:
            if result["forecasting_r2"] <= 0:
                self.log.warn(
                    f""" - Warning: Explained variance of forecasting model on validation split is negative ({result["forecasting_r2"]=}).
                    Indication of poor forecasting model fit."""
                )

            KerasModelDebugger(self.mngr).calculate_metrics_per_feature_forecasting(
                eval_outputs, ModeKeys.EVAL
            )

        elif self.mngr.train_phase == 1:
            # KerasModelDebugger(self.mngr).do_plots_reconstruction(
            #     self.mngr.model[self.mngr.tag_model], dataset_eval, data_split_tag=ModeKeys.EVAL
            # )
            KerasModelDebugger(self.mngr).calculate_metrics_per_feature_reconstruction(
                eval_outputs, ModeKeys.EVAL
            )

        KerasModelDebugger(self.mngr).run()

    def build_model(self, *args, **kwargs):
        """interface function for loading the model"""

        # new implementation
        model, losses = build_model_forecast_combined_vae(self.mngr)

        self.mngr.model = {
            self.mngr.tag_model: model,
        }

        # self.mngr.mlflow.tensorflow.autolog()
        # self.mngr.mlflow.tensorflow.log_model(model, "model")

        plot_tf_model_overview(self.mngr, model, "end_to_end_model")

    def save_model_file(self, tag_file, data):
        self.mngr.data.save_calib_file(
            self.mngr.tag_model_dir,
            tag_file,
            data,
            tag_prefix=None,
        )

    def load_model_file(self, tag_file):
        return self.mngr.data.load_calib_file(
            self.mngr.tag_model_dir,
            tag_file=tag_file,
            tag_prefix=None,
            is_verb=True,
        )

    def get_stopping_criteria_loss(self) -> str:
        return (
            "val_forecasting_loss"
            if self.mngr.train_phase == 0
            else "val_reconstruction_loss"
        )

    def get_user_keras_callbacks(self, dataset_train, dataset_eval):
        callbacks = []

        # callbacks.append(UpdateLossWeightsInitiallyCallbacks(self.mngr))

        # if self.model_mcmc_gaus_mix_is_active:
        #     callbacks.append(SaveMCMCGaussMixStateAfterTraining(self.mngr))

        if self.mngr.debug_create_model_output_plots:
            callbacks.append(
                CreateModelOutputPlotsCallback(self.mngr, dataset_train, dataset_eval)
            )

        if self.mngr.debug_create_model_output_csv_files:
            callbacks.append(
                WriteModelOutputsToCSVCallback(self.mngr, dataset_train, dataset_eval)
            )

        if self.mngr.debug_enable_comet:
            self.experiment = initialize_experiment("train", "flares")
            callbacks.append(CometLoggingCallback(self.mngr, self.experiment))

        if self.mngr.use_early_stopping > 0:
            self.log.info(
                [
                    self.tag,
                    [
                        "p",
                        f"early stopping after {self.mngr.use_early_stopping} epochs "
                        f"of val_loss not improving.",
                    ],
                ]
            )

            callbacks.append(
                tf.keras.callbacks.EarlyStopping(
                    monitor=self.get_stopping_criteria_loss(),
                    patience=self.mngr.use_early_stopping,
                    verbose=1,
                    mode="min",
                    baseline=None,
                    restore_best_weights=False,
                )
            )

        if self.mngr.debug_inspect_data_normalisation:
            callbacks.append(
                InspectDataNormalisationCallback(
                    mngr=self.mngr,
                    dataset_train=dataset_train,
                    dataset_val=dataset_eval,
                )
            )

        # add_aim_callback(self.mngr, callbacks)
        # add_custom_aim_callback(self.mngr, callbacks, dataset_eval)
        # callbacks.append(LossBalancingCallback())

        # this currently only works for 2d hidden representations
        # callbacks.append(PlottingHiddenReprCallback(dataset_train, self.mngr))

        return callbacks

    def get_full_model_version_name(self):
        # get full model version name
        if not hasattr(self.mngr, "full_model_version_name"):
            # load from disk or create it

            full_model_version_name_state = StateManager(
                self.mngr, self.mngr.tag_model_full_version_name
            )
            full_model_version_name = None

            if full_model_version_name_state.has_state():
                full_model_version_name = full_model_version_name_state.load_state()
                if (
                    self.mngr.model_version_name is not None
                    and not self.mngr.model_version_name in full_model_version_name
                ):
                    # self.mngr.model_version_name has changed in config, hence create new run name
                    full_model_version_name = None

            if full_model_version_name is None:
                version_name = self.mngr.model_version_name
                if version_name is None:
                    version_name = self.mngr.tag_model_base

                full_model_version_name = (
                    version_name + "-" + datetime.now().strftime("%Y%m%d-%H%M%S")
                )
                full_model_version_name_state.save_state(full_model_version_name)

            self.full_model_version_name = full_model_version_name

        return self.full_model_version_name

    def get_tensorboard_params(self):
        """the tensorboard output dir is based on the config flag model_version_name, if configured
        and has a timestamp to differentiate different training runs. The version name remains the same
        until the output dir is reset with reset_tag_dir."""
        tensorboard_root_dir = os.path.join(
            self.mngr.tag_output_dir,
            "tensorboard_logs",
        )

        tensorboard_dir = str(
            os.path.join(
                tensorboard_root_dir,
                self.get_full_model_version_name(),
            ),
        )

        return dict(
            log_dir=tensorboard_dir,
        )

    def loss_log_prob(self, y_true, y_pred):
        """function for calculating the loss"""

        loss = -1 * y_pred.log_prob(y_true)
        return tf.reduce_sum(loss, axis=[-1, -2])

    def save_model(self, model):
        model_dir = self.mngr.data.get_dir_name("model_export")
        model.save_weights(model_dir)
        self.log.info(
            [
                ["c", " - saved model to "],
                ["g", model_dir.replace(self.mngr.tag_base_output_dir, "")],
            ]
        )
        return


class UpdateLossWeightsInitiallyCallbacks(keras.callbacks.Callback):
    def __init__(self, mngr):
        self.mngr: Manager = mngr

    def on_train_begin(self, logs=None):
        # updating loss weights for a saved model is loaded

        # self.mngr.model.weight_loss_gaus_mix.assign(
        #     self.mngr.loss_scales_phase["gaus_mix"]
        # )

        self.mngr.model.weight_loss_bck_forecast.assign(
            self.mngr.loss_scales_phase["bck_forecast"]
        )
        self.mngr.model.weight_loss_auto_enc.assign(
            self.mngr.loss_scales_phase["auto_enc"]
        )

        return


class CometLoggingCallback(keras.callbacks.Callback):
    def __init__(self, mngr, experiment):
        self.mngr: Manager = mngr
        self.experiment = experiment

    def on_train_begin(self, logs=None):
        try:
            params = {k: str(v) for k, v in vars(self.mngr).items()}
            for k, v in params.items():
                if len(v) > 100:
                    params[k] = v[:100] + " [truncated]"

            self.experiment.log_parameters(params)
        except Exception as e:
            pass
