from __future__ import annotations

import os
from typing import TYPE_CHECKING

# import numpy as np
import pandas as pd
from sklearn.model_selection import StratifiedShuffleSplit, train_test_split

if TYPE_CHECKING:
    from trans_finder.flares.manager import Manager


class TrainValSplitter:
    def __init__(self, mngr):
        self.mngr: Manager = mngr
        self.all_realisation_ids = None
        self.val_split_realisation_ids = None

        # # load all input files to fit t
        # self.mngr.max_evt["train_scales"] = -1

    def after_load_single_data_file(self, df_in):
        pass
        # """register realisation ids from the input file to compile an array with all realisation ids"""
        # file_realisations = df_in["realisation_id"].unique()
        # if self.all_realisation_ids is None:
        #     self.all_realisation_ids = file_realisations
        # else:
        #     self.all_realisation_ids = np.concatenate(
        #         [self.all_realisation_ids, file_realisations]
        #     )

    def after_load_data_files(self):
        pass
        # val_split_realisation_ids_state = StateManager(
        #     self.mngr, self.mngr.tag_val_split_realisation_ids
        # )
        #
        # if val_split_realisation_ids_state.has_state():
        #     self.val_split_realisation_ids = (
        #         val_split_realisation_ids_state.load_state()
        #     )
        #
        # elif self.mngr.run_phase == "train" and self.mngr.only_do_train_scale:
        #     val_realisations = np.random.choice(
        #         self.all_realisation_ids,
        #         int(0.3 * len(self.all_realisation_ids)),
        #         replace=False,
        #     )
        #
        #     val_split_realisation_ids_state.save_state(val_realisations)
        #     self.val_split_realisation_ids = val_realisations

    def _split_stratified_into_train_val_test(
        self,
        df_input,
        stratify_colname="y",
        frac_train=0.7,
        frac_val=0.15,
        frac_test=0.15,
        random_state=None,
    ):
        """
        Splits a Pandas dataframe into three subsets (train, val, and test)
        following fractional ratios provided by the user, where each subset is
        stratified by the values in a specific column (that is, each subset has
        the same relative frequency of the values in the column). It performs this
        splitting by running train_test_split() twice.

        Parameters
        ----------
        df_input : Pandas dataframe
            Input dataframe to be split.
        stratify_colname : str
            The name of the column that will be used for stratification. Usually
            this column would be for the label.
        frac_train : float
        frac_val   : float
        frac_test  : float
            The ratios with which the dataframe will be split into train, val, and
            test data. The values should be expressed as float fractions and should
            sum to 1.0.
        random_state : int, None, or RandomStateInstance
            Value to be passed to train_test_split().

        Returns
        -------
        df_train, df_val, df_test :
            Dataframes containing the three splits.
        """

        if frac_train + frac_val + frac_test != 1.0:
            raise ValueError(
                "fractions %f, %f, %f do not add up to 1.0"
                % (frac_train, frac_val, frac_test)
            )

        if stratify_colname not in df_input.columns:
            raise ValueError("%s is not a column in the dataframe" % (stratify_colname))

        X = df_input  # Contains all columns.
        y = df_input[
            [stratify_colname]
        ]  # Dataframe of just the column on which to stratify.

        # Split original dataframe into train and temp dataframes.
        df_train, df_temp, y_train, y_temp = train_test_split(
            X, y, stratify=y, test_size=(1.0 - frac_train), random_state=random_state
        )

        # Split the temp dataframe into val and test dataframes.
        relative_frac_test = frac_test / (frac_val + frac_test)
        df_val, df_test, y_val, y_test = train_test_split(
            df_temp,
            y_temp,
            stratify=y_temp,
            test_size=relative_frac_test,
            random_state=random_state,
        )

        assert len(df_input) == len(df_train) + len(df_val) + len(df_test)

        return df_train, df_val, df_test

    def _df_to_mask(self, input_df, split_metadata_df):
        return input_df["realisation_id"].isin(split_metadata_df["id"].to_numpy())

    def get_masks_train_val_split(self, input_df: pd.DataFrame):
        metadata_out_dir = os.path.join(
            self.mngr.tag_input_dir, self.mngr.dir_output_prep_train
        )
        path_realisation_metdata = os.path.join(
            metadata_out_dir, self.mngr.tag_realisations_metadata_file
        )

        realisation_metadata = pd.read_csv(path_realisation_metdata)

        sss = StratifiedShuffleSplit(
            n_splits=1, test_size=1 - self.mngr.frac_train_eval, random_state=0
        )

        train_rows, temp_rows = list(
            sss.split(
                realisation_metadata["id"], realisation_metadata["has_injections"]
            )
        )[0]

        val_split_realisation_ids = realisation_metadata["id"][temp_rows].to_numpy()

        eval_mask = input_df["realisation_id"].isin(val_split_realisation_ids)
        train_mask = ~eval_mask

        return train_mask, eval_mask

    def get_masks_train_val_test_split(self, input_df: pd.DataFrame):
        metadata_out_dir = os.path.join(
            self.mngr.tag_input_dir, self.mngr.dir_output_prep_train
        )
        path_realisation_metdata = os.path.join(
            metadata_out_dir, self.mngr.tag_realisations_metadata_file
        )

        realisation_metadata = pd.read_csv(path_realisation_metdata)

        df_train, df_val, df_test = self._split_stratified_into_train_val_test(
            realisation_metadata,
            "has_injections",
            random_state=0,
        )

        masks = [
            self._df_to_mask(input_df, split_df)
            for split_df in [df_train, df_val, df_test]
        ]

        return masks
