from __future__ import annotations

import os
from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

from trans_finder.flares.utils import write_single_output_df

if TYPE_CHECKING:
    from trans_finder.flares.manager import Manager


class DataFrameBatchedHDFWriter:
    target_n_events_per_hdf = 10000

    def __init__(self, mngr, out_dir):
        self.mngr: Manager = mngr
        self.out_dir = out_dir
        self.df_out = None
        self.n_file = 0
        self.realisation_metadata_out = None
        self.event_id_offset = 0

    def add(self, df: pd.DataFrame):
        df[self.mngr.tag_event_id] = np.arange(
            self.event_id_offset, self.event_id_offset + len(df), 1
        )

        self.event_id_offset = df[self.mngr.tag_event_id].iat[-1] + 1

        if self.df_out is None:
            self.df_out = df
        else:
            self.df_out = pd.concat([self.df_out, df])

        if len(self.df_out) > self.target_n_events_per_hdf:
            self.flush()

    def add_realisation_metadata(self, realisation):
        metadata_out_dir = os.path.join(
            self.mngr.tag_input_dir, self.mngr.dir_output_prep_train
        )
        Path(metadata_out_dir).mkdir(parents=True, exist_ok=True)
        path_realisation_metdata = os.path.join(
            metadata_out_dir, self.mngr.tag_realisations_metadata_file
        )

        do_append = self.realisation_metadata_out is not None

        pd.DataFrame(
            [
                {
                    "id": realisation.id,
                    "has_injections": len(realisation.signal_injections) > 0,
                }
            ]
        ).to_csv(
            path_realisation_metdata,
            mode="a" if do_append else "w",
            header=False if do_append else True,
        )
        self.realisation_metadata_out = realisation

    def flush(self):
        if self.df_out is None:
            return

        Path(self.out_dir).mkdir(parents=True, exist_ok=True)
        self.n_file += 1

        write_single_output_df(
            self.df_out,
            output_dir=self.out_dir,
            in_file_filt_str="",
            out_file_patt=self.mngr.out_file_patt,
            tag_hdf=self.mngr.tag_hdf,
            n_file=self.n_file,
            log=self.mngr.log,
        )

        del self.df_out
        self.df_out = None
