import contextlib

from keras.callbacks import Callback

from trans_finder.flares.state_manager import StateManager

try:
    from aim import Run
    from aim.tensorflow import AimCallback
except ImportError:
    AimCallback = None
    Run = None


def add_aim_callback(mngr, callbacks):
    aim_kwargs = dict(
        repo=mngr.aim_repo if hasattr(mngr, "aim_repo") else None,
        experiment=mngr.model_version_name,
    )
    if Run is not None:
        run = _create_aim_run(mngr, aim_kwargs)

        callback = AimCallback(**aim_kwargs)
        callback._run_hash = run.hash
        callbacks.append(callback)


def get_aim_run(mngr):
    if hasattr(mngr, "_aim_run"):
        return mngr._aim_run

    elif Run is not None:
        aim_kwargs = dict(
            repo=mngr.aim_repo if hasattr(mngr, "aim_repo") else None,
            experiment=mngr.model_version_name,
        )
        mngr._aim_run = _create_aim_run(mngr, aim_kwargs)

        return mngr._aim_run
    else:
        return None


def log_params(mngr):
    run = get_aim_run(mngr)

    mngr.log.debug("Capturing current parameters in aim run")
    for k, v in vars(mngr).items():
        if (
            not k.startswith("tag")
            and v is not None
            and not callable(v)
            and k not in ["train_phase"]
        ):
            with contextlib.suppress(Exception):
                run[k] = v


def _create_aim_run(mngr, aim_kwargs):
    run_hash_state = StateManager(mngr, "aim_run_hash")
    is_run_hash_stored = run_hash_state.has_state()
    run_hash = run_hash_state.load_state() if is_run_hash_stored else None
    print(f"{run_hash=}")

    run = Run(**aim_kwargs, run_hash=run_hash)

    if not is_run_hash_stored:
        run_hash_state.save_state(run.hash)

    return run


def log_metrics_in_aim(
    mngr, metric_dict, data_split_tag, prefix=None, extra_context=None
):
    if extra_context is None:
        extra_context = {}

    run = get_aim_run(mngr)
    if run is None:
        return

    for k, v in metric_dict.items():
        run.track(
            v,
            k if prefix is None else f"{prefix}_{k}",
            context={
                "subset": data_split_tag,
                "train_phase": mngr.train_phase,
                **extra_context,
            },
        )


def add_custom_aim_callback(mngr, callbacks, dataset_eval):
    class NBatchLogger(Callback):
        """
        A Logger that log average performance per `display` steps.
        """

        def __init__(self, mngr):
            super().__init__()
            self.mngr = mngr
            self.global_step = {phase: 0 for phase in self.mngr.train_phases}
            self.global_step_state = StateManager(mngr, "global_train_step")
            self.val_freq = mngr.aim_track_val_metrics_every_n_steps

        def on_train_begin(self, logs=None):
            if self.global_step_state.has_state():
                self.global_step = self.global_step_state.load_state()

        def on_train_end(self, logs=None):
            self.global_step_state.save_state(self.global_step)

        def on_epoch_begin(self, epoch, logs=None):
            self.epoch = epoch

        def track_eval(self, results):
            for k, v in results:
                get_aim_run(self.mngr).track(
                    v,
                    k,
                    step=self.global_step[self.mngr.train_phase],
                    epoch=self.epoch,
                    context={
                        "subset": "val",
                        "train_phase": self.mngr.train_phase,
                    },
                )

        def on_train_batch_end(self, batch, logs=None):
            self.global_step[self.mngr.train_phase] = 1 + self.global_step.get(
                self.mngr.train_phase, 0
            )

            for k, v in logs.items():
                get_aim_run(self.mngr).track(
                    v,
                    k,
                    step=self.global_step[self.mngr.train_phase],
                    epoch=self.epoch,
                    context={
                        "subset": "train",
                        "train_phase": self.mngr.train_phase,
                    },
                )
            if (
                self.val_freq > 0
                and self.global_step[self.mngr.train_phase] % self.val_freq == 0
            ):
                val_metrics = self.mngr.model.evaluate(dataset_eval, verbose=0)
                metric_names = self.mngr.model.metrics_names

                self.track_eval(zip(metric_names, val_metrics))

        def on_epoch_end(self, epoch, logs=None):
            for k, v in logs.items():
                get_aim_run(self.mngr).track(
                    v,
                    f"epoch_{k}",
                    step=self.global_step[self.mngr.train_phase],
                    epoch=self.epoch,
                    context={
                        "train_phase": self.mngr.train_phase,
                    },
                )

    if get_aim_run(mngr) is not None:
        callbacks.append(NBatchLogger(mngr))
