import math

import numpy as np
import scipy.stats
from matplotlib import pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
from mpl_toolkits.axes_grid1 import make_axes_locatable
from scipy.stats import multivariate_normal
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE

from trans_finder.flares.plot_lib import PlotLib
from trans_finder.flares.plot_list.util.plot_config import tick_params


class PDFPlotCollection:
    """A class to create a multi-page pdf of matplotlib figures. A figure can have sub-figures arranged in a grid."""

    def __init__(self, mngr, figs=None):
        self.figs = [] if figs is None else figs
        self.plot_lib = PlotLib(mngr.log, tag_base_dir=mngr.tag_base_output_dir)
        self.fig_opts = {
            "gridspec_opts": {
                "hspace": 0.35,
                "wspace": 0.35,
            },
        }
        self.style_set = "pastels"

    def create_figure(self):
        fig, spec = self.create_figure_grid(1, 1)
        ax = self.get_figure_grid_axis(fig, spec, 0, 0)
        ax.tick_params(**tick_params)
        return fig, ax

    def create_figure_grid(self, n_rows, n_cols, opts=None):
        if opts is None:
            opts = {}

        plot_opts = {**self.fig_opts}
        plot_opts.update(opts)

        fig, spec = self.plot_lib.add_fig(
            n_rows=n_rows,
            n_cols=n_cols,
            opts=plot_opts,
        )
        self.figs.append(fig)
        return fig, spec

    def get_figure_grid_axis(self, fig, spec, n_row, n_col):
        ax = self.plot_lib.add_subplot(fig, spec[n_row, n_col])
        return ax

    def save_to_file(self, out_file_name, do_close_figs=True):
        with PdfPages(out_file_name) as pdf_pages:
            self.plot_lib.save_pdf_pages(pdf_pages, out_file_name, figs=self.figs)

        for fig in self.figs:
            plt.close(fig)

    @staticmethod
    def _create_color_by_label_options(labels):
        plot_opts = {}
        if labels is not None:
            label_listset = list(set(labels))
            plot_opts["c"] = [label_listset.index(l) for l in labels]
            plot_opts["label"] = labels
        return plot_opts

    def make_labeled_histogram(
        self,
        ax,
        data,
        labels,
        style_set="main_0",
        log_scale_x=True,
        log_scale_y=False,
        density=True,
        x_label=None,
    ):
        uniquelabels = set(labels)
        grouped_data = [
            [row[1:] for row in zip(labels, data) if row[0] == l] for l in uniquelabels
        ]

        for n_label, (data_label, data_segment) in enumerate(
            zip(uniquelabels, grouped_data)
        ):
            data_segment = next(zip(*data_segment))

            plot_style = self.plot_lib.get_plot_style(
                index=n_label,
                tag="main_data",
                style_set=style_set,
            )
            plot_style.pop("marker", None)
            plot_style["label"] = data_label
            if log_scale_x:
                plt.xscale("log")
            if log_scale_y:
                plt.yscale("log")

            if density:
                ax.set_ylabel("Density")

            if x_label is not None:
                ax.set_xlabel(x_label)

            ax.hist(data_segment, histtype="step", density=density, **plot_style)

        handles, labels = ax.get_legend_handles_labels()
        ax.legend(handles, labels)

    def plot_histogram_by_label(
        self,
        ax,
        data,
        labels,
        style_set="main_0",
        log_scale_x=True,
        log_scale_y=False,
    ):
        uniquelabels = set(labels)
        grouped_data = [
            [row[1:] for row in zip(labels, data) if row[0] == l] for l in uniquelabels
        ]

        plots_per_page = 4

        n_pages = int(math.ceil(len(uniquelabels) / 4))

        for n_label, (data_label, data_segment) in enumerate(
            zip(uniquelabels, grouped_data)
        ):
            data_segment = next(zip(*data_segment))

            plot_style = self.plot_lib.get_plot_style(
                index=n_label,
                tag="main_data",
                style_set=style_set,
            )
            plot_style.pop("marker", None)
            plot_style["label"] = data_label
            if log_scale_x:
                plt.xscale("log")
            if log_scale_y:
                plt.yscale("log")
            ax.hist(data_segment, histtype="step", density=True, **plot_style)

        handles, labels = ax.get_legend_handles_labels()
        ax.legend(handles, labels)
        pass

    def make_labeled_scatterplot(self, ax, x, y, labels, plot_opts, style_set="main_0"):
        uniquelabels = set(labels)
        grouped_x_and_y = [
            [row[1:] for row in zip(labels, x, y) if row[0] == l] for l in uniquelabels
        ]

        for n_label, (data_label, x_and_y) in enumerate(
            zip(uniquelabels, grouped_x_and_y)
        ):
            x, y = zip(*x_and_y)

            plot_style = self.plot_lib.get_plot_style(
                index=n_label,
                tag="main_data",
                style_set=style_set,
            )
            plot_style.update(plot_opts)

            plot_style["label"] = data_label

            if "marker" not in plot_opts:
                # only set the marker, if it has not been manually specified in the plot_opts
                plot_style["marker"] = "x" if "(training)" in data_label else "."

            ax.scatter(x, y, **plot_style)

        handles, labels = ax.get_legend_handles_labels()
        ax.legend(handles, labels)

    def make_contour_plot(
        self, ax, means, covs, alpha, title=None, delta=0.025, xymin=-2.0, xymax=8.0
    ):
        """visualize gaussian components as contours"""

        k = means.shape[0]

        x_grid, y_grid, coordinates = self.get_grid_values(ax, xymin, xymax, delta)

        for i in range(k):
            mean = means[i]
            cov = covs[i]
            z_grid = (
                multivariate_normal(mean, cov).pdf(coordinates).reshape(x_grid.shape)
            )
            ax.contour(
                x_grid,
                y_grid,
                z_grid,
                3,
                colors="black",
                alpha=alpha[i],
            )

        if title:
            ax.set_title(title)

    def get_grid_values(self, ax, xymin, xymax, delta):
        # determine the min and max values of the grid to sample the probability
        if xymax is None and xymin is None:
            ymin, ymax = ax.get_ylim()
            xmin, xmax = ax.get_xlim()

            # treat edge case where all values would fall into a single point in the sampling grid
            if xmax < xmin + delta:
                xmax += 10 * delta
                xmin -= 10 * delta
            if ymax < ymin + delta:
                ymax += 10 * delta
                ymin -= 10 * delta

            # make the  x and y symmetric
            xymax = max(xmax, ymax)
            xymin = min(xmin, ymin)

        elif xymax is None or xymin is None:
            raise ValueError("Both xymin and xymax have to be set, or none of them.")

        xmin = xymin
        xmax = xymax
        ymin = xymin
        ymax = xymax

        x = np.arange(xmin, xmax, delta)
        y = np.arange(ymin, ymax, delta)
        x_grid, y_grid = np.meshgrid(x, y)
        coordinates = np.array([x_grid.ravel(), y_grid.ravel()]).T

        return x_grid, y_grid, coordinates

    def make_contour_plot_with_filled_colors_and_colorbar(
        self,
        fig,
        ax,
        fn_get_score_from_coordinates,
        scatter_points=None,
        scatter_labels=None,
        delta=0.025,
        xymin=None,
        xymax=None,
        overlay_component_contours_prob_fn=None,
        overlay_component_contours_alpha=None,
        n_contours_filled=30,
    ):
        # make an invisible scatter plot just to set the dimensions of the plot
        if scatter_points is not None:
            ax.plot(scatter_points[:, 0], scatter_points[:, 1], "x", alpha=0)

        x_grid, y_grid, coordinates = self.get_grid_values(ax, xymin, xymax, delta)

        z_grid = fn_get_score_from_coordinates(coordinates).reshape(x_grid.shape)
        contourf = ax.contourf(
            x_grid, y_grid, z_grid, n_contours_filled, cmap="Greys"
        )  # , colors=col[i])

        # ax.colorbar()

        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="3%", pad="3%")
        fig.colorbar(contourf, cax=cax, orientation="vertical")

        # show contours of the joint probability
        if not overlay_component_contours_prob_fn:
            contours = ax.contour(
                x_grid,
                y_grid,
                z_grid,
                3,
                colors="black",
            )  # , colors=col[i])

            ax.clabel(contours, inline=True, fontsize=8)
            # plt.imshow(
            #     z_grid,
            #     extent=[xymin, xymax, xymin, xymax],
            #     origin="lower",
            #     cmap="RdGy",
            #     alpha=0.5,
            # )
        else:
            component_probability_list = overlay_component_contours_prob_fn(coordinates)
            for component_idx, component_probability in enumerate(
                component_probability_list
            ):
                alpha = (
                    None
                    if overlay_component_contours_alpha is None
                    else overlay_component_contours_alpha[component_idx]
                )
                contours = ax.contour(
                    x_grid,
                    y_grid,
                    component_probability.reshape(x_grid.shape),
                    3,
                    colors="black",
                    alpha=alpha,
                )  # , colors=col[i])

        # now make the real scatter plot on top
        if scatter_points is not None:
            self.make_labeled_scatterplot(
                ax,
                scatter_points[:, 0],
                scatter_points[:, 1],
                labels=scatter_labels,
                plot_opts={"marker": "x"},
            )

    def make_scatter_plot(self, ax, x, y, labels=None, plot_opts=None):
        if plot_opts is None:
            plot_opts = {}

        if labels is not None:
            self.make_labeled_scatterplot(ax, x, y, labels, plot_opts)

            # plot_opts = self._create_color_by_label_options(labels)
            # ax.scatter(x, y, **plot_opts)
        else:
            plot_opts["marker"] = "x"
            ax.scatter(x, y, **plot_opts)

    def make_tsne_plot(self, ax, data, data_title, labels=None):
        X_embedded = TSNE().fit_transform(data)
        self.make_scatter_plot(ax, X_embedded[:, 0], X_embedded[:, 1], labels)
        ax.set_title(f"2D t-SNE of {data_title}")
        return ax

    def try_profile_plot_otherwise_do_scatterplot(self, *args):
        self.make_scatter_plot(*args)
        return False
        # try:
        #     make_profile_plot_2d(*args)
        #     return True
        # except FloatingPointError:
        #     make_scatter_plot(*args)
        #     return False

    @staticmethod
    def make_profile_plot_2d(ax, x, y, plot_options):
        means_result = scipy.stats.binned_statistic(
            x, [y, y**2], bins=50, range=(min(x), max(x)), statistic="mean"
        )
        means, means2 = means_result.statistic
        temp = means2 - means**2

        temp = np.ma.array(temp, mask=np.isnan(temp))
        standard_deviations = np.sqrt(temp).filled(np.nan)
        bin_edges = means_result.bin_edges
        bin_centers = (bin_edges[:-1] + bin_edges[1:]) / 2.0

        ax.errorbar(
            x=bin_centers,
            y=means,
            yerr=standard_deviations,
            linestyle="none",
            marker=".",
            **plot_options,
        )

    def make_histogram(
        self,
        ax,
        data,
        labels=None,
        log_scale_x=True,
        log_scale_y=False,
        x_label=None,
    ):
        if labels:
            self.make_labeled_histogram(
                ax,
                data,
                labels,
                log_scale_x=log_scale_x,
                log_scale_y=log_scale_y,
                x_label=x_label,
            )
        else:
            hist = ax.hist(data, bins=50)

    def make_pca_plot(self, ax, data, data_title, n_comps=2, labels=None):
        pca = PCA(n_components=n_comps)

        X = pca.fit_transform(data)

        if n_comps == 1:
            self.make_histogram(ax, X)
            ax.set_title(
                f"PCA of {data_title}, explained variance: {pca.explained_variance_ratio_}"
            )

        elif n_comps == 2:
            self.make_scatter_plot(ax, X[:, 0], X[:, 1], labels)

            ax.set_title(f"PCA of {data_title}")
            ax.set_xlabel(f"explained variance: {pca.explained_variance_ratio_[0]}")
            ax.set_ylabel(f"explained variance: {pca.explained_variance_ratio_[1]}")

        else:
            raise ValueError("n_comps not supported")
