import argparse
import json
import os
import shutil
import time
from typing import Any, Dict, Iterable

import numpy as np


def log_table(log, data, title=None, add_space=False, sort_keys=True):
    clear_space = "\n" + (" " * 25) + "-" if add_space else "\n - "
    print_list = []
    if title is not None:
        print_list += [["b", title]]

    if isinstance(data, dict):
        if len(data) == 0:
            return

        max_key_len = len(max(data.keys(), key=len))
        keys = sorted(data.keys()) if sort_keys else data.keys()
        for key in keys:
            print_list += [
                ["y", str(clear_space + " %" + str(max_key_len) + "s") % str(key)],
                ["p", " ", data[key]],
            ]
            # for k, v in data.items():
            # print_list += [
            #     ['y', str(clear_space + ' %' + str(max_key_len) + 's') % str(k)],
            #     ['p', ' ', v],
            # ]
    elif isinstance(data, list):
        for ele in data:
            print_list += [["y", clear_space], ["p", ele]]

    log(print_list)

    return


def create_dir(
    log, dir_name, reset=False, verb=True, short_dir_name=None, keep_dirs=[]
):
    """create the directory dir_name (or clean it of all content if it already exists and reset)"""

    if dir_name == "":
        raise Exception("Tried to create_dir with empty directory name")
    if verb:
        if short_dir_name is None:
            short_dir_name = dir_name
        log.info(
            [
                ["b", " - create_dir("],
                [("r" if reset else "g"), ("reset" if reset else "exist")],
                ["b", "): "],
                ["y", short_dir_name],
            ]
        )

    if os.path.isdir(dir_name):
        if reset:
            files_in_dir = os.listdir(dir_name)
            for file_now in files_in_dir:
                if os.path.isdir(dir_name + "/" + file_now):
                    if file_now not in keep_dirs:
                        shutil.rmtree(dir_name + "/" + file_now, ignore_errors=True)
                    else:
                        log.info([["b", " - keeping dir: "], ["y", file_now]])
                else:
                    rm_cmnd = "rm "
                    os.system(rm_cmnd + dir_name + "/" + file_now)
    else:
        try:
            os.makedirs(dir_name)
        except FileExistsError:
            pass

    return


def str_to_bool(v):
    if v.lower() in ["yes", "true", "t", "y", "1"]:
        return True
    elif v.lower() in ["no", "false", "f", "n", "0"]:
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected")


def str_join(*keys, **kwargs):
    delim = kwargs["delim"] if "delim" in kwargs else "_"
    return str(delim).join([str(k) for k in keys])


def safe_np_div(a, b, def_val=0):
    def_val = np.full_like(a, def_val)
    return np.divide(a, b, out=def_val, where=(b != 0))


def recursive_replace(data, val_orig, val_new):
    if isinstance(data, dict):
        if val_orig in data:
            data[val_orig] = val_new

        for k, v in data.items():
            data[k] = recursive_replace(v, val_orig, val_new)

    elif isinstance(data, (list, tuple)):
        for n in range(len(data)):
            data[n] = recursive_replace(data[n], val_orig, val_new)
    elif data == val_orig:
        data = val_new

    return data


def try_and_sleep(
    f,
    args=(),
    kwargs={},
    max_tries=3,
    sleep_range=(1, 60),
    exception=Exception,
    log=None,
    verbose=False,
):
    """
    Function to try several times to execute function. Each time it fails,
    sleep for a while. When resources are busy when loading data from hdf5
    files in parallel, this function can help, by simply trying again after
    the pause.

    Parameters
    ----------
    f: callable
        Function that is tried to execute
    args: dict, optional
        arguments to pass to the function
    kwargs: dict, optional
        keyword arguments to pass to the function
    max_tries: int
        maximum number of tries till raising an error
    sleep_range: tuple, list
        range in seconds between the random sleeping duration is drawn
    exception: Exception
        type of exeption that are filtered. Default is `Exception`
        which will repeat the action for any occuring exception.
    log:
        Logger to be used.
    """
    failed_iteration = 0
    success = False
    while not success:
        try:
            result = f(*args, **kwargs)
            success = True
        except exception as e:
            failed_iteration += 1
            if log is not None:
                log_msg = [
                    ["r", f"Failed in {f.__name__}"],
                    ["y", f" Try: {failed_iteration} / {max_tries}"],
                ]

            if verbose:
                print(e)

            if failed_iteration >= max_tries:
                log_msg += [["y", f" --> Aborting..."]]
                log.debug(log_msg)
                raise Exception(e)
            else:
                log_msg += [["y", f" --> Trying again..."]]
                log.debug(log_msg)

            t_sleep = np.random.randint(sleep_range[0], sleep_range[1])
            time.sleep(t_sleep)

    return result


class LoggingContext:
    """
    Temporarily change the log level of the logger.
    """

    def __init__(self, logger, level=None):
        self.logger = logger
        if level is None:
            self.log_level = level
        else:
            self.log_level = level.upper()

    def __enter__(self):
        if self.log_level is not None:
            self.old_level = self.logger.log_level
            self.logger.log.setLevel(self.log_level)

    def __exit__(self, et, ev, tb):
        if self.log_level is not None:
            self.logger.log.setLevel(self.old_level)


def get_index_in_array(x, y):
    """
    Get the index in array y for all entries of x

    Parameters
    ----------
    x: numpy.ndarray
    y: numpy.ndarray

    Returns
    -------
    yindex: numpy.ndarray

    """
    index = np.argsort(x)
    sorted_x = x[index]
    sorted_index = np.searchsorted(sorted_x, y)

    yindex = np.take(index, sorted_index, mode="clip")
    mask = x[yindex] != y

    yindex = yindex.astype(np.float)
    yindex[mask] = np.nan
    return yindex


def get_columns_wildcard(df, col):
    """
    Get the column names which are matching with wildcards

    Parameters
    ----------
    df: pd.DataFrame
    col: str

    Returns
    -------
    columns: list
        list of columns that fullfill the wild card
    """

    if "*" not in col:
        if col not in df.columns:
            return []
        return [col]

    if (col[0] == "*") and (col[-1] == "*"):
        cols_now = df.columns[df.columns.str.contains(col[1:-1])]

    elif col[0] == "*":
        cols_now = df.columns[df.columns.str.endswith(col[1:])]

    elif col[-1] == "*":
        cols_now = df.columns[df.columns.str.startswith(col[:-1])]

    return list(cols_now)


def write_single_output_df(
    df_in, output_dir, in_file_filt_str, out_file_patt, tag_hdf, n_file=None, log=None
):
    if n_file is None:
        n_file = get_n_write_file(dir_path=output_dir, out_file_patt=out_file_patt)

    file_name = str_join(
        out_file_patt[0],
        in_file_filt_str,
        str("%08d" % n_file),
        out_file_patt[1],
    ).replace("__", "_")

    file_path = os.path.join(output_dir, file_name)

    df_in.to_hdf(
        file_path,
        key=tag_hdf,
        mode="w",
    )

    if log is not None:
        log.info(
            [
                ["y", " - # events: " + str(len(df_in)) + " written to: "],
                ["b", file_path],
            ]
        )


def get_n_write_file(dir_path, out_file_patt):
    if isinstance(out_file_patt, str) or not isinstance(out_file_patt, Iterable):
        out_file_patt = [out_file_patt]

    dir_files = sorted(os.listdir(dir_path))
    dir_files = [f for f in dir_files if all(o in f for o in out_file_patt)]
    n_file = len(dir_files)

    return n_file


class ParseKwargs(argparse.Action):
    """
    Helper for variable parsing of command line arguments depending on input.
    """

    def __call__(self, parser, namespace, values, *args, **kwargs):
        is_dict = all(["=" in v for v in values])

        if is_dict:
            setattr(namespace, self.dest, dict())
            for value in values:
                key, value = value.split("=")
                getattr(namespace, self.dest)[key] = value
        else:
            if len(values) == 1:
                # check if a list is passed
                is_list = values[0].startswith("[") and values[0].endswith("]")
                # list was given
                if is_list:
                    setattr(namespace, self.dest, json.loads(values[0]))

                # values[0] is defined as a string. Check some common cases:

                # see if boolean is given
                elif values[0] == "True":
                    setattr(namespace, self.dest, True)
                elif values[0] == "False":
                    setattr(namespace, self.dest, False)
                # value is None
                elif values[0] == "None":
                    setattr(namespace, self.dest, None)
                else:
                    # strings contraining characters `_` are ignored in python when converting it to integer or float.
                    # (E.g. int("1_0") = 10). However, theses are expected to be parsed as string. Thus we first check
                    # if the value contains `_` and always ares it as string in this case.

                    if "_" in values[0]:
                        is_int = False
                        is_float = False
                    else:
                        # try parsing as integer
                        try:
                            setattr(namespace, self.dest, int(values[0]))
                            is_int = True
                        except ValueError:
                            is_int = False

                        if not is_int:
                            # try parsing as float
                            try:
                                setattr(namespace, self.dest, float(values[0]))
                                is_float = True
                            except ValueError:
                                is_float = False

                    if not is_int and not is_float:
                        # parse as string
                        setattr(namespace, self.dest, values[0])

            else:
                setattr(namespace, self.dest, values)


def check_parameter_source_independence(sources: Dict[str, Dict[str, Any]]) -> bool:
    seen = {}
    for source_name, params in sources.items():
        for param_name in params.keys():
            if param_name not in seen:
                seen[param_name] = [source_name]
            else:
                seen[param_name].append(source_name)

    has_overlap = False
    for param_name, sources in sorted(seen.items()):
        if len(sources) > 1:
            print(
                f"'{param_name}' was defined {len(sources)} times: ({' + '.join(sources)})"
            )
            has_overlap = True

    return has_overlap
